#include <alloca.h>
#include <stack>
#include <assert.h>
#include <fenv.h>
#include "BIH.h"
#include "debug.h"
#include "Ray.h"
#include "Misc.h"
#include "Main.h"

#define BIHDEBUG 0

BIH::BIH(){
  numPrims = 0;
  maxSideLength = 0.0;
  bihDebug = 0;
  maxDepth = 8;
  maxItems = 10;
  rayId = 0;
  pthread_mutex_init(&lock, NULL);
  rays = 0;
  reset();
}

BIH::~BIH(){
  struct timeval start, end;
  cerr << "Deleting BIH..." << flush;
  gettimeofday(&start, NULL); // start delay measure

  // nothing to do...
  
  // destroy mutex
  pthread_mutex_destroy( &lock );

  gettimeofday(&end, NULL);   // end delay measure
  cerr << " done. (" << delay(start, end) << "ms)" << endl;
}


// Resets tree, deleting all nodes
// This is called by rescanScene() to regenerate an BIH
// because of changed geometry.
void BIH::reset(){

  // clear tree visualization
  showCube(SbVec3f(0.0, 0.0, 0.0), SbVec3f(0.0, 0.0, 0.0),
	   SbVec3f(0.0, 0.0, 0.0), 1);

  // clear tree
  primList.clear();
  nodeList.resize(1);
  vector<node>(nodeList).swap(nodeList);
  root = NULL;

  // reset variables
  root = &nodeList[0];
  root->nodeType = BIH_LEAF;
  root->index = 0;
  root->items = 0;
  
  min[X] = min[Y] = min[Z] = 1e30;
  max[X] = max[Y] = max[Z] = -1e30;
  shiny_min.setValue(1e30f, 1e30f, 1e30f);
  shiny_max.setValue(-1e30f, -1e30f, -1e30f);
  maxSideLength = 0.0;
  numPrims = 0;
}
  
// set triangle tree parameters
void BIH::setParams(int maxD, int maxI, int debug){
  maxDepth = maxD;
  maxItems = maxI;
  bihDebug = debug;
}

// creates the BIH structure
void BIH::refine(int simpleDiv){
    
  // adjust first node boundaries
  // (make global bounding box)
  for (int ti=0; ti<primList.size(); ti++){
    const Primitive *prim = primList[ti];
    min = BBox::vecMin(min, prim->getMin());
    max = BBox::vecMax(max, prim->getMax());
    shiny_min = BBox::vecMin(shiny_min, prim->getMinShiny());
    shiny_max = BBox::vecMax(shiny_max, prim->getMaxShiny());
  }

  // setup boundaries of first node
  for (int i=0; i<3; i++){
    if (max[i]<min[i]){
      min[i] = -1;
      max[i] = 1;
    }
    if (shiny_max[i]<shiny_min[i]){
      shiny_min[i] = -1;
      shiny_max[i] = 1;
    }
    if (max[i]-min[i] > maxSideLength)
      maxSideLength = max[i]-min[i];
  }

  // find bounding sphere
  center = (min+max)*.5;
  radius = (max-center).length();
  //showPoint(center, radius, SbColor(0.0, 0.0, 1.0));

  // shiny bounding sphere
  shiny_center = (shiny_min+shiny_max)*.5;
  shiny_radius = (shiny_max-shiny_center).length();
  //showPoint(shiny_center, shiny_radius, SbColor(1.0, 0.0, 0.0));
  
// #ifdef SHOWTRAV  
//   // debugging
//   root->min[0] = min[0];
//   root->min[1] = min[1];
//   root->min[2] = min[2];
//   root->max[0] = max[0];
//   root->max[1] = max[1];
//   root->max[2] = max[2];
//   root->depth = 0;
// #endif
  
//   if (min[X] == max[X] ||
//       min[Y] == max[Y] ||
//       min[Z] == max[Z]){
//     cerr << "You forgot to set the size of the tree before refine().\n";
//     exit(1);
//   }

  root->items = primList.size();
  refineNode();

  //reorderNodes();
}

namespace {
  typedef struct refineNode_st {
    BBox splitBox;
    BBox objsBox;
    int lIdx, rIdx;    // first and last primitive idxs
    int nodeIdx;       // current node index (itself valid leaf)
    int depth;
  } refineNode_t;
}

void BIH::refineNode(){

  SbMatrix scale;
  float nmin[3], nmax[3];

  stack<refineNode_t> subdivStack;
  subdivStack.push(refineNode_t());
  refineNode_t &rn = subdivStack.top();
  rn.splitBox = BBox(min, max);
  rn.objsBox = rn.splitBox;
  rn.lIdx = 0;
  rn.rIdx = primList.size()-1;
  rn.nodeIdx = 0;
  rn.depth=1;

//   if (bihDebug){
//     showCube(SbVec3f(min[X], min[Y], min[Z]),
// 	     SbVec3f(max[X], max[Y], max[Z]),
// 	     SbColor(0.0, 1.0, 1.0));
//   }

  while (!subdivStack.empty()){
    refineNode_t rn = subdivStack.top();
    subdivStack.pop();

    if (bihDebug){
      const SbVec3f &min = rn.objsBox.min;
      const SbVec3f &max = rn.objsBox.max;
      showCube(SbVec3f(min[X], min[Y], min[Z]),
               SbVec3f(max[X], max[Y], max[Z]),
               SbColor(0.0, 1.0, 1.0));
    }
    
    int nprims = rn.rIdx-rn.lIdx+1;
    if (rn.depth>=maxDepth || nprims<=maxItems){
      // leaf
      if (BIHDEBUG){
        cerr << "nprims=" << nprims <<"\n";
        if (nprims)
          showCube(rn.objsBox.min,
                   rn.objsBox.max,
                   SbColor(min2((nprims+1)/255.0,1.0),
                           0, //min2((rn.depth+1)/255.0,1.0),
                           0));
      }
      continue;
    }

    // inner node
//     showCube(rn.objsBox.min,
//              rn.objsBox.max,
//              SbColor(0.0, min2((rn.depth+1)/255.0,1.0), 0));
    

    // find split axis and plane position
    int sa;
    float sp;

    // subdivides until split plane overlaps objsBox
    while (1) {
      sa = rn.splitBox.longestAxis();
      sp = (rn.splitBox.max[sa]+rn.splitBox.min[sa])*0.5;

      if (rn.objsBox.max[sa]<=sp){        // all left
        rn.splitBox.max[sa] = sp;
        continue;
      } else if (rn.objsBox.min[sa]>sp){  // all right
        rn.splitBox.min[sa] = sp;
        continue;
      }
      break;
    }

    const char axis[] = "xyz";
    if (BIHDEBUG)
      cerr << "Sorting: " << axis[sa] << "=" << sp << "\n";

    // sort primitives according to split plane
    int li = rn.lIdx; // prims < are left (or equal)
    int ri = rn.rIdx; // prims > are right

    while (1){

      // increment li until prim is right
      while (li<primList.size() && li<=rn.rIdx &&
             (primList[li]->getMin(sa)+
              primList[li]->getMax(sa))*.5 <= sp)
        li++;

      // decrement ri until prim is left (or equal)
      while (ri>=0 && ri>=rn.lIdx &&
             (primList[ri]->getMin(sa)+
              primList[ri]->getMax(sa))*.5 > sp)
        ri--;

      if (li>=ri) break;
      
      // swap prims
      assert(li>=0 && li<primList.size());
      assert(ri>=0 && ri<primList.size());
      primList.swap(li, ri);
      li++;
      ri--;
    }

    // split node
    node n0, n1;  // rn.lIdx...li-1, ri+1...rn.rIdx
    n0.setLeaf();
    n0.index = rn.lIdx;
    n0.items = (li-1) - rn.lIdx + 1;
    assert(n0.items>=0);

    n1.setLeaf();
    n1.index = ri+1;
    n1.items = rn.rIdx - (ri+1) + 1;
    assert(n1.items>=0);
    
    // find new bboxes and split planes
    refineNode_t rn0, rn1;
    rn0.depth = rn.depth+1;
    rn0.lIdx = rn.lIdx;
    rn0.rIdx = ri;
    rn0.nodeIdx = nodeList.size()+0;
    rn0.splitBox = rn.splitBox;
    rn0.splitBox.max[sa] = sp;
    rn0.objsBox = rn.objsBox;
    float newMax = rn.objsBox.min[sa];
    for (int i=n0.index; i<n0.index+n0.items; i++)
      newMax = std::max(newMax, primList[i]->getMax(sa));
    rn0.objsBox.max[sa] = newMax;

    rn1.depth = rn.depth+1;
    rn1.lIdx = li;
    rn1.rIdx = rn.rIdx;
    rn1.nodeIdx = nodeList.size()+1;
    rn1.splitBox = rn.splitBox;
    rn1.splitBox.min[sa] = sp;
    rn1.objsBox = rn.objsBox;
    float newMin = rn.objsBox.max[sa];
    for (int i=n1.index; i<n1.index+n1.items; i++)
      newMin = std::min(newMin, primList[i]->getMin(sa));
    rn1.objsBox.min[sa] = newMin;

    // make sure objects list or bb is smaller
    if (n0.items == nodeList[rn.nodeIdx].items &&
        rn0.objsBox == rn.objsBox){
      if (BIHDEBUG)
        cerr << "retrying split (all left)\n";
      rn.splitBox.max[sa] = sp;
      rn.depth++;
      subdivStack.push(rn);  // try splitting again
      continue;
    } else if (n1.items == nodeList[rn.nodeIdx].items &&
               rn1.objsBox == rn.objsBox){
      if (BIHDEBUG)
        cerr << "retrying split (all right)\n";
      rn.splitBox.min[sa] = sp;
      rn.depth++;
      subdivStack.push(rn);  // try splitting again
      continue;
    }
    
    nodeList.push_back(n0);
    nodeList.push_back(n1);
    
    subdivStack.push(rn0);
    subdivStack.push(rn1);

//     if (bihDebug){
//       showCube(rn0.objsBox.min,
//                rn0.objsBox.max,
//                SbColor(0.0, 1.0, 1.0));
//       showCube(rn1.objsBox.min,
//                rn1.objsBox.max,
//                SbColor(0.0, 1.0, 1.0));
//     }
    

    // turn current node into inner node
    node &cn = nodeList[rn.nodeIdx];
    cn.setAxis(sa);
    cn.index = nodeList.size()-2;
    cn.clip[0] = rn0.objsBox.max[sa];
    cn.clip[1] = rn1.objsBox.min[sa];
  }

} // end of refineNode();

namespace {
  typedef struct {
//    BBox bb;             // distance to bbox faces
    int nodeIdx;
    double bmin, bmax;   // bbox min and max distance
    double tmin, tmax;   // bbox min and max distance including t0,t1
  } traverseNode_t;
}

bool BIH::intersectRay(Ray &r, double t0, double t1) {

  // count rays
  r.ctx->rays++;

  // set no hit by default
  r.t = -1;
  r.prim = NULL;

  // debugging
  r.accelTravs = 0;
  r.accelInts = 0;
  
  // return value pointers (private class members)
  SbVec3f invDir;
  BBox bb;   // NOTE! distance to {min,max}bounding_box{x,y,z}
  int f[3];  // "front"  0 or 1, depending on sign(dir)
  int b[3];  // "back"   0 or 1
  for (int i=0; i<3; i++){
    f[i] = r.dir[i]>=0 ? 0 : 1;
    b[i] = f[i]^1;
    invDir[i] = fabs(r.dir[i])<1e-9 ? 1e9 : 1./r.dir[i];
    bb[0][i] = (min[i]-r.pos[i])*invDir[i];
    bb[1][i] = (max[i]-r.pos[i])*invDir[i];
  }

  double tntmin = max2(t0, (double)max3(bb[f[0]][0], bb[f[1]][1], bb[f[2]][2]));
  double tntmax = min2(t1, (double)min3(bb[b[0]][0], bb[b[1]][1], bb[b[2]][2]));

  // AUGGGG
  // http://www.parashift.com/c++-faq-lite/newbie.html#faq-29.18
  // "Why is cos(x) != cos(y) even though x == y?"
  // Register truncation sometimes causes dist_to_plane to be DIFFERENT for the
  // same plane
#define EPSDIFF 0.000001
#define EPSMAX (1+EPSDIFF)
#define EPSMIN (1-EPSDIFF)
  
  if (tntmin*EPSMIN-EPSDIFF > tntmax*EPSMAX+EPSDIFF)
    return false;
  
  traverseNode_t *travStack = (traverseNode_t*)alloca(sizeof(traverseNode_t)*maxDepth);
  int tsi=0; // top (occupied) stack index
  traverseNode_t &tn = travStack[tsi];
//  tn.bb = bb;
  tn.nodeIdx = 0;
  tn.tmin = tntmin;
  tn.tmax = tntmax;

  // something is only on stack if it is intersected

  double minT=t1;
  Primitive *minPrim;
  SbVec3f minBc;
  FastPrimStack minPrim2;
  bool hit = false;
  
  // test nodes on stack
  while (tsi>=0){

    r.accelTravs++;

    traverseNode_t tn = travStack[tsi--];

//    cerr << "hello travs=" << r.accel_travs << " ints=" << r.accel_ints << "\n";

//     if (tn.bmin>tn.bmax || tn.bmin>minT)
//       continue;

    if (tn.tmin*EPSMIN-EPSDIFF>minT)
      continue;
    
//    BBox &bb = tn.bb;
    node &cn = nodeList[tn.nodeIdx];
    
//     if (bihDebug){
//       SbVec3f m0,m1;
//       for (int i=0; i<3; i++){
//         m0[i] = bb.min[i]*r.dir[i]+r.pos[i];
//         m1[i] = bb.max[i]*r.dir[i]+r.pos[i];
//       }
//       showCube(m0, m1, SbColor(0.0, 1.0, 1.0));
//       cerr << "Traversing " << min << " -> " << max << "\n";
//     }

    if (cn.isLeaf()){        // leaf
      
      // loop over prims
      double myt0 = tn.tmin*EPSMIN-EPSDIFF;
      double myt1 = tn.tmax*EPSMAX+EPSDIFF;
      for (int i=0; i<cn.items; i++){

        if (primList[cn.index+i]->intersectRay(r, myt0, myt1) &&
            r.t < minT){
          hit = true;
          minT = r.t;
          minPrim = r.prim;
          minBc = r.bc;
          minPrim2 = r.prim2;
        }
      }

      r.accelInts += cn.items;
      
    } else {               // inner node

      // check children
      int sa = cn.splitAxis();
      double lmin, lmax, rmin, rmax;
      double d0 = (cn.clip[0]-r.pos[sa])*invDir[sa];
      double d1 = (cn.clip[1]-r.pos[sa])*invDir[sa];

      if (f[sa]==0){  // "front" is left
        lmin = tn.tmin;
        lmax = min2(d0, tn.tmax);
        rmin = max2(d1, tn.tmin);
        rmax = tn.tmax;
      } else {
        lmin = max2(d0, tn.tmin);
        lmax = tn.tmax;
        rmin = tn.tmin;
        rmax = min2(d1, tn.tmax);
      }
      
      if (lmin*EPSMIN-EPSDIFF<=lmax*EPSMAX+EPSDIFF){
        tsi++;
        traverseNode_t &tn0 = travStack[tsi];
        tn0.nodeIdx = cn.index;
        tn0.tmin = lmin;
        tn0.tmax = lmax;
      }

      if (rmin*EPSMIN-EPSDIFF<=rmax*EPSMAX+EPSDIFF){
        tsi++;
        traverseNode_t &tn1 = travStack[tsi];
        tn1.nodeIdx = cn.index+1;
        tn1.tmin = rmin;
        tn1.tmax = rmax;
      }
    }
  }

  if (!hit) return false;
  
  r.t = minT;
  r.bc = minBc;
  r.prim = minPrim;
  r.prim2 = minPrim2;
  return true;
}

void BIH::printStats(){

  int inner=0, leafs=0, maxPrims=0, minPrims=0,
    totPrims=0, minDepth=9999, maxDepth=0, depth=1, avgDepth=0;
  
  stack<int> nodes;
  nodes.push(0); // root node

  while (!nodes.empty()) {

    int nodeIdx = nodes.top();
    nodes.pop();

    if (nodeIdx<0){
      //
      //     o     d=1
      //    / \
      //   o   o   d=2
      //  / \   \
      // o   o     d=3
      depth--;
      continue;
    }
    
    //if (nodeIdx>=nodeList.size()) return;
  
    // if leaf node, count prims
    if (nodeList[nodeIdx].isLeaf()){
      leafs++;
      avgDepth += depth;
      if (depth>maxDepth) maxDepth = depth;
      if (depth<minDepth) minDepth = depth;
      if (BIHDEBUG)
        cerr << "LEAF DEPTH=" << depth << "\n";
      int n = nodeList[nodeIdx].items;
      totPrims += n;
      if (n<minPrims) minPrims = n;
      if (n>maxPrims) maxPrims = n;
    }
    // otherwise, recurse over children
    else {
      inner++;
      depth++;
      nodes.push(-1); // signifies done we're with this child
                      // and we should decrement depth
      nodes.push(nodeList[nodeIdx].index+0);
      nodes.push(nodeList[nodeIdx].index+1);
    }
  }
  
  cerr << "BIH statistics: "
       << inner << " inner nodes, "
       << leafs << " leaf nodes, "
       << minPrims << " min prims/leaf, "
       << maxPrims << " max prims/leaf, "
       << totPrims/(float)leafs << " avg prims/leaf, "
       << avgDepth/float(leafs) << " avg leaf depth, "
       << minDepth << " min leaf depth, "
       << maxDepth << " max leaf depth, "
       << totPrims << " primitives, "
       << numPrims << " initially ("
       << totPrims/float(numPrims?numPrims:1)*100.0 << "%).\n";
  cerr << "  min:(" << min << ")  max=(" << max << "\n";
  return;
}

void BIH::printStatsTrav(int nodeIdx, int &numInner,
                         int &numLeafs, int &maxPrims,
                         int &minPrims, int &numPrims){


  return;
}
