//WWWW''''''''WW''''''''WWW''''''''WW''''''''''WWW''''''''''''''''''
//WW  W       WW       WW  W       WW           WW                  
//WW  W  WWW WWWW  WWW  WW    WW  WWWW     WWW  WW   WWW  WWWW  WWWW
//WW  W W WW  WW  W WW   WW  WWWW  WW     WW W  WW  W WW WWW   WWW  
//WW  W W WW  WW  W WW W  WW WW    WW     WW    WW  W WW   WWW   WWW
//WWWW   WWW  WW   WWW  WWW   WWW  WW      WWW WWWW  WWW WWWW  WWWW 
//                                                                  
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#ifndef DATASETCLASS
#define DATASETCLASS

#include <string>
#include <Inventor/SbLinear.h>
#include <Inventor/SbColor.h>
#include "Main.h"

using namespace std;

class Ray; // forward declaration of Ray

class DataSet {
  public:

  // This class uses a voxel centers representation.
  // I.e., if dim[]={20,20,20}, there are 20 voxels along each dimension,
  // and we store the value for the center of each of those voxels.
  
  int dim[3];              // dimensions of volume data
  int sdim[3];             // storage dimensions for bricking
  float *data;             // pointer to volume data
  SbVec3f min, max;        // min and max XYZ in spatial coordinates
  SbVec3f voxel, invVoxel; // size of each voxel (deltas)
  int *Fx, *Fy, *Fz;       // precompute index tables
  SbColor diffColor;       // color of dataset
  SbColor specColor;
  float transparency;      // transparency of dataset
  float shininess;
  float voxelVol;          // voxel volume
  float minValue;          // minimum value
  float maxValue;          // maximum value
  float totalDensity;      // sum of values
  float totalBrightness;   // maximum brightness (if value is temperature)
  double defaultStepSize;  // default step size, about size of 1 voxel
                           // in local (post-transform) coordinates
  float ***weightsX, **weightsY, *weightsZ; // for storing brightness PMF
                           // BTW, CDF is faster to sample than PMF
                           // CDF = Cumulative Density Function
                           // PMF = Probablity Mass Function
  int *colors;             // Precomputed emission colors (indices)
  SbColor *colorsTable;
  float dataScaleVal;      // Cumulative scale on the data, for special
                           // purposes. Stored data has scale baked in.
  float volMaxDist;        // maximum distance through volume
  
  const static int colorSamples;
  const static float minTemp;
  const static float maxTemp;
  const static float tempDeltaInv;
  SbMatrix transform;
  void setRotation(SbRotation r);
  
//  int offset[2][2][2]; // [i][j][k]

    /*     ____
       /|4  /|6    z
      / |  / |     |
     /__|0=min2    |__y
    5|  / |7=max  /
     | /  | /    x
    1|/___|3

          012
          ---
    0 = p 000
    1 = p 100
    2 = p 010
    3 = p 110
    4 = p 001
    5 = p 101
    6 = p 011
    7 = p 111

  */

  
  // constructor
  DataSet();
  
  ~DataSet(){
    delete [] data;
    delete [] Fx;
    delete [] Fy;
    delete [] Fz;
    delete [] colors;
  }
  
//  inline float val(int i, int j, int k) const {
//    return(data[ dim[0]*(dim[1]*k + j) + i] );
//  }

  static const int m = 2;   // for intel machines
  static const int n = 5;
//  static const int m = 4;
//  static const int n = 4;
  virtual float val(int x, int y, int z) const {
    int q = Fx[x] + Fy[y] +Fz[z];
	/*
      ((x/n)/m)* n*n*n* m*m*m* ((dim[2]/n)/m)* ((dim[1]/n)/m) +
      ((y/n)/m)* n*n*n* m*m*m* ((dim[2]/n)/m) +
      ((z/n)/m)* n*n*n* m*m*m +
      ((x/n)%m)* n*n*n* m*m +
      ((y/n)%m)* n*n*n* m +
      ((z/n)%m)* n*n*n +
      (x%(n*n))* n*n +
      (y%n)*n +
      (z%n);
	*/
    
    return data[q];
  }
  virtual float val(const SbVec3f &x_) const {
    SbVec3f x;
    transform.multVecMatrix(x_, x);
    return valNoXform(x);
  }
  float valBound(const SbVec3f &x_) const{

    SbVec3f x;
    transform.multVecMatrix(x_, x);
    
    if (x[0]>=min[0] && x[0]<=max[0] &&
	x[1]>=min[1] && x[1]<=max[1] &&
	x[2]>=min[2] && x[2]<=max[2])
      return valNoXform(x);
    else
      return 0.;
  }
  virtual float valNoXform(const SbVec3f &x) const;
  virtual float valNoInterp(const SbVec3f &x_) const {

    SbVec3f x;
    transform.multVecMatrix(x_, x);
  
    // convert position into index
    int i, j, k;
    xToijk(x, i, j, k);

    return val(i,j,k);
  }
  
  
  SbVec4f calcBaryCent(SbVec3f x);
  SbVec3f getGradient(SbVec3f x) const;
  SbVec3f getIntNormal(SbVec3f x) const;
  
  inline SbVec3f pos(int i, int j, int k){
    return(SbVec3f((i+0.5f)*voxel[0]+min[0],
		   (j+0.5f)*voxel[1]+min[1],
		   (k+0.5f)*voxel[2]+min[2]));
  }
  
  inline SbVec3f pos(int p[3]){
    return(SbVec3f((p[0]+0.5f)*voxel[0]+min[0],
		   (p[1]+0.5f)*voxel[1]+min[1],
		   (p[2]+0.5f)*voxel[2]+min[2]));
  }

  inline SbVec3f pos(float i, float j, float k){
    return(SbVec3f((i+0.5f)*voxel[0]+min[0],
		   (j+0.5f)*voxel[1]+min[1],
		   (k*0.5f)*voxel[2]+min[2]));
  }
  
  virtual void useData(const string &filename, int _dim[],
		       int header=0, int fourBytes=0,
		       SbVec3f _min = SbVec3f(0.0, 0.0, 0.0),
		       SbVec3f _max = SbVec3f(100.0, 100.0, 100.0));
  virtual void scaleData(float mul);
  virtual void scaleMaxToOne();
  virtual float getMaxValue(){  return maxValue;  }
  void computeWeights(float scale);
  SbVec3f impSample(float x, float r1, float r2, float r3);
  void drawGrid();
  
  // convert position into cell index
  inline void xToijk(const SbVec3f &x, int &i, int &j, int &k) const{
    float q = (x[0]-min[0])*invVoxel[0] - 0.5f;
    float r = (x[1]-min[1])*invVoxel[1] - 0.5f;
    float s = (x[2]-min[2])*invVoxel[2] - 0.5f;

    i = (int)floor(q);
    j = (int)floor(r);
    k = (int)floor(s);
  
    i = clamp(i, 0, dim[0]-2);
    j = clamp(j, 0, dim[1]-2);
    k = clamp(k, 0, dim[2]-2);
  }

  // convert position into cell index and offset
  inline void xToijkabc(const SbVec3f &x,
			int &i, int &j, int &k,
			float &a, float &b, float &c) const{
    
    float q = (x[0]-min[0])*invVoxel[0] - 0.5f;
    float r = (x[1]-min[1])*invVoxel[1] - 0.5f;
    float s = (x[2]-min[2])*invVoxel[2] - 0.5f;

    i = (int)floor(q);
    j = (int)floor(r);
    k = (int)floor(s);

    i = clamp(i, 0, dim[0]-2);
    j = clamp(j, 0, dim[1]-2);
    k = clamp(k, 0, dim[2]-2);
    
    a = q-i;
    b = r-j;
    c = s-k;

    a = clamp(a, 0.f, 1.f);
    b = clamp(b, 0.f, 1.f);
    c = clamp(c, 0.f, 1.f);
  }
  
  inline SbVec3f ijkabcToPos(int i, int j, int k,
			     float a, float b, float c) const {
    return SbVec3f(min[0]+(i+a)*voxel[0],
		   min[1]+(j+b)*voxel[1],
		   min[2]+(k+c)*voxel[2]);
  }
  
  float bresenSum(const SbVec3f &p1, const SbVec3f &p2,
		  float *t=NULL, const float stopAmt=1e30);
  float normalSum(const SbVec3f &p1, const SbVec3f &p2,
		  float *t=NULL, const float stopAmt=1e30);
  void clipToVolume(const SbVec3f &p1_, const SbVec3f &p2_,
		    SbVec3f &p1, SbVec3f &p2, float &t1, float &t2);

  void writeData(const string &fileName, float trimFactor=1.0f);

  // color storage
  void initColors(float mpu);

  // for setting color values
  SbColor& colorVal(int x, int y, int z) {
    const int q = Fx[x] + Fy[y] +Fz[z];
    return(colorsTable[colors[q]]);
  }
  
  const SbColor& colorVal(int x, int y, int z) const {
    const int q = Fx[x] + Fy[y] +Fz[z];
    return(colorsTable[colors[q]]);
  }

  SbColor colorVal(const SbVec3f &x_) const;
  
  const SbColor& colorValPoint(const SbVec3f &x_) const {

    SbVec3f x;
    transform.multVecMatrix(x_, x);
  
    // convert position into index
    int i, j, k;
    xToijk(x, i, j, k);

    return colorsTable[colors[Fx[i]+Fy[j]+Fz[k]]];
  }

  const int colorIdxLookup(const float temp) const {
    int i= (int)((temp-minTemp)*tempDeltaInv);
    return clamp(i, 0, colorSamples);
  }

  const SbColor& colorLookup(const float temp) const {
    return colorsTable[colorIdxLookup(temp)];
  }

  const float getVolMaxDist() { return volMaxDist; }
  
};

class MultiDataSet : public DataSet {
public:
  vector<DataSet *> sets;
  MultiDataSet(DataSet *ds) {
    *(DataSet*)this = *ds;
    
    sets.push_back(ds);
  }
  ~MultiDataSet(){
    // prevent base dtor from double dtor'ing,
    // since we copied a dataSet to local vars
    data = NULL;
    Fx = Fy = Fz = NULL;
    colors = NULL; 
  }
  float valNoXform(const SbVec3f &x) const;

};

#endif /* DATASETCLASS */
