#include "Fractal.h"

#include <Inventor/SbLinear.h>
#include <algorithm>
#include "Main.h"
#include "debug.h"

Fractal::Fractal(Primitive *p, SoLinearFractal *soFract) : prim(p) {
  init(soFract);
}

static Instance localInst[16];        // per-thread temp storage

void Fractal::init(SoLinearFractal *soFract){

  // set Max Depth
  static int maxD = getenv("PANE_MD") ? atoi(getenv("PANE_MD")) : 10;
  maxDepth = maxD; // depends on resolution

  string xformsFile = 
    //"sier_xforms.iv"
    //"pound_xforms3D.iv"
    "ompf_xforms2.iv"  // don't forget to unset skipIthXform and set intersectIntermediate
    //"randh_xforms.iv"
    //"tree_xforms.iv"
    ;
  onlyIntersectBV = false;
  skipIthXform = false;
  intersectIntermediate = true;

  if (soFract){
    maxDepth              = soFract->depth.getValue();
    xformsFile            = soFract->xformsFile.getValue().getString();
    onlyIntersectBV       = soFract->isectBVOnly.getValue();
    intersectIntermediate = soFract->isectIntermediate.getValue();
    skipIthXform          = soFract->skipIthXform.getValue();
  }

  loadXforms(xformsFile.c_str());
  
  int numXforms = xforms.size();
  if (!numXforms){
    cerr << __FUNCTION__ << ": No transforms found! Using identity.\n";
    xforms.push_back(SbMatrix::identity());
  }

  // compute scales
  scales.resize(numXforms);
  for (int i=0; i<numXforms; i++){
    SbVec3f t, s;
    SbRotation r, so;
    xforms[i].getTransform(t,r,s,so);
    scales[i] = MAX3(s[0], s[1], s[2]);
  }

  // compute initial bounding sphere
  // using algorithm from S.3.3
  // "Efficient Antialiased Rendering of 3-D Linear Fractals" by John C. Hart
  bsPos = SbVec3f(0,0,0);
  bsRad = 1; // amazingly, it does not matter what the initial value is

  int iters=MAX2(10000,maxDepth);
  for (int j=0; j<iters; j++){
    SbVec3f avgPos(0,0,0);
    float maxRad = 0;
    for (int i=0; i<numXforms; i++){

      // avg position
      SbVec3f posNew;
      xforms[i].multVecMatrix(bsPos, posNew);
      avgPos += posNew;

      // maximum distance
      float radi = (bsPos-posNew).length() + scales[i]*bsRad;
      if (radi>maxRad)
        maxRad = radi;
    }

    avgPos /= numXforms;
    bsPos = avgPos;
    bsRad = maxRad;
    //cerr << "bdRad=" << bsRad << endl;
  }

  static float bsRadOverride = getenv("PANE_FRACT_RAD") ?
    atof(getenv("PANE_FRACT_RAD")) : -1;

  if (bsRadOverride>=0)
    bsRad = bsRadOverride;
  cerr << "Bounding center=" << bsPos << " radius=" << bsRad << "\n";
}

bool Fractal::intersectSphere(Ray &r, const SbVec3f &pos, const float rad,
                              const double &t0, const double &t1,
                              double &tNear, double &tFar){
  
  Vec3d oc = Vec3d(r.pos) - Vec3d(pos);
  Vec3d dd = r.dir;
  double ddoc = dd.dot(oc);
  double dd2 = dd.dot(dd);
  double det = ddoc*ddoc - dd2*(oc.dot(oc)-rad*rad);
  
  if (det<0) return false;
  det = sqrt(det);
  tNear = (-ddoc - det);
  if (tNear>t1) return false;
  if (tNear>=t0) return true;
  tFar = (-ddoc + det);
  if (tFar>=t0 && tFar<=t1) return true;
  return false;
}

void Fractal::loadXforms(const char *file){

  if (!file){      // no file given
    return;
  }

  // load .iv file
  SoInput *sceneIn = new SoInput();
  if (!sceneIn->openFile(file)){
    cerr << "Unable to open scene file: " << file << endl;
    exit(1);
  }
  SoSeparator *sep = SoDB::readAll(sceneIn);
  sep->ref();
  delete sceneIn;

  // loop over xforms
  SoSearchAction sa;
  sa.setType(SoTransform::getClassTypeId(), TRUE);
  sa.setSearchingAll(TRUE);
  sa.setFind(SoSearchAction::TYPE);
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(sep);
  SoPathList pList = sa.getPaths();
  cerr << __FUNCTION__ << ": " << pList.getLength() << " transform paths found\n";
  
  for (int i=0; i<pList.getLength(); i++){
    SoTransform *xform = (SoTransform *)pList[i]->getTail();
    SbMatrix result;
    result.setTransform(xform->translation.getValue(),
                        xform->rotation.getValue(),
                        xform->scaleFactor.getValue(),
                        xform->scaleOrientation.getValue(),
                        xform->center.getValue());
    xforms.push_back(result);
  }
}

typedef struct fractal_stackElem {
  int oldIdx;
  int depth;
  float rad;
  SbVec3f pos;
  double t0, t1;
  SbMatrix mat;
  fractal_stackElem(const double &t0_, const double &t1_, int oldIdx_,
                    const SbVec3f &pos_, float rad_, int depth_,
                    const SbMatrix &mat_) :
    t0(t0_), t1(t1_), oldIdx(oldIdx_),
    pos(pos_), rad(rad_), depth(depth_), mat(mat_) {}

  fractal_stackElem() { } 
  
} fractal_stackElem;

namespace {

  typedef struct {
    float d;
    int i;
  } sortRec;

  bool operator<(const sortRec &a, const sortRec &b){
    return a.d > b.d;
  }

}

bool Fractal::intersectRay(Ray &r, double t0, double t1){


   if (r.ignorePrim == this)
     t0 = 0.00002;  // works for fractal pound (.002 produces artifacts)
  
  r.t = -1;
  r.prim = NULL;

  // count rays
  if (onlyIntersectBV)
    r.ctx->rays++;
  
  // recursively intersect until maxdepth
  // then return either intersection or intersection with prim

  double tNear, tFar;
  if (!intersectSphere(r, bsPos, bsRad, t0, t1, tNear, tFar))
    return false;
  
  int numXforms = xforms.size();
  
  fractal_stackElem myStack[200];  // have seen it go as high as 59 so far
                                   // make that 110
  
  int stackIdx = -1; // "top" of stack
  myStack[++stackIdx] = fractal_stackElem(tNear, tFar, -1, bsPos, bsRad, 0,
                                          SbMatrix::identity());

  SbVec3f minBc;
  double minT=1e8;
  Primitive *minPrim=NULL;
  FastPrimStack minPrim2;
  Instance minInst;
  
  while (stackIdx>=0){

    // move closest node to top of stack and pop
    // Unfortunately this uses a lot of memory (stack size of ~3000)
    // and is consequently slow.
//     double minD=1e8;
//     int minIdx=0;
//     for (int i=0; i<=stackIdx; i++){
//       if (myStack[i].t0<minD){
//         minD = myStack[i].t0*0.9999999;
//         minIdx = i;
//       }
//     }
//     fractal_stackElem bv = myStack[minIdx];
//     if (minIdx != stackIdx)
//       myStack[minIdx] = myStack[stackIdx];
//     stackIdx--;
    
//    cerr << "stack item...\n";
    fractal_stackElem bv = myStack[stackIdx--];

    if (bv.t0>minT)
      continue;

    if (bv.depth == maxDepth){

      if (onlyIntersectBV) {
        
        // return the sphere intersection
        minT = bv.t0 >= t0 ? bv.t0 : bv.t1;
        minPrim = this;
      
        // compute normal
        SbVec3f pnt = r.pos + minT*r.dir;
        pnt -= bv.pos;
        pnt.normalize();
        minBc = pnt;
        continue;
      }
      else {
      
        // intersect primitive
        Instance &inst = localInst[r.ctx->id];
        inst = Instance(prim, bv.mat);
        if (inst.intersectRay(r, t0, t1) && r.t<minT){
          minT = r.t;
          minBc = r.bc;
          minPrim = r.prim;
          minPrim2 = r.prim2;
          minInst = inst;
        }
        continue;
      }

    }

    sortRec recs[16];
    
    fractal_stackElem newStack[16];
    int newStackIdx = -1; // "top" of stack

    fractal_stackElem childbv(0, 0, 0, bv.pos, 0, bv.depth+1, bv.mat);
    
    for (int i=0; i<numXforms; i++){

      // skip i==j xform (pound fractal)
      if (skipIthXform && i==bv.oldIdx) continue;

      // intersect child
      SbMatrix mat(bv.mat);
      mat.multLeft(xforms[i]);
      mat.multVecMatrix(bsPos, childbv.pos);
      double childRad = bv.rad*scales[i];

      // temp debugging
      static double minRad=1e30;
      if (childRad<minRad*0.9999999) {
        minRad = childRad;
        cerr << "minRad=" << minRad << "\n";
      }
      static int maxStackIdx=-1;
      if (stackIdx>maxStackIdx){
        maxStackIdx = stackIdx;
        print(" stackIdx=%d\n", stackIdx);
      }
      
      if (intersectSphere(r, childbv.pos, childRad, t0, t1, tNear, tFar)){
        //cerr << "parent rad=" << bv.rad << " child[" << i << "] rad = " << childRad << endl;
        childbv.oldIdx = i;
        childbv.rad = childRad;
        childbv.t0 = tNear;
        childbv.t1 = tFar;
        childbv.mat = mat;
        newStack[++newStackIdx] = childbv;

        recs[newStackIdx].d = tNear;
        recs[newStackIdx].i = newStackIdx;
      }

    }

    // put this node in stack, so intermediate nodes are intersected
    // (ompf fractal)
    if (intersectIntermediate)
      myStack[++stackIdx] = fractal_stackElem(0, 0, -1, bv.pos, bv.rad, maxDepth, bv.mat);

    // or put bounding sphere
//     childbv.rad = bv.rad*.005f;
//     if (intersectSphere(r, bv.pos, childbv.rad, t0, t1, tNear, tFar)){
//       childbv.oldIdx = -1;
//       childbv.t0 = tNear;
//       childbv.t1 = tFar;
//       childbv.depth = maxDepth;
//       myStack[++stackIdx] = childbv;
//     }

    // sort recs ascending order and insert
    sort(recs, recs+newStackIdx+1);
    for (int i=0; i<=newStackIdx; i++){
      myStack[++stackIdx] = newStack[recs[i].i];
    }
    
  }

  if (minT<1e8){
    r.t = minT;
    r.bc = minBc;
    r.prim = minPrim;
    r.prim2 = minPrim2;
    localInst[r.ctx->id] = minInst;
    return true;
  }

  return false;
}


// SoLinearFractal

SO_NODE_SOURCE(SoLinearFractal);

void SoLinearFractal::initClass(){
  SO_NODE_INIT_CLASS(SoLinearFractal, SoNode, "Node");
}

// constructor
SoLinearFractal::SoLinearFractal(){
  SO_NODE_CONSTRUCTOR(SoLinearFractal);

  SO_NODE_ADD_FIELD(depth, (10));  
  SO_NODE_ADD_FIELD(xformsFile, ("xforms.iv"));     // perfectly diffuse
  SO_NODE_ADD_FIELD(isectBVOnly, (FALSE));
  SO_NODE_ADD_FIELD(isectIntermediate, (FALSE));  // perfectly isotropic
  SO_NODE_ADD_FIELD(skipIthXform, (FALSE));

  isBuiltIn = TRUE;   // pertains to whether .iv files need extra [...] field
}

SoLinearFractal::~SoLinearFractal(){
}

void SoLinearFractal::GLRender(SoGLRenderAction *action){
//  doAction(action, "GLRender");
}

void SoLinearFractal::callback(SoCallbackAction *action){
  doAction(action, "SoLinearFractal callback");
}

void SoLinearFractal::doAction(SoAction *action, const char *s){
//  cerr << "hi from SoLinearFractal::doAction(" << s << ")\n";
}

int linearFractal_main(int argc, char *argv[]){
//int main(int argc, char *argv[]){
  SoLinearFractal *hi[50];
  for (int i=0; i<50; i++)
    hi[i] = new SoLinearFractal;
  for (int i=0; i<50; i++)
    hi[i]->ref();
  for (int i=0; i<50; i++)
    hi[i]->unref();
  return(0);
}

ostream& operator<< (ostream& os, const SoLinearFractal & m){
  os << "LinearFractal ["
     << " depth=";

  os << " depth="             << m.depth.getValue()
     << " xformsFile="        << m.xformsFile.getValue().getString()
     << " isectBVOnly="       << m.isectBVOnly.getValue()
     << " isectIntermediate=" << m.isectIntermediate.getValue()
     << " skipIthXform="      << m.skipIthXform.getValue()
     << " ]";
  
  return(os);
}
