////////////////////////////////////////////////////////////////////////////
//
//WW''''WW''''''WW''''''''''''''WWW''''''''WW''''''''WWW'''''''
//WWW  WWW                       WW        WW         WW       
//W WWW WW  WWW WW WWW      WWW  WW   WW   WWW   WWW  WW   WWWW
//W  W  WW W WW WW WW W    W WW  WW  WW W  WW W W WW  WW  WWW  
//W     WW W WW WW WW W    W WW  WW  WW W  WW W W WW  WW    WWW
//W     WW  WWW WW WW W     WWW WWWW  WW   WWW   WWW WWWW WWWW 
//                           WW                                
//'''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''
//                                                                           
// File: Main.h
// Description: Global includes, variables, and function definitions
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Main header for the Pane ray tracer.
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// Usage: pane [OPTION]... [IV FILE]
//
// Synopsis: pane is a ray tracer.
//           Input is a scene in the Open Inventor format (scene.iv)
//           Output is an image (scene.png).
//
// $Id: Main.h 1271 2011-07-31 01:40:30Z beason $
//
////////////////////////////////////////////////////////////////////////////

#ifndef MAINGLOBALS
#define MAINGLOBALS

#include <unistd.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <list>
#include <string>
#include <errno.h>
#include <vector>
#include <getopt.h>
#include <string.h>
#include <pthread.h>
#include <semaphore.h>
#include <regex.h>
#include <sstream>
#include <map>
//#include <mathimf.h>

// environ on OS X
#ifdef __darwin__
# include <crt_externs.h>
# define environ (*_NSGetEnviron())
#endif

#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoFile.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoInfo.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoLOD.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/nodes/SoText3.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoTriangleStripSet.h>
#include <Inventor/SbColor.h>
#include <Inventor/SbLinear.h>
#include <Inventor/SbViewportRegion.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoInteraction.h>
#include <Inventor/SoOutput.h>
#include <Inventor/SoPrimitiveVertex.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>

// memory leak detector
#ifdef DEBUG_MEM
#include "memory_debug.h"
#endif 

// random number generator
#include "RNG.h"

#ifdef SUNOS
#include <alloca.h>
#include <ieeefp.h>

inline int isinf(double x) { return !finite(x) && x==x; }

#ifdef PNG_SETJMP_SUPPORTED
#  define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
#else
#  define png_jmpbuf(png_ptr) \
   (LIBPNG_WAS_COMPILED_WITH__PNG_SETJMP_NOT_SUPPORTED)
#endif

char* strsep(char** str, const char* delims);

#endif /* SUNOS settings */

using namespace std;

extern int DEBUG_STATE;                         // defined in Main.cxx
extern int STOP_RENDERING;                      // defined in Main.cxx
extern bool IS_RENDERING;                       // defined in Main.cxx
extern SoSeparator *scnRoot;                    // defined in Main.cxx
extern RNG mydrand48;                           // defined in Main.cxx
extern const int mutexNum;                      // defined in Main.cxx
extern pthread_mutex_t mutexes[];               // defined in Main.cxx

namespace sigactionHandlerNS {
  extern int pixel[32][2];
  extern pthread_t threadIds[32];
}

// structure for threads maintaining state.
// was in Pane.h, moved to here
// so it can be used in Ray.h as well.
typedef struct thread_context_st {
  RNG mydrand48;  // random number generator
  int id;         // thread id
  unsigned long long rays;       // rays counter
  unsigned long long isects;     // intersections counter
  clock_t ticks;  // ticks for thread
  void *other;    // for passing miscellaneous data

  // these could be stored somewhere else i suppose
  int subPixSampId;
  int pixel[2];

  void setPixel(int x, int y) {
    pixel[0] = x;
    pixel[1] = y;
    sigactionHandlerNS::pixel[id][0] = x;
    sigactionHandlerNS::pixel[id][1] = y;
  }
  const int* getPixel() const { return pixel; }

  void registerThread() {
    sigactionHandlerNS::threadIds[id] = pthread_self();
  }
  
} thread_context;

extern thread_context defaultCtx;

enum { X, Y, Z };                               // defs for XYZ
enum { R, G, B, E };                            // defs for RGBExponent
enum { LOCK_STDERR, LOCK_RAND, LOCK_ICACHE, LOCK_ICREJECTS, LOCK_ENV }; // thread locks
enum { HL_OFF, HL_CENTER, HL_OFFCENTER };

#include "Misc.h"

#endif /* MAINGLOBALS */


