//WW''''WW'''''''WW''''''''''''WW''''''WWW''''''''''WWW''''''''''''''''''
//WWW  WWW       WW                     WW           WW                  
//W WWW WW  WWW WWWW  WW  W WW WW  WWW  WW      WWW  WW   WWW  WWWW  WWWW
//W  W  WW W WW  WW  WWWW WW   WW W WW  WW     WW W  WW  W WW WWW   WWW  
//W     WW W WW  WW  WW   WW   WW W WW  WW     WW    WW  W WW   WWW   WWW
//W     WW  WWW  WW   WWW WW   WW  WWW WWWW     WWW WWWW  WWW WWWW  WWWW 
//                                                                       
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include <stdlib.h>
#include <assert.h>
#include <Inventor/nodes/SoMaterial.h>
#include "Material.h"
#include "SoMatPlus.h"
#include "debug.h"
#include "Main.h"
#include "Misc.h"
#include "Ray.h"
#include "Primitive.h"

using namespace std;

static DIFFUSE_BRDF defDiffuseBRDF;
static SCHLICK_BRDF defSchlickBRDF;
static ASHIKHMIN_BRDF defAshikhminBRDF;

Material::Material(){
  init();
}

Material::Material(SoMaterial *mat, SoMatPlus *matPlus) {
  init(mat, matPlus);
}

void Material::init(SoMaterial *mat, SoMatPlus *matPlus) {
  static int newMats = 0;

  bool delMat = false;
  bool delMatPlus = false;
  if (!mat) {
    mat = new SoMaterial();
    mat->ref();
    delMat = true;
  }
  if (!matPlus) {
    matPlus = new SoMatPlus();
    matPlus->ref();
    delMatPlus = true;
  }
  
  ambient      = mat->ambientColor[0];
  diffuse      = mat->diffuseColor[0];

  static float saturation =
    getenv("PANE_SATURATION") ? atof(getenv("PANE_SATURATION")) : 1.0;
  if (saturation!=1.0f){
    float h, s, v;
    diffuse.getHSVValue(h, s, v);
    s *= saturation;
    diffuse.setHSVValue(h,s, v);
  }
  
  setShininess(mat->shininess[0]);
  setGlossiness(0);
  setSpecular(mat->specularColor[0]);
  setEmission(mat->emissiveColor[0]);
  transparency = mat->transparency[0];
  brdf = &defDiffuseBRDF;
  
  // mat plus
  model     = (ReflModel) matPlus->model.getValue();
  rough1    = matPlus->rough1.getValue();
  rough2    = matPlus->rough2.getValue();
  isotropy1 = matPlus->isotropy1.getValue();
  isotropy2 = matPlus->isotropy2.getValue();
  R0        = matPlus->R0.getValue();
  layers    = matPlus->layers.getValue();
  ior1      = matPlus->ior1.getValue();
  ior2      = matPlus->ior2.getValue();
  atten     = matPlus->atten.getValue();
  emisEcc   = matPlus->emisEcc.getValue();
  minDist   = matPlus->minDist.getValue();

  resolveParams();
  
  for (int j=0; j<3; j++)
    if (!finite(diffuse[j]))
      diffuse = SbColor(1.0, 0.0, 0.0);

  // temporary reflection capping
  const static bool reflCap = (bool)getenv("PANE_REFLECTION_CAP");
  if (reflCap) diffuse = BBox::vecMin(diffuse, SbVec3f(.9,.9,.9));

  const static bool reflScale = (bool)getenv("PANE_REFLECTION_SCALE");
  if (reflScale)
    diffuse *= .8;
  
  if (delMat) mat->unref();
  if (delMatPlus) matPlus->unref();
  
  //cerr << "Material():" << *this;
}

Material::~Material(){
}

void Material::resolveParams(){
  
  if (model == SCHLICK) {   // Override parameters if using Schlick

    brdf = &defSchlickBRDF;

    // Schlick model (single layer) only has one color
    // so just use diffuse.
    specular = diffuse;
    
    float g = 0, s = 0, d = 0;
    g = 4*rough1*(1-rough1);
    
    // compute g, s, d
    if (rough1<0.5){
      s = 1-g;
      d = 0;
    } else {
      s = 0;
      d = 1-g;
    }

#ifdef SCHLICK_CYLINDER_TEST    
    // for debugging only
    s = d = 0;
    g = 1;
#endif
    
    shininess = s;
    glossiness = g;
    diffuseness = d;
  } else if (model == ASHIKHMIN){

    brdf = &defAshikhminBRDF;

    // scale diffuse by (1-R_s) now
    vecScaleItself(diffuse, SbVec3f(1,1,1)-specular);

//     // compute splitting probabilities
//     float specAvg = vecAvg(specular);
//     float diffAvg = vecAvg(diffuse);
//     float total = specAvg + diffAvg;

//     // smooth
//     specAvg += .25*total;
//     diffAvg += .25*total;
//     total += .5*total;
    
//     shininess = 0;
//     glossiness = specAvg/total;  // dont forget to divide by these, because each
//     diffuseness = diffAvg/total; // should be 1!

    // Even weights seems to produce a lot less noise for
    // bunny on plane scene with:
    // bunny: Rs=.6  Rd=0  nu=nv=100
    // plane: Rs=.05 Rd=.5 nu=nv=40000
    shininess = 0;
    glossiness = .5;    // these are counteracted for in
    diffuseness = .5;   // shading so each is really 1.

    if (vecAvg(diffuse) <= 0){
      glossiness = 1;
      diffuseness = 0;
    }

    // CAN'T set glossiness to 0 if specular is 0 because of fresnel

  } else if (model == SEELIGER) {
    shininess = 0;
    glossiness = 0;
    diffuseness = 1;
  }
}

ostream& operator<< (ostream& os, const Material& m){
  os << "Material [" << "\n"
     << "  ambient      = " << m.ambient << "\n"
     << "  diffuse      = " << m.diffuse << "\n"
     << "  specular     = " << m.specular << "\n"
     << "  emission     = " << m.emission << "\n"
     << "  emissive     = " << (m.emissive?"yes":"no") << "\n"
     << "  shininess    = " << m.shininess << "\n"
     << "  transparency = " << m.transparency << "\n"
     << "  emisAvg      = " << m.emisAvg << "\n"
     << "  model        = " << m.model << "\n"
     << "  rough1       = " << m.rough1 << "\n"
     << "  rough2       = " << m.rough2 << "\n"
     << "  isotropy1    = " << m.isotropy1 << "\n"
     << "  isotropy2    = " << m.isotropy2 << "\n"
     << "  R0           = " << m.R0 << "\n"
     << "  layers       = " << m.layers << "\n"
     << "  ior1         = " << m.ior1 << "\n"
     << "  ior2         = " << m.ior2 << "\n"
     << "  atten        = " << m.atten << "\n"
     << "  emisEcc      = " << m.emisEcc << "\n"
     << "  minDist      = " << m.minDist << "\n"
     << "]\n";
  
  return os;
}

// z (or "w") point up
void BRDF::GetSampleDir(double eps1, double eps2, double eps3,
                        const SbVec3f &in, const SbVec3f &nl,
                        const SbVec3f &ngl, const Ray &r,
                        SbVec3f &out, double &brdf, double &pdf,
                        double &brdfOverPdf, double &brdfOverPdfCosT,
                        SbVec3f &color, bool internalCall){

  if (internalCall) {
    GIVE_ERROR("recursion");
    return;
  }

  SbVec3f T;
  PrimDerivs pDer;
  if (r.prim && r.prim->getDerivs(r.bc, pDer)){
    T = pDer.dPdU;
  } else {
    T = in.cross(nl);
  }

  // Rotate (T,TxUp,Up) to (X,Y,Z)
  ONB onb;
  onb.constructFromUW(T, ngl);
  SbVec3f inA = onb.toONBSpace(in);
  SbVec3f outA;
  color = SbColor(1,1,1);
  GetSampleDirCanonicalBasis(eps1, eps2, eps3, r, inA, outA, brdf, pdf,
                             brdfOverPdf, brdfOverPdfCosT, color, true);
  out = onb.toWorldSpace(outA);
  if (out.dot(ngl)<0)
    brdf = pdf = brdfOverPdf = brdfOverPdfCosT = 0;
}

// Canonical basis is z axis being up, and x axis being along dp/du,
// and y axis along dp/dv.
void
BRDF::GetSampleDirCanonicalBasis(double eps1, double eps2, double eps3,
                                 const Ray &r, const SbVec3f &in, SbVec3f &out,
                                 double &brdf, double &pdf,
                                 double &brdfOverPdf, double &brdfOverPdfCosT,
                                 SbVec3f &color, bool internalCall) {
    
  if (internalCall) {
    GIVE_ERROR("recursion");
    return;
  }

  SbVec3f n(0,0,1);
  color = SbColor(1,1,1);
  GetSampleDir(eps1, eps2, eps3, in, n, n, r, out, brdf, pdf,
               brdfOverPdf, brdfOverPdfCosT, color, true);
}

SbVec3f BRDF::GetBRDFCanonicalBasis(const SbVec3f &in, const SbVec3f &out,
                                    const Ray &r, SbVec3f &brdfCosT, float &pdf){
  SbVec3f n(0,0,1);
  return GetBRDF(in, out, n, n, r, brdfCosT, pdf);
}

void DIFFUSE_BRDF::GetSampleDirCanonicalBasis(double eps1, double eps2, double eps3,
                                              const Ray &r, const SbVec3f &in, SbVec3f &out,
                                              double &brdf, double &pdf,
                                              double &brdfOverPdf, double &brdfOverPdfCosT,
                                              SbVec3f &color, bool internalCall) {

  // cos(theta) probability distribution
  double theta = acos(sqrt( eps1 ));
  double phi = (2.0 * M_PI) * eps2;
    
  out.setValue(sin(theta)*cos(phi),
               sin(theta)*sin(phi),
               cos(theta));

  double cosT = cos(theta);
  
  Material *mtl = r.prim->getMat(r);
  float value = mtl->diffuseness;
    
  brdf            = M_1_PI              * value; // 1/pi
  pdf             = cosT*M_1_PI;                 // cos(theta)/pi
  brdfOverPdf     = (cosT ? 1/cosT : 0) * value; //
  brdfOverPdfCosT = 1                   * value;
  color = r.diffCol;
}

SbVec3f DIFFUSE_BRDF::GetBRDF(const SbVec3f &in, const SbVec3f &out,
                              const SbVec3f &nl, const SbVec3f &ngl,
                              const Ray &r, SbVec3f &brdfCosT, float &pdf){

  float dot1 = out.dot(nl);
  if (dot1<0 || out.dot(ngl)<0){
    brdfCosT = 0;
    pdf = 0;
    return SbVec3f(0,0,0);
  }
  
  Material *mtl = r.prim->getMat(r);
  SbVec3f brdf = r.diffCol*(mtl->diffuseness*M_1_PI);                     // cached
//  SbVec3f brdf = r.prim->getIntDiffColor(r)*(mtl->diffuseness*M_1_PI);  // uncached
  brdfCosT = brdf * dot1;
  pdf = dot1*M_1_PI;
  return brdf;
}

void SEELIGER_BRDF::GetSampleDirCanonicalBasis(double eps1, double eps2, double eps3,
                                               const Ray &r, const SbVec3f &in, SbVec3f &out,
                                               double &brdf, double &pdf,
                                               double &brdfOverPdf, double &brdfOverPdfCosT,
                                               SbVec3f &color, bool internalCall) {

  defDiffuseBRDF.GetSampleDirCanonicalBasis(eps1, eps2, eps3,
                                            r, in, out,
                                            brdf, pdf, brdfOverPdf, brdfOverPdfCosT,
                                            color, internalCall);

  SbVec3f n(0,0,1);
  float dot1 = -(in.dot(n));
  float dot2 = out.dot(n);
  
  float value = dot1/(dot1+dot2) * M_PI; // pi for 1/pdf
 
  brdf            *= value;
  brdfOverPdf     *= value;
  brdfOverPdfCosT *= value;
}

SbVec3f SEELIGER_BRDF::GetBRDF(const SbVec3f &in, const SbVec3f &out,
                               const SbVec3f &nl, const SbVec3f &ngl,
                               const Ray &r, SbVec3f &brdfCosT, float &pdf){

  float dot1 = -(in.dot(nl));
  float dot2 = out.dot(nl);
  if (dot1<=0 || dot2<=0 || out.dot(ngl)<0){
    brdfCosT = 0;
    pdf = 0;
    return 0;
  }
  
  Material *mtl = r.prim->getMat(r);
  SbVec3f brdf = mtl->diffuse*(mtl->diffuseness*M_1_PI);
  brdfCosT =  brdf * dot2;
  pdf = dot1*M_1_PI;
  return brdf;
}


// From "An Inexpsensive BRDF Model for Physically-based Rendering"
// by Christophe Schlick
void
SCHLICK_BRDF::GetSampleDir(double eps1, double eps2, double eps3,
                           const SbVec3f &in, const SbVec3f &nl,
                           const SbVec3f &ngl, const Ray &r,
                           SbVec3f &out, double &brdf, double &pdf,
                           double &brdfOverPdf, double &brdfOverPdfCosT,
                           SbVec3f &color, bool internalCall) {

  Material *mtl = r.prim->getMat(r);
  
  float diffComp = mtl->getDiffuseness();
  float glossyComp = mtl->getGlossiness();
  float total = diffComp + glossyComp;

  if (!total) {
    brdf = pdf = brdfOverPdf = brdfOverPdfCosT = 0;
    color = SbColor(1,1,1);
    return;
  }
  
  float diffProb = diffComp/total;
  float glossyProb = glossyComp/total;
  
  if (eps3<=diffProb && diffProb){

    defDiffuseBRDF.GetSampleDir(eps1, eps2, 0, in, nl, ngl, r, out, brdf, pdf,
                                brdfOverPdf, brdfOverPdfCosT, color);

    float value = 1/diffProb;  // diffComp multiplied in DIFFUSE_BRDF::GetSampleDir

    brdf            *= value;
    brdfOverPdf     *= value;
    brdfOverPdfCosT *= value;

    return;
  }

  
            // choose new direction according to cos(theta)
            //ONB onb(nl);

//             SbVec3f rdir = dir-2.0*n.dot(dir)*n;

//             // grazing angle problem
//             // don't let reflected ray go beneath surface
//             if (rdir.dot(ngl)<0.0){
//               rdir += (-rdir.dot(ngl)+0.0015) * ngl;
//               rdir.normalize();
//             }

            // calculate schlick constants
            // from "An Inexpsensive BRDF Model for Physically-based Rendering"
            // by Christophe Schlick, page 2.
            SbVec3f V = -in;  // incoming dir
            SbVec3f N = nl;       // normal
            SbVec3f Vp;// = rdir;     // reflection dir (LIGHT DIR)
            SbVec3f H;// = Vp + V;   // halfway vector
            //H.normalize();

            double rough = mtl->getRough1();
             
//             if (r.ctx->mydrand48()<0.5)
//               rough = 1;
          
            double theta = acos(sqrt(eps1/(rough - eps1*rough + eps1)));
            double p = mtl->getIsotropy1();
            bool otherDir = false;
            if (p<0){
              otherDir = true;
              p = -p;
            }
            
            // calculate phi, the angle above u (or x) axis
            double phi;
            if (eps2<.25){           // [0, .25)
              eps2 = eps2*4;
              phi = M_PI_2*sqrt(p*p*eps2*eps2/(1-eps2*eps2+p*p*eps2*eps2));
            } else if (eps2<.5) {
              eps2 = 1-4*(0.5-eps2);
              phi = M_PI_2*sqrt(p*p*eps2*eps2/(1-eps2*eps2+p*p*eps2*eps2));
              phi = M_PI-phi;
            } else if (eps2<.75) {
              eps2 = (eps2-.5)*4;
              phi = M_PI_2*sqrt(p*p*eps2*eps2/(1-eps2*eps2+p*p*eps2*eps2));
              phi = -phi;
            } else {
              eps2 = 1-4*(1-eps2);
              phi = M_PI_2*sqrt(p*p*eps2*eps2/(1-eps2*eps2+p*p*eps2*eps2));
              phi = M_PI+phi;
            }

            // The paper (formula 29) and above generates phi's mostly
            // along 0 or pi (the scratch direction), 
            // when it should be mostly along pi/2 or -pi/2 (ortho-scratch),
            // so I fix that here.
            phi += M_PI_2;

//             // dont worry about isotropy for now
//             if (r.ctx->mydrand48()<0.5)
//               phi = eps2*2*M_PI;
            
//             theta = acos(sqrt( eps1 ));
//             phi = 2.0 * M_PI * eps2;

            SbVec3f T;
            PrimDerivs pDer;
            if (r.prim->getDerivs(r.bc, pDer)){
              T = otherDir ? pDer.dPdV : pDer.dPdU;
            } else {
              //T = in.cross(nl);
              T = (fabs(nl[0])<.99) ? SbVec3f(1,0,0) : SbVec3f(0,1,0);
              T -= T.dot(nl)*nl;
            }
            T.normalize();

            //  cerr << "phi=" << phi << " T=" << T << "\n";

            ONB onbH;
            onbH.constructFromUW(T, nl);
          
            SbVec3f neww;
            neww.setValue(sin(theta)*cos(phi),
                          sin(theta)*sin(phi),
                          cos(theta));
    
            SbVec3f newH = onbH.toWorldSpace(neww);
            newH.normalize();
            out = in-2.0*newH.dot(in)*newH;
            //bounce.dir = newH;  // diffuse brdf
            out.normalize();
          
            Vp = out;
            H = newH;
            double vp = Vp.dot(N);
            double t = H.dot(N);

            // make sure vector goes up
            if (vp<=0 || out.dot(ngl)<=0 || t<=0){
              brdf = pdf = brdfOverPdf = brdfOverPdfCosT = 0;
              color = SbColor(1,1,1);
              return;
            }
            
            //H.normalize();
          
            SbVec3f Hb = H - N.dot(H)*N;
            double u = H.dot(V);
            double v = V.dot(N);
            double w = T.dot(Hb);
//           double alpha = acos(t);
//           double beta = acos(u);
//           double theta = acos(v);
//           double thetap = acos(vp);
//           double phi = acos(w);
          
            double Zt = rough/( (1 + rough*t*t - t*t) * (1 + rough*t*t - t*t) ) ;
            double Aw = p ? sqrt(p/((p*p) - (p*p)*(w*w) + w*w)) : 0;
            double Gv = v/(rough - rough*v + v);
            double Gvp = vp/(rough - rough*vp + vp);
//             double Dtuvpw = Gv*Gvp/(4*M_PI*v*vp)*Zt*Aw + (1-Gv*Gvp)*M_1_PI*Aw;

//             double pdf = 1./(4*M_PI*v*vp)*Zt*Aw;

//            double brdfPdf = Gv*Gvp + (1-Gv*Gvp)*M_1_PI*Aw;
//            double brdfPdf = Gv*Gvp + (1-Gv*Gvp)*Zt*4*v*vp;    // was working okay

            // testing
//            double Dtuvpw = Gv*Gvp/(4*M_PI*v*vp)*Zt*Aw + (1-Gv*Gvp)/(4*M_PI*v*vp);
//             double pdf = 1./(4*M_PI*v*vp)*Zt*Aw;
//            double brdfPdf = Gv*Gvp + (1-Gv*Gvp)/(Zt*Aw);
//            double brdfPdf = H.dot(V); // eqn 25
//            double brdfPdf = (Gv*Gvp + (1-Gv*Gvp)/(Zt*Aw))*H.dot(V)/max2(H.dot(N),.00001f);
//            double brdfPdf = H.dot(V)/H.dot(N);//*Gv*Gvp;
//            double brdfPdf = 4*H.dot(V)/H.dot(N);//Zt*Aw;//*H.dot(N);


  double Dtuvpw = 1/(4*M_PI)*(Gv*Gvp*Zt*Aw + (1-Gv*Gvp));
            
//            double brdfPdf = u/t;//H.dot(V); ///H.dot(N);//Zt*Aw;//*H.dot(N);
            
//            double brdfPdf = 1/(4*M_PI)*(Gv_*Gvp_) + (1-Gv*Gvp)/Zt;

//             if (bounce.dir.dot(ngl)<0){
//               bounce.dir -= 2*bounce.dir.dot(ngl)*ngl;
//             }
            
            
            // grazing angle problem
            // ...treat as black.

            float value = glossyComp/glossyProb;
            
            brdf            = Dtuvpw            * value;
            pdf             = 1/(4*M_PI)*Zt*Aw  * (t/u);
            brdfOverPdf     = Dtuvpw/pdf        * value;
            brdfOverPdfCosT = Dtuvpw*vp/pdf     * value;
            
            SbColor specCol = mtl->getSpecular();
            color = fresnel(specCol, u);
            
            return;
}

// This doesn't include ideal specular component, that's handled separately.
// It does include the diffuse component.
SbVec3f SCHLICK_BRDF::GetBRDF(const SbVec3f &in, const SbVec3f &out,
                              const SbVec3f &nl, const SbVec3f &ngl,
                              const Ray &r, SbVec3f &brdfCosT, float &pdf){

  Material *mtl = r.prim->getMat(r);
    
  // calculate schlick constants
  // from "An Inexpsensive BRDF Model for Physically-based Rendering"
  // by Christophe Schlick, page 2.
  SbVec3f V = -in;      // outcoming direction of light / incoming dir
  SbVec3f Vp = out;     // incoming direction of light
  SbVec3f N = nl;       // normal
  SbVec3f H = Vp + V;   // halfway vector
  H.normalize();

  double rough = mtl->getRough1();
          
  double p = mtl->getIsotropy1();
  bool otherDir = false;
  if (p<0){
    otherDir = true;
    p = -p;
  }
            
  // The paper (formula 29) and above generates phi's mostly
  // along 0 or pi (the scratch direction), 
  // when it should be mostly along pi/2 or -pi/2 (ortho-scratch),
  // so I fix that here.
  //phi += M_PI_2;
  //cerr << "phi=" << phi << " T=" << T << "\n";

  SbVec3f T;
  PrimDerivs pDer;
  if (r.prim->getDerivs(r.bc, pDer)){
    T = otherDir ? pDer.dPdV : pDer.dPdU;
  } else {
    //T = in.cross(nl);
    T = (fabs(nl[0])<.99) ? SbVec3f(1,0,0) : SbVec3f(0,1,0);
    T -= T.dot(nl)*nl;
  }
  T.normalize();

  SbVec3f Hb = H - N.dot(H)*N;
    
  double t = H.dot(N);
  double u = H.dot(V);
  double v = V.dot(N);
  double vp = Vp.dot(N);
  double w = T.dot(Hb);
  // double alpha = acos(t);
  // double beta = acos(u);
  // double theta = acos(v);
  // double thetap = acos(vp);
  // double phi = acos(w);

  if (v<=0 || vp<=0){
    brdfCosT.setValue(0,0,0);
    pdf = 0;
    return SbVec3f(0,0,0);
  }
          
  double Zt = rough/( (1 + rough*t*t - t*t) * (1 + rough*t*t - t*t) ) ;
  double Aw = p ? sqrt(p/((p*p) - (p*p)*(w*w) + w*w)) : 0;
  double Gv = v/(rough - rough*v + v);
  double Gvp = vp/(rough - rough*vp + vp);

  // double Dtuvpw = /*1/(4*M_PI)* */ Zt*Aw;  // schlick cylinder test
  double Dtuvpw = 1/(4*M_PI)*(Gv*Gvp*Zt*Aw + (1-Gv*Gvp));
  

  SbColor diffColor = r.diffCol; //mtl->diffuse;
  SbColor specColFresnel = fresnel(mtl->specular, u);

  SbVec3f D = SbVec3f(0,0,0);
  
  brdfCosT =
    diffColor*(mtl->diffuseness*M_1_PI)*vp +
    specColFresnel*(mtl->glossiness*Dtuvpw)*vp;

  pdf =
    mtl->diffuseness * vp*M_1_PI +
    mtl->glossiness  * 1/(4*M_PI)*Zt*Aw;

  // grazing angle problem
  // ...treat as black.
  return D;
}

SbVec3f ASHIKHMIN_BRDF::GetBRDF(const SbVec3f &in, const SbVec3f &out,
                                const SbVec3f &nl, const SbVec3f &ngl,
                                const Ray &r, SbVec3f &brdfCosT, float &pdf){

  SbVec3f zero(0,0,0);
  brdfCosT = zero;
  pdf = 0;
  
  // diffuse part

  // "An Ansiotropic Phong Light Reflection Model"
  // by Ashikhmin and Shirley, page 7, Equation 14
  Material *mtl = r.prim->getMat(r);
  SbColor diffColor = r.diffCol; //mtl->diffuse;
  float dot1 = -(in.dot(nl));
  float dot2 = out.dot(nl);

  float val1 = 1-dot1*0.5;
  float val2 = 1-dot2*0.5;
  float diffVal = 28./(23.*M_PI) *
    (1 - val1*val1*val1*val1*val1) *
    (1 - val2*val2*val2*val2*val2);

  if (dot1<0 || dot2<0 || ngl.dot(out)<0)
    diffVal = 0;
  
  // glossy part

  double nu = mtl->getRough1();
  double nv = mtl->getRough2();

  PrimDerivs pDer;
  SbVec3f u, v;
  if (r.prim->getDerivs(r.bc, pDer)){
    u = pDer.dPdU;

    // check for zero derivative
    if (u.dot(u)<0.0001){
      u = (fabs(nl[0])<.99) ? SbVec3f(1,0,0) : SbVec3f(0,1,0);
      u -= u.dot(nl)*nl;
      u.normalize();
    }

    u.normalize();
    
    if (DEBUG_STATE)
      Ray(r.x,u,r.ctx).show(1, SbColor(1,0,0));
              
  } else {
    u = (fabs(nl[0])<.99) ? SbVec3f(1,0,0) : SbVec3f(0,1,0);
    u -= u.dot(nl)*nl;
    u.normalize();
  }
  v = nl.cross(u);
  v.normalize();

  SbVec3f k1 = out;      // to light
  SbVec3f k2 = -in;      // to viewer
  SbVec3f H = k1 + k2;   // halfway vector
  H.normalize();
            
  float ndotk1 = nl.dot(k1);
  float ndotk2 = nl.dot(k2);

  // make sure vector goes up
  if (out.dot(ngl)<=0 || ndotk1<=0 || ndotk2<=0)
    return zero;
            
  // calculate brdf
  float hdotn = H.dot(nl);
  float hdotu = H.dot(u);
  float hdotv = H.dot(v);
  float maxnk = max(ndotk1, ndotk2);
  float hdotk = H.dot(k2);    // k1 or k2, its the same

  if (maxnk<1e-10 || hdotk<1e-10)
    return zero;
                
  float brdf, brdfcos, exponent, brdfhdotk;
  float brdfPart1 = sqrtf((nu+1)*(nv+1))*(1/(M_PI*8));
  if (hdotn<.99){
    exponent = (nu*hdotu*hdotu + nv*hdotv*hdotv) / (1 - hdotn*hdotn);
  } else {
    SbVec3f Hb = H-H.dot(nl)*nl;
    Hb.normalize();
    double cos2phi = u.dot(Hb)*u.dot(Hb);
    exponent = nu*cos2phi+nv*(1-cos2phi);
  }
  brdfPart1 *= powf(hdotn, exponent);
  brdfPart1 /= hdotk;
  brdf  = brdfPart1 / maxnk;
  brdfcos = brdfPart1 * (dot2/maxnk);

  SbColor specCol = mtl->specular;
  SbVec3f Fkh = fresnel(specCol, hdotk); // matches explicit version
  
  SbVec3f D = diffColor*diffVal      + brdf*Fkh;
  brdfCosT  = diffColor*diffVal*dot2 + brdfcos*Fkh;
  pdf =
    mtl->diffuseness * dot2*M_1_PI +
    mtl->glossiness  * brdfPart1;

  // Just ignore mtl->diffuseness and mtl->glossiness

  // grazing angle problem
  // ...treat as black.
  return D;
}

void
ASHIKHMIN_BRDF::GetSampleDir(double eps1, double eps2, double eps3,
                             const SbVec3f &in, const SbVec3f &nl,
                             const SbVec3f &ngl, const Ray &r,
                             SbVec3f &out, double &brdf, double &pdf,
                             double &brdfOverPdf, double &brdfOverPdfCosT,
                             SbVec3f &color, bool internalCall) {

  Material *mtl = r.prim->getMat(r);
  SbColor diffColor = r.diffCol; //mtl->diffuse;

  float diffComp = mtl->getDiffuseness();
  float glossyComp = mtl->getGlossiness();
  if (eps3<diffComp){

    defDiffuseBRDF.GetSampleDir(eps1, eps2, 0, in, nl, ngl, r, out, brdf, pdf,
                                brdfOverPdf, brdfOverPdfCosT, color);

    // "An Ansiotropic Phong Light Reflection Model"
    // by Ashikhmin and Shirley, page 7, Equation 14
    float dot1 = -(in.dot(nl));
    float dot2 = out.dot(nl);
    float val1 = 1-dot1*0.5;
    float val2 = 1-dot2*0.5;
    float value = 28./23. *
      (1 - val1*val1*val1*val1*val1) *
      (1 - val2*val2*val2*val2*val2);

    // 1/pi is cancelled by 1/pdf numerator (pi).
    // *= (SbVec3f(1,1,1)-specCol) is done in Material::init

    value /= diffComp*diffComp; // first to compensate for russian roulette,
                                // second to cancel multiplication in DIFFUSE_BRDF::GetSampleDir.

    brdf            *= value;
    brdfOverPdf     *= value;
    brdfOverPdfCosT *= value;

    return;
  }

  
  double nu = mtl->getRough1();
  double nv = mtl->getRough2();

  PrimDerivs pDer;
  SbVec3f u, v;
  if (r.prim->getDerivs(r.bc, pDer)){
    u = pDer.dPdU;

    // check for zero derivative
    if (u.dot(u)<0.0001) {
      u = (fabs(nl[0])<.99) ? SbVec3f(1,0,0) : SbVec3f(0,1,0);
      u -= u.dot(nl)*nl;
      u.normalize();
    }              
              
    if (DEBUG_STATE)
      Ray(r.x,u,r.ctx).show(1, SbColor(1,0,0));
              
  } else {
    u = (fabs(nl[0])<.99) ? SbVec3f(1,0,0) : SbVec3f(0,1,0);
    u -= u.dot(nl)*nl;
    u.normalize();
  }
  v = nl.cross(u);
  v.normalize();
            
  ONB onbH;
  onbH.constructFromUW(u, nl);
            
  // calculate phi, the angle above u (or x) axis
  float phi;
  if (eps2<.25){           // [0, .25)
    eps2 = eps2*4;
    phi = atanf(sqrtf((nu+1)/(nv+1))*tan((0.5*M_PI)*eps2));
  } else if (eps2<.5) {
    eps2 = 1-4*(0.5-eps2);
    phi = atanf(sqrtf((nu+1)/(nv+1))*tan((0.5*M_PI)*eps2));
    phi = M_PI-phi;
  } else if (eps2<.75) {
    eps2 = (eps2-.5)*4;
    phi = atanf(sqrtf((nu+1)/(nv+1))*tan((0.5*M_PI)*eps2));
    phi = -phi;
  } else {
    eps2 = 1-4*(1-eps2);
    phi = atanf(sqrtf((nu+1)/(nv+1))*tan((0.5*M_PI)*eps2));
    phi = M_PI+phi;
  }

  float cosPhi = cosf(phi);
  float sinPhi = sinf(phi);
  float cosTheta = powf(1-eps1, 
                        1/(nu*cosPhi*cosPhi + nv*sinPhi*sinPhi + 1));
  float sinTheta = sqrt(1-cosTheta*cosTheta);

  SbVec3f neww;
  neww.setValue(sinTheta*cosPhi,
                sinTheta*sinPhi,
                cosTheta);
    
  SbVec3f H = onbH.toWorldSpace(neww);
  out = in-2.0*H.dot(in)*H;
  out.normalize();

  SbVec3f k1 = out;   // to light
  SbVec3f k2 = -in;       // to viewer
            
  float ndotk1 = nl.dot(k1);
  float ndotk2 = nl.dot(k2);

  // make sure vector goes up
  if (out.dot(ngl)<=0 || ndotk1<=0 || ndotk2<=0){
    brdf = pdf = brdfOverPdf = brdfOverPdfCosT = 0;
    color = SbColor(1,1,1);
    return;
  }

  // calculate brdf
  float hdotn = H.dot(nl);
  float hdotu = H.dot(u);
  float hdotv = H.dot(v);
  float maxnk = MAX2(ndotk1, ndotk2);
  float hdotk = H.dot(k2);    // k1 or k2, its the same
  
  float pk2, exponent;
  brdf = sqrtf((nu+1)*(nv+1))*(1/(M_PI*8));
  if (hdotn<.99){
    exponent = (nu*hdotu*hdotu + nv*hdotv*hdotv) / (1 - hdotn*hdotn);
  } else {
    SbVec3f Hb = H-H.dot(nl)*nl;
    Hb.normalize();
    double cos2phi = u.dot(Hb)*u.dot(Hb);
    exponent = nu*cos2phi+nv*(1-cos2phi);
  }
  brdf *= powf(hdotn, exponent);
  pk2  = brdf/hdotk;
  brdf = brdf/(hdotk * maxnk);

  SbColor specCol = mtl->specular;
  SbVec3f Fkh = fresnel(specCol, hdotk); // matches explicit version
  
//   SbVec3f D = diffColor*diffVal      + brdf*Fkh;
//   brdfCosT  = diffColor*diffVal*dot2 + brdfcos*Fkh;
  pdf = pk2;
//    mtl->diffuseness * dot2*M_1_PI +
//    mtl->glossiness  * brdfcos;
  
//               // calculate brdf
//               float hdotn = newH.dot(nl);
//               float nhRaised = powf(hdotn, nu*cosPhi*cosPhi + nv*sinPhi*sinPhi);
//               float pdf_h = sqrtf((nu+1)*(nv+1))*(0.5*M_1_PI) * nhRaised;
//               float pdf_k1 = 4*k2.dot(newH)/pdf_h;  // equation 18 (k1 and k2 reversed)

//               // calculate brdf
//               float hdotu = newH.dot(onbH.getU());
//               float hdotv = newH.dot(onbH.getV());
//               float ndotk1 = nl.dot(k1);
//               float ndotk2 = nl.dot(k2);
//              float maxnk = MAX2(ndotk1, ndotk2);
//              float hdotk = newH.dot(k2);    // k1 or k2, its the same
                
//               SbVec3f Fkh = fresnel(specCol, hdotk);
              
//               float brdf = sqrt((nu+1)*(nv+1))*(1/(M_PI*8));
//               //float exponent = (nu*hdotu*hdotu + nv*hdotv*hdotv) / (1 - hdotn*hdotn);
//               //brdf *= powf(hdotn, exponent) / (hdotk * maxnk);
//               brdf *= nhRaised / (hdotk * maxnk);
              
//               SbVec3f brdfVec = (2*brdf) * Fkh; // mult by 2 since half the
//                                                 // time we evaluate diffuse

              // divide by mtl->getGlossiness() since that is just RR prob
              // of choosing glossy instead of diffuse

              float value = 1/glossyComp;

              //brdfOverPdf = hdotk; // brdf/pdf = brdf/pk2 = brdf/(brdf/hdotk)
              brdf        =               brdf * value;
              brdfOverPdf =          1 / maxnk * value;
              brdfOverPdfCosT = ndotk1 / maxnk * value;           // ok
              
              color = Fkh;
}
