//WW''''WW'''''''WW''''''''''''WW''''''WWW''''''''''WWW''''''''''''''''''
//WWW  WWW       WW                     WW           WW                  
//W WWW WW  WWW WWWW  WW  W WW WW  WWW  WW      WWW  WW   WWW  WWWW  WWWW
//W  W  WW W WW  WW  WWWW WW   WW W WW  WW     WW W  WW  W WW WWW   WWW  
//W     WW W WW  WW  WW   WW   WW W WW  WW     WW    WW  W WW   WWW   WWW
//W     WW  WWW  WW   WWW WW   WW  WWW WWWW     WWW WWWW  WWW WWWW  WWWW 
//                                                                       
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

// Contains copy of SoMaterial info plus pointer to
// SoMatPlus node. Used by pane to contain all material
// properties. SoMatPlus is a custom node for storing
// additional material descriptions in an inventor (.iv)
// file that are not already contained within the SoMaterial
// node.

#ifndef MATERIALCLASS
#define MATERIALCLASS

#include <iostream>
#include <Inventor/SbColor.h>

using namespace std;

class SoMaterial;
class SoMatPlus;
class BRDF;

class Material { 
  
  private:

  void init(SoMaterial *mat=NULL, SoMatPlus *matPlus=NULL);
  BRDF *brdf;
  
  public:
  
  SbColor ambient;
  SbColor diffuse;
  SbColor specular;
  SbColor emission;
  bool emissive;

  // f_r = s*f_s + g*f_g + d*f_d        s+g+d = 1
  float shininess;         // proportion of 1 that brdf is shiny component (s)
  float glossiness;        // proportion of 1 that brdf is glossy (g)
  float diffuseness;       // proportion of 1 that brdf is diffuse (d)
  float transparency;      // proportion of 1 that surface is transparent
  float emisAvg;           // average of R,G,B emission values
  
  // mat plus
  enum ReflModel {
    DEFAULT,
    SCHLICK,
    DATA,
    LIBRARY,
    BSSRDF,
    HIERSUBSURF,
    ASHIKHMIN,
    SEELIGER
  };
  ReflModel model;       // model (schlick, data, library, ?)

  // Schlick
  // F0_r,g,b for top surface from Material diffuseColor
  // F0_r,g,b fro bottom surface from Material specularColor
  float rough1, rough2;  // roughness  (0=mirror, 1=diffuse (default))
                         // for double layers of heterogenous material
  float isotropy1;       // anisotropy for double layers of heterogenous material
  float isotropy2;       // (0=anisotropic, 1=isotropic (symmetrical, default))
  float R0;              // reflection at normal incidence
                         // Default: 1
                         // -1 means auto-calculated based on index of refractions

  int layers;            // 1 (default) or 2
  

  float ior1;            // OUTSIDE index of refraction Default: 1.0 (air)
  float ior2;            // INSIDE index of refraction Default: 1.5 (glass)
  SbColor atten;         // attenuation (fraction) at 1m when going through inside
                         // Default (0.98, 0.99, 0.98)
                         // from specularColor ?


  
  float emisEcc;         // Light eccentricity (default 1 = diffuse)
  float minDist;         // minimum distance for intersection (default 0.0001)
  

  Material();
  ~Material();
  Material(SoMaterial *mat, SoMatPlus *matPlus);

  void resolveParams();
  
  void setShininess(float shininess_){
    shininess = shininess_;
    if (shininess<0.00001)
      shininess = 0;
    diffuseness = 1-shininess;
  }
  void setGlossiness(float glossiness_){ glossiness = glossiness_; }
  void setSpecular(const SbColor &spec){ specular = spec; }
  void setDiffuse(const SbColor &diff){ diffuse = diff; }
  void setEmission(const SbColor &e){
    emission = e;
    emisAvg = (emission[0]+emission[1]+emission[2])/3.0;
    emissive = emisAvg>0;
  }
  bool isShiny() const { return shininess; }
  bool isEmissive() const { return emissive; }
  bool isGlossy() const { return glossiness; }
  bool isDiffuse() const { return diffuseness; }
  bool isAtten() const {
    return atten[0] != 1 || atten[1] != 1 || atten[2] !=1;
  }

  float    getTransparency()         const { return transparency; }
  const SbColor& getSpecular()       const { return specular;     }
  float    getEmisAvg()              const { return emisAvg;  }
  SbColor  getEmission()             const { return emission; }
  float    getIndexOfRefractionTop() const { return ior1;  }
  float    getIndexOfRefractionBot() const { return ior2;  }
  const SbColor& getAttenuation()    const { return atten; }
  float getShininess() const { return shininess; }
  float getGlossiness() const { return glossiness; }
  float getDiffuseness() const { return diffuseness; }
  float getRough1() const { return rough1; }
  float getRough2() const { return rough2; }
  float getIsotropy1() const { return isotropy1; }
  ReflModel getModel() const { return model; }
  BRDF* getBRDF() { return brdf; }
  const BRDF* getBRDF() const { return brdf; }
  
  friend ostream& operator<< (ostream& os, const Material& m);
};

class Ray;

class BRDF {
  
public:
  
  // "in" is the viewing direction, points toward point
  // "out" is the reflected direction (where the light comes from)
  // This is kind of backwards physics-wise but it shouldn't matter
  // since the BRDF should be symmetric.


  // Materials can implement either. The default version of these
  // functions calls the other version, so if one is missing it will still work.
  
  // z (or "w") point up
  virtual void GetSampleDir(double eps1, double eps2, double eps3,
                            const SbVec3f &in, const SbVec3f &nl,
                            const SbVec3f &ngl, const Ray &r,
                            SbVec3f &out, double &brdf, double &pdf,
                            double &brdfOverPdf, double &brdfOverPdfCosT,
                            SbVec3f &color, bool internalCall = false);
  
  // Canonical basis is z axis being up, and x axis being along dp/du,
  // and y axis along dp/dv.
  virtual void
  GetSampleDirCanonicalBasis(double eps1, double eps2, double eps3,
                             const Ray &r, const SbVec3f &in, SbVec3f &out,
                             double &brdf, double &pdf,
                             double &brdfOverPdf, double &brdfOverPdfCosT,
                             SbVec3f &color, bool internalCall = false);
  
  virtual SbVec3f GetBRDFCanonicalBasis(const SbVec3f &in, const SbVec3f &out,
                                        const Ray &r, SbVec3f &brdfCosT, float &pdf);
  
  virtual SbVec3f GetBRDF(const SbVec3f &in, const SbVec3f &out,
                          const SbVec3f &nl, const SbVec3f &ngl,
                          const Ray &r, SbVec3f &brdfCosT, float &pdf) = 0;
  
  virtual bool IsDirac() const = 0;
};

class DIFFUSE_BRDF : public BRDF {
  
public:
  virtual bool IsDirac() const { return false; }
  virtual void GetSampleDirCanonicalBasis(double eps1, double eps2, double eps3,
                                          const Ray &r, const SbVec3f &in, SbVec3f &out,
                                          double &brdf, double &pdf,
                                          double &brdfOverPdf, double &brdfOverPdfCosT,
                                          SbVec3f &color, bool internalCall = false);  
  virtual SbVec3f GetBRDF(const SbVec3f &in, const SbVec3f &out,
                          const SbVec3f &nl, const SbVec3f &ngl,
                          const Ray &r, SbVec3f &brdfCosT, float &pdf);
};

class SEELIGER_BRDF : public BRDF {
  
public:
  virtual void GetSampleDirCanonicalBasis(double eps1, double eps2, double eps3,
                                          const Ray &r, const SbVec3f &in, SbVec3f &out,
                                          double &brdf, double &pdf,
                                          double &brdfOverPdf, double &brdfOverPdfCosT,
                                          SbVec3f &color, bool internalCall = false);  
  virtual SbVec3f GetBRDF(const SbVec3f &in, const SbVec3f &out,
                          const SbVec3f &nl, const SbVec3f &ngl,
                          const Ray &r, SbVec3f &brdfCosT, float &pdf);
};

class SCHLICK_BRDF : public BRDF {
public:
  // From "An Inexpsensive BRDF Model for Physically-based Rendering"
  // by Christophe Schlick

  virtual void GetSampleDir(double eps1, double eps2, double eps3,
                            const SbVec3f &in, const SbVec3f &nl,
                            const SbVec3f &ngl, const Ray &r,
                            SbVec3f &out, double &brdf, double &pdf,
                            double &brdfOverPdf, double &brdfOverPdfCosT,
                            SbVec3f &color, bool internalCall = false);
  
  // This doesn't include ideal specular component, that's handled separately.
  // It does include the diffuse component.
  virtual SbVec3f GetBRDF(const SbVec3f &in, const SbVec3f &out,
                          const SbVec3f &nl, const SbVec3f &ngl,
                          const Ray &r, SbVec3f &brdfCosT, float &pdf);
  virtual bool IsDirac() const { return false; }
};

class ASHIKHMIN_BRDF : public BRDF {
  
public:
  virtual void GetSampleDir(double eps1, double eps2, double eps3,
                            const SbVec3f &in, const SbVec3f &nl,
                            const SbVec3f &ngl, const Ray &r,
                            SbVec3f &out, double &brdf, double &pdf,
                            double &brdfOverPdf, double &brdfOverPdfCosT,
                            SbVec3f &color, bool internalCall);
  
  virtual SbVec3f GetBRDF(const SbVec3f &in, const SbVec3f &out,
                          const SbVec3f &nl, const SbVec3f &ngl,
                          const Ray &r, SbVec3f &brdfCosT, float &pdf);
  virtual bool IsDirac() const { return false; }  
};

#endif /* MATERIALCLASS */

