// Header for Misc.cxx

#ifndef MISCFUNCS
#define MISCFUNCS

#include <vector>
#include <float.h>
#include <iostream>
#include <iomanip>
#include <Inventor/SbLinear.h>

class SoNode;

using namespace std;

// find min/max of three/two vals
// Erkk! Make these inline functions! Or else they could be executing arguments twice.
#define MIN3(a,b,c) ((((a)<(b))&&((a)<(c))) ? (a) : (((b)<(c)) ? (b) : (c)))
#define MAX3(a,b,c) ((((a)>(b))&&((a)>(c))) ? (a) : (((b)>(c)) ? (b) : (c)))
#define MAX2(a,b) ( ((a)>(b)) ? (a) : (b) )
#define MIN2(a,b) ( ((a)<(b)) ? (a) : (b) )

template <typename T>
inline T max3(T a, T b, T c){
  return MAX3(a,b,c);
}

template <typename T>
inline T min3(T a, T b, T c){
  return MIN3(a,b,c);
}

template <typename T>
inline T max2(T a, T b){
  return MAX2(a,b);
}

template <typename T>
inline T min2(T a, T b){
  return MIN2(a,b);
}

#define SAY_HELLO cerr << "\033[1;32mHELLO\033[0m: " << __FILE__ << ":" << __LINE__ \
  << " (" << __FUNCTION__ << ")" << "\n";

#define GIVE_ERROR(x) cerr << "\033[1;32mERROR\033[0m: " << __FILE__ << ":" << __LINE__  \
  << " (" << __FUNCTION__ << ") " << (x) << "\n";

//fprintf(stderr, "%s:%d (%s) HELLO: \n", __FILE__, __LINE__, __FUNCTION__);

// vector mul and finite (defined in Main.cxx for some reason)
SbVec3f mul(SbVec3f &x, SbVec3f &y, float c=1.0);
int finite(SbVec3f &c);
inline SbVec3f scalarToVec(float c) { return SbVec3f(c,c,c); }
  
// find absolute value
#define ABSF(x) ( ((x)<0.0) ? -(x) : (x) )

// time a set of instructions
#define TIME(x) { struct timeval start, end;  \
                gettimeofday(&start, NULL); \
                x;                          \
 	        gettimeofday(&end, NULL);   \
	        cerr << "Time: "  << delay(start, end) << "ms" << endl; }


// brightness with weights tuned to human visual system
inline float brightness(SbColor const &c){
  return(0.27*c[0] + 0.67*c[1] + 0.06*c[2]);
}
inline float brightness(const unsigned char *i){
  return((0.27*i[0] + 0.67*i[1] + 0.06*i[2])/255.0);
}

// clamp definition
// restricts a number to a range [min, max]
template <class T> inline
T clamp(T x, T min, T max){
  if (x<min)
    return(min);
  else if (x>max)
    return(max);
  else
    return(x);
}

// Compute the delay between t1 and t2 in milliseconds
inline long delay (struct timeval t1, struct timeval t2){
  long d;
    
  d = (t2.tv_sec - t1.tv_sec) * 1000;
  d+= ((t2.tv_usec - t1.tv_usec + 500) / 1000);
  return(d);
}

inline void resetios(ostream &s){
  const int defaultPrc = 6;
  s << resetiosflags(ios::floatfield) << setprecision(defaultPrc);
}

inline int nearestPowerOf2(unsigned int x){
  unsigned int i=0, y=x;                                                                           
  while (y=y>>1) i++;                                                                 
  int r = 1<<i;                                                                       
  if (r!=x)                                                                           
    r = r<<1;
  return r;
}

template <typename T>
inline void vecCopy2(T *dst, T *src){
  dst[0] = src[0];
  dst[1] = src[1];
}

template <typename T>
inline void vecCopy3(T *dst, T *src){
  dst[0] = src[0];
  dst[1] = src[1];
  dst[2] = src[2];
}

inline double fresnel(const double &Refl0, const double &cosTheta){
  double c = 1-cosTheta;
  double cc = c*c;
  return Refl0 + (1-Refl0) * (cc*cc*c);
}
inline SbVec3f fresnel(const SbVec3f &Refl0, const double &cosTheta){
  double c = 1-cosTheta;
  double cc = c*c;
  return Refl0 + (SbVec3f(1,1,1)-Refl0) * (cc*cc*c);
}

// Multiply two SbVec3f's piecewise
inline SbVec3f vecScale(const SbVec3f &a, const SbVec3f &b) {
  return SbVec3f(a[0]*b[0], a[1]*b[1], a[2]*b[2]);
}

inline SbVec3f vecDiv(const SbVec3f &a, const SbVec3f &b) {
  return SbVec3f(a[0]/b[0], a[1]/b[1], a[2]/b[2]);
}

inline SbVec3f& vecScaleItself(SbVec3f &a, const SbVec3f &b) {
  a[0] *= b[0];
  a[1] *= b[1];
  a[2] *= b[2];
  return a;
}

inline SbVec3f vecExp(const SbVec3f &v){
  return SbVec3f(exp(v[0]),
                 exp(v[1]),
                 exp(v[2]));
}

inline float vecAvg(const SbVec3f &v) {
  return (v[0]+v[1]+v[2])*(1.f/3.f);
}

inline float vecMax(const SbVec3f &v) {
  return MAX3(v[0],v[1],v[2]);
}

inline bool isWhite(const SbVec3f &x){
  return x[0] == 1 && x[1] == 1 && x[2] == 1;
}
inline bool isWhite(const SbVec3f &x, float tol){
  return
    fabs(x[0]-1)<=tol &&
    fabs(x[1]-1)<=tol &&
    fabs(x[2]-1)<=tol;
}
inline bool isBlack(const SbVec3f &x){
  return x[0] == 0 && x[1] == 0 && x[2] == 0;
}
inline bool isBlack(const SbVec3f &x, float tol){
  return
    fabs(x[0]-0)<=tol &&
    fabs(x[1]-0)<=tol &&
    fabs(x[2]-0)<=tol;
}

// Idea from article by Altair in vol. 14 of Hugi diskmag:
// http://www.hugi.scene.org/main.php?page=hugi14
//
// 18 times faster than logf(x)*(1/logf(2)) in my testing.
// inline int fastlog2(float x){
// //  cerr << __LINE__ << " (int &)x >> 23=" << ((int &)x >> 23) << "\n";
//   cerr << "x=" << x << " ";
// //  cerr << "(int&)x=" << (int&)x << " ";
//   int a = ((int &)x >> 23);
//   cerr << "a=" << a << " ";
//   a -= 127;
//   cerr << "a=" << a << " ";
//   return a;
// }


// this function refuses to work anywhere near properly for me???????????????
inline int fastlog2(float x){
  return ((int &)x >> 23)-127;
}

float cauchy(float n, float w);

inline SbVec3f vecClamp(const SbVec3f &x, float a, float b){
  SbVec3f y;
  y[0] = clamp(x[0], a, b);
  y[1] = clamp(x[1], a, b);
  y[2] = clamp(x[2], a, b);
  return y;
}

class Vec3d {
public:

  double vec[3];          // position, also color (r,g,b)
    
  Vec3d(double x_=0, double y_=0, double z_=0){
    vec[0] = x_;
    vec[1] = y_;
    vec[2] = z_;
  }
  
  Vec3d(const SbVec3f &v) {
    vec[0] = v[0];
    vec[1] = v[1];
    vec[2] = v[2];
  }
  
  operator SbVec3f() const {
    return SbVec3f(vec[0], vec[1], vec[2]);
  }
  
  Vec3d operator+(const Vec3d &b) const {
    return Vec3d(vec[0]+b.vec[0],
                 vec[1]+b.vec[1],
                 vec[2]+b.vec[2]);
  }
  
  Vec3d operator-(const Vec3d &b) const {
    return Vec3d(vec[0]-b.vec[0],
                 vec[1]-b.vec[1],
                 vec[2]-b.vec[2]);
  }
  
  Vec3d operator*(double b) const {
    return Vec3d(vec[0]*b,
                 vec[1]*b,
                 vec[2]*b);
  }

  Vec3d operator/(double b) const {
    double inv = 1/b;
    return Vec3d(vec[0]*inv,
                 vec[1]*inv,
                 vec[2]*inv);
  }
  
  Vec3d& operator*=(double b) {
    vec[0] *= b;
    vec[1] *= b;
    vec[2] *= b;
    return *this;
  }

  Vec3d& operator/=(double b) {
    double inv = 1/b;
    return (*this)*=inv;
  }
  
  Vec3d scale(const Vec3d &b) const {
    return Vec3d(vec[0]*b.vec[0],
                 vec[1]*b.vec[1],
                 vec[2]*b.vec[2]);
  }
  
  Vec3d norm(){
    double len2 = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
    double len;
    Vec3d n(0,0,0);
    if (!len2) return n;
    len = sqrt(len2);
    n = *this * (1/len);
    return n;
  }

  double normalize(){
    double len2 = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
    double len;
    if (len2){
      len = sqrt(len2);
      *this = *this * (1/len);
    }
    return len;
  }
  
  double length() const {
    return sqrt(vec[0]*vec[0] +
                vec[1]*vec[1] +
                vec[2]*vec[2]);
  }
  
  double dot(const Vec3d &b) const {
    return
      vec[0]*b.vec[0] +
      vec[1]*b.vec[1] +
      vec[2]*b.vec[2];
  }
  
  // cross prod  
  Vec3d operator%(const Vec3d &b) const {
    return Vec3d(vec[1]*b.vec[2] - vec[2]*b.vec[1],
                 vec[2]*b.vec[0] - vec[0]*b.vec[2],
                 vec[0]*b.vec[1] - vec[1]*b.vec[0]);
  }
  
  Vec3d cross(const Vec3d &b) const { return *this % b; }

  double       & operator [](int i) 	  { return vec[i]; }
  const double & operator [](int i) const { return vec[i]; }
};


int computeNormalsIv(SoNode *scene);


//'WWW''WW'''W'WWWW''''''''''WWW''''''''''''''''''
//WW  W WWW  W WW  W          WW                  
//WW  W W WW W WWWW      WWW  WW   WWW  WWWW  WWWW
//WW  W W  WWW WW  W    WW W  WW  W WW WWW   WWW  
//WW  W W   WW WW  W    WW    WW  W WW   WWW   WWW
// WWW  W    W WWWW      WWW WWWW  WWW WWWW  WWWW 
//                                                
//''''''''''''''''''''''''''''''''''''''''''''''''

// ortho-normal basis class
class ONB {

private:
  
  SbVec3f u, v, w; // w "points up"

public:

  ONB(){}
  ONB(SbVec3f &c) {
    constructFromW(c);
  }
  
  void init(){
    u.setValue(1,0,0);
    v.setValue(0,1,0);
    w.setValue(0,0,1);
  }

  // from two triangle edges
  void constructFromUV(const SbVec3f &a, const SbVec3f &b){
    u = a;
    u.normalize();

    w = a.cross(b);
    w.normalize();
    
    v = w.cross(u);
    v.normalize();
  }

  // from up direction
  void constructFromW(const SbVec3f &c){
    w = c;
    w.normalize();

    u = SbVec3f(w[1], w[2], w[0]+0.5); // Hopefully this !~ w
    u -= u.dot(w)*w;
    u.normalize();

    v = w.cross(u);
    v.normalize();
  }

  // from U and W directions
  void constructFromUW(const SbVec3f &a, const SbVec3f &c){
    w = c;
    w.normalize();

    u = a;
    u -= u.dot(w)*w;
    u.normalize();

    v = w.cross(u);
    v.normalize();
  }
  
  SbVec3f toONBSpace(const SbVec3f &a) const {
    SbVec3f x;
    x[0] = u.dot(a);
    x[1] = v.dot(a);
    x[2] = w.dot(a);
    return x;
  }

  Vec3d toONBSpace(const Vec3d &a) const {
    Vec3d x;
    x[0] = a.dot(u); //u.dot(a);
    x[1] = a.dot(v); //v.dot(a);
    x[2] = a.dot(w); //w.dot(a);
    return x;
  }
  
  SbVec3f toWorldSpace(const SbVec3f &a) const {
    return SbVec3f(a[0]*u + a[1]*v + a[2]*w);
  }
  
  Vec3d toWorldSpace(const Vec3d &a) const {
    return Vec3d(Vec3d(u)*a[0] + Vec3d(v)*a[1] + Vec3d(w)*a[2]);
  }

  const SbVec3f& getU() const { return u; }
  const SbVec3f& getV() const { return v; }
  const SbVec3f& getW() const { return w; }

};

class BBox {

private:
  void init() {
    min = SbVec3f(FLT_MAX, FLT_MAX, FLT_MAX);
    max = SbVec3f(-FLT_MAX, -FLT_MAX, -FLT_MAX);
  }
  
public:
  SbVec3f min, max;

  SbVec3f& operator[](int i) {
    return ((SbVec3f*)this)[i];
  }

  bool operator==(const BBox &b) const {
    return
      min[0] == b.min[0] &&
      min[1] == b.min[1] &&
      min[2] == b.min[2] &&
      max[0] == b.max[0] &&
      max[1] == b.max[1] &&
      max[2] == b.max[2];
  }
  
  BBox(){
    init();
  }
  
  BBox(SbVec3f *p, int n){
    init();
    grow(p, n);
  }

  BBox(const SbVec3f &min_, const SbVec3f &max_) :
    min(min_), max(max_) {}

  SbVec3f size(){
    return max-min;
  }
  
  void grow(const SbVec3f *vert, int n){
    for (int v=0; v<n; v++){
      for (int d=0; d<3; d++){
        if (vert[v][d] < min[d])
          min[d] = vert[v][d];
        if (vert[v][d] > max[d])
          max[d] = vert[v][d];
      }
    }
  }

  void grow(const BBox &x) {
    grow(&x.min, 1);
    grow(&x.max, 1);
  }
  
  BBox formUnion(const BBox &x) const {
    BBox u = *this;
    u.grow(x);
    return u;
  }

  static SbVec3f vecMin(const SbVec3f &a, const SbVec3f &b) {
    SbVec3f min;
    for (int i=0; i<3; i++)
      min[i] = a[i]<b[i] ? a[i] : b[i];
    return min;
  }
  
  static SbVec3f vecMax(const SbVec3f &a, const SbVec3f &b) {
    SbVec3f max;
    for (int i=0; i<3; i++)
      max[i] = a[i]>b[i] ? a[i] : b[i];
    return max;
  }

  int longestAxis() {
    SbVec3f diff = max-min;
    if (diff[0]>diff[1] && diff[0]>diff[2])
      return 0;
    else if (diff[1]>diff[2])
      return 1;
    else
      return 2;
  }
  
};

template <typename T>
class Spline {
private:
  vector<T> items;
  vector<float> vals;
  float min, max;
  bool interp;

  void AutoRemap(){
    vals.resize(items.size());
    int num = items.size();
    for (int i=0; i<num; i++){
      vals[i] = i/(float)num*(max-min)+min;
    }
  }

public:
  Spline(float min_=0, float max_=1) : interp(true) {
    Init(min_, max_);
  }

  void Init(float min_=0, float max_=1){
    min = min_;
    max = max_;

    items.clear();
    vals.clear();
  }
  
  void Add(const T& x){
    items.push_back(x);
    AutoRemap();
  }

  void SetInterp(bool b){
    interp = b;
  }

  // v is when the added item takes full effect
  void Add(float v, const T& x){

    items.resize(items.size()+1);
    vals.resize(items.size(), 1);

    // find insertion point
    int i=0;
    while (i<items.size()-1 && vals[i]<v)
      i++;

    // move items after it
    int j = items.size()-1;
    while (j>i){
      items[j] = items[j-1];
      vals[j] = vals[j-1];
      j--;
    }

    // insert item
    items[i] = x;
    vals[i] = v;
  }
  
  T eval(float x) const;

};


#endif /* MISCFUNCS */
