
#ifndef NOISECLASS
#define NOISECLASS

#include <Inventor/SbLinear.h>
#include <stdlib.h>
#include <math.h>
#include "Main.h"
#include "Misc.h"

// Perlin noise,
// based on "Hypertextures" by Ken Perlin

// #define float double
// #define powf pow
// #define logf log

enum NoiseFuncType {
  MONO = 0,
  FBM,
  MULTI,
  TERRAIN,
  HYBRID_MULTI,
  RIDGED_MULTI
};

class NoiseParams {
public:
  int octaves;
  int seed;
  bool sign;
  float gain;
  float bias;
  float amplitude;
  SbVec3f translation;  // pre noise translation
  SbVec3f frequency;
  float offset;         // post noise
  float freqJump;       // lacunarity (lambda)
  float ampJump;        // roughness  (H?)
  NoiseFuncType funcType;         // noise function
};

class Noise {

  private:

  static const int n=16;
  SbVec3f G[n];
  int P[n];

  float TauDotXYZ(int i, int j, int k, float x, float y, float z) const {
    const SbVec3f t = Tau(i,j,k);
    return t[0]*x + t[1]*y + t[2]*z;
  }
  
  float omega(float x) const {
    if (x<0) x = -x;
    // dont need to check if x<1 because it should be already
    const float x2 = x*x;
    const float x3 = x2*x;
    return  -6.0f* x2*x3 + 15.0f* x2*x2 - 10.0f* x3 + 1.0f;
  }

  SbVec3f Tau(int i, int j, int k) const {
    return G[ phi( i+phi( j+phi(k) ) ) ];
  }

  int phi(int i) const {
    return P[i&15]; // fast mod 16
    // return P[(i%n+n)%n];
  }

  // slow, use TauDOTXYZ instead
  float Omega(int i, int j, int k, const SbVec3f &x) const {
    return omega(x[0])*omega(x[1])*omega(x[2]) * Tau(i,j,k).dot(x);
  }

  // slow, use TauDOTXYZ instead
  float Omega(int i, int j, int k, float x, float y, float z) const {
    const SbVec3f t = Tau(i,j,k);
    return
      omega(x)*
      omega(y)*
      omega(z)*
      (t[0]*x + t[1]*y + t[2]*z);
  }


  // precomputed exponents
  vector<float> exponentArray;
  float lastAmpJump, lastFreqJump;

public:

  // noise function options
  int octaves;
  int seed;
  bool sign;            // IF TRUE: returned value can be + and -
  float gain;
  float bias;
  float amplitude;
  SbVec3f translation;  // pre noise translation
  SbVec3f frequency;
  float offset;         // post noise
  float freqJump;       // lacunarity (lambda)
  float ampJump;        // roughness  (H?)
  
  float (Noise::*funcPtr)(const SbVec3f &x) const;
  
  Noise(){  init(); }
  void init();

  float noise(const SbVec3f &x) const {
    return (*this.*funcPtr)(x);
  }
  float noise(const double &x, const double &y, const double &z) const {
    return (*this.*funcPtr)(SbVec3f(x,y,z));
  }
  float noise(const float x, const float y, const float z) const {
    return (*this.*funcPtr)(SbVec3f(x,y,z));
  }

  void initExponents(); // call after changing octaves/ampJump/freqJump

  void setParams(const NoiseParams &np);
  
  float perlin(const SbVec3f &x) const;

  // vector values noise, from HyperTexture pg. 7 (pg. 259)
  SbVec3f dnoise(const SbVec3f &x) const {
    SbVec3f ofs(100.4384, -904.3082, 534.34932);
    return SbVec3f(
      (*this.*funcPtr)(x-ofs),
      (*this.*funcPtr)(x),
      (*this.*funcPtr)(x+ofs)
      );
  }

  // variable lacunarity noise
  // From Methods for Realistic Landscape Imaging by K. Musgrave
  float vlnoise(const SbVec3f &x) const {
    return perlin(x + dnoise(x)*0.5);
  }
  
  // moves midpoint of [0,1] curve
  // down if b<0.5
  // up   if b>0.5
  float biasFunc(float x, float b) const {
    return powf(x, logf(b)*(1.0f/logf(0.5f)));
  }

  // changes contrast of [0,1] curve about 0.5
  // higher if g>0.5
  // lower  if g<0.5
  float gainFunc(float x, float g) const {
    if (x<0.5f)
      return (biasFunc(2.0f*x, 1.0f-g)/2.0f);
    else
      return (1.0f - biasFunc(2.0f - 2.0f*x, 1.0f-g)/2.0f);
  }
  // changes contrast of [-inf,inf] curve about 0
  // higher if b>0.5
  // lower  if b<0.5
  float gainFunc_0(float x, float b) const {
    if (x<0.0)
      return -powf(-x, logf(b)*(1.0f/logf(0.5f)));
    else
      return powf(x, logf(b)*(1.0f/logf(0.5f)));
  }

  float turbulence(const SbVec3f &x) const {
    float y = 0.0;
    int a = 1;
    for (int j=0; j<octaves; j++){
      y += fabs(perlin(x*a)/a);
      a <<= 1;
    }

    return y;
  }

  float applyRest(float val) const {

    if (!sign){
      val = fabs(val);
      val = gainFunc(biasFunc(val, bias), gain);
    } else if (val>=0) {
      val = gainFunc(biasFunc(val, bias), gain);
    } else  {
      val = -gainFunc(biasFunc(-val, bias), gain);
    };
    
    return val*amplitude+offset;
  }
  
  
  float val(const SbVec3f &x, float b=0.5, float g=0.5) const {
    return biasFunc(gainFunc(turbulence(x), g), b);
    //return fabs(perlin(x/100.0f));
  }


  // basic monofractal
  float val_fractal(const SbVec3f &x, float b=0.5, float g=0.5, float f=1.0) const {
    return gainFunc_0(gainFunc_0(perlin(x), g), b);
  }


  float monofractal(const SbVec3f &x_) const;
  // from Texturing & Modeling: A Procedural Approach pg. 440
  float fBm(const SbVec3f &x_) const;
  // from Texturing & Modeling: A Procedural Approach pg. 440
  float multiFractal(const SbVec3f &x_) const;
  // Methods for Realistic Lanscape Imaging by Ken Musgrave
  // Terrain Model p. 55
  float terrain(const SbVec3f &x_) const;
  // from Texturing & Modeling: A Procedural Approach pg. 502
  float hybridMultiFractal(const SbVec3f &x_) const;
  // from Texturing & Modeling: A Procedural Approach pg. 504
  float ridgedMultiFractal(const SbVec3f &x_) const;

  float (Noise::*funcPtrArray[6])(const SbVec3f &x) const;

};

#undef float
#undef powf
#undef logf

#endif /* NOISECLASS */
