////////////////////////////////////////////////////////////////////////////
//
//WWWW'''''''''''''''''''''''''WWW''''''''''''''''''
//WW  W                         WW                  
//WW  W  WWW WWW   WW      WWW  WW   WWW  WWWW  WWWW
//WWWW  W WW WW W WWWW    WW W  WW  W WW WWW   WWW  
//WW    W WW WW W WW      WW    WW  W WW   WWW   WWW
//WW     WWW WW W  WWW     WWW WWWW  WWW WWWW  WWWW 
//                                                  
//''''''''''''''''''''''''''''''''''''''''''''''''''
//
// File: Pane.cxx
// Description: Pane class (Core Ray Tracing Engine) implementation
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Rendering engine for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// 
// $Id: Pane.cxx 1272 2011-09-06 00:51:58Z beason $
//
////////////////////////////////////////////////////////////////////////////

#include "Pane.h"
#include "debug.h"
#include "DataSet.h"
#include "NoiseDataSet.h"
#include "EnvironMap.h"
#include "Sky.h"
#include "RayAccel.h"
#include "Octree.h"
#include "BIH.h"
#include "Photon_map.h"
#include <set>

using namespace std;

static char rcsid[] = "@(#) $Id: Pane.cxx 1272 2011-09-06 00:51:58Z beason $";
const double Pane::N_AIR = 1.0;   // refr. index for air, 1.00029;
const double Pane::N_GLASS = 1.7; // refr. index for glass, 1.5-1.7

//'WWW''''''''''''''''''''WW''''''''''''''''''WW''''''''''''''''''
//WW  W                   WW                  WW                  
//WW     WW   WWW   WWWW WWWW W WW W WW  WWW WWWW  WW   W WW  WWWW
//WW    WW W  WW W WWW    WW  WW   W WW WW W  WW  WW W  WW   WWW  
//WW  W WW W  WW W   WWW  WW  WW   W WW WW    WW  WW W  WW     WWW
// WWW   WW   WW W WWWW   WW  WW    WWW  WWW  WW   WW   WW   WWWW 
//                                                                
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Constructors
//

// Constructor
Pane::Pane(GlobalVars *globVar){
  gv = globVar;
  pix = NULL;
  pixDepth = NULL;
  pixWeight = NULL;
  gPhotonMap = NULL;
  cPhotonMap = NULL;
  importanceMap = NULL;
  volumeMap = NULL;
  toneMapped = 0;
  sampMap = NULL;
  envMap = NULL;
  perspectiveCamera = NULL;
  headLight = NULL;
  dataSet1 = NULL;
  dataSet2 = NULL;
  dataSet3 = NULL;
  allDataSets = NULL;
  noiseDS = NULL;
  filtpix = NULL;
  pixMutexLocks = NULL;
  pixMutexes = NULL;

  totalPower = 0.0;
  max_photons = 0;
  firedPhotons = 0;
  minApproxDistance = 0.0;
  rejectedICPoints = 0;
  numLights = 0;
  maxEmisAvg = 0.0;
  totalEmisAvg = 0.0;
  lastImageYi = 0.0;
  lastImageAvgYi = 0.0;
  renderRaysFired = 0;
  renderIsects = 0;
  photonRaysFired = 0;
  currentlyRendering = false;
  currentlyShooting = false;
  currentlyRenderingPPMRange = false;
  writeWhenDone = false;
  ppmPix = NULL;
  ppmPixImages = 0;
  ppmImgOffset[0] = ppmImgOffset[1] = 0;

  // Scene processing stuff
  processTriangles = true;

  // create default materials
  sepListItem_t *item = new sepListItem_t;
  item->mat = new Material();
  item->matPlus = NULL;
  item->verts = NULL;
  item->norms = NULL;
  item->soMat = NULL;
  item->vProp = NULL;
  sepStack.push_front(item);
  matList.push_back(item->mat);
}


//WWWW''''''''''''''''''''''''''''''WW''''''''''''''''''WW''''''''''''
//WW  W                             WW                  WW            
//WW  W  WW   WWW  WW   WWW   WWWW WWWW W WW W WW  WWW WWWW  WW   W WW
//WW  W WWWW WW W WW W  WW W WWW    WW  WW   W WW WW W  WW  WW W  WW  
//WW  W WW   WW   WW W  WW W   WWW  WW  WW   W WW WW    WW  WW W  WW  
//WWWW   WWW  WWW  WW   WW W WWWW   WW  WW    WWW  WWW  WW   WW   WW  
//                                                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Deconstructor
//
Pane::~Pane(){

  // remove headlight (at front of litPnts)
  headLight->unref();

  delete [] pix;
  delete [] pixDepth;
  delete [] pixWeight;
  delete [] pixMutexLocks;
  delete [] pixMutexes;
  delete [] sampMap;
  delete envMap;
  if (allDataSets != dataSet1)
    delete allDataSets;
  delete dataSet1;
  delete dataSet2;
  delete dataSet3;
  delete rayAccel;
  delete scenePrim;
  delete gPhotonMap;
  delete cPhotonMap;
  delete importanceMap;
  delete volumeMap;
  delete filtpix;
  delete noiseDS;

  // remove last sepStack
  if (sepStack.size() == 1){
    delete sepStack.front();
  } else {
    cerr << __FUNCTION__ << ": Bad sepStack count ("
         << sepStack.size() << ")\n";
  }
  
  // clear matList
  while (!matList.empty()){
    delete matList.back();
    matList.pop_back();
  }

  // clear meshs and free textures
  TriMesh::deleteMeshs();

  // clear instances
  set<Primitive*> deletedPrims;         // don't delete instanced prims more
                                        // than once!
  for (int i=instances.size(); i-->0;){
    Primitive *prim = instances[i].getPrim();
    if (deletedPrims.find(prim) == deletedPrims.end()){
      instances[i].deletePrim();
      deletedPrims.insert(prim);
    }
  }

  // clear any nodes that have been scheduled to be removed
  garbageDump->removeAllChildren();
  garbageDump->unref();
}
  

//WW''''WW'''''''WW''WW'''''''''''WW''''''
//WWW  WWW       WW  WW           WW      
//W WWW WW  WW  WWWW WWW   WW    WWW  WWWW
//W  W  WW WWWW  WW  WW W WW W  W WW WWW  
//W     WW WW    WW  WW W WW W  W WW   WWW
//W     WW  WWW  WW  WW W  WW    WWW WWWW 
//                                        
//''''''''''''''''''''''''''''''''''''''''


// Return string containing version number
const char* getVersion() {
  static char versionInfo[80] =
    "$Revision: 1272 $";
  static bool verSet = false;
  static string version;

  static char dateInfo[80] =
    "$Date:$";
  
  if (!verSet){
    verSet = true;
    
    char *p = strchr(versionInfo, ' ')+1;
    int lenp = strlen(p);
    char *q = strsep(&p, " ");
    int lenq = strlen(q);

    // insert "." after first digit
    if (lenq<lenp && lenq>=4){
      string s = q;
      sprintf(q, "%c.%s", s[0], s.c_str()+1);
      version = q;
    } else {
      cerr << "Revision string error.\n";
      version = "Unknown";
    }

  }
  
  return version.c_str();
}

string getNameVersionDate() {
  const char *moduleBuildDate = __DATE__;
  static string buildDate = string("built approx. ") + moduleBuildDate;
  const char *revision = getVersion();

  ostringstream out;
  out << "Pane v" << revision << " (" << buildDate << ")";

  return out.str();
}

string getAuthorName() {
  ostringstream str;
  str << "Kevin Beason <kevin.beason@gmail.com> (c) " << getCurrentYear();
  return str.str();
}

string getAuthorUrl(){
  return "http://kevinbeason.com/scs/pane/";
}

string getCurrentTime() {
  time_t myTime = time(NULL);
  string curTime = ctime(&myTime);
  return curTime;
}

int getCurrentYear(){
  time_t rawtime;
  struct tm *timeinfo;
  time(&rawtime);
  timeinfo = localtime(&rawtime);  // get year
  int year = timeinfo->tm_year + 1900;
  return year;
}

//WW''''''WW''WW'
//            WW 
//WW WWW  WW WWWW
//WW WW W WW  WW 
//WW WW W WW  WW 
//WW WW W WW  WW 
//               
//'''''''''''''''
//
// class initializer
// builds octree, creates headlight, pixel array
//
int Pane::init(){

  // Create dumping ground for nodes to be automatically removed on exit
  garbageDump = new SoSeparator();
  garbageDump->ref();

  // create photon maps
  if (gv->numPhotons==-1)
    gv->numPhotons = gv->algorithm == 1 ? 90000 : 0;
  if (gv->numCPhotons==-1)
    gv->numCPhotons = gv->algorithm == 1 ? 2000 : 0;
  max_photons = max2(max2(gv->numPhotons, gv->numCPhotons)*2, 0);
  switch (gv->accelType){
  case GlobalVars::ACC_OCTREE:
    rayAccel = (RayAccel*)(new Octree());
    rayAccelName = "octree";
     break;
  case GlobalVars::ACC_KDTREE:
    rayAccel = NULL; //(RayAccel*)(new KdTree());
    rayAccelName = "kdtree";
    break;
  case GlobalVars::ACC_BIH:
    rayAccel = (RayAccel*)(new BIH());
    rayAccelName = "BIH";
    break;
  case GlobalVars::ACC_TEST:
    rayAccel = (RayAccel*)(new DualAccel());
    rayAccelName = "DualAccel";
    break;
  case GlobalVars::ACC_LIST:
    rayAccel = (RayAccel*)(new AccelList());
    rayAccelName = "AccelList";
    break;
  default:
    GIVE_ERROR("Unknown ray acceleration type.")
    break;
  }
  scenePrim = new PrimList();
  ((PrimList*)scenePrim)->add(rayAccel);
  
  if (!rayAccel){
    cerr << "ERROR: That acceleration structure is not available.\n";
    exit(1);
  }

  // create headlight
  headLight = new SoPointLight();
  headLight->ref();
  headLight->intensity.setValue(1.0);
  litPnts.insert(litPnts.begin(), headLight);

  // set default sub pixel dim
  if (gv->subPixDim<=0){
    if (gv->numSamples>=16)
      gv->subPixDim = 4;
    //else if (gv->numSamples>=4)
    // ...no. The tradeoff at -s4 --subpix=2 vs -s4 --subpix=1 is too costly.
    // Without pixel centers: noisy edges of lights
    // With pixel centers: aliased edges
    // To benefit from subpixel samples go to at least -s9.
    else if (gv->numSamples>4)
      gv->subPixDim = 2;
    else
      gv->subPixDim = 1;
  }
  cerr << "Using " << gv->subPixDim << " subpixel resolution.\n";
  
  // calculate total pixel size
  gv->imgTotalDim[0] = gv->imgDim[0]*gv->subPixDim;
  gv->imgTotalDim[1] = gv->imgDim[1]*gv->subPixDim;
  gv->subPixFilterSize = 1.0; // in subpixels
  gv->invSubPixDim2 = 1.f/(gv->subPixDim*gv->subPixDim);
  sigactionHandlerNS::imgHeight = gv->imgDim[Y];

  // precompute total number of samples
  const int sqrtSamples = (int)ceil(sqrt((float)gv->numSamples)); 
  const int numSamples = sqrtSamples*sqrtSamples;
  gv->numSamplesActual = numSamples;
  
  int subPix = gv->subPixDim*gv->subPixDim;
  if (gv->numSamples<subPix && gv->inImgFile.empty()){
    cerr << endl << endl;
    cerr << "Warning: You have less samples (" << gv->numSamples
         << ") than subpixels (" << subPix << ")!\n"
         << "\tIncreasing pixel filter size to compensate. "
         << "You should probably increase the number of samples "
         << " to -s" << subPix << "." << endl;
    cerr << endl << endl;
    gv->subPixFilterSize = gv->subPixDim;
  }
  if (getenv("PANE_SUBPIX_FILTER"))
    gv->subPixFilterSize = atof(getenv("PANE_SUBPIX_FILTER"));
  gv->invSubPixFilterSize = 1.f/gv->subPixFilterSize;

  // initialize pixel locks
  pixMutexLocks = new lock_spaced_t[gv->imgTotalDim[Y]];
  pixMutexes = new pthread_mutex_t*[gv->imgTotalDim[Y]];
  for (int i=0; i<gv->imgTotalDim[Y]; i++){
    pthread_mutex_init(&(pixMutexLocks[i].lock), NULL);
    pixMutexes[i] = &(pixMutexLocks[i].lock);
  }
  pthread_mutex_init(&completionMutex, NULL);
 
  // initialize pix array
  if (gv->inImgFile.empty()){
    int size = gv->imgTotalDim[X]*gv->imgTotalDim[Y];
    pix = new SbColor[size]; 
    pixDepth = new float[size];
    pixWeight = new float[size];
    for (int i=0; i<size; i++){
      pix[i].setValue(0,0,0);
      pixDepth[i] = pixWeight[i] = 0.0;
    }
  }
  // load image into pix array (aka framebuffer)
  else 
    readHDR();

  // Initialize PPM buffer
  if (gv->ppmIter>=0) {
    int size = gv->imgTotalDim[X]*gv->imgTotalDim[Y];
    ppmPix = new SbColor[size];
    for (int i=0; i<size; i++){
      ppmPix[i].setValue(0,0,0);
    }
  }

  // adjust worldBG (unless it's negative... which is a secret code to skip
  // this adjustment)
  if (vecAvg(gv->worldBG) < 0) {
    gv->worldBG = -gv->worldBG;
  } else {
    gv->worldBG[R] = pow(gv->worldBG[R], gv->gamma);
    gv->worldBG[G] = pow(gv->worldBG[G], gv->gamma);
    gv->worldBG[B] = pow(gv->worldBG[B], gv->gamma);
  }

  // set intelligent default of lightsamples:
  // roughly 100/pixelsamples. however, since
  // both the pixel and the light sampling
  // use stratified sampling, the number of
  // pixel samples and light samples used are the
  // next largest perfect square. this code chooses a
  // perfect square for the light samples in order
  // to gaurantee at least 100 light samples per pixel.
  // 
  if (gv->lightSamples == -1){
    if ( (gv->algorithm == 1 || gv->algorithm == 0)
	 && gv->areaLights == 1 ){
      
      int samples = (int)sqrt((float)gv->numSamples);
      samples *= samples;
      gv->lightSamples = MAX2(100/samples, 1); // at least 1
    }
    else
      gv->lightSamples = 1;
  }
  // find next largest perfect square
  int tmpSqrt = (int)sqrt((float)gv->lightSamples);
  if ( (gv->lightSamples-tmpSqrt*tmpSqrt) > 0.01 )
    tmpSqrt++;
  gv->lightSamples = tmpSqrt*tmpSqrt;
  gv->lightSqrtSamples = tmpSqrt;
  cerr << "Using " << gv->lightSamples << " light samples.\n";

  // Check environment map
  //
  //   Four possible valid inputs:
  //
  //   empty        : no map
  //   -1           : constant bg color
  //   filename.hdr : texture
  //   0.0-24.0     : time of day for sky model

  // check empty
  if (!gv->environMapFile.empty()){

    // check for .hdr extension
    string::size_type d = gv->environMapFile.find_last_of(".");
    if (d != string::npos) { // has extension

      // check extension
      string ext =  gv->environMapFile.substr(d);
      
      if (ext == ".hdr"){          // env map
        envMap = new EnvironMap();
        envMap->init(gv);
      }
    }

    // check for float
    if (!envMap) {
      
      const char *p = gv->environMapFile.c_str();
      char *endp;
      float t = strtof(p, &endp);

      // float found
      if (p != endp){

        // time of day (float in [0,24]) ?
        if (t>=0 && t<=24) {
          envMap = new Sky();
          ((Sky*)envMap)->init(gv, t);
          gv->sceneHasLights++;  // the sky part already does this, but 
                                 // lets do it for the sun too anyway
        }
        else if (t < 0) {
          envMap = new EnvironMap(); // constant color map
          envMap->init(gv);
          envMap->setSingleColor(true);
        }
      }
    }

    if (!envMap) {
      cerr << "ERROR: envMap: Invalid environment map option: "
           << gv->environMapFile.c_str() << "\n"
           << "Please provide [ filename.hdr | [0,24] | -1 ] to "
           << "create an environment map." << endl;
    }
  }

  // read volumetric dataset
  noiseDS = new NoiseDataSet();
  if (gv->tempInt==28)
    noiseDS->writeData("texture.dat");
    
  if (gv->stepSize>0.0){
    //dataSet1 = new DataSet();
    dataSet1 = new NoiseDataSet();
    
    dataSet1->useData(gv->scalarDataFile, gv->scalarDataDim,
		     gv->scalarDataHeader, 1,
		     SbVec3f(0,0,0), gv->scalarDataMax);

    // trim data by factor of 2
    if (getenv("PANE_DATA_DECIMATE")){
      dataSet1->writeData("half.dat", 0.5);
      dataSet1->writeData("quarter.dat", 0.25);
      exit(1);
    }
    
    dataSet1->scaleData(gv->scalarDataMult);
    
    if (!gv->scalarDataFile2.empty()){     
      dataSet2 = new DataSet();
      dataSet2->useData(gv->scalarDataFile2, gv->scalarDataDim,
			gv->scalarDataHeader, 1,
			SbVec3f(0,0,0), gv->scalarDataMax);
      dataSet2->scaleData(gv->scalarDataMult2);

      float maxTemp = dataSet2->maxValue;
      Spectrum::setMaxLMS(maxTemp);
      //dataSet2->computeWeights(gv->tempFloat);
      cerr << "Max temp: " << maxTemp << " (K) = "
	   << Spectrum(maxTemp).getAdaptedRGB() << " (adapted) = "
	   << Spectrum(maxTemp).getRGB() << " W/(m^2 sr Hz)\n";
      dataSet2->initColors(gv->meters(1.0));

      // experimental
      if (1||dataSet2->totalBrightness>0.0f)
	gv->sceneHasLights=1;
      totalEmisAvg += dataSet2->totalBrightness;
    }

    // this dataset only does absorbtion
    if (!gv->scalarDataFile3.empty()){
      dataSet3 = new DataSet();
      dataSet3->useData(gv->scalarDataFile3, gv->scalarDataDim,
			gv->scalarDataHeader, 1,
			SbVec3f(0,0,0), gv->scalarDataMax);
      dataSet3->scaleData(gv->scalarDataMult3);
    }
    
    //dataSet->useData("brain.mri", 109, 512);
    // dataSet->useData("/home/u2/users/beason/c/josh_gi/colin_brain.mri", 217, 512);
    //dataSet->scaleMaxToOne();

    // rotate
//     SbRotation rot(SbVec3f(0,1,0), SbVec3f(0,0,-1));
//     rot *= SbRotation(SbVec3f(0,0,1), SbVec3f(-1,0,0));
    SbRotation rot;

    rot.setValue(SbVec3f(0,1,0), gv->scalarDataRot*M_PI/180.);
    dataSet1->setRotation(rot);
    if (dataSet2)
      dataSet2->setRotation(rot);
    if (dataSet3)
      dataSet3->setRotation(rot);

    // create allDataSets
    if (dataSet3) {    
      MultiDataSet *multiDataSet = new MultiDataSet(dataSet1);
      multiDataSet->sets.push_back(dataSet3);
      allDataSets = multiDataSet;
    } else
      allDataSets = dataSet1;
  }

  readSampleMap();
  
  // read scene
  cerr << "Inserting primitives into " << rayAccelName.c_str()
       << " acceleration structure...\n" << flush;
  return readScene();
}


//'''''''''''''''W'''''''''''''''''''''''WW'''''''''''''''''''''''
//               W                        W                       
// WW  WW   WW  WW     WWW  WW WW W  WW   W   WW     WW W   WW WW 
//W   Wmmm W W W W     WW  W W W W W W W  W  Wmmm    W W W W W W W
//W    WW   WW  WW    WWW   WW W W W WW  WWW  WW     W W W  WW WW 
//                                   W                         W  
//'''''''''''''''''''''''''''''''''''W'''''''''''''''''''''''''W''
// read sample map
void Pane::readSampleMap(){
  
  if (gv->sampleMap.empty())
    return;
  
  ifstream pgm;
  cerr << "Reading sample map " << gv->sampleMap << "...";
  pgm.open(gv->sampleMap.c_str());
  if (!pgm){
    cerr << "Error: Could not open sample map " << gv->sampleMap << ".\n";
    exit(1);
  }
  char buf[100];
  pgm >> buf;
  if (strcmp(buf, "P2")){
    cerr << "Error: Sample map " << gv->sampleMap << " does not appear "
      " to be a PGM file.\n";
    cerr << "Could not recognize: \"" << buf << "\"\n";
    exit(1);
  }

  // read width
  while (pgm >> buf){
    if (buf[0] == '#')
      pgm.getline(buf, 100);
    else
      break;
  }
  sampMapWidth = atoi(buf);

  // read height
  while (pgm >> buf){
    if (buf[0] == '#')
      pgm.getline(buf, 100);
    else
      break;
  }
  sampMapHeight = atoi(buf);

  // read max value
  while (pgm >> buf){
    if (buf[0] == '#')
      pgm.getline(buf, 100);
    else
      break;
  }
  int maxVal = atoi(buf);

  // create space for map
  sampMap = new float[sampMapWidth * sampMapHeight];
    
  // read values
  int f;
  for (int j=sampMapHeight; j>0; j--)
    for (int i=0; i<sampMapWidth; i++){
      pgm >> f;
      sampMap[j*sampMapWidth+i] = f/(float)maxVal;
    }

  // finish up
  pgm.close();
  cerr << "done.\n";
}


//''''''''''''''''''''''''''''''''WWW''''''''WW'''''''''''
//                               WW  W       WW           
// WWW  WWW WWW  WW W  WWW  WWWW  WW    WW  WWWW W WW WWW 
//WW W W WW WW W WW W W WW WWW     WW  WWWW  WW  W WW WW W
//WW   W WW WW W WW W W WW   WWW W  WW WW    WW  W WW WW W
// WWW  WWW WW W  WW   WWW WWWW   WWW   WWW  WW   WWW WWW 
//                                                    WW  
//''''''''''''''''''''''''''''''''''''''''''''''''''''WW''
// sets basic canvas parameters, such as viewing direction,
// frustrum
void Pane::canvasSetup(CanvasParameters &canvas, SoCamera *cam){

  // ripped from Realistic Ray Tracing p.40 by Peter Shirley
  if (cam){
    
    // Up vector
    cam->orientation.getValue().multVec(SbVec3f(0.0, 1.0, 0.0),
					    canvas.vup);

    // Camera center
    canvas.e = SbVec3f(cam->position.getValue());

    // Camera direction
    cam->orientation.getValue().multVec(SbVec3f(0.0, 0.0, 1.0), canvas.w);
    canvas.w.normalize();

    // Distance to viewplane
    canvas.s = cam->focalDistance.getValue();

    // height, width
    float aspectRatio = ABSF(gv->imgDim[X]/(double)gv->imgDim[Y]);
    if (cam->isOfType(SoPerspectiveCamera::getClassTypeId())){
      float angle = ((SoPerspectiveCamera *)cam)->heightAngle.getValue();
      gv->heightAngle = angle; // store for tone mapping
      float dist = ((SoPerspectiveCamera *)cam)->focalDistance.getValue();
      canvas.height = tanf(angle/2.0)*dist*2.0;
      if (aspectRatio<1.0)
	canvas.height /= aspectRatio;
      canvas.orthographic = 0;
    }
    else if (cam->isOfType(SoOrthographicCamera::getClassTypeId())){
      canvas.height = ((SoOrthographicCamera *)cam)->height.getValue();
      canvas.orthographic = 1;
    }
    else {
      cerr << "WARNING: Unknown camera type!\n";
      canvas.height = 45.0;
    }
    canvas.width = aspectRatio*canvas.height;  
  }

  // Pointless, since Main.cxx gaurantees a camera
  else {
    canvas.vup = SbVec3f(0.0, 1.0, 0.0);
    canvas.e = SbVec3f(0.0, 0.0, rayAccel->getMaxSideLength()*2.0);
    canvas.w = SbVec3f(0.0, 0.0, 1.0);
    canvas.s = rayAccel->getMaxSideLength()*2.0;
    canvas.height = 45.0;
    canvas.width = canvas.height * ABSF(gv->imgDim[X]/(double)gv->imgDim[Y]);
  }

  // Orthonormal basis for viewplane [0]
  // canvas.w set above

  // Orthonormal basis for viewplane [1]
  canvas.u = SbVec3f(canvas.vup.cross(canvas.w));
  canvas.u.normalize();

  // Orthonormal basis for viewplane [2]
  canvas.v = SbVec3f(canvas.w.cross(canvas.u));
  canvas.v.normalize();

  if (!gv->fNumber){
    canvas.aperture = 0;
  } else {

    //                                 |
    //              |      lens        |
    // focus plane  |-------()---------|    image plane
    // (film)       |       e          |
    // H=24mm                          |
    //
    //              |---I---||----S----|
    //                  |-F-|
    //
    // 1/I + 1/S = 1/F
    // 
    // f/# = N = f/D = f/(2R)
    //
    // H/2 / I = tan(theta/2)
    
    // Compute DOF constants
    float N = gv->fNumber; // 1 1.4 2.8 4 5.6 8 11 16 22 32 45 64 90 128
    float S = gv->meters(canvas.s);
    float H = 24e-3; // 24 mm  height of film
    float theta = gv->heightAngle;
    float I = H/(2*tan(theta*0.5));
    float F = I*S/(S+I);
    float R = F/(2*N);
    R = gv->units(R);
//     cerr << "R=" << R << endl;
//     cerr << "F=" << F << endl;
//     cerr << "I=" << I << endl;

    canvas.aperture = R;
  }
  
}
