////////////////////////////////////////////////////////////////////////////
//
//WWWW'''''''''''''''''''''''''WWW''''''''''''''''''
//WW  W                         WW                  
//WW  W  WWW WWW   WW      WWW  WW   WWW  WWWW  WWWW
//WWWW  W WW WW W WWWW    WW W  WW  W WW WWW   WWW  
//WW    W WW WW W WW      WW    WW  W WW   WWW   WWW
//WW     WWW WW W  WWW     WWW WWWW  WWW WWWW  WWWW 
//                                                  
//''''''''''''''''''''''''''''''''''''''''''''''''''
//
// File: Pane.h
// Description: Pane class (Core Ray Tracing Engine) header
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Rendering engine for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// class Pane implements core ray tracing routines
// such as computing pixel colors and writing
// images to files. These routines are called
// from progMain() in Main.cxx.
// 
// $Id: Pane.h 1273 2011-09-07 05:42:25Z beason $
//
////////////////////////////////////////////////////////////////////////////

#ifndef PANECLASS
#define PANECLASS

#include "GlobalVars.h"
#include "Ray.h"
#include "Triangle.h"
#include "Julia.h"
#include "Sphere.h"
#include "RayAccel.h"
#include "IrradianceCache.h"
#include "Fractal.h"

// forward declarations
class EnvironMap;   
class DataSet;      
class NoiseDataSet;
class Photon_map;

enum {FRONTSIDE, BACKSIDE};

// for containing image plane information
typedef struct CanvasParameters {
  double s;       // focal distance
  double height, width;
  float aperture; // radius of lens
  SbVec3f e;    // eye position
  SbVec3f vup;  // up dir
  SbVec3f w;    // camera orthonormal basis forward
  SbVec3f u;    // camara orthonormal basis sideways
  SbVec3f v;    // camara orthonormal basis up
  int prepass;
  int trace;
  int flags;
  int orthographic;
  int atten;               // for setting up camera inside attenuating material
  SbColor attenColor;
  CanvasParameters() : atten(0), attenColor(0,0,0) {}
} CanvasParameters;

// photon types
const int P_GLOBAL     = 1<<0;
const int P_CAUSTIC    = 1<<1;
const int P_IMPORTANCE = 1<<2;
const int P_VOLUME     = 1<<3;
typedef int photon_t;

// for passing object functions to threads
typedef struct MemberFunction {
  Pane *object;
  void (Pane::*pmf)(thread_context*);
  void (Pane::*pmf2)(thread_context*, float, photon_t);
  float numToFire;
  thread_context *ctx;
  photon_t phot_type;
} MemberFunction;


// Core ray tracing engine class
class Pane {

  public:
  
  static const int maxThreads = 16;  // maximum number of threads
  
  private:

  vector<Triangle> regTris;             // regular triangles
  vector<DMTri> dmTris;                 // displacement mapped triangles
  vector<EmissiveTriangle> litTris;     // emissive triangles
  vector<Julia> julias;                 // scene julia quaternion fractal primitives
  vector<Sphere> spheres;               // scene spheres
  vector<Instance> instances;           // instanced primitives
  bool processTriangles;                // whether to process triangles during
                                        // scene traversal
  vector<SoPointLight *> litPnts;       // list of point luminaires
  vector<int> litSpheres;               // list of sphere luminaires
  vector<Triangle> instanceTris;        // triangles in instances
  vector<Fractal> fractals; 
  
  int numLights;                        // number of lights
  float maxEmisAvg;                     // maximum emissive average
  float totalEmisAvg;                   // total emissive average for all lights (hW)
  SbColor *pix;                         // image ptr, full resolution
                // (0,0) = (X,Y) is bottom left of pixmap, X is horiz
  float *pixDepth;                      // image depth map (<0=matte)
  float *pixWeight;                     // weights of pixels
  SbColor *filtpix;                     // filter pixels for rendering glare
  SbColor *ppmPix;                      // average buffer for PPM
  int ppmPixImages;                     // images accumulated for PPM
  float ppmImgOffset[2];                // pixel shift for PPM
      
  double totalPower;                    // power emitted from lights in W
  unsigned long long firedPhotons;      // number of emitted photons
  unsigned long long max_photons;       // maximum number of photons in each map
  IrradianceCache irradCache;           // irradiance cache
  CanvasParameters canvas;              // image plane information
  float completion[maxThreads];         // thread rendering completion meters
  char completion_stage[maxThreads];    // thread rendering stage indicators
  int completionOrder[maxThreads];      // for storing order of who finishes stages 
  float minApproxDistance;              // minimum length of ray to use approx. sol.
  int rejectedICPoints;                 // number of points rejected for cache
  int toneMapped;                       // true if img tone mapped, false otherwise
  EnvironMap *envMap;                   // environment map for image based lighting
  int sampMapHeight;                    // sample map height
  int sampMapWidth;                     // sample map width
  float *sampMap;                       // sample map pixel values
  static const int Mmax=100, Nmax=300;  // maximum (phi,theta) indirect samples
  int Mmdc, Nmdc;                       // (phi, theta) indirect samples to use
  float sinTxcos2T[Mmax], sinTmsinT[Mmax]; // holds precompute values for mdc
  float vk_X[Nmax], vk_Z[Nmax];            // holds ... for irr. grad.
  unsigned long long renderRaysFired;      // counter for rays fired for rendering
  unsigned long long renderIsects;         // counter for intersections for rendering
  unsigned long long photonRaysFired;      // counter for rays fired for photon tracing
  float lastImageYi;
  float lastImageAvgYi;
  string inputFileComment;                 // Comment from read-in file

  typedef struct sepListItem {
    SoMatPlus   *matPlus;
    Material    *mat;
    SoMFVec3f *verts;     // Rather than keeping track of SoCoordinate3, SoNormal,
    SoMFVec3f *norms;     // SoVertexProperty and what fields override what,
                          // just update field pointer as non-empty fields are
                          // seen.

    SoMaterial  *soMat;        // need these two to track material 
    SoVertexProperty *vProp;
  } sepListItem_t;
  
  list <sepListItem_t *> sepStack; // stack of separators during scene traversalo
  list <Material *> matList;
  SoSeparator *garbageDump;   // place nodes here that should be removed on exit

  // Scene processing stuff
  static SoCallbackAction::Response
  sepPreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  sepPostCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  matPreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  matPlusPreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  spherePreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  spherePostCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  indexedShapePreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  indexedShapePostCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  coord3PreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  normalPreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  vPropPreCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  turnOffNotifyCB(void *data, SoCallbackAction *action, const SoNode *node);
  static SoCallbackAction::Response
  linearFractalPreCB(void *data, SoCallbackAction *action, const SoNode *node);
  

  float sampMapValue(float x, float y); // returns value of sample map
  float jnd(float La);
  void histogramAdjust(int human=0); // Ward's histogram adjustment for tonemapping
  void allocatePhotonMaps();

  // if out is null, memory is allocated.
  SbColor* downSampleImg(SbColor *in, int inX, int inY,
                         SbColor *&out, int outX, int outY, int subPixDim);

  void gammaCorrect(SbColor *buf, int x, int y);

  void convertYxyToRgb(SbColor *buf, int x, int y);
  
  void convolveFFT(SbColor *in, int inX, int inY,
                   SbColor *filter, int filtX, int filtY,
                   SbColor *out, int subPixDim);

  public:

  RayAccel *rayAccel;      // main primitive acceleration structure
  string rayAccelName;
  Primitive *scenePrim;
  DataSet *dataSet1;
  DataSet *dataSet2;
  DataSet *dataSet3;
  DataSet *allDataSets;

  GlobalVars* gv;                           // global variables
  SoPerspectiveCamera *perspectiveCamera;   // camera
  SoPointLight *headLight;                  // camera mounted point light
  Photon_map *importanceMap;                // visual importance map
  Photon_map *gPhotonMap;                   // global photon map 
  Photon_map *cPhotonMap;                   // caustic photon map
  Photon_map *volumeMap;                    // volume photon map
  NoiseDataSet *noiseDS;                    // Perlin noise hyper texture
  bool currentlyRendering;                  // Are we currently rendering?
  bool currentlyShooting;                   // Are we shooting photons?
  bool currentlyRenderingPPMRange;          // Are we currently rendering
                                            // multiple PPM frames?
  bool writeWhenDone;                       // Should we write file when done?
  
  Pane(GlobalVars *globVar = NULL);         // constructor
  ~Pane();                                  // deconstuctor

  
  // flags
  static const int C_APPROX    = 1;
  static const int C_SPECULAR  = 2;
  static const int C_DIRECT    = 4;
  static const int C_EMITTED   = 8;
  static const int C_CAUSTIC   = 16;
  static const int C_MDIFFUSE  = 32;
  static const int C_SIMPLE    = 64;
  static const int C_PREPASS   = 128;
  static const int C_RUSSIANR  = 256;
  static const int C_EXPLICITL = 512;
  static const int C_VOLDIR    = 1024;
  static const int C_VOLINDIR  = 2048;
  static const int C_NORMAL    = C_EMITTED | C_DIRECT | C_SPECULAR | C_CAUSTIC | C_EXPLICITL;
  static const int C_FULL      = C_NORMAL | C_MDIFFUSE;

  // refraction constants
  // From http://www.is.kiruna.se/~cjo/d2i/REFRACTION.INDEX3.html
  static const double N_AIR;  // = 1.0; // refr. index for air, 1.00029;
  static const double N_GLASS;// = 1.7; // refr. index for glass, 1.5-1.7


  
  // returns maximum xyz coordinate of triangle vertex in tree
  double getMaxDist(){
    return rayAccel->getRadius()*2;
  }
  
  // create lock for each pixel row, but
  // space them apart so they don't share the same cacheline
  typedef struct lock_spaced {
    pthread_mutex_t lock;
    char offset[128-sizeof(pthread_mutex_t)];
  } lock_spaced_t;
  lock_spaced_t *pixMutexLocks;
  pthread_mutex_t **pixMutexes;
  pthread_mutex_t completionMutex;
  float mitchellFilter(float r);
  float boxFilter(float r);
  float tentFilter(float r);
  // Sets a pixel in the image
  void setPixWeighted(SbVec2f &pixCrd, const SbColor &color);
  void setPix(int x, int y, const SbColor &c, float depth=0.0f, float weight=1.0f){
    for (int i=0; i<gv->subPixDim; i++){
      for (int j=0; i<gv->subPixDim; i++){
        int idx = (y*gv->subPixDim+j)*gv->imgTotalDim[X] + x*gv->subPixDim+i;
        pix[idx]       = c;
        pixDepth[idx]  = depth;
        pixWeight[idx] = weight;
      }
    }
  }
  void setSubPix(int x, int y, const SbColor &c, float depth=0.0f, float weight=1.0f){
    int idx = y*gv->imgTotalDim[X] + x;
    pix[idx]       = c;
    pixDepth[idx]  = depth;
    pixWeight[idx] = weight;
  }

  SbColor getPix(int x, int y){
    SbColor avg(0,0,0);
    for (int i=0; i<gv->subPixDim; i++){
      for (int j=0; i<gv->subPixDim; i++){
        int idx = (y*gv->subPixDim+j)*gv->imgTotalDim[X] + x*gv->subPixDim+i;
        float weight = pixWeight[idx]?pixWeight[idx]:1.0;
        avg += pix[idx]/weight;
      }
    }
    avg *= gv->invSubPixDim2;
    return avg;
  }
  void setDepth(int x, int y, float depth=0.0f){
    int idx = gv->imgDim[X]*y+x;
    pixDepth[idx] = depth;
  }
  
  int init();                          // initalizes rendering engine
  int readScene();                      // reads scene graph into octtree
  void rescanScene();                  // re-reads scene graph
  void readSampleMap();
  
  // Scene scanning methods
  Material* getCurrentMat(){ return sepStack.front()->mat; }
  SoMFVec3f* getCurrentVerts(){ return sepStack.front()->verts; }
  SoMFVec3f* getCurrentNorms(){ return sepStack.front()->norms; }
  SoMaterial* getCurrentSoMat(){ return sepStack.front()->soMat; }
  SoVertexProperty* getCurrentVProp(){ return sepStack.front()->vProp; }

  // adds triangle to octtree
  static void triCBHandler(void *data, SoCallbackAction *action,
			   const SoPrimitiveVertex *v0,
			   const SoPrimitiveVertex *v1,
			   const SoPrimitiveVertex *v2);

  // adds point light into point light list
  static SoCallbackAction::Response
    pntLtCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node);

  // decreases transparency of objects for interactive viewing
  static SoCallbackAction::Response
    transMatCBHandler(void *data, SoCallbackAction *action,
		      const SoNode *node);

  // stores pointer to most recent SoMaterial node
  static SoCallbackAction::Response
    matBlackCBHandler(void *data, SoCallbackAction *action,
		 const SoNode *node);

  
  // computes refractive direction of ray or signals total internal refl.
  void refract(SbVec3f &ret, int &tir,
               const SbVec3f &d, const SbVec3f &n, const double &nc_nt); // nc/nt
  
  // computes reflected radiance due to point lights
  SbColor pointLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng);

  SbColor sphereLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng, bool simple=false);
  
  // experimental volume scattering effect
  SbColor volPointLightContrib(const Ray &r);
  SbColor volumeIntegrate(Ray &r, const SbColor &c);
  
  // computes reflected radiance due to area lights, in 2D
  SbColor areaLightContrib2D(Ray &r, SbVec3f n, SbVec3f ng);

  // computes reflected radiance due to area lights, sampling every light
  SbColor areaLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng);
  SbColor volAreaImpLightContrib(Ray &r, int samples);

  // computes reflected radiance due to area lights, importance sampling the lights
  SbColor areaImpLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng, int simple=0);

  // precomputes tables needed for multipleDiffuseContrib()
  void precomputeMDCTables();

  // computes reflected radiance due to indirect light
  SbColor multipleDiffuseContrib(Ray &r, SbVec3f &n, SbVec3f &ngl,
				 int simpleCalc, int prepass, int inOrOut);

  // computes (refracted+reflected) radiance due to transparent dielectric
  SbColor refractionContrib(Ray &r, const SbVec3f &n,
                            /* const SbVec3f &ng, const SbVec3f &nl,*/
                            const SbVec3f &ngl, int flags);

  // function pointer to preferred radiance computing member
  SbColor (Pane::*getRadiance)(Ray &r, int flags);

  // photon mapping radiance computation
  SbColor getRadiancePM(Ray &r, int flags=C_NORMAL);

  // path tracing radiance computation
  SbColor getRadiancePT(Ray &r, int flags=C_NORMAL);

  // depth computation
  SbColor getRadianceDepth(Ray &r, int flags=C_NORMAL);


  void applyAttenuation(const Ray &r, const SbVec3f &a, const SbVec3f &b, float t,
                        SbVec3f &color);
  
  void findCamera();                     // set camera pointer
  void configureHeadlight();             // update camera-mounted light in lit pts. list
  void canvasSetup(CanvasParameters &canvas, SoCamera *cam); // sets up canvas camera settings
  void render(int trace=0);              // compute image pixel radiance values
  void renderTile(thread_context *ctx);  // compute image region's pixel values
  void renderVerts(thread_context *ctx); // computes vertex colors
  void renderIv(thread_context *ctx);    // compute per-vertex radiance values
  void renderPPM(int trace=0);           // render several passes for PPM
  void toneMap(SbColor *&pixOut);        // tone maps the image (pix), possibly
                                         // returning result in pixOut
  void writeHDR();
  void readHDR();
  void writeDepth();
  int readGlareFilter(const string &fn, int *fdim);
  void glareFilter(float Yi, float Yw, SbColor *&pixOut); // applies glare, may return 
                                                          // output in pixOut
  string createComment();
  void writeImg();                  // writes image as PNG file
  void writeIv();                   // outputs scene to inventor file
  void writeIvFile(const string &fileName, SoGroup *scene); // writes single .iv file
  void shootPhotons();              // distribute photons throughout scene


  // saving parts of iv files
  typedef struct fileNode_t {
    string name;
    SoGroup *parent;
    SoGroup *copy;
  } fileNode_t;

  list <fileNode_t> fileNodeList;
  void makeEditableChildren(SoNode *scene);
    
  // shoot photons from lights
  void emitPhotons(thread_context *ctx, float numToFire,
		   photon_t phot_type);
  
  // propogate photons through scene
  void bouncePhoton(thread_context *ctx, Ray photon, SbColor emission,
		    photon_t phot_type=P_GLOBAL, float maxPhotonComponent = 1.0,
		    int light_id=1);

  // wrapper function, glue for threading
  static void *renderTileWrapper(void *arg){
    
    MemberFunction *memberFunc = static_cast <MemberFunction*> (arg);
    Pane *pane = memberFunc->object;
    void (Pane::*pmf)(thread_context*) = memberFunc->pmf;
    (pane->*pmf)(memberFunc->ctx);
    return(NULL);
  }

  // wrapper function, glue for threading
  static void *emitPhotonsWrapper(void *arg){
    
    MemberFunction *memberFunc = static_cast <MemberFunction*> (arg);
    Pane *pane = memberFunc->object;
    void (Pane::*pmf)(thread_context*, float numToFire, photon_t) =
      memberFunc->pmf2;
    (pane->*pmf)(memberFunc->ctx, memberFunc->numToFire, memberFunc->phot_type);
    return(NULL);
  }

};

extern const char* getVersion();     // returns source code version
extern string getNameVersionDate();  // returns app name + version + build date
extern string getAuthorName();
extern string getAuthorUrl();
extern string getCurrentTime();
extern int getCurrentYear();
extern bool getIsAuthor();

#endif /* PANECLASS */
