#include "debug.h"
#include "Pane.h"
#include "DataSet.h"
#include "EnvironMap.h"
#include "Halton.h"
#include "Spectrum.h"
#include "Photon_map.h"

using namespace std;

void Pane::allocatePhotonMaps(){

  if (!gPhotonMap){ 
    gPhotonMap    = new Photon_map(max_photons, gv);
    gPhotonMap->compress_sphere(gv->compressEst); // shrink sphere volume into disc
  }
  if (!cPhotonMap) {
    cPhotonMap    = new Photon_map(max_photons, gv);
    cPhotonMap->compress_sphere(gv->compressEst);
  }
  if (!importanceMap) {
    importanceMap = new Photon_map(gv->numIPhotons*2, gv);
    importanceMap->compress_sphere(gv->compressEst);
  }
  if (!volumeMap)
    volumeMap     = new Photon_map(gv->numVPhotons*2, gv);

}

//''''''WW''''''''''''''''WW''WWWW''WW''''''''''WW''''''''''''''''''''WW'WW'
//      WW                WW  WW  W WW          WW                   WW   WW
// WWWW WWW   WW    WW   WWWW WW  W WWW   WW   WWWW  WW   WWW   WWWW WW   WW
//WWW   WW W WW W  WW W   WW  WWWW  WW W WW W   WW  WW W  WW W WWW   WW   WW
//  WWW WW W WW W  WW W   WW  WW    WW W WW W   WW  WW W  WW W   WWW WW   WW
//WWWW  WW W  WW    WW    WW  WW    WW W  WW    WW   WW   WW W WWWW   WW WW 
//                                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Starts threads to shoot photons from the lights into the scene.
// Each thread calls emitPhotons(), which in turn calls bouncePhoton().
//
void Pane::shootPhotons(){

  if (!gv->sceneHasLights){
    cerr << "WARNING: Scene has no lights: not shooting any photons.\n";
    return;
  }

  currentlyShooting = true;
  
  allocatePhotonMaps();

  // ppm setup
  if (gv->ppmIter>=0){
    RNG mydrand48(gv->ppmIter);
    ppmImgOffset[0] = mydrand48();
    ppmImgOffset[1] = mydrand48();
  }
  
  int logNumToFire = -1;
  float numToFire;
  photon_t phot_type = P_GLOBAL;
  MemberFunction *memberFunc = (MemberFunction *)alloca(
    sizeof(MemberFunction)*gv->maxThreads); // used to pass object pointer
  thread_context *ctx = (thread_context *)alloca(
    sizeof(thread_context)*gv->maxThreads);   // random number generators
  
  // start timer
  struct timeval start, end;
  cerr << "Shooting photons...\n" << flush;
  gettimeofday(&start, NULL); // start delay measure

  // finish calculating totalEmisAvg. (scene sum of average(RGB) emission))
  // adjust max emissive avg for environment map
  if (envMap){ // Phi = L * A * cosTheta * Omega
    float receiverArea = M_PI*gv->meters2(rayAccel->getRadius()*rayAccel->getRadius());
    SbColor emisFlux = envMap->getEmissiveFlux(receiverArea);
    float emisAvg = vecAvg(emisFlux);
    totalEmisAvg += emisAvg;
    if (emisAvg > maxEmisAvg)
      maxEmisAvg = emisAvg;
  }

  // adjust max emissive avg for point lights
  vector<SoPointLight *>::iterator pI = litPnts.begin();
    
  if (gv->headLightPos == HL_OFF)
    pI++;
      
  for (; pI != litPnts.end(); pI++){

    // get emission
    SbColor emission = (*pI)->color.getValue()*(*pI)->intensity.getValue();
    float pntEmisAvg = vecAvg(emission);

    totalEmisAvg += pntEmisAvg;
    if (pntEmisAvg > maxEmisAvg)
      maxEmisAvg = pntEmisAvg;
  }

  // adjust max emissive avg for sphere lights
  for (int pI=0; pI<litSpheres.size(); pI++){
    float sphEmisAvg = spheres[litSpheres[pI]].getMat()->getEmisAvg();
    totalEmisAvg += sphEmisAvg;
    if (sphEmisAvg > maxEmisAvg)
      maxEmisAvg = sphEmisAvg;
  }
    
  
  // position headlight
  findCamera();
  configureHeadlight();

  // generate projection maps
  if (gv->useProjectionMaps){
    for (int ai=0; ai<litTris.size(); ai++){
      EmissiveTriangle &tri = litTris[ai];
      SbVec3f n = tri.getNormalPre();

      if (gv->sceneHasShininess)
        tri.genImpMap(rayAccel->getCenterShiny(),
                      rayAccel->getRadiusShiny());
    }
  }
  
  // set up canvas
  canvasSetup(canvas, gv->cam);
  
  // pack object pointer, function pointer, and argument into struct
  for (int i=0; i<gv->numThreads; i++){
    memberFunc[i].object = this;
    memberFunc[i].pmf2 = &Pane::emitPhotons;
    memberFunc[i].ctx = &ctx[i];
    ctx[i].mydrand48 = RNG(mydrand48.lrand48());
    ctx[i].id = i;
    ctx[i].rays = 0;
    ctx[i].ticks = 0;
  }

  // dont bother with caustic photons if rendering an approximation
  if (gv->approximate)
    gv->numCPhotons = 0;
  else if (gv->algorithm != 1 ||
           gv->renderFlags && (gv->renderFlags & C_MDIFFUSE) == 0){
    gv->numPhotons = 0;
  }

  int importance_done = gv->numIPhotons == 0;
  int global_done = 0;
  int caustic_done = 0;
  while (gPhotonMap->get_stored_photons()   <gv->numPhotons ||
         cPhotonMap->get_stored_photons()   <gv->numCPhotons ||
	 importanceMap->get_stored_photons()<gv->numIPhotons ||
	 volumeMap->get_stored_photons()    <gv->numVPhotons
    ){

    if (!importance_done &&
        importanceMap->get_stored_photons()>=gv->numIPhotons){
      cerr << "\n" << importanceMap->get_stored_photons()
	   << " photons stored in importance map.\n";
      cerr << "Importance map done. Continuing to other maps...\n";
      logNumToFire = -1;

      // scale photon power
      float proj_area = M_PI * gv->meters2(rayAccel->getRadius()*
					   rayAccel->getRadius());      
      float scale = importanceMap->get_stored_photons() ?
	proj_area/importanceMap->get_stored_photons() : 1;
      importanceMap->scale_photon_power(scale);

      // balance
      cerr << "Balancing importance photon map...";
      TIME(importanceMap->balance(gv->iPhotonEstimate));
      
      importance_done = 1;
    }
    
    if (!global_done && gPhotonMap->get_stored_photons()>=gv->numPhotons){
      cerr << "\n" << gPhotonMap->get_stored_photons()
	   << " photons stored in global map.\n";
      cerr << "Global map done. Continuing to build caustic map...\n";
      global_done = 1;
    }

    if (global_done &&
	!caustic_done && cPhotonMap->get_stored_photons()>=gv->numCPhotons){
      cerr << "\n" << cPhotonMap->get_stored_photons()
	   << " photons stored in caustic map.\n";
      cerr << "Caustic map done. Continuing to build volume map...\n";
      caustic_done = 1;
    }
    
    if (!importance_done)
      phot_type = P_IMPORTANCE;
    else if (!caustic_done){
      
      // cut power in map in half
      cPhotonMap->scale_photon_power(0.5);
      
      if (global_done)
	phot_type = P_CAUSTIC;
      else {
	phot_type = P_GLOBAL;
	gPhotonMap->scale_photon_power(0.5);
      }
      totalPower /= 2.0;
    }
    else
      phot_type = P_VOLUME;

    volumeMap->scale_photon_power(0.5);
    

    logNumToFire++;
    numToFire = powf(2.0f, (float)logNumToFire);
    //cerr << "numToFire=" << numToFire << "\n";

    if (gv->numThreads == 1){
      emitPhotons(&ctx[0], numToFire, phot_type);
    }
    else {
    
      // launch threads
      for (int i=0; i<gv->numThreads; i++){

	// call wrapper function with struct
	memberFunc[i].numToFire = numToFire;
	memberFunc[i].phot_type = phot_type;
	pthread_create(&gv->threads[i], NULL, Pane::emitPhotonsWrapper,
		       (void*)&memberFunc[i]);
      }

      // wait for threads to finish and flush buffers
      for (int i=0; i<gv->numThreads; i++){
	pthread_join(gv->threads[i], NULL);
      }
    }

    // flush buffered stores
    for (int i=0; i<gv->numThreads; i++){
      photonRaysFired += ctx[i].rays;
      gv->times.totalTicks += ctx[i].ticks;

      cPhotonMap->   buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
      gPhotonMap->   buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
      importanceMap->buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
      volumeMap->    buffered_store(NULL, NULL, NULL, NULL, 0, 0, i, 1);
    }


  }

  gettimeofday(&end, NULL);   // end delay measure
  gv->times.photonShooting = delay(start, end);
  cerr << "\n" << firedPhotons << " total photons fired, for "
       << setiosflags(ios::showpoint | ios::fixed) << setprecision(5)
       << totalPower << "W (" << delay(start, end) << "ms)" << endl;
  cerr << gPhotonMap->get_stored_photons()
       << " photons stored in global map.\n";
  cerr << cPhotonMap->get_stored_photons()
       << " photons stored in caustic map.\n";
  cerr << volumeMap->get_stored_photons()
       << " photons stored in volume map.\n";
  if (gPhotonMap->get_stored_photons()>=max_photons)
    cerr << " WARNING: global photon map full, some photons probably not stored!\n";
  if (volumeMap->get_stored_photons()>=2*gv->numVPhotons && gv->numVPhotons>0)
    cerr << " WARNING: volume photon map full, some photons probably not stored!\n";
  cerr << flush;
  resetios(cerr);

  currentlyShooting = false;
}

//'''''''''''''WW''WW''WWWW''WW''''''''''WW''''''''''''''''''
//                 WW  WW  W WW          WW                  
// WW  WWWW W  WW WWWW WW  W WWW   WW   WWWW  WW   WWW   WWWW
//WWWW WW WW W WW  WW  WWWW  WW W WW W   WW  WW W  WW W WWW  
//WW   WW WW W WW  WW  WW    WW W WW W   WW  WW W  WW W   WWW
// WWW WW WW W WW  WW  WW    WW W  WW    WW   WW   WW W WWWW 
//                                                           
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Loops over lights, shooting out photons from each.
// Uses an advanced Quasi-Monte-Carlo technique called
// Trajectory Splitting, described by Kollig and Keller
// in their paper Efficient Multidimensional Sampling.
//
void Pane::emitPhotons(thread_context *ctx, float numToFire,
		       photon_t phot_type){
  
  unsigned long long addNum;
  double cA, cB, cC;
  SbVec3f x;
  SbVec3f hemiDir, diffuseDir;
  SbVec3f bc;
  double phi, theta, rho;
  //MultiHalton h;
  //sobinit(2);
  //double eps[4];


  //WWW'''''''''''''''''''''''''''''''''''
  // W                     W              
  // W  WW W  WW   W   WW WWW  W  WW   WWW
  // W  W W W W W W W W    W  W W W W  WW 
  //WWW W W W WW   W  W    W   W  W W WWW 
  //''''''''''W'''''''''''''''''''''''''''
  //          W                           

  if (phot_type == P_IMPORTANCE){

    Ray photon;
    photon.ctx = ctx;
    photon.pos = canvas.e;
    int sqrt_phots = (int)sqrt((float)numToFire);
    int total_phots = sqrt_phots*sqrt_phots;

    // one thread only calculations
    if (ctx->id == 0){

      firedPhotons += total_phots;
      
      // status report
      cerr << "\r" << total_phots
	   << " firing, " << importanceMap->get_stored_photons()
	   << " stored in importance map."
	   << flush;
    }

    for (int j=0+ctx->id; j<sqrt_phots; j+=gv->numThreads){

      for (int i=0; i<sqrt_phots; i++){

	double eps1, eps2;
	int n = j*sqrt_phots+i;
	eps1 = (double)n/total_phots;
	eps2 = RI_LP(n);
	
	// create Ray
	photon.dir = SbVec3f(

	  // sideways component
	  ( -canvas.width/2.0 + eps1 * canvas.width ) * canvas.u +

	  // vertical component
	  ( -canvas.height/2.0 + eps2 * canvas.height ) * canvas.v +

	  // distance ahead component
	  -canvas.s*canvas.w  
	  );

	// adjust for orthographic camera
	if (canvas.orthographic){
	  photon.pos = canvas.e + photon.dir + canvas.s*canvas.w;
	  photon.dir = -canvas.w;
	}
	photon.dir.normalize();

	// send photon
        SbColor emission(1.0, 1.0, 1.0);
	bouncePhoton(ctx, photon, emission, phot_type, 1.0, 1);

      } // end for (i=...
    } // end for (j=...

    if (gv->numThreads > 1)
      pthread_exit(0);
    return;
  }
  
  float maxPhotonComponent = 0.0;
  int raysPerPoint = 4;

  
    //'WW''''''''''''''''''WW''m'''''W'''''''''''
    //W  W                  W        W    W      
    //WWWW  WW  WW   WW     W  W  WW WW  WWW  WWW
    //W  W W   Wmmm W W     W  W W W W W  W   WW 
    //W  W W    WW   WW    WWW W  WW W W  W  WWW 
    //'''''''''''''''''''''''''''''W'''''''''''''
    //                           WW              
    for (int ai=0; ai<litTris.size(); ai++){

      EmissiveTriangle *tri = &litTris[ai];

      // calculate average emission (power) for R,G,B
      double emisAvg = tri->getEmisAvg();
      if (emisAvg == 0.0)
	continue;

      // photon power
      SbColor emission = tri->getEmission();
      
      // emisAvg is in W
      // totalPower is in W
      float approxNum = emisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                numToFire;

      unsigned long long points = (unsigned long long)(approxNum/raysPerPoint+0.5);
      addNum = points*raysPerPoint;
      if (!addNum) continue;
      float scale = 1.0/(addNum*2.0);

      unsigned int index_p, index_r;

      // choose emission point
      emission *= scale; // scale photon power now

      // one thread only calculations
      if (ctx->id == 0){

	firedPhotons += addNum;
	totalPower += addNum*emisAvg*3.0*scale;

	// cause caller to stop shooting photons / calling me
	// after threshold reached
	if (addNum > gv->numPhotonsFireMax){
	  gv->numPhotons = gPhotonMap->get_stored_photons();
	  gv->numCPhotons = cPhotonMap->get_stored_photons();
	  gv->numIPhotons = importanceMap->get_stored_photons();
	  gv->numVPhotons = volumeMap->get_stored_photons();
	}
	
	// status report
	cerr << "\r" << addNum
	     << " firing, "
	     << gPhotonMap->get_stored_photons() << "g/"
	     << cPhotonMap->get_stored_photons() << "c/"
	     << volumeMap->get_stored_photons()  << "v "
	     << "stored          " << flush;
      }

      // photon power
      for (int i=0; i<3; i++)
	if (emission[i] > maxPhotonComponent)
	  maxPhotonComponent = emission[i];

      // emit photon diffusely
      SbVec3f ng = tri->getNormalPre();
      ONB &onb = tri->onb;
      
      // targetNum 1..16
      // points (8+) 1..8
      // real 8*4..16*4
      for (unsigned long long n=ctx->id; n<points; n+=gv->numThreads){

	// calculate index into quasi random sequence.
	// we have already fired addNum.
	// we are now shooting the second half.
	// therefore, the index starts at addNum.
	index_p = addNum+n;
    
	double eps1, eps2;
	//eps1 = h.next(0); eps2 = h.next(1);
	//eps1 = eps[0]; eps2 = eps[1];
	//eps1 = 0.5; eps2 = 0.5;
	if (gv->quasiRandom==0){
	  eps1 = ctx->mydrand48();
	  eps2 = ctx->mydrand48();
	}
	else {
	  eps1 = (double)n/points;//points==1?0.5:(double)n/(points-1);
	  eps2 = RI_LP(index_p);

          // shift position for ppm
          // otherwise we will get same position every pass
          //
          // Is this adequate?
          if (gv->ppmIter>=0){
            eps1 += ppmImgOffset[0];
            if (eps1>=1.f) eps1 -= 1.f;
            eps2 += ppmImgOffset[1];
            if (eps2>=1.f) eps2 -= 1.f;
          }
          
	  //eps1 = RI_vdC(n, r1);
	  //eps2 = RI_S(n, r2);
	}
	
	// choose random position inside light
	cA = 1.0-sqrt(eps1);
	cB = (1.0-cA)*eps2;
	cC = 1.0-cA-cB;


// 	// do a texture lookup
// 	SbColor diffCol = (*aI)->getIntDiffColor(SbVec3f(cA, cB, cC));
// 	emission[0] = origEmission[0] * diffCol[0];
// 	emission[1] = origEmission[1] * diffCol[1];
// 	emission[2] = origEmission[2] * diffCol[2];
// 	if (emission[0]<0.000001)
// 	  continue;

	
	// 2d adjustment
	if (gv->useDimensions<3 && ng[1]<1.0){
	  ng[1] = 0.0;
	  ng.normalize();
	}

	// create randomized realization of (t,m,s)-net using
	// random digit scrambling
	// (Efficient Multidimensional Sampling - Kollig and Keller)
 	unsigned int r3 = (unsigned int)(ctx->mydrand48()*double(UINT_MAX));
 	unsigned int r4 = (unsigned int)(ctx->mydrand48()*double(UINT_MAX));
        // The mult above needs to be in double precision because
        // (int)(float)INT_MAX => INT_MAX+1

	for (int m=0; m<raysPerPoint; m++){
	  // choose random value
	  double eps1, eps2;
	  //eps1 = h.next(2); eps2 = h.next(3);
	  //eps1 = h.next(0); eps2 = h.next(1);
	  //sobvect(2, eps);
	  //eps1 = eps[0]; eps2 = eps[1];
	  //eps1 = (double)n/maxPhots;
	  //eps2 = RI_vdC(n);
	  //eps1 = actualPhotons==1?0.5:(double)n/(actualPhotons-1);
	  //eps2 = RI_LP(maxPhots/2+n);
	  //eps1 = RI_vdC(maxPhots/2+n, r2);
	  //eps2 = RI_S(maxPhots/2+n, r2);
	  index_r = raysPerPoint*index_p+m;
	
	  if (gv->quasiRandom==0){
	    eps1 = ctx->mydrand48();
	    eps2 = ctx->mydrand48();
	  }
	  else {
	    eps1 = RI_vdC(index_r, r3);
	    eps2 = RI_S(index_r, r4);
	  }
	  //eps1 = eps2 = 1.0;

          // shift direction for ppm
          // otherwise we will get same photon distribution every pass
          //
          // Is this adequate?
          if (gv->ppmIter>=0){
            eps1 += ppmImgOffset[0];
            if (eps1>=1.f) eps1 -= 1.f;
            eps2 += ppmImgOffset[1];
            if (eps2>=1.f) eps2 -= 1.f;
          }
          
	  // sample with projection map
	  float pdf = 1.0f;
	  if (phot_type == P_CAUSTIC && gv->useProjectionMaps && tri->UseMap()){

	    tri->sampleDir(diffuseDir, pdf, eps1, eps2);
	    if (pdf<=0){
              cerr << "ERROR: " << __FILE__ << ":" <<  __LINE__ << ": pdf 0!\n";
              continue;
            }

	  }
	  else {

            // phong dist from RRT pg. 98
            static float lightExp = getenv("PANE_LIGHT_EXP") ? atof(getenv("PANE_LIGHT_EXP")) : 1;
	    phi = eps1 * 2.0 * M_PI;   // rotation about normal
	    theta = acos(powf(eps2, 1/(lightExp+1)));  // angle from normal
	    hemiDir.setValue(sin(theta)*cos(phi),
			     sin(theta)*sin(phi),
                             cos(theta));
    
            diffuseDir = onb.toWorldSpace(hemiDir);
	  }
	  
	  // 2d adjustment
	  if (gv->useDimensions<3){
	    SbVec3f ng_perp = SbVec3f(ng[2], 0.0, -ng[0]);
	    theta = asin(2.0*eps1-1.0);
	    diffuseDir = ng*cos(theta) + ng_perp*sin(theta);
	  }
	  
	  diffuseDir.normalize();
	  x = cA*tri->vert(0) + cB*tri->vert(1) + cC*tri->vert(2);

	  // emit photon
	  Ray photon(x, diffuseDir, ctx, tri);
          emission /= pdf;
	  bouncePhoton(ctx, photon, emission, phot_type,
		       maxPhotonComponent, 1);

	} // end for (int m=0; m<rays...
      } // end for (n=0; n<pnts...

    } // end for (; aI != litTris.end(); aI++){    light loop



    //WWW''''''m''''''''''''WW''m'''''W'''''''''''
    //W  W            W      W        W    W      
    //WWW   W  W WW  WWW     W  W  WW WW  WWW  WWW
    //W    W W W W W  W      W  W W W W W  W   WW 
    //W     W  W W W  W     WWW W  WW W W  W  WWW 
    //'''''''''''''''''''''''''''''W''''''''''''''
    //                            WW
    vector<SoPointLight *>::iterator pI = litPnts.begin();
    
    if (gv->headLightPos == HL_OFF)
      pI++;
      
    for (; pI != litPnts.end(); pI++){

      // get emission
      SbColor emission = (*pI)->color.getValue()*(*pI)->intensity.getValue();
      float pntEmisAvg = vecAvg(emission);

      float approxNum = pntEmisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                numToFire;

      addNum = (unsigned long long)(approxNum+0.5);
      if (!addNum) continue;
      float scale = 1.0/(addNum*2.0);
      
      firedPhotons += addNum;
      totalPower += addNum*pntEmisAvg*3.0*scale;

      emission *= scale;
      
      // one thread only calculations
      if (ctx->id == 0){
      
	firedPhotons += addNum;
	totalPower += addNum*pntEmisAvg*3.0*scale;

	// cause caller to stop shooting photons / calling me
	// after threshold reached
	if (addNum >= gv->numPhotonsFireMax){
	  gv->numPhotons = gPhotonMap->get_stored_photons();
	  gv->numCPhotons = cPhotonMap->get_stored_photons();
	  gv->numIPhotons = importanceMap->get_stored_photons();
	  gv->numVPhotons = volumeMap->get_stored_photons();
	}
	
	// status report
	cerr << "\r" << addNum
	     << " firing, "
	     << gPhotonMap->get_stored_photons() << "g/"
	     << cPhotonMap->get_stored_photons() << "c/"
	     << volumeMap->get_stored_photons()  << "v "
	     << "stored          " << flush;
      
      }
      
      // photon power
      for (int i=0; i<3; i++)
	if (emission[i] > maxPhotonComponent)
	  maxPhotonComponent = emission[i];
    
      // compute light location
      x = (*pI)->location.getValue();
      
      // emit numPhots photons
      for (unsigned long long n=ctx->id; n<addNum; n+=gv->numThreads){

	// find emission point with rejection sampling
	float length2;
	do {
	  phi   = ctx->mydrand48()*2.0-1.0;
	  theta = ctx->mydrand48()*2.0-1.0;
	  rho   = ctx->mydrand48()*2.0-1.0;
	  length2 = (phi*phi+theta*theta+rho*rho);
	} while (length2>1.0);
	diffuseDir.setValue(phi, theta, rho);
	diffuseDir /= sqrt(length2);

	// emit photon
	Ray photon(x, diffuseDir, ctx);
      
	// global photon
	bouncePhoton(ctx, photon, emission, phot_type, maxPhotonComponent);

      } // end for (n=0; n<numPhots; n++){   photon loop

    } // end for (; pI != litPnts.end(); aI++){    light loop



    // Spheres
    for (int pI=0; pI<litSpheres.size(); pI++){

      // get emission
      Sphere &sph = spheres[litSpheres[pI]];
      SbColor emission = sph.getMat()->getEmission();
      float sphEmisAvg = sph.getMat()->getEmisAvg();

      float approxNum = sphEmisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                numToFire;

      addNum = (unsigned long long)(approxNum+0.5);
      if (!addNum) continue;
      float scale = 1.0/(addNum*2.0);
      
      firedPhotons += addNum;
      totalPower += addNum*sphEmisAvg*3.0*scale;

      emission *= scale;
      
      // one thread only calculations
      if (ctx->id == 0){
      
	firedPhotons += addNum;
	totalPower += addNum*sphEmisAvg*3.0*scale;

	// cause caller to stop shooting photons / calling me
	// after threshold reached
	if (addNum >= gv->numPhotonsFireMax){
	  gv->numPhotons = gPhotonMap->get_stored_photons();
	  gv->numCPhotons = cPhotonMap->get_stored_photons();
	  gv->numIPhotons = importanceMap->get_stored_photons();
	  gv->numVPhotons = volumeMap->get_stored_photons();
	}
	
	// status report
	cerr << "\r" << addNum
	     << " firing, "
	     << gPhotonMap->get_stored_photons() << "g/"
	     << cPhotonMap->get_stored_photons() << "c/"
	     << volumeMap->get_stored_photons()  << "v "
	     << "stored          " << flush;
      
      }
      
      // photon power
      for (int i=0; i<3; i++)
	if (emission[i] > maxPhotonComponent)
	  maxPhotonComponent = emission[i];
          
      // emit numPhots photons
      for (unsigned long long n=ctx->id; n<addNum; n+=gv->numThreads){

        // compute light location and position
        SbColor emissionTmp;
        float scale=1;
        if (gv->useProjectionMaps){
          sph.getPhotonImpSample(ctx->mydrand48(), ctx->mydrand48(),
                                 ctx->mydrand48(), ctx->mydrand48(),
                                 rayAccel, phot_type == P_CAUSTIC,
                                 x, diffuseDir, emissionTmp, scale);
          if (scale<=0) continue;
        } else {
          sph.getPhoton(ctx->mydrand48(), ctx->mydrand48(),
                        ctx->mydrand48(), ctx->mydrand48(),
                        x, diffuseDir, emissionTmp);
        }

	// emit photon
	Ray photon(x, diffuseDir, ctx);

	// global photon
	bouncePhoton(ctx, photon, emission*scale,
                     phot_type, maxPhotonComponent);

      } // end for (n=0; n<numPhots; n++){   photon loop

    } // end for (int pI=0; pI<litSpheres.size(); aI++){    light loop
    


    //WWWW'''''''''W''''W''''''''''''WW''m'''''W'''''''''''
    //W            WW  WW             W        W    W      
    //WWW  WW  W W W WW W  WW WW      W  W  WW WW  WWW  WWW
    //W    W W W W W    W W W W W     W  W W W W W  W   WW 
    //WWWW W W  W  W    W  WW WW     WWW W  WW W W  W  WWW 
    //''''''''''''''''''''''''W''''''''''''''W'''''''''''''
    //                        W            WW              

    // calculate average emission (power)
    if (envMap){

      SbVec3f center;
      double radius, proj_area;

      if (phot_type == P_CAUSTIC){
	center = rayAccel->getCenterShiny();
	radius = rayAccel->getRadiusShiny();
      }
      else {
	center = rayAccel->getCenter();
	radius = rayAccel->getRadius();
      }
      proj_area = M_PI * gv->meters2(radius*radius);

      SbColor emisFlux = envMap->getEmissiveFlux(proj_area);
      double emisAvg = (emisFlux[0] + emisFlux[1] + emisFlux[2])/3.0;
    
      if (emisAvg > 0.0){

	// emisAvg is in W
	// totalPower is in W
	float approxNum = emisAvg/totalEmisAvg *    // ~emisAvg but in [0,1]
	                  numToFire;

	unsigned long long points = (unsigned long long)(approxNum/raysPerPoint+0.5);
	addNum = points*raysPerPoint;
	if (!addNum) goto end_envMap;
	float scale = 1.0/(addNum*2.0);
	
	unsigned int index_p, index_r;

	// one thread only calculations
	if (ctx->id == 0){

	  firedPhotons += addNum;
	  totalPower += addNum*emisAvg*3.0*scale;

	  // cause caller to stop shooting photons / calling me
	  // after threshold reached
	  if (addNum > gv->numPhotonsFireMax){
	    gv->numPhotons = gPhotonMap->get_stored_photons();
	    gv->numCPhotons = cPhotonMap->get_stored_photons();
	    gv->numIPhotons = importanceMap->get_stored_photons();
	    gv->numVPhotons = volumeMap->get_stored_photons();
	  }
	
	  // status report
	  cerr << "\r" << addNum
	       << " firing, "
	       << gPhotonMap->get_stored_photons() << "g/"
	       << cPhotonMap->get_stored_photons() << "c/"
	       << volumeMap->get_stored_photons()  << "v "
	       << "stored          " << flush;

	}


	// choose points spots
	for (unsigned long long n=ctx->id; n<points; n+=gv->numThreads){

	  // calculate index into quasi random sequence.
	  // we have already fired half of addNum.
	  // we are now shooting the second half.
	  // therefore, the index starts at addNum.
	  index_p = points+n;
	  
	  double eps1, eps2;
	  if (gv->quasiRandom==0){
	    eps1 = ctx->mydrand48();
	    eps2 = ctx->mydrand48();
	  }
	  else {
	    eps1 = (double)n/points;//points==1?0.5:(double)n/(points-1);
	    eps2 = RI_LP(index_p);
	  }


	  // choose emission point
	  SbColor emission;
	  float pdf = envMap->getSample(eps1, eps2, hemiDir, emission);
          if (!pdf) continue;
	  emission *= proj_area/pdf;
	  emission *= scale; // scale photon power now

	  // adjust maximum photon emission component
	  for (int i=0; i<3; i++)
	    if (emission[i] > maxPhotonComponent)
	      maxPhotonComponent = emission[i];
	  
	  hemiDir = -hemiDir; // flip direction around

	  // create randomized realization of (t,m,s)-net using
	  // random digit scrambling
	  // (Efficient Multidimensional Sampling - Kollig and Keller)
	  unsigned int r3 = (unsigned int)(ctx->mydrand48()*double(UINT_MAX));
	  unsigned int r4 = (unsigned int)(ctx->mydrand48()*double(UINT_MAX));
          // The mult above needs to be in double precision because
          // (int)(float)INT_MAX => INT_MAX+1

	  // chose points
	  for (int m=0; m<raysPerPoint; m++){
	    
	    // choose random value
	    double eps1, eps2;
	    index_r = raysPerPoint*index_p+m;
	
	    if (gv->quasiRandom==0){
	      eps1 = ctx->mydrand48();
	      eps2 = ctx->mydrand48();
	    }
	    else {
	      eps1 = RI_vdC(index_r, r3);
	      eps2 = RI_S(index_r, r4);
	    }


	    // choose point on unit disc
	    double r = sqrt(eps1);
	    theta = 2.0*M_PI*eps2;

	    // scale to radius of bounding sphere
	    r *= radius;

	    x[X] = r*cos(theta);
	    x[Y] = r*sin(theta);
	    x[Z] = 0.0;

	    // rotate disc normal to l
            ONB onb(hemiDir);
            x = onb.toWorldSpace(x);

	    // position outside bounding sphere
	    x += center;
	    x -= hemiDir*radius;
	  
	    // emit photon
	    Ray photon(x, hemiDir, ctx);
	    bouncePhoton(ctx, photon, emission, phot_type, maxPhotonComponent, 1);

	  } // end for (int m=0; m<rays...
	} // end for (n=0; n<pnts...
	
      }
    }
  end_envMap:


    // volume lights
    if (dataSet2){

      // emisAvg is in W
      // totalPower is in W
      float approxNum =
	(dataSet2->totalBrightness*gv->tempFloat3) /
	totalEmisAvg *    // ~emisAvg but in [0,1]
	numToFire;

      // debugging
      //approxNum=500;
      if (!approxNum) goto end_volLight;

	  // status report
	  cerr << "\r" << approxNum
	       << " firing (volume emission), "
	       << gPhotonMap->get_stored_photons() << "g/"
	       << cPhotonMap->get_stored_photons() << "c/"
	       << volumeMap->get_stored_photons()  << "v "
	       << "stored          " << flush;
      
      
      for (unsigned long long i=0; i<approxNum; i++){

	// pick random point
	SbVec3f x = dataSet2->impSample(ctx->mydrand48(),
					ctx->mydrand48(),
					ctx->mydrand48(),
					ctx->mydrand48());

	// get blackbody color
	float temp = dataSet2->val(x)*gv->tempFloat;
	SbColor emission = Spectrum(temp).getRGB()*(1.0f/(4.0f*M_PI));
	emission *= gv->tempFloat2*gv->tempFloat3;
	//emission *= dataSet2->totalBrightness
	//  25.f/brightness(emission);

	// pick random direction on sphere
	double eps1 = ctx->mydrand48();
	double eps2 = ctx->mydrand48();
	phi = eps1 * 2.0 * M_PI;   // rotation about normal
	float z = 2.0*eps2-1.0;
	theta = acos(z);           // angle from normal
	SbVec3f sphereDir(sin(theta)*cos(phi),
			  z, // cos(theta)
			  sin(theta)*sin(phi));
	sphereDir.normalize();

	// debugging
	//showPhoton(x, SbColor(1.0, 1.0, 1.0));

	Ray photon(x, sphereDir, ctx);
	bouncePhoton(ctx, photon, emission, phot_type, maxPhotonComponent, 1);

      }
    }
  end_volLight:

    if (gv->numThreads == 1)
      return;
    
    // time statistics
    struct tms renderTimes;
    times(&renderTimes);
    ctx->ticks += renderTimes.tms_utime + renderTimes.tms_cutime;
    
    pthread_exit(0);
}


//WW'''''''''''''''''''''''''''''WWWW''WW''''''''''WW''''''''''''''WW'WW'
//WW                             WW  W WW          WW             WW   WW
//WWW   WW   W WW WWW   WWW  WW  WW  W WWW   WW   WWWW  WW   WWW  WW   WW
//WW W WW W  W WW WW W WW W WWWW WWWW  WW W WW W   WW  WW W  WW W WW   WW
//WW W WW W  W WW WW W WW   WW   WW    WW W WW W   WW  WW W  WW W WW   WW
//WWW   WW    WWW WW W  WWW  WWW WW    WW W  WW    WW   WW   WW W  WW WW 
//                                                                       
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// bounce photon around until it absorbs, disappears, gets lost, etc.
//
// Light paths and the appropriate map:
// global map (D+S)*D
// caustic map: LS+D
//
void Pane::bouncePhoton(thread_context *ctx,
			Ray photon, SbColor emission, photon_t phot_type,
			float maxPhotonComponent,
			int light_id){

  // for tracking specular-ness of photons for caustic map gen.      
  int hit_specular = 0;
  int hit_diffuse = 0;
  int bounce = -1;
  double phi, theta;
  SbVec3f hemiDir, diffuseDir;

  static bool spectralPhot = getenv("PANE_SPECTRAL_PHOTONS");
  if (gv->spectral || spectralPhot){
    SbColor mult;
    Spectrum::getSpectralMult(ctx->mydrand48(), mult, photon.wavelen);
    vecScaleItself(emission, mult);
  }


//   Reasoning
//   ==========
//
//
//   OLD WAY - caustics contains L(S|D)*SD photons
//   ---------------------------------------------------------------------
//
//   Photons that hit a diffuse after a specular surface are stored
//   in the caustic map.
//
//   Because of this, when doing a final gather for indirect illumination,
//   gather rays should not follow specular bounces because this light
//   contribution will come from the caustic photon map.
//
//   Similarly, when importance sampling for indirect illumination,
//   we would like to find areas where indirect and non-specular lighting
//   is coming from, in other words, areas that contain diffuse surfaces.
//
//   Therefore, we store with a photon whether or not it came from a
//   diffuse surface, or in other words, from an area we might want
//   to sample from in the final gather.
//
//   We also store the light_id if it's directly from a light.
//   Currently this is unused, but it may be used in the future
//   for some type of direct lighting test.
//
//
//   NEW WAY (actually, original way, again) - caustic photons are LS+D
//   ---------------------------------------------------------------------
//
//   Photons that hit one or more specular surfaces and then a diffuse
//   surface are stored in the caustic map.
//
//   Final gather rays are then responsible for following specular paths
//   until they hit a diffuse surface, to account for L(S|D)*DS+DE paths.
//
//   Therefore, we tag flags with from_diffuse=1 after they hit at least
//   one diffuse surface.
//
//   Photons that come directly from a light are stored with a
//   non-zero light_id.


  // debugging
  int tir_bounces = 0;

  // Figure purposes - show emitted photons
  int showIt = 0;
//   const int numang = 40;
//   static int angles[numang] = {0};
//   const double targetHeight = 29.0;
  
  // if (ctx->mydrand48()<15.0/gv->numPhotons*8.0*1.0)
//   if (gv->figureMode && ctx->id==0 && ABSF(photon.pos[1]-targetHeight)<8.0
//       && gv->tempInt!=30){
    
//     float angle = atan2(-photon.dir[2], photon.dir[0]);
//     int iangle = (int)(angle*M_1_PI*numang);
//     if (angle>=0){
//       iangle = clamp(iangle, 0, numang);
//       if ((iangle+4)%7==0 && angles[iangle]==0 && ctx->mydrand48()<0.3){
// 	cerr << "SHOWING ANGLE: " << iangle << "(" << angle << ")" << endl;
// 	photon.pos[0] = 50.0+(angle*M_1_PI-0.5)*-20.0;
// 	photon.pos[2] -= 0.001;
// 	angles[iangle]=1;
// 	showIt = 1;
//       }
//     }
//     photon.pos[1] = targetHeight;
//   }
  
  while (++bounce<1000){

    // debugging
//     if (0&&bounce>=50){
//       if (bounce == 50 || bounce == 100 || bounce == 200 ||
// 	  bounce == 400 || bounce == 800){
// 	cerr << "Warning: " << bounce << " bounces.\n";
// 	printVec(photon.pos); cerr << "\n";
// 	showPoint(photon.pos, photon.t);
// 	photon.tri->show();
//       }
//     }

    int miss = 0;

    // don't bother with dark photons (ones that bounce off red then green wall)
    if ((emission[0]+emission[1]+emission[2])<=0.0)
      break;

    // shoot photon
    if (! scenePrim->intersectRay(photon) )
      miss = 1; // does not hit anything

    photon.ignore(photon.prim);

    // Super surfaces (double surfaces)
    //
    // Sometimes surfaces are on top of another.
    // this is a difficult situation, and should be avoided.
    // it not obvious what to do in this case, but if
    // we let photons bounce off of overlapping surfaces
    // as many times as they want, we will have a thousand
    // photons in one spot.
    //
    // To prevent this, we effectively police a mininum hit_t
    // by pushing the photon epsilon distance along the ray, and then let
    // the new minimum hit_t be zero.

    const float epsilon = 1.5e-4f;
    if (!miss && photon.t<epsilon){
      photon.pos += epsilon*photon.dir; // barely over overlapping edge tolerance

      //cerr << "close hit! t = " << photon.t << "\n\n";
      //showPoint(x, 15.0);
      
      // shoot photon
      if (! scenePrim->intersectRay(photon) )
	miss = 1; // does not hit anything

      //if (photon.t<epsilon)
      //cerr << "STILL close hit! t = " << photon.t << "\n\n";
      
      photon.ignore(photon.prim);
    }

    // Figure purposes - show photon paths
    if (showIt){
      // if (emission[0]==0.0)
      // showIt = 0;
      // else
      photon.show(miss?100:photon.t, emission*gv->numPhotons/8.0);
      // if (ctx->mydrand48()<0.4)
      // showIt = 0;
    }


    // compute point of intersection
    if (miss) {
      photon.t = (photon.pos-rayAccel->getCenter()).length();
      if (dataSet1) photon.t += dataSet1->getVolMaxDist();
    }
    SbVec3f x = photon.pos + photon.t*photon.dir;

    
    // volume scattering
    if (gv->stepSize>0){
      double eps = ctx->mydrand48();
      float stopAmt = (eps ? -log(eps) : 1000) *gv->units(1); // convert to units
      // stopping tau, as based on reformulation of Eqn 10.22 from Photon Mapping
      // Eqn 10.22:
      //     d         * m/u = -log(eps)/sigma_t
      //     d*sigma_t       = -log(eps)         * u/m
      //     tau(0,t)        = -log(eps)         * units(1)
      
      float dist = 0.0; //ctx->mydrand48()-0.5; // ret val plus voxel offset for bresenHam
      float tau = allDataSets->bresenSum(photon.pos, x, &dist, stopAmt);

      // photon interaction
      if (tau>=stopAmt && dist<photon.t){

	SbVec3f vx = photon.pos+dist*photon.dir;
        static SbVec3f up(0,1,0);   // arbitrary direction for normal. Fix this if using
                                    // anisotropic scattering.

	// absorb
	if (ctx->mydrand48() > gv->albedo)
	  return;
        
	// store indirect photons
	if (bounce>=1){

	  if (gv->showPhotons)
	    showPhoton(vx, up);
	
	  volumeMap->buffered_store(emission.getValue(),
				    vx.getValue(),
				    photon.dir.getValue(),
				    up.getValue(), light_id, hit_diffuse,
				    ctx->id);
	}

	// scatter photon
	// pick random direction on sphere
	double eps1 = ctx->mydrand48();
	double eps2 = ctx->mydrand48();
	phi = eps1 * 2.0 * M_PI;   // rotation about normal
	float z = 2.0*eps2-1.0;
	theta = acos(z);           // angle from normal
	SbVec3f sphereDir;
	sphereDir.setValue(sin(theta)*cos(phi),
			   z, // cos(theta)
			   sin(theta)*sin(phi));
	sphereDir.normalize();

	photon.dir = sphereDir;
	photon.pos = vx;
	light_id = 0;
	continue;
      }
    }
    
    // stop if didn't hit anything
    if (miss)
      break;

    // calculate interpolated normal
    SbVec3f n = photon.prim->getIntNormal(photon);
    
    // calculate surface normal
    SbVec3f ngl = photon.prim->getNormal(photon);
    int flipNormal = 0;
    if (photon.dir.dot(ngl)>0.0){      // outward ng
      ngl.negate();
      flipNormal = 1;
    }

    // do Beer's law attenuation
    Material *mtl = photon.prim->getMat(photon);
    SbColor attenColor;
    bool atten=false;
    if (photon.atten){
      attenColor = photon.attenColor;
      atten = true;
    } else if (n.dot(ngl) < 0) {  // ray strikes inside of surface
      attenColor = mtl->getAttenuation();
      atten = !isWhite(attenColor);
    }

    if (atten) {
      
      SbColor att;
      att[R] = attenColor[R] ? exp(log(attenColor[R])*gv->meters(photon.t)) : 0;
      att[G] = attenColor[G] ? exp(log(attenColor[G])*gv->meters(photon.t)) : 0;
      att[B] = attenColor[B] ? exp(log(attenColor[B])*gv->meters(photon.t)) : 0;

      vecScaleItself(emission, att);
      
      if (DEBUG_STATE){
        cerr << "\n" << __LINE__ << ": att = " << att << "\n";
      }

      // russian roulette
      if (emission[0]<maxPhotonComponent &&
	  emission[1]<maxPhotonComponent &&
	  emission[2]<maxPhotonComponent){

	//float attAvg = (att[0]+att[1]+att[2])*(1.0/3.0);
        float attAvg = MAX2(att[0], MAX2(att[1], att[2]));
	if (ctx->mydrand48() < attAvg){
	  float inv_attAvg = 1.0/attAvg;
	  emission *= inv_attAvg;
	}
	else
	  return;
      }
    }
          
    float shiny_comp = mtl->getShininess();
    float diff_comp = 1.0 - shiny_comp;

    // temporary test
    SbColor tempCol = photon.prim->getIntDiffColor(photon);
    float totalRefl = tempCol[0]+tempCol[1]+tempCol[2];
    
    //'''''''''''''''''''''''''''''W''''''''''''''''''
    //      W                      W        W         
    // WWW WWW  W   WW  WW     WW  WW   W  WWW  W  WW 
    // WW   W  W W W   Wmmm    W W W W W W  W  W W W W
    //WWW   W   W  W    WW     WW  W W  W   W   W  W W
    //'''''''''''''''''''''''''W''''''''''''''''''''''

    if (diff_comp > 0.01 && totalRefl > 0.0001){

      if (phot_type == P_IMPORTANCE){
	
	// store in importance map
	if (gv->showPhotons)
	  showPhoton(x, SbColor(1.0, 1.0, 0.0));
	
	importanceMap->buffered_store(emission.getValue(),
				      x.getValue(),
				      photon.dir.getValue(),
				      ngl.getValue(), light_id, hit_diffuse,
				      ctx->id);

	if (bounce >= 1)
	  return;
      }
      // before doing importance check, verify photon may possibly be stored
      else if (phot_type == P_GLOBAL || (phot_type == P_CAUSTIC && hit_specular)) {

	// do importance sampling
	// (throw away unimportant photons)
	int store_okay = 1;
	if (gv->numIPhotons){
	  
	  store_okay = 0;
	  
	  // is it even close ?
	  SbVec3f diff = x-importanceMap->bbox_cen;
	  if (diff.dot(diff) < importanceMap->bound_rad2){
	    
	    float irrad[3];
 	    importanceMap->irradiance_estimate(
 	      irrad, x.getValue(), ngl.getValue(),
 	      gv->iPhotonEstimate);
// 	    importanceMap->irradiance_estimate_pre(
// 	      irrad, x.getValue(), ngl.getValue());

	    if (irrad[0]>0.0)
	      store_okay = 1;
	  }
	}

	if (store_okay) {
      
	  // store in caustic map if hit specular surface
	  int shown = 0;
	  if (!gv->approximate &&
              hit_specular && !hit_diffuse){  
	    if (gv->showPhotons){
	      showPhoton(x, SbColor(1.0, 0.0, 0.0));
	      shown = 1;
	    }
	    cPhotonMap->buffered_store(emission.getValue(),
				       x.getValue(),
				       photon.dir.getValue(),
				       ngl.getValue(), light_id, hit_diffuse,
				       ctx->id);
	    // hit_specular = 0;
	  }

	  // stop here if we're only building caustics map
	  // if (caustic_only)
	  // return;

	  if (phot_type == P_GLOBAL){
	    
	    // store in global map
	    if (gv->showPhotons && !shown)
	      showPhoton(x, SbColor(0.0, 0.0, 1.0));

	    // figure purposes, for restricting photons to brain streamlines
	    //if (!gv->figureMode ||
	    // (x[0]>0.0 && x[0]<133.3 &&
	    // x[1]>0.0 && x[1]<200 &&
	    // x[2]>0.0 && x[2]<167))
	    
	    gPhotonMap->buffered_store(emission.getValue(),
				       x.getValue(),
				       photon.dir.getValue(),
				       ngl.getValue(), light_id, hit_diffuse,
				       ctx->id);
	  }
	}
      }
    }

    
    //'''''''''''''''''''''''''''''''''''''W''''''''''''''''''
    //              W   W                  W        W         
    // WWW  WW  WW WWW WWW  WW   WW    WW  WW   W  WWW  W  WW 
    // WW  W   W W  W   W  Wmmm W      W W W W W W  W  W W W W
    //WWW   WW  WW  W   W   WW  W      WW  W W  W   W   W  W W
    //'''''''''''''''''''''''''''''''''W''''''''''''''''''''''
    
    if (gv->useGlassModel == 0){

      // transmit in straight line
      if (ctx->mydrand48() < mtl->getTransparency()){
	photon.pos = x;

	hit_specular = 1;
	continue;
      }
    }
    
    float eps = ctx->mydrand48();

    // calculate upward pointing normal
    SbVec3f nl = n;
    if (flipNormal)
      nl.negate();
    
    // calculate specular reflectance
    SbColor pS = mtl->getSpecular();

    if (mtl->isShiny()){
      // fresnel effect (schlick approx & H=n for specular directions)
      pS = fresnel(pS, -photon.dir.dot(nl));
    }
    
    double pSavg = vecAvg(pS);

    // don't use russian roulette on high power photons
    if (emission[R]>maxPhotonComponent ||
	emission[G]>maxPhotonComponent ||
	emission[B]>maxPhotonComponent)
      pSavg = 1.0;
    
    ////////////////////////////
    // specular reflection
    ////////////////////////////

    if (eps < pSavg*shiny_comp){
      
      // track caustic photons
      hit_specular = 1;
      light_id = 0;      

      ////////////////////////////
      // transmit
      ////////////////////////////
      if (mtl->getTransparency() && gv->useGlassModel &&
	  ctx->mydrand48() < mtl->getTransparency()){

	SbVec3f tdir;
	double c;
	double R0;
	double Re;

	// From RRT by PS
	double nc, nt;
	int tir;
        bool atten, inside;
        SbColor attenColor;
        
	if (n.dot(ngl)>0.0){  // into object
          nc = mtl->getIndexOfRefractionTop();
          nt = mtl->getIndexOfRefractionBot();
          if (gv->spectral || spectralPhot){
            nc = cauchy(nc, photon.wavelen);
            nt = cauchy(nt, photon.wavelen);
          }
	  refract(tdir, tir, photon.dir, n, nc/nt);
          if (!tir)
            c = -photon.dir.dot(n);
          attenColor = mtl->getAttenuation();
          atten = !isWhite(attenColor);
          inside = true;
	}
	else {                // out of object
          nc = mtl->getIndexOfRefractionBot();
          nt = mtl->getIndexOfRefractionTop();
          if (gv->spectral || spectralPhot){
            nc = cauchy(nc, photon.wavelen);
            nt = cauchy(nt, photon.wavelen);
          }
	  refract(tdir, tir, photon.dir, -n, nc/nt);
	  float t = gv->meters(photon.t);
          if (!tir)
            c = tdir.dot(n);
          attenColor = SbVec3f(1,1,1);
          atten = false;
          inside = false;
	}

	// total internal reflection
	if (tir){
	  tir_bounces++;
	  Re = 1.0f;

	  // use russian roulette to prevent infinite loops
	  if (tir_bounces>20){

	    // double power
	    const float extinction = 0.9;
	    if (ctx->mydrand48()<extinction)
	      emission /= extinction;

	    // or absorb
	    else
	      break;
	  }
	  
	}
	else {

	  // grazing angle problem (normal interpolation error)
	  // don't let transmitted ray be on same side
	  if (tdir.dot(ngl)>0.0){
	    tdir -= (tdir.dot(ngl)+0.0015) * ngl;
	    tdir.normalize();
	  }
    
	  // due to errors for grazing angles
	  // with normal interpolation, c may
	  // incorrectly be < 0.0. This clamps it.
	  if (c<0.0)
	    c = 0.0;
	  
	  R0 = (nt-nc)/(nt+nc);
	  R0 *= R0;
          Re = fresnel(R0, c);
	}
      
	// transmit
	if (ctx->mydrand48() > Re){
	      
	  // scale power
	  for (int i=0; i<3; i++)
	    emission[i] *= pS[i] * (1./pSavg);

	  // bounce photon
	  photon.atten = atten;    
          photon.attenColor = attenColor;
	  photon.inside = inside;
	  photon.dir = tdir;
	  photon.pos = x;

	  continue;
	}

	// reflect
	else {
	      
	  // scale power
	  for (int i=0; i<3; i++)
	    emission[i] *= pS[i] * (1./pSavg);

	  // normal correction (two sided lighting)
	  SbVec3f rdir = photon.dir-2.0*n.dot(photon.dir)*n;

	  // grazing angle problem
	  // don't let reflected ray go beneath surface
	  if (rdir.dot(ngl)<0.0)
	    rdir += (-rdir.dot(ngl)+0.0015) * ngl;

          rdir.normalize();
          
	  // bounce photon
	  photon.dir = rdir;
	  photon.pos = x;

	  continue;
	}

      }

      ////////////////////////////
      // reflect
      ////////////////////////////
      else {

	// reflect specularly
	SbVec3f rdir = photon.dir-2.0*n.dot(photon.dir)*n;

	// grazing angle problem
	// don't let reflected photon go beneath surface
	if (rdir.dot(ngl)<0.0)
	  rdir += (-rdir.dot(ngl)+0.0015) * ngl;

        rdir.normalize();
	  
	// scale power
	for (int i=0; i<3; i++)
	  emission[i] *= pS[i] * (1./pSavg);
	    
	// bounce photon
	photon.dir = rdir;
	photon.pos = x;

	continue;
      }

    }

    // specular photons do not bounce diffusely
    if (phot_type == P_CAUSTIC)
      return;
    
    // calculate diffuse reflectance
    SbColor pD = photon.prim->getIntDiffColor(photon);
    //double pDavg = (pD[0]+pD[1]+pD[2]) * 1.0/3;
    double pDavg = MAX3(pD[0],pD[1],pD[2]); // experimental

    // don't use russian roulette on high power photons
    if (emission[R]>maxPhotonComponent ||
	emission[G]>maxPhotonComponent ||
	emission[B]>maxPhotonComponent)
      pDavg = 1.0;

    ////////////////////////////
    // diffuse reflection
    ////////////////////////////
    
    if (eps < ((pDavg*diff_comp)+(pSavg*shiny_comp))){
    
      // flip normal if transmitting
      int transp = mtl->getTransparency() && gv->useGlassModel &&
	ctx->mydrand48() < mtl->getTransparency();
      if (transp)
	ngl = -ngl;

      // calculate new direction
      ONB onb(ngl);

      double eps1 = ctx->mydrand48();
      double eps2 = ctx->mydrand48();
      phi = eps1 * 2.0 * M_PI;   // rotation about normal
      theta = acos(sqrt(eps2));  // angle from normal
      hemiDir.setValue(sin(theta)*cos(phi),
		       sin(theta)*sin(phi),
                       cos(theta));
    
      diffuseDir = onb.toWorldSpace(hemiDir);

      // 2d adjustment
      if (gv->useDimensions<3){
	ngl[1] = 0.0;
	ngl.normalize();
	SbVec3f ngl_perp = SbVec3f(ngl[2], 0.0, -ngl[0]);
	theta = asin(2.0*eps1-1.0);
	diffuseDir = ngl*cos(theta) + ngl_perp*sin(theta);
      }
      
      diffuseDir.normalize();

      // scale power
      for (int i=0; i<3; i++)
	emission[i] *= pD[i] * (1/pDavg);
	    
      // bounce photon
      photon.dir = diffuseDir;
      photon.pos = x;

      hit_diffuse = 1;
      light_id = 0;
      continue;
	  
    } // end trasmit/spec/diff/absorb decision

    ////////////////////////////
    // absorb
    ////////////////////////////

    return;
    
  } // end while() loop for bouncing   bounce loop
}


