#include "Pane.h"
#include "debug.h"
#include "DataSet.h"
#include "EnvironMap.h"
#include "Sky.h"
#include "Photon_map.h"

using namespace std;


SbColor Pane::volumeIntegrate(Ray &r, const SbColor &c){

  // ray march
  float sigma_t = gv->sigma_t;      //0.5;   // in meters!
  float sigma_s =      gv->albedo *sigma_t;
  float sigma_a = (1.0-gv->albedo)*sigma_t;

  // s_t = inverse of mean free path length
  // albedo = prob of scattering vs absorbing
  // s_s =    albedo *s_t
  // s_a = (1-albedo)*s_t
  
  const float p = 1.0/(4.0*M_PI);

  // experimental
  //if (r.importance < 0.1)
  //  return(c);
  float dt = gv->stepSize; ///r.importance;
  if (r.depth>1) dt = 1.0;
  //dt = gv->units(dt); // convert meters to units
  float fdt = r.ctx->mydrand48()*dt;

  // iterate over length of ray
  // (from back to front)
  SbColor color = c;

  float t=r.t;
  if (t<0)
    t=(r.pos-rayAccel->getCenter()).length()+dataSet1->getVolMaxDist();

  float t1, t2;
  SbVec3f end(r.pos+t*r.dir), foo1, foo2;

  // path tracing solution
  static const bool pathTracing = getenv("PANE_VOL_PT") != 0x0;
  if (pathTracing) {

    float stopAmt = -log(r.ctx->mydrand48()) *gv->units(1); // convert to units
    float dist = 0.0;
    float tau = dataSet1->bresenSum(r.pos, end, &dist, stopAmt);

    // By traveling distance dist, we do not need to scale the radiance
    // by the transmittance Tr=exp^-tau where tau=int(sigma_t(t)*dt)

    // photon interaction
    if (tau<stopAmt || dist>=t)
      return color;

    // absorb
    if (r.ctx->mydrand48() >= gv->albedo)
      return SbColor(0,0,0); // black!
    // now we don't need to scale by albedo

    Ray r2(r); // direct copy, so i have to adjust values
    SbVec3f vx = r.pos + dist*r.dir;
    r2.x = vx;
//     sigma_t = dataSet1->valBound(r2.x);
//     sigma_s = gv->albedo*sigma_t;
//     sigma_a = (1.f-gv->albedo)*sigma_t;

//     if (sigma_s<=0.0)
//       return SbColor(0,0,0);
      
    // compute scattered light
    SbColor dcolor(0,0,0);

    // point lights
    dcolor += volPointLightContrib(r2);

    // area lights
    dcolor += volAreaImpLightContrib(r2, 4); 

    // infinite area lights
    if (envMap){
      dcolor += envMap->volImpSample(r2, 4);
    }

    // indirect illumination
    SbColor icolor(0,0,0);
    int numIndSamples = r.depth == 1 ? 4 : 1;

    static const bool directOnly = getenv("PANE_VOL_PT_DIRECTONLY") != 0x0;
    if (directOnly)
      numIndSamples = 0;
    
    // allocate buffer dim
    r2.ldNextDim();
    int ldDim = r2.ldDimension; // save vals
    int ldIns = r2.ldInstance+1;
      
    for (int i=0; i<numIndSamples; i++){

      // pick random direction on sphere
      double eps1 = r.ctx->mydrand48();
      double eps2 = r.ctx->mydrand48();
      double phi = eps1 * 2.0 * M_PI;   // rotation about normal
      float z = 2.0*eps2-1.0;
      double theta = acos(z);           // angle from normal
      SbVec3f sphereDir;
      sphereDir.setValue(sin(theta)*cos(phi),
                         z, // cos(theta)
                         sin(theta)*sin(phi));
      sphereDir.normalize();

      r2.pos = vx;
      r2.dir = sphereDir;
      r2.depth = r.depth+1;
      
      r2.ldDimension = ldDim;
      r2.ldInstance = ldIns + i;

      if (gv->algorithm <= 1)
        icolor += (*this.*getRadiance)(r2, (canvas.flags | C_SIMPLE) ^ C_EMITTED ); //-a1
      else if (gv->algorithm==2)
        icolor += (*this.*getRadiance)(r2, canvas.flags ^ C_EMITTED ); // -a2
      else if (gv->algorithm==3)
        icolor += (*this.*getRadiance)(r2, canvas.flags ); // -a3
      else
        DBG("WTF!");
    }
    if (numIndSamples)
      icolor *= (1./numIndSamples);

//    return dcolor*sigma_t*p + icolor*sigma_t;
    SbColor ret = dcolor*p;
    if (!directOnly)
      ret += icolor;
    return ret;
    
    // sigma_t and not simga_s because we are importance sampling albedo.
    // icolor 1/pdf cancels multiplying by p
  }

  dataSet1->clipToVolume(r.pos, end, foo1, foo2, t1, t2);
  if (t2<=t1)
    return(color);

  // start at end
  t = t2;

  int first = 0;
  float edt;
  Ray r2(r);
  int done = 0;

  // tempInt    -> PANE_TEMP_CUTOFF   / temp_cutOff   = 600
  // tempFloat2 -> PANE_FLAME_COLMUL  / flame_colMul  = .0000000000005
  // tempFloat3 -> PANE_SMOKE_INDMUL  / smoke_indMul  = 1000000.0
  //               PANE_SMOKE_MUL     / smoke_mul     = 1.0
  //               PANE_SMOKE_ALB     / smoke_alb     = 0.2
  //               PANE_FLAME_ALB     / flame_alb     = 0.2
  
  static const float smoke_mul = getenv("PANE_SMOKE_MUL") ?
    atof(getenv("PANE_SMOKE_MUL")) : 1.0;
  static const float smoke_alb = getenv("PANE_SMOKE_ALB") ?
    atof(getenv("PANE_SMOKE_ALB")) : 0.6;
  static const float smoke_indMul = getenv("PANE_SMOKE_INDMUL") ?
    atof(getenv("PANE_SMOKE_INDMUL")) : 1.0;

  static const float flame_mul = getenv("PANE_FLAME_MUL") ?
    atof(getenv("PANE_FLAME_MUL")) : 0.02;
  static const float flame_colMul = getenv("PANE_FLAME_COLMUL") ?
    atof(getenv("PANE_FLAME_COLMUL")) : 1.0;
  
  static const float temp_cutOff = getenv("PANE_TEMP_CUTOFF") ?
    atof(getenv("PANE_TEMP_CUTOFF")) : 1150.0;

  while (! done){

    // random first length
    if (first++ == 0)
      edt=fdt;
    else 
      edt=dt;

//     // experimental adaptive ray marching
//     float sst = -r.ctx->mydrand48()/sigma_t;
//     sst = gv->units(sst);
//     if (sst<edt)
//       edt = sst;

    float newt = t-edt;
    if (newt<t1){
      edt = t-1.0000001*t1;
      done = 1;
    }

    t -= edt;
    r2.x = r.pos+t*r.dir;

    // convert to meters
    edt = gv->meters(edt);

    // experimental
    if (dataSet1){
      sigma_t = dataSet1->valBound(r2.x);
      sigma_s = gv->albedo*sigma_t;
      sigma_a = (1.f-gv->albedo)*sigma_t;

      // record nearest interaction
      if (sigma_t>0.0f)
	r.tMedia = t;
    }

    SbColor dcolor(0,0,0), icolor(0,0,0), ecolor(0,0,0);

    // experimental
    if (sigma_s>0.0){

      // point lights
      dcolor += volPointLightContrib(r2);

      // area lights
      dcolor += volAreaImpLightContrib(r2, 4); // was 4. Depends on solid angle
                                               // subtended by light.
      //r2.ldDimension += 2;

      // infinite area lights
      if (envMap){
	dcolor += envMap->volImpSample(r2, 4);
        //r2.ldDimension += 2; // remove this when the above func does it
      }

      // indirect illumination
      float irrad[3];
      volumeMap->volume_irradiance_estimate(irrad, r2.x.getValue(),
					    gv->vPhotonEstimate);
      icolor += SbColor(irrad[R], irrad[G], irrad[B])*gv->unitsPerMeter3;
    }

    // emission
    //sigma_a = gv->tempFloat2;
    //sigma_t = gv->tempInt/1024.;
    float flame_sigma_t = 0.0;
    float flame_sigma_a = 0.0;
    float flame_sigma_s = 0.0;
    if (sigma_a>0.0 && dataSet2){
//      float temp = dataSet2->valBound(r2.x);
//       if (temp>1000.0f){
// 	sigma_s = 0.0;
// 	sigma_t = sigma_a;
//       }
//      if (temp>temp_cutOff){ // min temp

      //SbColor emisCol = Spectrum(temp).getRGB()/(4.0*M_PI);
      SbColor emisCol = dataSet2->colorValPoint(r2.x);
      //SbColor emisCol = dataSet2->colorLookup(temp);
	//emisCol = Spectrum(temp).getAdaptedRGB();
	ecolor = emisCol*flame_colMul;
// 	sigma_t *= smoke_mul;
//         sigma_s = sigma_t*smoke_alb;
//         sigma_a = sigma_t*(1.0-smoke_alb);

        // my fire sim
//	flame_sigma_t = sigma_t/gv->scalarDataMult*flame_mul; // good fire
//	flame_sigma_a = flame_sigma_t;

        // universe
        flame_sigma_t = 0;
        flame_sigma_a = sigma_a*flame_mul;
        
	//if (rd.depth==1)
	//continue;
	//ecolor.setValue(0,0,0);
	//dcolor.setValue(0,0,0);
//       }
//       else {
	//ecolor.setValue(0,.1,0);
//	sigma_t *= smoke_mul;
//	sigma_s = sigma_t*smoke_alb;
//	sigma_a = sigma_t*(1.0f-smoke_alb);
//       }
     }

//     if (!ecolor[0]){
//  	//sigma_t *= smoke_mul;
// 	sigma_t = smoke_mul;
// 	sigma_s = sigma_t*smoke_alb;
// 	sigma_a = sigma_t*(1.0-smoke_alb);
//     }
    

    // color scattering
    dcolor[0] *= gv->scalarColor[0];
    dcolor[1] *= gv->scalarColor[1];
    dcolor[2] *= gv->scalarColor[2];
    icolor[0] *= gv->scalarColor[0];
    icolor[1] *= gv->scalarColor[1];
    icolor[2] *= gv->scalarColor[2];
    
    // random scatter
    // find dir with rejection sampling
    //dcolor.setValue(0.0, 0.0, 0.0);
    if (0&&sigma_s>0.0 && r.depth<=1){ // ==1
      float length2, phi, theta, rho;
      do {
	phi   = r.ctx->mydrand48()*2.0-1.0;
	theta = r.ctx->mydrand48()*2.0-1.0;
	rho   = r.ctx->mydrand48()*2.0-1.0;
	length2 = (phi*phi+theta*theta+rho*rho);
      } while (length2>1.0);
      SbVec3f neww(phi, theta, rho);
      neww /= sqrt(length2);

      // scatter light
      Ray wp(r);
      wp.pos = r2.x;
      wp.dir = neww;
      r2.ldNextDim();
      wp.ldDimension = r2.ldDimension;
      wp.ldInstance = r2.ldInstance+1;
      wp.depth = wp.depth+1;

      // 11/18/08 (07?)
      //
      // This seems unnecessary if I'm using the volume photon map.
      // Plus, shouldnt I remove C_EMITTED from flags?
      
      icolor = (*this.*getRadiance)(wp, canvas.flags|C_SIMPLE);///gv->albedo;
      icolor *= smoke_indMul;
      icolor *= sigma_s;
    }

    if (dataSet3)
      sigma_t += dataSet3->valBound(r2.x);

    
    // for doing volumetric photon mapping + indirect scattering
    color = ( (dcolor*p)*sigma_s + icolor*p + ecolor*flame_sigma_a )*edt +
      exp(-(sigma_t+flame_sigma_t)*edt)*color;
  }

  return(color);
}


//'''''''''''''''''''''''''''''''''''''''''''''WW'''''''WW''''WW'WWWW'''WWW''
//                                             WW       WWW  WWW WW  W WW  W 
//WWW  W WW  WW   WWW  WW   WWWW W  WWW  W WW WWWW  WW  W WWW WW WW  W WW    
//WW W WW   WWWW WW W WW W  WW WW W WW W W WW  WW  WWWW W  W  WW WW  W WW    
//WW W WW   WW   WW   WW W  WW WW W WW W W WW  WW  WW   W     WW WW  W WW  W 
//WWW  WW    WWW  WWW  WW   WW WW W WWW   WWW  WW   WWW W     WW WWWW   WWW  
//WW                                WW                                       
//WW''''''''''''''''''''''''''''''''WW'''''''''''''''''''''''''''''''''''''''
//
// precompute look up tables
//
void Pane::precomputeMDCTables(){

  // Calculate M and N to match a user specified total # of gather rays
  // Mmdc = MIN2((int)sqrt(gv->gatherRays/2.0), Mmax);
  // Nmdc = 2*Mmdc;
  Mmdc = MIN2((int)sqrt(gv->gatherRays/3.0), Mmax);
  Nmdc = 3*Mmdc;

  if (Mmdc<=0) return;
  
  // theta, M, j
  for (int j=0; j<Mmdc; j++){
    float sinTm = sqrt( j/(double)Mmdc );
    float sinTp = sqrt( (j+1)/(double)Mmdc );
    float thetam = asin(sinTm);
    float thetap = asin(sinTp);
    float cosTm = cos(thetam);

    // sin(Oj_)*cos^2(Oj_)
    // sin(O_) = sinTm
    sinTxcos2T[j] = sinTm*cosTm*cosTm;

    // sin(Oj+)-sin(Oj_)
    sinTmsinT[j] = sinTp - sinTm;
  }

  // phi, N, k
  float phi;
  for (int k=0; k<Nmdc; k++){
    // phi_k_ + pi/2
    phi = 2.0 * M_PI * k/(float)Nmdc + M_PI/2.0;
    vk_X[k] = cos(phi);
    vk_Z[k] = sin(phi);
  }
	
}


//'''''''''''''WWW'''WW''WW''''''WWW'''''''WWWW''WW'''WW'''WW'''''''''''''''''
//              WW   WW           WW       WW  W     WW   WW                  
//WWWW W  W WW  WW  WWWW WW WWW   WW   WW  WW  W WW WWWW WWWW W WW  WWWW  WW  
//WW WW W W WW  WW   WW  WW WW W  WW  WWWW WW  W WW  WW   WW  W WW WWW   WWWW 
//WW WW W W WW  WW   WW  WW WW W  WW  WW   WW  W WW  WW   WW  W WW   WWW WW   
//WW WW W  WWW WWWW  WW  WW WWW  WWWW  WWW WWWW  WW  WW   WW   WWW WWWW   WWW 
//                          WW                                                
//''''''''''''''''''''''''''WW''''''''''''''''''''''''''''''''''''''''''''''''
// (mdc)
//
// calculates the contribution from multiple diffuse reflections
// (indirect illumination). Uses Monte Carlo integration
// and the Photon Mapping radiance estimate to estimate
// indirect illumination (no lights). The irradiance is then
// multiplied by the BRDF to return the reflected radiance.
//
// Uses Irradiance Caching with Irradiance Gradients to cache and
// interpolate the results, since this is an expensive computation.
//
SbColor Pane::multipleDiffuseContrib(Ray &r, SbVec3f &n, SbVec3f &ngl,
				     int simpleCalc,
				     int prepass,
				     int inOrOut){

  if (Mmdc<=0) return SbColor(0,0,0);

  // Cached value exist?
  SbColor ptcolor = SbColor(0.0, 0.0, 0.0);
  SbColor *ptcolorp = &ptcolor;
  
  int cache_result = 0;
  if (gv->useIrradCache)
    cache_result = irradCache.getIrradiance(&ptcolor, r, &n, &ngl, prepass);

  if (cache_result == 0){

    const int n_theta = 4;
    const int n_phi = 16;
    float *flux_map;

    // flux map section disabled
    if (gv->useIndirectImpSampling){
      
      flux_map = (float *)alloca(sizeof(float)*n_theta*n_phi);
      // get flux map for importance sampling
      const int n_photons = gv->gPhotonEstimate;


      gPhotonMap->get_flux_map((float *)flux_map, n_theta, n_phi, r.x.getValue(),
			       n.getValue(), n_photons);

      // find zero_regions and total_energy
      int zero_regions = 0;
      float total_energy = 0.0;
      for (int u=0; u<n_theta; u++)
	for (int v=0; v<n_phi; v++){
	  if (flux_map[u*n_phi+v] > 0.0)
	    total_energy += flux_map[u*n_phi+v];
	  else
	    zero_regions++;
	}

      // add small fraction of energy to zero regions
      float de;
      if (total_energy>0.0)
	de = total_energy / (n_phi*n_theta); // * 0.5;
      else
	de = 1.0;

      total_energy += de*(n_phi*n_theta);

      // scale flux map to create c.d.f.
      for (int u=0; u<n_theta; u++)
	for (int v=0; v<n_phi; v++){

	  // directional
	  flux_map[u*n_phi+v] += de;
	  flux_map[u*n_phi+v] /= total_energy;

	  // stratified
          // flux_map[u][v] = 1.0/(n_phi*n_theta);
	}

      // draw debugging histogram
      if (DEBUG_STATE)
	drawHist((float *) flux_map, n_phi, n_theta,
		 n_phi*n_theta*getMaxDist()/10.0, r.x, n);
    }

    
    // integrate hemisphere
    float phi, theta;
    SbVec3f neww;
    ONB onb(ngl);

    // stratified sampling
    ptcolor = SbColor(0.0, 0.0, 0.0);

    // number of stratifications in theta, and phi
    int M, N;
    if (simpleCalc){
      if (r.depth<=2){
	M=3;
	N=4;
      }
      else {
	M = N = 1;
      }
    }
    else {
      M=Mmdc;
      N=Nmdc;
    }

    // 2d adjustment
    if (gv->useDimensions<3){
      M=MIN2(gv->gatherRays, Mmax);
      N=1;
      if (simpleCalc)
	M=4;
    }

    Ray bounce(r.x, ngl, r);
    float harm_mean = 0.0;
    SbVec3f Er = SbVec3f(0.0, 0.0, 0.0);
    SbVec3f Et = SbVec3f(0.0, 0.0, 0.0);
    SbVec3f v_k;
    float *Ljk   = (float *)alloca(sizeof(float)*N*M);
    float *rjk   = (float *)alloca(sizeof(float)*N*M);
    float *phijk = (float *)alloca(sizeof(float)*N*M);
    int rejectPoint = 0; // for trashing (not storing) statistically bad irrpnt

    // debugging
    float maxrad = -1.0;
    int maxk=-1, maxj=-1;

    // allocate ld dimension even though we don't use ld numbers,
    // to decorrelate child rays
    r.ldNextDim();
    
    // loop over phi
    for (int k=0; k<N; k++){

      float rot_sum = 0.0;

      // loop over theta
      for (int j=0; j<M; j++){

	// search for new direction using importance sampling
	double eps1 = (j + r.ctx->mydrand48()) / M;  // stratified
	double eps2 = (k + r.ctx->mydrand48()) / N;  // stratified
//	double eps1 = (j+0.5)/M;
//	double eps2 = (k+0.5)/N;

	float imp_scale;
	
	if (gv->useIndirectImpSampling){

	  // use eps1 to pick region
	  double eSum = 0.0;
	  int u, v;
	  int found = 0;
	  for (u=0; u<n_theta; u++){
	    for (v=0; v<n_phi; v++)
	      if (eps2<(eSum+flux_map[u*n_phi+v])){
		found = 1;
		break;
	      }
	      else
		eSum += flux_map[u*n_phi+v];

	    if (found)
	      break;
	  }

	  // check if reached end (shouldn't, but might...)
	  if (!found){
	    u = n_theta-1;
	    v = n_phi-1;
	  }

	  // compute scaling factor
	  imp_scale = 1.0/(flux_map[u*n_phi+v]*n_theta*n_phi);

	  // adjust eps
	  eps2 = (eps2 - eSum) / flux_map[u*n_phi+v];

	  // calculate phi and theta from eps1 and eps2
	  // this uses a non-linear transformation to give
	  // a cosine distrubition in theta, that is
	  // theta takes [0,1]->[0,pi/2], and if
	  // [0,1] is uniform, theta is ~ cos(theta).
	  // the actual mapping is [0,1]->[pi/2,0].
	  // i reversed it to save a 1-eps subtraction operation.

	  // importance
	  theta = acos(sqrt( (u+eps1)/n_theta ));
	  phi = 2.0 * M_PI * (v+eps2)/n_phi;

	}
	else {
	
	  // stratified
	  theta = acos(sqrt( eps1 ));
	  phi = 2.0 * M_PI * eps2;
	  imp_scale = 1.0;
	}
	
	neww.setValue(sin(theta)*cos(phi),
		      sin(theta)*sin(phi),
                      cos(theta));
    
        bounce.dir = onb.toWorldSpace(neww);
	
	// 2d adjustment
	if (gv->useDimensions<3){
	  SbVec3f fngl = SbVec3f(ngl[0], 0.0, ngl[2]);
	  fngl.normalize();
	  SbVec3f fngl_perp = SbVec3f(fngl[2], 0.0, -fngl[0]);
	  theta = asin(2.0*eps1-1.0);
	  bounce.dir = fngl*cos(theta) + fngl_perp*sin(theta);
	  // if (DEBUG_STATE && !simpleCalc)
	  // bounce.show(200, SbColor(1.0, 1.0, 1.0));
	}
	
	bounce.dir.normalize();

	SbVec3f Lrad;
	int useApprox = 0; //(simpleCalc)?1:0;

	bounce.tMedia = -1.0;
	const int myFlags = C_SIMPLE | C_SPECULAR | (useApprox?C_APPROX:0) |
	  C_DIRECT | C_CAUSTIC | C_MDIFFUSE;
	bounce.ignore(r.prim);

        // each call to getRadiance modifies ld values, so reset them.
        bounce.ldDimension = r.ldDimension;
        bounce.ldInstance = r.ldInstance + k*M+j + 1;
        
	Lrad = getRadiancePM(bounce, myFlags);

	// if hit something, either surface or participating media
	if (bounce.t>0.0f || bounce.tMedia>0.0f){

	  // use gas distance if no surface is it
	  if (bounce.t<0.0f)
	    bounce.t = bounce.tMedia;

	  // bump map
	  float bumpVal = 1.0;
	  if (gv->bumpMapping && simpleCalc == 0){
	    bumpVal = bounce.dir.dot(n)/bounce.dir.dot(ngl);
// 	    if (bumpVal>gv->bumpScale)
// 	      bumpVal = gv->bumpScale;
	    if (bumpVal<0.0)
	      continue;
	  }
	  
	  // don't do irradiance_estimate on non-diffuse surfaces - pointless&slow!!
	  ptcolor += Lrad*imp_scale*bumpVal;

	  // continue with phi iterations if simple calculation
	  if (simpleCalc || gv->useIrradCache == 0)
	    continue;
	  
	  // update harmonic mean
	  harm_mean += 1/bounce.t;
	  
	  // for rotational gradient calculation
	  float l_avg = brightness(Lrad);
	  rot_sum += - tan(theta)*l_avg;	  

	  // Reject ridiculously close to pi/2 values of theta.
	  // The following line of code rejects points if a sample
	  // is used, satisfying the condition:
	  // Prob(x>sample)*num_samples < 1/500.
	  // In other words, if the probability of having a
	  // sample closer to Pi/2 in this set is less than 0.2%,
	  // junk the whole set.
	  // This effectively skips over the craziest 0.2%.
	  // The reasoning behind this is, if
	  // Prob(x) = cos(x), then
	  // Prob(x>sample) = 1-sin(sample)  (by integrating pdf)
	  // If #samples*above < 1/500, then this sample is too close.
	  // Close samples make tan() (see previous line of code)explode.
	  if ((1.-1./(500*M*N))<sin(theta)){
	    rejectPoint = 1;

	    // record keeping
	    pthread_mutex_lock(&mutexes[LOCK_ICREJECTS]);
	    rejectedICPoints++;
	    pthread_mutex_unlock(&mutexes[LOCK_ICREJECTS]);
	  }

	  // for translational gradient
	  // [j][k] = N*k+j
	  Ljk[k*M+j] = l_avg;
	  rjk[k*M+j] = bounce.t;
	  phijk[k*M+j] = phi;
        }

	// ray did not hit geometry
	else { 

	  // for translational gradient
	  Ljk[k*M+j] = 0;
	  rjk[k*M+j] = 1e8;
	  phijk[k*M+j] = phi;

	  // show missed ray
	  if (0 &&DEBUG_STATE){ 
	    SbColor Lrad = SbColor(0,0,0);
	    bounce.show(getMaxDist()/10.0, Lrad);
	    cerr << "missed: " << "??" << endl;
	  }
	}
      
      } // end theta loop

      // skip IC stuff
      if (simpleCalc || gv->useIrradCache == 0)
	continue;
      
      // calculate v_k (theta=pi/2 phi=phi+pi/2)
      double phi_kp2 = phi+M_PI/2.0;
      v_k = SbVec3f(cos(phi_kp2), sin(phi_kp2), 0.0);
	    
      // update rotational gradient
      Er += v_k*rot_sum;

      if (0&&DEBUG_STATE){
	cerr << "Er += " << rot_sum << " * "; printVec(v_k);
	cerr << " = "; printVec(Er); cerr << "\n";
      }
	
    } // end phi loop

    // Calculate irradiance (and gradients)
    double factor = M_PI/(M*N);

    // 2d adjustment
    if (gv->useDimensions<3)
      factor = 2.0/(M*N);
    
    ptcolor *= factor;

    // skip the gradient calulcation stuff if simple calculation
    if (gv->useIrradCache && simpleCalc == 0){

      Er = onb.toWorldSpace(Er);
      Er *= factor;

      // Calculate translational gradient
      SbVec3f u_k, v_k_;
      for (int k=0; k<N; k++){

	SbVec3f tran_sum1 = SbVec3f(0.0, 0.0, 0.0);

	// compute right sum for j = 0
	int km1 = (k==0)?N-1:k-1;
	double tran_sum2 =
	  sinTmsinT[0] /
	  MIN2(rjk[k*M], rjk[km1*M]) *
	  (Ljk[k*M] - Ljk[km1*M]);

	// perform inner sum
	for (int j=1; j<M; j++){

	  // left sum
	  u_k.setValue(cos(phijk[k*M+j]), sin(phijk[k*M+j]), 0.0);
          u_k = onb.toWorldSpace(u_k);
	  //u_k.normalize();        // accurate computation not needed
	  
	  int jm1 = j-1;
	  double s = sinTxcos2T[j] *
	    (Ljk[k*M+j] - Ljk[k*M+jm1]) /
	    MIN2(rjk[k*M+j], rjk[k*M+jm1]);
	  tran_sum1 += u_k*s;

	  // right sum
	  tran_sum2 +=
	    sinTmsinT[j] /
	    MIN2(rjk[k*M+j], rjk[km1*M+j]) *
	    (Ljk[k*M+j] - Ljk[km1*M+j]);
	}

	// finish inner sum
	
	// compute v_k_
	v_k_.setValue(vk_X[k], 0.0, vk_Z[k]);
        v_k_ = onb.toWorldSpace(v_k_);
	//v_k_.normalize();      // accurate computation not needed

	// add to final
	Et += (2.0*M_PI/N)*tran_sum1 + v_k_*tran_sum2;

      }
      
      // Calculate harmonic mean
      // 1/harmonic mean = 1/n * Sum[1/x] => H = n/Sum[1/x]
      if (harm_mean >0.0)
	harm_mean = (M*N) / harm_mean;
      else // if no objects hit, set mean to bounding box size
	harm_mean = irradCache.size();

      if (!rejectPoint){

	// store irradiance
	if (DEBUG_STATE){
	  cerr << "STORING: r=" << harm_mean << " E="; printCol(ptcolor);
	  cerr << " Et="; printVec(Et);
	  cerr << " Er="; printVec(Er);
	  cerr << " n="; printVec(n);
	  cerr << " x="; printVec(r.x);
	  cerr << "\n";
	}

	// 2d adjustment
	if (gv->useDimensions<3)
	  Er = Et = SbVec3f(0.0, 0.0, 0.0);

	// store point
	pthread_mutex_lock(&mutexes[LOCK_ICACHE]);
	bool stored =
          irradCache.storeIrradiance(ptcolor, r.x, n, ngl, harm_mean, Er, Et);
	pthread_mutex_unlock(&mutexes[LOCK_ICACHE]);    

	// draw point
 	if (gv->renderQuit == 0 && stored)
 	  showPhoton(r.x, SbColor(0.0, 1.0, 0.0));
      }
    }

  }
      
  // multiply irradiance by brdf to get radiance
  if (gv->useDimensions<3){
    ptcolor[R] *= r.diffCol[R]*1.0f/2.0f;
    ptcolor[G] *= r.diffCol[G]*1.0f/2.0f;
    ptcolor[B] *= r.diffCol[B]*1.0f/2.0f;
  }
  else {
    ptcolor[R] *= r.diffCol[R]*M_1_PI;
    ptcolor[G] *= r.diffCol[G]*M_1_PI;
    ptcolor[B] *= r.diffCol[B]*M_1_PI;
  }

  // return radiance
  return(ptcolor);
}

// refract ray
// Computes refracted ray direction or signals
// Total Inteneral Reflection (tir) if that occurs
void Pane::refract(SbVec3f &ret, int &tir,
		   const SbVec3f &d, const SbVec3f &n, const double &nc_nt){ // nc/nt
  
  double ddn = d.dot(n);
  double costsq = 1 - nc_nt*nc_nt*(1 - ddn*ddn);
  
  if (costsq<0) {  // total internal reflection
    tir = 1;       // signal tir
    return;
  }

  tir = 0;
  ret = nc_nt*d - (ddn*nc_nt + sqrt(costsq))*n;
  ret.normalize();
}


//''''''''''''WW'''''''''''''''''WW''WW'''''''''''''WWW''''''''''''''WW'''''''
//           WW                  WW                WW  W             WW       
//W WW  WW  WWWW W WW  WWW  WWW WWWW WW  WW   WWW  WW     WW   WWW  WWWW W WW 
//WW   WWWW  WW  WW   W WW WW W  WW  WW WW W  WW W WW    WW W  WW W  WW  WW   
//WW   WW    WW  WW   W WW WW    WW  WW WW W  WW W WW  W WW W  WW W  WW  WW   
//WW    WWW  WW  WW    WWW  WWW  WW  WW  WW   WW W  WWW   WW   WW W  WW  WW   
//                                                                            
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates refracted component contribution
// Returns radiance due to transparent dielectric,
// specifically the specularly reflected and refracted components.
//
SbColor Pane::refractionContrib(Ray &r, const SbVec3f &n,
                                /*const SbVec3f &ng, const SbVec3f &nl,*/
                                const SbVec3f &ngl, int flags){
  
  SbColor ncolor;
  SbVec3f tdir;
  double cosTheta, R0, Re;
  bool atten, inside;
  SbColor attenColor;
  
  SbVec3f rdir = r.dir-2.0*n.dot(r.dir)*n;
  
  // grazing angle problem
  // don't let reflected ray go beneath surface
  if (rdir.dot(ngl)<0.0){
    rdir += (-rdir.dot(ngl)+0.0015) * ngl;
    rdir.normalize();
  }
  r.ldNextDim();
  Ray reflect(r.x, rdir, r, r.prim);
  reflect.ldInstance++;

  // From RRT by PS
  double nc, nt;
  Material *mtl = r.prim->getMat(r);
  float transparency = mtl->getTransparency();
  int tir;

  if (n.dot(ngl)>0.0){  // into object
    nc = mtl->getIndexOfRefractionTop();
    nt = mtl->getIndexOfRefractionBot();
    if (gv->spectral){
      nc = cauchy(nc, r.wavelen);
      nt = cauchy(nt, r.wavelen);
    }
    refract(tdir, tir, r.dir, n, nc/nt);
    
    if (tir)
      goto totalInternalReflection;
    
    cosTheta = -r.dir.dot(n);
    attenColor = mtl->getAttenuation();
    atten = !isWhite(attenColor);
    inside = true;
    if (0&&DEBUG_STATE)
    cerr << "IN : ndoti=" << n.dot(r.dir)
         << "OUT: ndoto=" << n.dot(tdir)
         << "\n";
  }
  else {                // out of object
    nc = mtl->getIndexOfRefractionBot();
    nt = mtl->getIndexOfRefractionTop();
    if (gv->spectral){
      nc = cauchy(nc, r.wavelen);
      nt = cauchy(nt, r.wavelen);
    }
    refract(tdir, tir, r.dir, -n, nc/nt);

    if (tir)
      goto totalInternalReflection;
    
    cosTheta = tdir.dot(n); // schlick approximation relies on outside(?) dir
    attenColor = SbVec3f(1,1,1);
    atten = false;
    inside = false;
    if (0&&DEBUG_STATE)
    cerr << "\n"
         << "IN : ndoti=" << n.dot(r.dir) << " "
         << "OUT: ndoto=" << n.dot(tdir)
         << "\n";
  }

  // total internal reflection
  if (tir){
    totalInternalReflection:
    SbColor rcolor = (*this.*getRadiance)(reflect, flags);
    return rcolor;
  }

  // reflect+refract (Fresnel)

  // grazing angle problem (normal interpolation error)
  // don't let transmitted ray be on same side
  if (tdir.dot(ngl)>0.0){
    tdir -= (tdir.dot(ngl)+0.0015) * ngl;
    tdir.normalize();
  }
    
  // due to errors for grazing angles
  // with normal interpolation, c may
  // incorrectly be < 0.0. This clamps it.
  if (cosTheta<0)
    cosTheta = 0;
    
  R0 = (nt-nc)/(nt+nc);
  R0 *= R0;     // same no matter which way (for n=1.7, R0~0.67
  Re = fresnel(R0, cosTheta);

  // use russian roulette
  // (after a few bounces, prescribed in getRadiancePM)
  if (flags & C_RUSSIANR){
    double heur = 0.5;  // interpolate between Re (heur=0) and 0.5 (heur=1)
    double P = (1-heur)*Re + heur*0.5;
      
    // reflect
    // OPTIMIZE: remove transparency check below (added for debugging)
    if (r.ctx->mydrand48()<P || transparency==0.0){
      reflect.importance = r.importance*Re;
      if (DEBUG_STATE) cerr << "REFLECTrr depth=" << r.depth << "\n";
      SbColor rcolor = /*(DEBUG_STATE) ? SbVec3f(0,0,0) : */(*this.*getRadiance)(reflect, flags);
      ncolor = Re*rcolor;
      ncolor /= P;
    }

    // transmit
    else {
      Ray transmit(r.x, tdir, r, r.prim);
      transmit.atten = atten;
      transmit.attenColor = attenColor;
      transmit.inside = inside;
      transmit.importance = r.importance*(1-Re);
      if (DEBUG_STATE) cerr << "REFRACTrr depth=" << r.depth << "\n";
      SbColor tcolor = (*this.*getRadiance)(transmit, flags);
      ncolor = (1-Re)*tcolor;
      ncolor /= (1-P);
    }
  }

  // no russian roulette
  else {

    Ray transmit(r.x, tdir, r, r.prim);
    transmit.atten = atten;
    transmit.attenColor = attenColor;
    transmit.inside = inside;
    transmit.ldInstance+=2; // reflect is +1

    SbColor rcolor(0,0,0);
//    if (DEBUG_STATE) cerr << "REFLECT depth=" << r.depth << "\n";
    rcolor = /*(DEBUG_STATE) ? SbColor(0,0,0) : */(*this.*getRadiance)(reflect, flags);
    SbColor tcolor(0,0,0);

//    if (DEBUG_STATE) cerr << "REFRACT depth=" << r.depth << "\n";
    static int refractionRays = getenv("PANE_REFRACTION_RAYS") ?
      atof(getenv("PANE_REFRACTION_RAYS")) : 1;
    int samples = r.depth==1 ? refractionRays : 1;
    
    int transmitDim = transmit.ldDimension;
    int transmitInst = transmit.ldInstance;
//    cerr << "\n\nBEGIN\n";
    for (int i=0; i<samples; i++){
//      cerr << "tdim=" << transmit.ldDimension << endl;
      tcolor += (*this.*getRadiance)(transmit, flags);
      transmit.ldDimension = transmitDim;
      transmit.ldInstance = ++transmitInst;
    }
    tcolor /= samples;
      
    ncolor = Re*rcolor + (1-Re)*tcolor;
  }
  
  return ncolor;
}


//WWWW''WW''''''''''WW''''''''''''''''WW''''WW''''''''''''''''WW''''''''''
//WW  W WW          WW                WWW  WWW                            
//WW  W WWW   WW   WWWW  WW   WWW     W WWW WW  WWW WWW  WWW  WW WWW   WWW
//WWWW  WW W WW W   WW  WW W  WW W    W  W  WW W WW WW W WW W WW WW W W WW
//WW    WW W WW W   WW  WW W  WW W    W     WW W WW WW W WW W WW WW W W WW
//WW    WW W  WW    WW   WW   WW W    W     WW  WWW WWW  WWW  WW WW W  WWW
//                                                  WW   WW             WW
//''''''''''''''''''''''''''''''''''''''''''''''''''WW'''WW'''''''''''WWW'
// (getRadiancePM)
//
// Returns radiance for a ray, using Photon-mapping path tracing.
// This is vastly more efficient than pure Monte Carlo path tracing,
// which is available in getRadiancePT.
//
// Accurate and Approxmate solutions are available, specified
// using the flags variable, which is the binary OR of the following
// available flags:
//
// C_APPROX, C_PREPASS, C_SIMPLE
// C_EMITTED, C_MDIFFUSE,  C_CAUSTIC, C_SPECULAR, C_DIRECT,  
//
// Note the last five flags above refer to the
// the emitted radiance component, which is trivial,
// plus the four rendering components 
// described in Henrik Jensen's Two-Pass Photon Mapping
// Algorithm, in his book "Realistic Image Synthesis Using
// Photon Mapping" Ch.9.
//
// In a nutshell, incoming radiance is split into
// direct, indirect caustic, and indirect diffuse components,
// while the BRDF is split into specular and diffuse reflection
// components.
//
// Use accurate (no C_APPROX) if:
// seen directly by the eye or maybe a few sepcular
// reflections, or if intersection ray is short
// (to prevent inaccurate color bleeding in corners)
//
// Use approximate (C_APPROX) if:
// ray has been reflected diffusely or ray
// contributes little to pixel radiance
//
// Used for pixel color algorithms (-aN):
// 0: Ray tracing (no indirect illumination, but w/ caustics)
// 1: Photon mapping (all four components, using Photon Mapping)
//
SbColor Pane::getRadiancePM(Ray &r, int flags){

  SbColor color(0, 0, 0);

  // check length of ray isn't accidentally > 1
  if (r.dir.dot(r.dir)>1)
    r.dir.normalize();
  
  // intersection?
  if (!scenePrim->intersectRaySS(r, gv->seeSingleSided)){

    NOHIT_PM:
    
    // debugging
    if (DEBUG_STATE)
      r.show2(getMaxDist()*4.0, color);

    if (flags & C_EMITTED){
      if (envMap)
	color = envMap->getRadiance(r.dir);
      else
	color = gv->worldBG;
    }
    
    r.t = -1.0;

    // testing
//  static bool addAerial = getenv("PANE_SKY_AERIAL");
//     if (addAerial && envMap && envMap->getType() == EnvironMap::SKY && r.depth==1)
//       color = ((Sky*)envMap)->ComputeAerialPersp(r, color);
    
    // experimental
    if (r.depth<gv->maxDepth && gv->stepSize>0) // /*&& !(flags&C_PREPASS)*/ && !r.inside)
      color = volumeIntegrate(r, color);

    return color;
  }

  // set flags
  int c_approx   = flags & C_APPROX;
  int c_emitted  = flags & C_EMITTED;
  int c_explicitl= flags & C_EXPLICITL;
  int c_mdiffuse = flags & C_MDIFFUSE;
  int c_caustic  = flags & C_CAUSTIC;
  int c_specular = flags & C_SPECULAR;
  int c_direct   = flags & C_DIRECT;
  int c_prepass  = flags & C_PREPASS;
  int c_simple   = flags & C_SIMPLE;

  // more flags
  int c_russianr  = c_prepass || c_simple || r.depth>=gv->rrStartDepth;
  if (c_russianr)
    flags |= C_RUSSIANR;
  int c_recurse  = (r.depth < gv->maxDepth)?1:0;
    
  // use approximate values if several specular bounces
  // (if doing global illumination)
  if (r.importance<0.01 && gv->algorithm==1){
    c_approx = 1;
    flags |= C_APPROX;
  }
    
  // use approximate calculation for final gather
  // unless t is small.
  if (c_simple){
    if (r.t>minApproxDistance){
      c_approx = 1;
      flags |= C_APPROX;
    }
  }

  Material *mtl = r.prim->getMat(r);
  
  // compute point of intersection
  r.x = r.point();

  // compute normal
  SbVec3f n = r.prim->getIntNormal(r);   // interpolated normal
  SbVec3f ng = r.prim->getNormal(r);     // geometry normal    
  SbVec3f nl = n;                           // points toward ray source
  SbVec3f ngl = ng;                         // points toward ray source

  // normal correction (two sided lighting)
  if (ng.dot(r.dir)>0.0){
    nl = -nl;
    ngl = -ngl;
  }
    
  if (DEBUG_STATE && gv->showNormals){
    print("t=%f\n",r.t);
    Ray(r.x,n,r.ctx).show(r.t/7.0, SbColor(1.0, 1.0, 0.0));      // yellow
    Ray(r.x,ng,r.ctx).show(r.t/6.0, SbColor(0.0, 1.0, 1.0));     // green
    Ray(r.x,ngl,r.ctx).show(r.t/5.0, SbColor(1.0, 0.0, 1.0));    // purple
    if (r.prim->isOfType(Primitive::PRIMTYPE_TRIANGLE))
      ((Triangle*)r.prim)->show();
    //printVec(n);
    //printVec(ng);
    //cerr << *(r.tri);
  }

  // frequently used constants
  double trans_comp = mtl->getTransparency();
  double reg_comp = 1.0-trans_comp;    
  double shiny_comp = mtl->getShininess();
  double glossy_comp = mtl->getGlossiness();
  double diff_comp = mtl->getDiffuseness();

  // resolve QMC dimension for deterministic ray tree
  int deltaDim = 0;
  //if (trans_comp)
  //  deltaDim += 4;    // glass
  //if (glossy_comp)
  //  deltaDim += 2;    // gloss
  if (diff_comp){
    //deltaDim += 2;    // diffuse reflection
    //deltaDim += 2;    // area lights
    if (envMap)       // env map
      deltaDim += 2;
  }
  if (trans_comp)
    deltaDim *= 2;
  //r.ldDimensionNext += deltaDim;

  double visComp = 1;
  bool skipDiffuse = false;
  
  // do plain OpenGL transparency
  if (gv->useGlassModel == 0){
    if (trans_comp>0.01){
      r.ldNextDim();
      Ray transmit(r.x, r.dir, r);
      transmit.ldInstance++;
      color += getRadiancePM(transmit, flags) * trans_comp;
    }

    visComp *= reg_comp;
    reg_comp = 1.0;
    trans_comp = 0.0;
  }

  // fake alpha
  if (gv->figureMode && gv->tempInt==31){
    SbColor surf = r.prim->getIntDiffColor(r);
    if ((surf-SbColor(0.9, 0.9, 0.99)).length()<0.01){
      const float alpha = gv->tempFloat;
      r.ldNextDim();
      Ray transmit(r.x, r.dir, r, r.prim);
      transmit.ldInstance++;
      color += getRadiancePM(transmit, flags) * alpha;

      visComp *= 1.0-alpha;
    }
  }

  // some constants
  const double invMetersConversion2 = 1.0/gv->meters2(1.0);
  double invPi= M_1_PI; // 1.0/M_PI;
    
  // different brdf
  if (gv->useDimensions<3)
    invPi = 1.0/2.0;
    
  // Emitted (emissive)
  if (c_emitted && mtl->isEmissive() && ng.dot(r.dir)<0.0) {

    SbColor emis = r.prim->getIntEmission(r); // flux in W

    // convert to radiance (divide by area*pi)
    if (r.prim->isOfType(Primitive::PRIMTYPE_EMISTRIANGLE))
      emis *= invPi/gv->meters2( ((EmissiveTriangle*)r.prim)->area );

    if (r.prim->isOfType(Primitive::PRIMTYPE_SPHERE)){
      float rad = ((Sphere*)r.prim)->rad();
      emis *= M_1_PI/gv->meters2((4*M_PI)*rad*rad);
    }
    
    color += emis;
  }

  // volume scattering code may have turned off c_emitted
  if (!c_emitted && c_explicitl)
    flags |= C_EMITTED; // reset flag for future rays
  
  
  //'WWW''''''''''''''''''WW'''''''''''''''''''''''W'''''WWW'WW'''''''''''''
  //W                      W                       W    W     W             
  // WW  WW   WW   WW W W  W   WW  WW     WW WW   WW    W WW  W   W   WWW  W
  //   W W W Wmmm W   W W  W  W W W      W W W W W W    W  W  W  W W  WW   W
  //WWW  WW   WW   WW  WW WWW  WW W       WW W W  WW     WWW WWW  W  WWW  WW
  //'''''W''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  // (specular)

  if (c_specular && c_recurse && mtl->isShiny()){

    SbColor specCol = mtl->getSpecular();

    // fresnel effect
    specCol = fresnel(specCol, -r.dir.dot(nl));

    // russian roulette
    float prob_recurse;
    if (c_russianr){

      // find maximum component
      float maxComp = 0.0;
      for (int i=0; i<3; i++)
        if (specCol[i]>maxComp)
          maxComp = specCol[i];
	
      prob_recurse = shiny_comp*maxComp;

      // increase chance of ray termination
      // this ensures termination, eventually
      if (r.depth>15 || (r.depth>10 && (c_prepass || c_simple))  )
        prob_recurse *= 0.9;
    }
    else
      prob_recurse = 1;

    // recurse?
    if (prob_recurse == 1 || r.ctx->mydrand48()<prob_recurse){

      // russian roullette
      float prob_refract;
      float prob_reflect;
      if (c_russianr){

        // reflect OR refract
        if (trans_comp != 0 && r.ctx->mydrand48() < trans_comp){
          prob_refract = trans_comp;
          prob_reflect = 0.;
        }
        else {
          prob_refract = 0.;
          prob_reflect = reg_comp;
        }

        // do NOT do any diffuse computation
        skipDiffuse = true;
      }
      else {
        prob_refract = trans_comp ? 1 : 0;
        prob_reflect = reg_comp ? 1 : 0;
      }
	
      // Refraction component contribution
      if (prob_refract > 0.0){

        SbColor tcolor = refractionContrib(r, n, ngl, flags);
        vecScaleItself(tcolor, specCol);
        color += tcolor * trans_comp * shiny_comp / prob_refract;
      }

      // Reflection component
      if (prob_reflect > 0.0){
	  
        SbVec3f rdir = r.dir-2.0*n.dot(r.dir)*n;

        // grazing angle problem
        // don't let reflected ray go beneath surface
        if (rdir.dot(ngl)<0.0){
          rdir += (-rdir.dot(ngl)+0.0015) * ngl;
          rdir.normalize();
        }

        r.ldNextDim();
        Ray reflect(r.x, rdir, r, r.prim);
        reflect.ldInstance++;
        SbColor rcolor = getRadiancePM(reflect, flags);
        vecScaleItself(rcolor, specCol);
        color += rcolor * reg_comp * shiny_comp / prob_reflect;
	
      }// end prob_reflect test

    }// end prob_recurse test

  } // end specular component

  // fetch diffuse color
  SbColor surf = r.prim->getIntDiffColor(r);
  r.diffCol = surf;

  // Diffuse (and glossy) contribution

  // loop over top (side==0) and bottom (side==1) hemispheres 
  for (int side=0; side<2; side++){

    if (shiny_comp>=1) continue;

    // Compute diffuse contribution for this side
    double myVisComp = visComp;
    if (side == 0){
      
      if (reg_comp<0.01) continue;      
      myVisComp *= reg_comp;
      
    } else {

      if (trans_comp<0.01) continue;
      myVisComp *= trans_comp;
      n = -n;
      nl = -nl;
      ng = -ng;
      ngl = -ngl;
    }

    // Only do following computations if reasonable amount of diff_comp
    if (skipDiffuse || visComp<0.01 || diff_comp<=0.005 && glossy_comp<=0.005) continue;

    // Approximate solution to Direct, Caustic, and MDiffuse (diffuse) contribution
    if (c_approx && (c_direct || c_caustic || c_mdiffuse)){
	
      float irrad[3];
      if (gv->precomputeIrr)   // precomputed estimate
        gPhotonMap->irradiance_estimate_pre(
          irrad, r.x.getValue(), ngl.getValue());
      else                     // regular estimate
        gPhotonMap->irradiance_estimate(
          irrad, r.x.getValue(), nl.getValue(), gv->gPhotonEstimate);

      // frd * E(x) = p/pi * E(x)         from jensen p.21 eq 2.22
      color += vecScale(surf, SbVec3f(irrad)) *
        ( invPi * invMetersConversion2 * myVisComp*diff_comp );

      continue;  // we're done
    }

    // Accurate solution to diffuse contribution

    // Direct illumination
    if (c_direct){ 

      // Point lights
      color += pointLightContrib(r, nl, ngl) * (diff_comp+glossy_comp)*myVisComp;//FIXME

      // Area lights contribution (incident)
      if (gv->areaLights) {
        color += areaImpLightContrib(r, nl, ngl, c_simple) * myVisComp;
	    
        if (envMap)
          color += envMap->impSample(r, nl, ngl, c_simple) * myVisComp;

        color += sphereLightContrib(r, nl, ngl, c_simple) * myVisComp;
      }
      else {
        if (gv->ambientIllum)
          color += (diff_comp+glossy_comp)*myVisComp * r.diffCol;
        else
          color += areaLightContrib(r, nl, ngl) * myVisComp*(diff_comp+glossy_comp);
      }
    }

    // Caustics
    if (c_caustic){
      float irrad[3];

      cPhotonMap->irradiance_estimate(
        irrad, r.x.getValue(), ngl.getValue(), gv->cPhotonEstimate);

      // f_rd * E(x) = p/pi * E(x)          from jensen p.21 eq 2.22
      color += vecScale(surf, SbVec3f(irrad)) *
        ( invPi * invMetersConversion2 * diff_comp*myVisComp );
    }

    // Multiple diffuse contribution
    if (c_mdiffuse && c_recurse)
        color += multipleDiffuseContrib(r, nl, ngl, c_simple, c_prepass,
                                        FRONTSIDE) * diff_comp*myVisComp;

  } // end hemisphere sides

  // do Beer's law attenuation
  SbColor attenColor;
  bool atten=false;
  if (r.atten){
    attenColor = r.attenColor;
    atten = true;
  } else if (n.dot(ngl) < 0) {  // ray strikes inside of surface
    attenColor = mtl->getAttenuation();
    atten = !isWhite(attenColor);
  }

  if (atten) {

    SbColor att;
    att[R] = attenColor[R] ? exp(log(attenColor[R])*gv->meters(r.t)) : 0;
    att[G] = attenColor[G] ? exp(log(attenColor[G])*gv->meters(r.t)) : 0;
    att[B] = attenColor[B] ? exp(log(attenColor[B])*gv->meters(r.t)) : 0;

    vecScaleItself(color, att);
    
    if (DEBUG_STATE){
      cerr << "\natt = " << att << "\n";
    }
  }
    
  // debugging
  if (DEBUG_STATE)
    r.show2(r.t, color);

  // testing
  static bool addAerial = getenv("PANE_SKY_AERIAL");
  if (addAerial && envMap && envMap->getType() == EnvironMap::SKY && r.depth==1)
    color = ((Sky*)envMap)->ComputeAerialPersp(r, color);
  
  // experimental
  if (r.depth<gv->maxDepth && gv->stepSize>0) // /*&& !c_prepass*/ && !r.inside)
    color = volumeIntegrate(r, color);

  return color;
}


//WWWW''''''''WW''WW''''''WWWWWW'''''''''''''''''WW''''''''''
//WW  W       WW  WW        WW                               
//WW  W  WWW WWWW WWW       WW    W WW  WWW  WWW WW WWW   WWW
//WWWW  W WW  WW  WW W      WW    WW   W WW WW W WW WW W W WW
//WW    W WW  WW  WW W      WW    WW   W WW WW   WW WW W W WW
//WW     WWW  WW  WW W      WW    WW    WWW  WWW WW WW W  WWW
//                                                         WW
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''WWW'
//  
// Used for pixel color algorithms (-aN):
// 2: MCPT w/ Explicit Lighting test, and   (faster)
// 3: MCPT w/o
//
SbColor Pane::getRadiancePT(Ray &r, int flags){
  
  SbColor color = SbColor(0.0, 0.0, 0.0);

  // nan generator for debugging
//   if (r.ctx->mydrand48()>0.999){
//     cerr << "pixel=" << r.ctx->getPixel[0] << "," << r.ctx->getPixel[1] << endl;
//     cerr << r.depth/0;
//   }

//   // debugging
//   if (DEBUG_STATE==2){
//     for (int i=0; i<r.depth; i++)
//       cerr << "  ";
//     cerr << "(" << r.depth << ") getRadiancePT\n";
//   }
  
  // check length of ray isn't accidentally > 1
  if (r.dir.dot(r.dir)>1)
    r.dir.normalize();
  
  // intersection?
  if (scenePrim->intersectRaySS(r, gv->seeSingleSided)){
    
    // set flags
    int c_emitted  = flags & C_EMITTED;
    int c_explicitl  = flags & C_EXPLICITL;
    
    // compute point of intersection
    r.x = r.point();

    // compute normal
    SbVec3f n = r.prim->getIntNormal(r);  // interpolated normal
    SbVec3f ng = r.prim->getNormal(r);      // geometry normal    
    SbVec3f nl = n;                         // points toward ray source
    SbVec3f ngl = ng;                       // points toward ray source

    // normal correction (two sided lighting)
    bool front=true;
    if (ng.dot(r.dir)>0.0){
      nl = -nl;
      ngl = -ngl;
      front = false;
    }

    if (DEBUG_STATE) 
    cerr << "hit: o=" << r.pos << " "
         << "d=" << r.dir << " "
         << "t=" << r.t << " "
         << r.prim->getTypeName() << " "
         << (front ? "front" : "back") << " "
         << "x=" << r.x
         << endl;
    
    if (DEBUG_STATE && gv->showNormals){
      Ray(r.x,n,r.ctx).show(r.t/7.0, SbColor(1.0, 1.0, 0.0));     // yellow
      Ray(r.x,ng,r.ctx).show(r.t/6.0, SbColor(0.0, 1.0, 1.0));    // green
      Ray(r.x,ngl,r.ctx).show(r.t/5.0, SbColor(1.0, 0.0, 1.0));    // purple
      if (r.prim->isOfType(Primitive::PRIMTYPE_TRIANGLE))
        ((Triangle*)r.prim)->show();
      //printVec(n);
      //printVec(ng);
    }

    // frequently used constants
    Material *mtl = r.prim->getMat(r);
    double trans_comp = mtl->getTransparency();
    double reg_comp = 1.0-trans_comp;    
    double shiny_comp = mtl->getShininess();
    double glossy_comp = mtl->getGlossiness();
    double diff_comp = mtl->getDiffuseness();
    double visComp = 1;

    // resolve QMC dimension for deterministic ray tree
    int deltaDim = 0;
    //if (trans_comp)
    //  deltaDim += 4;    // glass
    //if (glossy_comp)
    //  deltaDim += 2;    // gloss
    if (diff_comp){
      //deltaDim += 2;    // diffuse reflection
      //deltaDim += 2;    // area lights
      if (envMap)       // env map
        deltaDim += 2;
    }
    if (trans_comp)
      deltaDim *= 2;
    //r.ldDimensionNext += deltaDim;

    // do plain OpenGL transparency
    if (gv->useGlassModel == 0){
      if (trans_comp>0.01){
        r.ldNextDim();
        Ray transmit(r.x, r.dir, r, r.prim);
        transmit.ldInstance++;
	color += getRadiancePT(transmit, flags) * trans_comp;
      }

      visComp *= reg_comp;
      reg_comp = 1.0;
      trans_comp = 0.0;
    }

    // get colors
    SbColor specCol = mtl->getSpecular();
    SbColor specColFresnel = (mtl->isShiny() || mtl->isGlossy()) ?
      fresnel(specCol, -r.dir.dot(nl)) : specCol;
//    SbColor specColFresnel = specCol;
    SbColor surf = r.diffCol = r.prim->getIntDiffColor(r);

    if (c_explicitl && shiny_comp<1){

      // transmit
      SbVec3f mynl, myngl;
      if (r.ctx->mydrand48()<trans_comp){
        myngl = -ngl;
        mynl = -nl;
      } else {
        myngl = ngl;
        mynl = nl;
      }
      
      // Point lights contribution
      color += pointLightContrib(r, mynl, myngl) * (diff_comp+glossy_comp)*visComp;

      // Area lights contribution (incident)
      static bool skipUnderwaterLights = getenv("PANE_SKIP_UNDERWATER_LIGHTS");
      if (!skipUnderwaterLights || !r.atten) 
        color += areaImpLightContrib(r, mynl, myngl, C_SIMPLE) * visComp;

      // Area lights contribution (incident)
      if (envMap)
	color += envMap->impSample(r, mynl, myngl, C_SIMPLE) * visComp;

      color += sphereLightContrib(r, mynl, myngl, C_SIMPLE) * visComp;
    }

    // Emissive contribution
    if (c_emitted && r.dir.dot(ng)<0.0f && mtl->getEmisAvg()>0.0f){

      SbColor emis = mtl->getEmission(); // flux in W

      // convert to radiance (divide by area*pi)
      if (r.prim->isOfType(Primitive::PRIMTYPE_EMISTRIANGLE))
        emis *= M_1_PI/gv->meters2( ((EmissiveTriangle*)r.prim)->area );

      if (r.prim->isOfType(Primitive::PRIMTYPE_SPHERE)){
        float rad = ((Sphere*)r.prim)->rad();
        emis *= M_1_PI/gv->meters2((4*M_PI)*rad*rad);
      }
      
      color += emis;
    }
    else if (!c_emitted && c_explicitl)
      flags |= C_EMITTED; // reset flag for future rays

    static bool disableRR = getenv("PANE_MCPT_RR_OFF");
    if (disableRR){
      // re-enable RR
      int c_russianr  = r.depth >= 2; //gv->rrStartDepth;
      if (c_russianr)
        flags |= C_RUSSIANR;
    }
    
    int c_recurse = r.depth<gv->maxDepth;
    double prob_recurse;
    if (c_recurse){

      // choose reflection type
      double u = r.ctx->mydrand48();
      double choiceCDF[4];
      choiceCDF[0] = trans_comp*shiny_comp;                // refraction
      choiceCDF[1] = reg_comp*shiny_comp   + choiceCDF[0]; // reflection
      choiceCDF[2] = 1;                                    // diffuse + glossy
      int choice = 0;
      while (u>choiceCDF[choice]){
        choice++;
      }

      // Refraction component contribution
      if (choice == 0){
	//prob_recurse = vecAvg(specCol)*.75 + .25;
        float avgSpecCol = vecAvg(specCol);
        prob_recurse = avgSpecCol ? (r.depth<gv->rrStartDepth ? 1 : avgSpecCol*.75 + .25) : 0;
	if (r.ctx->mydrand48()<prob_recurse){
	  SbColor rcolor = refractionContrib(r, n, ngl, flags);
          vecScaleItself(rcolor, specCol);
	  color += rcolor * 1.0/prob_recurse;
	}
        else if (DEBUG_STATE) cerr << "ABORT recursion!\n\n";
      }

      // Shiny component contribution
      else if (choice == 1){

        float avgCol = vecAvg(specColFresnel);
	prob_recurse = avgCol ? (r.depth<gv->rrStartDepth ? 1 : avgCol*.75 + .25 ) : 0;
	if (r.ctx->mydrand48()<prob_recurse){
	
	  SbVec3f rdir = r.dir-2.0*n.dot(r.dir)*n;

	  // grazing angle problem
	  // don't let reflected ray go beneath surface
	  if (rdir.dot(ngl)<0.0){
	    rdir += (-rdir.dot(ngl)+0.0015) * ngl;
	    rdir.normalize();
	  }

          r.ldNextDim();
          Ray reflect(r.x, rdir, r, r.prim);
          reflect.ldInstance++;
	  SbColor rcolor = getRadiancePT(reflect, flags);
          vecScaleItself(rcolor, specColFresnel);
	  color += rcolor * 1.0/prob_recurse * visComp;
	}
        else if (DEBUG_STATE) cerr << "ABORT recursion!\n\n";
      }

      // diffuse component
      else if (choice == 2){

        float avgDiffCol = vecAvg(surf);
        float avgSpecColFres = vecAvg(specColFresnel);
	float prob_recurse_diff = avgDiffCol*.5 + .5;
	float prob_recurse_glossy = avgSpecColFres*.5 + .5;
        float prob_recurse =  // weighted average
          (avgDiffCol || avgSpecColFres) ?
          (r.depth<gv->rrStartDepth ? 1 :
           ((diff_comp*prob_recurse_diff + glossy_comp*prob_recurse_glossy)/
            (diff_comp+glossy_comp))) : 0;
	if (r.ctx->mydrand48()<prob_recurse){

	  // transmit
          SbVec3f dir = r.dir;
	  if (r.ctx->mydrand48()<trans_comp){
            dir = r.dir-2.0*n.dot(r.dir)*n;
	    ngl = -ngl;
            nl = -nl;
          }

          r.ldNextDim();
          Ray bounce(r.x, ngl, r, r.prim);
          bounce.ldInstance++;

          double eps1 = r.ctx->mydrand48();  // FIXME:: use quasirandom sequence
          double eps2 = r.ctx->mydrand48();
          double eps3 = r.ctx->mydrand48();

          BRDF *brdf = mtl->getBRDF();
          double brdfVal, pdfVal, brdfOverPdfVal, brdfOverPdfCosTval;
          SbVec3f brdfColor;
          brdf->GetSampleDir(eps1, eps2, eps3, dir, nl, ngl, r, bounce.dir,
                             brdfVal, pdfVal, brdfOverPdfVal, brdfOverPdfCosTval,
                             brdfColor);
            
          // make sure vector goes up
          if (brdfOverPdfCosTval>0){
            if (c_explicitl)
              flags ^= C_EMITTED;
            SbColor gcolor = getRadiancePT(bounce, flags);
            vecScaleItself(gcolor, brdfColor);
            gcolor *= (1.0/prob_recurse) * brdfOverPdfCosTval * visComp / (diff_comp+glossy_comp);
            color += gcolor;
          }
        }
        else if (DEBUG_STATE) cerr << "ABORT recursion!\n\n";
      } else {
        GIVE_ERROR("invalid choice\n.");
      }
    }
    else if (DEBUG_STATE) cerr << "MAXDEPTH abort\n\n";
    
    // debugging
    if (DEBUG_STATE)
      r.show(r.t, color);

    // do Beer's law attenuation
    SbColor attenColor;
    bool atten=false;
    if (r.atten){
      attenColor = r.attenColor;
      atten = true;
    } else if (n.dot(ngl) < 0) {  // ray strikes inside of surface (exiting)
      attenColor = mtl->getAttenuation();
      atten = !isWhite(attenColor);
    }

    if (atten) {

      SbColor att;
      att[R] = attenColor[R] ? exp(log(attenColor[R])*gv->meters(r.t)) : 0;
      att[G] = attenColor[G] ? exp(log(attenColor[G])*gv->meters(r.t)) : 0;
      att[B] = attenColor[B] ? exp(log(attenColor[B])*gv->meters(r.t)) : 0;

      vecScaleItself(color, att);

      if (DEBUG_STATE){
        cerr << "att = " << att << "\n";
      }
    }
    
  }
  // light did not intersect anything
  else {
    NOHIT_MCPT:

    if (DEBUG_STATE) cerr << "MISS!\n\n";

    r.t = -1;
    if (!gv->environMapFile.empty()){
      if (flags & C_EMITTED)
	color = envMap->getRadiance(r.dir);
    }
    else
      color = gv->worldBG;
    
    // debugging
    if (DEBUG_STATE)
      r.show(getMaxDist()/4.0, color);

    // do Beer's law attenuation
    if (r.atten)
      color = SbVec3f(0,0,0);
  }

  static bool addAerial = getenv("PANE_SKY_AERIAL");
  if (addAerial && envMap && envMap->getType() == EnvironMap::SKY && r.depth==1)
    color = ((Sky*)envMap)->ComputeAerialPersp(r, color);
  
  // experimental
  if (r.depth<gv->maxDepth && gv->stepSize>0) // /*&& !(flags&C_PREPASS)*/ && !r.inside)
    color = volumeIntegrate(r, color);
  
  return color;
}


SbColor Pane::getRadianceDepth(Ray &r, int flags){

  SbColor color = SbColor(0.0, 0.0, 0.0);
  
  // intersection?
  if (scenePrim->intersectRaySS(r, gv->seeSingleSided)){

    // compute normal
    SbVec3f n = r.prim->getIntNormal(r);  // interpolated normal
    
    // debugging
    if (DEBUG_STATE)
      r.show(r.t, color);

    float ndotl = fabs(n.dot(r.dir));
    float depth = (2*scenePrim->getRadius()-r.t)/(2*scenePrim->getRadius());
    depth = powf(depth, 16.f);
    float c = ndotl*depth;
    color.setValue(c,c,c);

    if (gv->ambientIllum){
      SbColor surf = r.prim->getIntDiffColor(r);
      vecScaleItself(color, surf);
    }

  } else {

    r.t = -1;
    color = gv->worldBG;
    
    // debugging
    if (DEBUG_STATE)
      r.show(getMaxDist()/4.0, color);
  }

  return color;
}

// Sets the gv->cam variable.
void Pane::findCamera(){

  // Search for camera in scene graph
  SoSearchAction sa;
  sa.setType(SoCamera::getClassTypeId());
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.apply(gv->root);
  if (sa.getPath())
    gv->cam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();

  // else Look in root at second child (put there in Main.cxx)
  else if (gv->root->getChild(1)->isOfType(SoCamera::getClassTypeId()))
    gv->cam = (SoCamera *)gv->root->getChild(1);

  // or give up
  else {
    cerr << "Could not find camera!" << endl;
    return;
  }
}

// Positions the headlight in the litPts list.
// Call findCamera() first.
void Pane::configureHeadlight(){

  // position headlight (to above and left of camera)
  if (gv->headLightPos==HL_OFFCENTER){
    SbVec3f e, g, vup, w, u, v, x, hl;
    SbRotation sbr;
    
    gv->cam->orientation.getValue().multVec(SbVec3f(0.0, 1.0, 0.0), vup);
    e = SbVec3f(gv->cam->position.getValue());
    sbr = gv->cam->orientation.getValue();
    sbr.multVec(SbVec3f(0.0, 0.0, -1.0), g);
    g.normalize();
    w = SbVec3f(-g);
    x = SbVec3f(vup.cross(w));
    u = SbVec3f(x/x.length());
    v = SbVec3f(w.cross(u));
    float maxDist = getMaxDist();
    if (maxDist<=0) maxDist = 1;
    hl = -u*maxDist/30.0 + v*maxDist/30.0 + e;
    (*(litPnts.begin()))->location.setValue(hl);
  }
  else if (gv->headLightPos != HL_OFF){
    (*(litPnts.begin()))->location.setValue(gv->cam->position.getValue());
  }

  // still need to account for orthographic camera
}


void Pane::applyAttenuation(const Ray &r, const SbVec3f &a, const SbVec3f &b, float t,
                            SbVec3f &color){

  // ray marching absorbtion
  if (gv->stepSize>0.0f){
    float tau = 0.0;
    tau = allDataSets->bresenSum(a, b);
    tau = gv->meters(tau);
    color *= exp(-tau);
  }
  
  // do Beer's law attenuation
  if (!r.atten || t<=0) return;
  if (t>0.5*1e6) {
    color = SbVec3f(0,0,0);
    return;
  }
    
  SbColor attenColor = r.attenColor;
  SbColor att;
  att[R] = attenColor[R] ? exp(log(attenColor[R])*gv->meters(r.t)) : 0;
  att[G] = attenColor[G] ? exp(log(attenColor[G])*gv->meters(r.t)) : 0;
  att[B] = attenColor[B] ? exp(log(attenColor[B])*gv->meters(r.t)) : 0;

  vecScaleItself(color, att);
}
