#include "Pane.h"
#include "debug.h"
#include "CameraList.h"
#include "Halton.h"
#include "Photon_map.h"
#include "PixelIterator.h"
#include "PreviewQuad.h"
#include "Spectrum.h"
#include "sobol.h"
#include <memory>

using namespace std;

//'''''''''''''''''WW''''''''''''WW'WW'
//                 WW           WW   WW
//W WW  WW  WWW   WWW  WW  W WW WW   WW
//WW   WWWW WW W W WW WWWW WW   WW   WW
//WW   WW   WW W W WW WW   WW   WW   WW
//WW    WWW WW W  WWW  WWW WW    WW WW 
//                                     
//'''''''''''''''''''''''''''''''''''''
// (render)
//
// Called via a keypress or the main program.
//
// Prepares to compute the image pixel values, then
// creates threads which call renderTile() to compute the values
// for each sub-region.
//
// Assigns each thread one row, except in the case of
// using a sample map, where each thread uses random samples.
void Pane::render(int trace){

  // exit cleanup code (up here to prevent "crosses initialization" compiler
  // errors
  if (0){
    renderExit:
    
    IS_RENDERING=false;
    return;
  }

  cerr << "Preparing to render...\n";
  IS_RENDERING=true;

  // reset image
  for (int j=0; j<gv->imgDim[Y]; j++)
    for (int i=0; i<gv->imgDim[X]; i++)
      setPix(i, j, SbColor(0.0, 0.0, 0.0), 0.0, 0.0);
  
  // check for lights
  static bool checkLights = !getenv("PANE_SKIP_LIGHT_CHECK");
  if (gv->environMapFile.empty() &&
      gv->sceneHasLights==0 &&
      isBlack(gv->worldBG) &&
      checkLights &&
      gv->algorithm != 4 && gv->algorithm != 5){
    cerr << "ERROR: Scene has no lights!! Aborting render!\n";
    goto renderExit;
  }

  // ppm setup
  if (gv->ppmIter>=0){
    RNG mydrand48(gv->ppmIter);
    ppmImgOffset[0] = mydrand48();
    ppmImgOffset[1] = mydrand48();
  }

  allocatePhotonMaps();

  currentlyRendering = true;
  
  // Set default photon estimate
  // better to have too many that not enough for given radius size
  if (gv->gPhotonEstimate<0){
    gv->gPhotonEstimate = clamp(gPhotonMap->get_stored_photons()/100, 8, 100);
    cerr << "Using global photon estimate of " << gv->gPhotonEstimate << ".\n";
  }
  if (gv->cPhotonEstimate<0){
    gv->cPhotonEstimate = clamp(cPhotonMap->get_stored_photons()/100, 8, 50);
    cerr << "Using caustic photon estimate of " << gv->cPhotonEstimate << ".\n";
  }
  if (gv->vPhotonEstimate<0){
    gv->vPhotonEstimate = clamp(cPhotonMap->get_stored_photons()/100, 8, 100);
    cerr << "Using volume photon estimate of " << gv->cPhotonEstimate << ".\n";
  }

  // balance Photon map
  static int last_stored_photons = 0;
  if (last_stored_photons !=
      (gPhotonMap->get_stored_photons() +
       cPhotonMap->get_stored_photons() +
       volumeMap->get_stored_photons() ) ){
    
    last_stored_photons = gPhotonMap->get_stored_photons() +
      cPhotonMap->get_stored_photons() +
      volumeMap->get_stored_photons();

    // start balancing timer
    struct timeval start, end;
    gettimeofday(&start, NULL); 
    
    cerr << "Balancing global photon map...";
    TIME(gPhotonMap->balance(gv->gPhotonEstimate));
    //gPhotonMap->print_stats();
    //gPhotonMap->dump_map();

    if (cPhotonMap->get_stored_photons()){
      cerr << "Balancing caustic photon map...";
      TIME(cPhotonMap->balance(gv->cPhotonEstimate));
    }

    if (volumeMap->get_stored_photons()){
      cerr << "Balancing volume photon map...";
      TIME(volumeMap->balance(gv->vPhotonEstimate));
    }

    // stop balancing timer
    gettimeofday(&end, NULL);   // end delay measure
    gv->times.photonBalancing = delay(start, end);

    // set mininum distance for ray to travel to use approximation
    minApproxDistance = gPhotonMap->get_max_radius();

    // precompute irradiance
    if ((gv->algorithm==1 && gv->precomputeIrr) || gv->showPhotons){

      cerr << "Precomputing irradiance for global photon map...\n";
      gettimeofday(&start, NULL); // start delay measure

      // this stage crashes the previewer
      if (gv->preview && gv->xtViewer)
        gv->xtViewer->setAutoRedraw(false);  // disable preview
      float estdist = gPhotonMap->precompute_irradiance(gv->gPhotonEstimate,
							gv->numThreads,
							gv->showPhotons);
      if (gv->preview && gv->xtViewer)
        gv->xtViewer->setAutoRedraw(true);   // re-enable preview
      
      minApproxDistance = estdist;
      
      cerr << "Average maximum photon distance in estimate is " << estdist << "\n";
      gettimeofday(&end, NULL);   // end delay measure
      gv->times.photonIrradiance = delay(start, end);
      cerr << "Time: " << gv->times.photonIrradiance << "ms" << endl;
    }

    // report minimum approx distance
    cerr << "Using minimum approximation distance " << minApproxDistance << ".\n";
  }

  // re-initialize preview buffer
  if (gv->preview) {
    bool resetTexture = gv->ppmIter<0;
    gv->preview->preRender(resetTexture);
  }

  // feedback
  struct timeval start, end;
  cerr << "Rendering...\n";
  gettimeofday(&start, NULL); // start delay measure
  struct tms renderTimes;
  times(&renderTimes);
  clock_t startTicks = renderTimes.tms_utime + renderTimes.tms_cutime;
    
  // find camera
  findCamera();

  // positon headlight
  configureHeadlight();
  
  // precompute multipleDiffuseContrib tables
  precomputeMDCTables();
  
  // basic canvas setup, mostly camera settings
  canvasSetup(canvas, gv->cam);

  // further canvas setup
  // set up flags
  if (gv->renderFlags){
    canvas.flags = gv->renderFlags | gv->approximate * Pane::C_APPROX;
    cerr << "Rendering with: ";
    if (canvas.flags & C_APPROX)   cerr << "approx ";
    if (canvas.flags & C_SPECULAR) cerr << "specular ";
    if (canvas.flags & C_DIRECT)   cerr << "direct ";
    if (canvas.flags & C_EMITTED)  cerr << "emitted ";
    if (canvas.flags & C_CAUSTIC)  cerr << "caustic ";
    if (canvas.flags & C_MDIFFUSE) cerr << "mdiffuse ";
    if (canvas.flags & C_VOLDIR)   cerr << "voldir ";
    if (canvas.flags & C_VOLINDIR) cerr << "volindir ";
    cerr << "\n";
  }
  else {
    canvas.flags = Pane::C_NORMAL;
    if (gv->algorithm==1) canvas.flags |= Pane::C_MDIFFUSE;
    if (gv->approximate)  canvas.flags |= Pane::C_APPROX;
    if (gv->stepSize)     canvas.flags |= Pane::C_VOLDIR;
    if (gv->stepSize && gv->numVPhotons>0) canvas.flags |= Pane::C_VOLINDIR;
  }
  
  // set function pointer for getRadiance()
  if (gv->algorithm==0 ||           // Ray tracing
      gv->algorithm==1)             // Photon Mapping
    getRadiance = &Pane::getRadiancePM;
  else if (gv->algorithm==2 ||      // MCPT w/EL
	   gv->algorithm==3){       // MCPT
    getRadiance = &Pane::getRadiancePT;

    // this violates this idea of mono-path tracing.. but it might look better
    static bool disableRR = getenv("PANE_MCPT_RR_OFF");
    if (!disableRR)
      canvas.flags |= C_RUSSIANR;     // No branching
    
    if (gv->algorithm==2)
      canvas.flags |= C_EXPLICITL;
    else
      canvas.flags &= ~C_EXPLICITL;
  } else if (gv->algorithm==5){
    getRadiance = &Pane::getRadianceDepth;
  }
  else 
    getRadiance = &Pane::getRadiancePM;    // default
  
  // further setup
  gv->useIrradCache = gv->approximate?0:1 &&
    canvas.flags&C_MDIFFUSE &&
    gv->algorithm==1 &&
    gv->irradCacheError>0.0 &&
    gv->useIndirectImpSampling == 0 &&
    gv->bumpMapping == 0;
  canvas.prepass = gv->useIrradCache && canvas.flags && gv->prepass;
  
  // trace pixel 
  canvas.trace = trace;
  if (trace){      // clear previous rays
    Ray r;
    r.show(0.0, SbColor(0.0, 0.0, 0.0), 1); // clears previous rays
  }

  // reset completion meters
  for (int i=0; i<maxThreads; i++){
    completion[i] = 0.0;
    completion_stage[i] = ' ';
    completionOrder[i] = -1;
  }
  
  // initialize mutexes
  for (int i=0; i<mutexNum; i++)
    pthread_mutex_init(&mutexes[i], NULL);

  // launch threads
  MemberFunction *memberFunc = (MemberFunction *)alloca(
    sizeof(MemberFunction)*gv->maxThreads); // used to pass object pointer
  thread_context *ctx = (thread_context *)alloca(
    sizeof(thread_context)*gv->maxThreads);   // random number generators


  // RE-seed random number generator
  int seed = gv->srandSeed;
  if (gv->ppmIter >= 0)
    seed += gv->ppmIter;

  srand48(seed);             // deterministic (default 0)
  mydrand48.setSeed(seed);   // platform independent RNG
  sobinit(20);               // initialize sobol point generator
  
  // scramble random number generators
  for (int i=0; i<gv->numThreads; i++){
    ctx[i].mydrand48 = RNG(mydrand48.lrand48());
    ctx[i].id = i;
    ctx[i].rays = 0;
    ctx[i].isects = 0;
    ctx[i].ticks = 0;
  }
  
  // render iv
  if (gv->renderIv)
    renderIv(&ctx[0]);
  
  // render with one thread
  else if (gv->numThreads == 1){
    renderTile(&ctx[0]);
  }
  
  // render with multiple threads
  else {

    for (int i=0; i<gv->numThreads; i++){

      // pack object pointer, function pointer, and argument into struct
      memberFunc[i].object = this;
      memberFunc[i].pmf = &Pane::renderTile;
      memberFunc[i].ctx = &ctx[i];

      // call wrapper function with struct
      pthread_create(&gv->threads[i], NULL, Pane::renderTileWrapper,
		     (void*)&memberFunc[i]);
    }
  
    // wait for threads to finish
    for (int i=0; i<gv->numThreads; i++){
      pthread_join(gv->threads[i], NULL);
    }
  }
  
  for (int i=0; i<gv->numThreads; i++){
    renderRaysFired += ctx[i].rays;
    renderIsects    += ctx[i].isects;
    gv->times.renderTicks += ctx[i].ticks;
    gv->times.totalTicks += ctx[i].ticks;
  }
  
  // destroy mutexes
  for (int i=0; i<mutexNum; i++)           
    pthread_mutex_destroy(&mutexes[i]);

  if (gv->algorithm != 4){
    
    // normalize sample values
    for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++)
      if (pixWeight[i]>0.0f)
	pix[i] /= pixWeight[i];
    
    // render finished, update preview
    int lowerLeft[2] = {0};
    if (gv->preview && gv->ppmIter<0)
      gv->preview->updatePreview(lowerLeft, gv->imgDim);
  }
  
  print("\n%d stored points in irradiance cache (%d rejected).\n",
	irradCache.numPoints(), rejectedICPoints);

  // timing
  times(&renderTimes);
  clock_t endTicks = renderTimes.tms_utime + renderTimes.tms_cutime;
  gv->times.renderTicks += endTicks - startTicks + 1;
  long ticksPerSec = sysconf(_SC_CLK_TCK);
  float ticksMs = gv->times.renderTicks/(double)ticksPerSec*1000.0;
  print ("\nRendering CPU ticks: %d, Ticks/s: %d\n", gv->times.renderTicks, ticksPerSec);
  print("Rendering CPU time:    (%f ms, %f s, %f m, %f h)\n",
	ticksMs, ticksMs/1000.0, ticksMs/1000.0/60.0, ticksMs/1000.0/60.0/60.0);
  
  gettimeofday(&end, NULL);   // end delay measure
  gv->times.render = delay(start, end); // renderTime - photon balancing + irradiance
  int ms = MAX2(1,gv->times.render);

  print("Rendering actual Time: (%d ms, %f s, %f m, %f h)\n", ms, ms/1000.0, ms/1000.0/60.0,
	ms/1000.0/60.0/60.0);

  // rays statistics
  print("%llu total rays fired, for %.2f rays/sec (cpu time), "
	"%.2f rays/sec (actual time)\n"
	"%.2f intersections/ray average\n",
	renderRaysFired,
	renderRaysFired/ticksMs*1000.0,
	renderRaysFired/(ms/1000.0),
	renderRaysFired ? renderIsects/(float)renderRaysFired : 0
	);

  renderRaysFired = 0;
  renderIsects = 0;
  photonRaysFired = 0;
  
  // bell
  if (!gv->silent && !STOP_RENDERING)
    print("\a");
  
  print("Done.\n");
  
  toneMapped = 0;

  currentlyRendering = false;

  // update ppm buffer
  if (gv->ppmIter>=0){

    // add image to ppm buffer
    int size = gv->imgTotalDim[X]*gv->imgTotalDim[Y];
    for (int i=0; i<size; i++)
      ppmPix[i] += pix[i];
    ppmPixImages++;

    // normalize sum back to original image
    if (ppmPixImages>1) {
      float inv = 1.f/ppmPixImages;
      for (int i=0; i<size; i++) {
        pix[i] = ppmPix[i] * inv;
        pixWeight[i] = 1;
      }
    }

    // update preview
    if (gv->preview){
      gv->preview->setMult(1.0);
      int lowerLeft[2] = {0};
      int *upperRight = gv->imgDim;
      gv->preview->copyPixToPreviewPPM(lowerLeft, upperRight,
                                       pix, gv->subPixDim);
      gv->preview->redrawPreview(lowerLeft, upperRight);
    }

    gv->ppmIter++;

    // clear photon maps
    cerr << "Clearing photon maps.\n";
    if (gPhotonMap) gPhotonMap->clear();
    if (cPhotonMap) cPhotonMap->clear();
    if (importanceMap) importanceMap->clear();
    if (volumeMap) volumeMap->clear();
  }
  
  if (writeWhenDone && !currentlyRenderingPPMRange) {
    writeWhenDone = false;
    writeImg();
  }

  goto renderExit;
} // end render()


//'''''''''''''''''WW'''''''''''WWWWWW''WW'WWW''''''''WW'WW'
//                 WW             WW        WW       WW   WW
//W WW  WW  WWW   WWW  WW  W WW   WW    WW  WW   WW  WW   WW
//WW   WWWW WW W W WW WWWW WW     WW    WW  WW  WWWW WW   WW
//WW   WW   WW W W WW WW   WW     WW    WW  WW  WW   WW   WW
//WW    WWW WW W  WWW  WWW WW     WW    WW WWWW  WWW  WW WW 
//                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
// (renderTile)
//
// Renders a tile (row) of the image. Computes pixel
// color for each pixel in tile (except when using a
// sample map, then computes values for random pixels).
//
// Computing pixel colors is done in two passes.
// This is to reduce artifacts from irradiance caching.
// With one pass, subsequent pixels will see and interpolate
// indirect lighting samples that previous pixels didn't have yet,
// creating visible lines in the indirect illumination.
// 
// The first pass computes at least one indrect illumination
// sample for each pixel, roughly populating the cache.
//
// The second, final pass, computes the color for each pixel.
//
void Pane::renderTile(thread_context *ctx){

  // for each pixel
  int i, j;
  Ray r;
  r.ctx = ctx;
  r.depth = 1;    // primary ray
  r.pos = canvas.e;
  r.prim = NULL;
  r.ignorePrim = NULL;
  r.ldInstance = 0;
  r.ldDimension = 0;
  r.ctx->registerThread();
  
  // image size calc (in image space)
  SbColor color = SbColor(0.0, 0.0, 0.0);
  int lowerLeft[2];
  int upperRight[2];
  double offset[2];

  // render range
  lowerLeft[X] = 0;
  lowerLeft[Y] = 0;
  upperRight[X] = gv->imgDim[X];
  upperRight[Y] = gv->imgDim[Y];

  // Restrict pixels to debugging range
  if (gv->debugRange >= 0){

    // get pixel center
    int center[2];
    if (gv->debugPixel[X]<0 && gv->debugPixel[Y]<0){
      // trace center
      center[X] = gv->imgDim[X]/2;
      center[Y] = gv->imgDim[Y]/2;
    }
    else {
      center[X] = gv->debugPixel[X];
      center[Y] = gv->imgDim[Y]-gv->debugPixel[Y]-1;
    }
    
    int size = gv->debugRange;
    lowerLeft[X]  = center[X] -size;
    lowerLeft[Y]  = center[Y] -size;
    upperRight[X] = center[X] +size+1;
    upperRight[Y] = center[Y] +size+1;
  }

  // Trace a single pixel
  if (canvas.trace){

    if (gv->debugPixel[X]<0 && gv->debugPixel[Y]<0){
      // trace center of image
      i = gv->imgDim[X]/2;
      j = gv->imgDim[Y]/2;
    }
    else {
      i = gv->debugPixel[X];
      j = (gv->imgDim[Y]-1)-gv->debugPixel[Y];
    }

    lowerLeft[X] = i;
    lowerLeft[Y] = j;
    upperRight[X] = i+1;      
    upperRight[Y] = j+1;
  }
      
  // bounds checking
  lowerLeft[X] = MAX2(lowerLeft[X], 0);
  lowerLeft[Y] = MAX2(lowerLeft[Y], 0);
  upperRight[X] = MIN2(upperRight[X], gv->imgDim[X]);
  upperRight[Y] = MIN2(upperRight[Y], gv->imgDim[Y]);    

  float pixDepthAvg = 0;
  int pixDepthCnt = 0;

  //'WWW'''''''''''''''WW''''''''''W''''W''''''''
  //W                   W          WW  WW        
  // WW   WW WW W  WW   W   WW     W WW W  WW WW 
  //   W W W W W W W W  W  Wmmm    W    W W W W W
  //WWW   WW W W W WW  WWW  WW     W    W  WW WW 
  //               W                          W  
  //'''''''''''''''W''''''''''''''''''''''''''W''
  //
  // If a sample map is provided, use rejection sampling
  // of the map (scaled) to choose pixels to sample.
  if (sampMap){

    // divvy up samples
    for (int i=0; i<gv->numSamples/gv->numThreads; i++){

      // completion meter
      completion[ctx->id] = 100.0*i/(gv->numSamples/gv->numThreads);
      if (ctx->id==0 && i%100==0){
	pthread_mutex_lock(&mutexes[LOCK_STDERR]);
	print("\r");
	for (int i=0; i<gv->numThreads; i++)
	  print("%6.2f%%%c ", completion[i], completion_stage[i]);
	pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
      }
    
      // rejection sample the map
      float du, dv;
      do {
	du = ctx->mydrand48();
	dv = ctx->mydrand48();
      } while (ctx->mydrand48()>sampMapValue(du, dv));

      // set direction
      r.dir = SbVec3f(

	// sideways component
	( -canvas.width/2.0 + du * canvas.width ) *
	canvas.u +

	// vertical component
	( -canvas.height/2.0 + dv * canvas.height ) *
	canvas.v +
	
	// distance ahead component
	-canvas.s*canvas.w  
	);
      
      // adjust for orthographic camera
      if (canvas.orthographic){
	r.pos = canvas.e + r.dir + canvas.s*canvas.w;
	r.dir = -canvas.w;
      }
      r.dir.normalize();
      r.ldDimension = 0;
      r.ldInstance = i;

      // debugging a single ray (samp map)
      SbColor c;
      if (canvas.trace){
	DEBUG_STATE=1;
	c = (*this.*getRadiance)(r, canvas.flags);
	DEBUG_STATE=0;

	pthread_mutex_lock(&mutexes[LOCK_STDERR]);
	print("Sample color: ");
	printCol(c);
	print("\n");
	pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
      }
      else
	c = (*this.*getRadiance)(r, canvas.flags);

      int x = (int)(du*gv->imgDim[X]);
      int y = (int)(dv*gv->imgDim[Y]);
      setPix(x, y, c, r.t);
    }
    
  }

  // max frames per sec
  else if (gv->algorithm==4){

    // trace 2x2 blocks
    for (j=lowerLeft[Y]; j<upperRight[Y]; j+=2){

      // progress meter
//       if (ctx->id==0){
// 	pthread_mutex_lock(&mutexes[LOCK_STDERR]);
// 	print("\r");
// 	for (int i=0; i<gv->numThreads; i++)
// 	  print("%6.2f%%%c ", completion[i], completion_stage[i]);
// 	pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
//       }
      
      for (i=lowerLeft[X]+2*ctx->id; i<upperRight[X]; i+=2*gv->numThreads){

	for (int tileY=0; tileY<2; tileY++){
	  for (int tileX=0; tileX<2; tileX++){

	    const int ii = i+tileX;
	    const int jj = j+tileY;

	    // create Ray
	    r.dir = SbVec3f(

	      // sideways component
	      ( -canvas.width/2.0 +
		( (double)ii + 0.5) * canvas.width /
		gv->imgDim[X] ) *
	      canvas.u +

	      // vertical component
	      ( -canvas.height/2.0 +
		( (double)jj + 0.5) * canvas.height /
		gv->imgDim[Y] ) *
	      canvas.v +

	      // distance ahead component
	      -canvas.s*canvas.w  
	      );

	    // adjust for orthographic camera
	    if (canvas.orthographic){
	      r.pos = canvas.e + r.dir + canvas.s*canvas.w;
	      r.dir = -canvas.w;
	    }
	    r.dir.normalize();

            bool is = scenePrim->intersectRay(r);

            SbColor c;
            if (gv->octreeDebug){
              c.setValue(r.accelInts/255., r.accelTravs/255.,0);
              //cerr << "c=" << c << "\n";
              r.t = 0.1;
            }
            else if (!is){
	      c.setValue(0,0,0);
              r.t = 0.1;
	    }
            else {
              // compute n.dot(l)
              //SbVec3f n = r.prim->getNormal(r);
              SbVec3f n = r.prim->getIntNormal(r);
              float v = n.dot(r.dir);
              if (v<0.0) v = -v;
              //float v=r.t*(1.f/200.f);
              c.setValue(v,v,v);
            }

            setPix(ii, jj, c, r.t);

            if (gv->preview && gv->ppmIter<0)
              gv->preview->setPreviewPix(ii, jj, ctx->id, c,
                                         1, 1,
                                         lowerLeft, upperRight);
            
	  }
	}
      }
    }

    if (gv->preview){
      gv->preview->setMult(1.0);
      gv->preview->copyPixToPreview(lowerLeft, upperRight, pix);
      gv->preview->redrawPreview(lowerLeft, upperRight);
    }
  }

  //WWW'''''''''''''W'''''''''''''WWWW'''''''''''''''''''
  //W  W            W             W                      
  //WWW   WW  WW   WW  WW   WW    WWW   WW  WW WW W   WW 
  //W  W Wmmm W W W W Wmmm W      W    W   W W W W W Wmmm
  //W  W  WW  W W  WW  WW  W      W    W    WW W W W  WW 
  //'''''''''''''''''''''''''''''''''''''''''''''''''''''
  else {
    
  
    //WWW''''''''''''''W''''''WWW'''''''''''''''
    //W  W             W      W  W              
    //WWW   W  W W  WW WW     WWW   WW  WWW  WWW
    //W  W W W W W W W W W    W    W W  WW   WW 
    //W  W  W   WW  WW W W    W     WW WWW  WWW 
    //               W                          
    //'''''''''''''WW'''''''''''''''''''''''''''
    // Rough pass to precalculate most irradiance values
  
    if (canvas.prepass && !canvas.trace){

      if (ctx->id==0)
	print("Precalculating indirect illumination (p)...\n");

      // Todo:
      // If we buffer irradiance cache samples and then sync
      // when each thread finishes a pass, the results
      // will be repeatable.
      // See this post by Vladimir Koylazov for more info:
      // http://www.fxshare.com/list_servers/render_theory/archive/2344.html
      ProgressiveIterator pIter(gv->imgDim, lowerLeft, upperRight,
                                ctx->id, gv->numThreads,
                                completion, completion_stage);

      pIter.setStage('p');

      while (pIter.next(i,j)){

        // store pixel info for diagnostic purposes
        ctx->setPixel(i,j);
        
        // for removing randomness
        if (gv->nonRandomPixels)
          ctx->mydrand48.setSeed((i+1)*(j+1));
	  
        // create Ray
        r.dir = SbVec3f(

          // sideways component
          ( -canvas.width/2.0 +
            ( (double)i + 0.5) * canvas.width /
            gv->imgDim[X] ) *
          canvas.u +

          // vertical component
          ( -canvas.height/2.0 +
            ( (double)j + 0.5) * canvas.height /
            gv->imgDim[Y] ) *
          canvas.v +

          // distance ahead component
          -canvas.s*canvas.w  
          );

        // adjust for orthographic camera
        if (canvas.orthographic){
          r.pos = canvas.e + r.dir + canvas.s*canvas.w;
          r.dir = -canvas.w;
        }
        
        r.dir.normalize();
        r.ldDimension = 0;
	r.ldInstance = gv->imgDim[X]*j+i;

        // trace ray

        //debugging
        SbColor color = getRadiancePM(r, C_SPECULAR|C_MDIFFUSE|C_PREPASS);

        if (gv->preview && gv->ppmIter<0)
          gv->preview->setPreviewPix(i, j, ctx->id, color,
                                     pIter.getTileHalfSize(),
                                     pIter.getTileIteration(),
                                     lowerLeft, upperRight);
        

      }
      pIter.printStatus(100);
    
    }

    if (ctx->id==0)
      print("\n%d stored points in irradiance cache (%d rejected).\n",
	    irradCache.numPoints(), rejectedICPoints);


    //WWWW'm'''''''''WW'''''WWW'''''''''''''''''''
    //W               W      W                    
    //WWW  W WW   WW  W      W  WW W   WW  WW  WW 
    //W    W W W W W  W      W  W W W W W W W Wmmm
    //W    W W W  WW WWW    WWW W W W  WW  WW  WW 
    //                                      W     
    //''''''''''''''''''''''''''''''''''''WW''''''
    // Make final image

    if (ctx->id==0)
      print("Making final render pass (f)...\n");
    
    const int sqrtSamples = (int)ceil(sqrt((float)gv->numSamples)); 
    const int numSamples = sqrtSamples*sqrtSamples;
    const float invNumSamples = 1./numSamples;
    const float invSqrtSamples = 1.f/sqrtSamples;
    const SbVec2f pixelDelta(1.f/gv->imgDim[X], 1.f/gv->imgDim[Y]);
    const int tSamps = MAX2(1, gv->timeSamples);
    const float invtSamps = 1./tSamps;

    if (ctx->id==0)
      print("Using %d samples.\n", sqrtSamples*sqrtSamples);
    
    // initialize integer permutation for anti aliasing
    // from Strictly Deterministic Sampling, Sec. 4.1,
    // by Alexander Keller
    int n = (int)(log(3.f*sqrtSamples)/log(2.f)+0.5f); // n spans 3 pixels
    int Period = 1 << n;
    vector<unsigned short> Sigma(Period);
    unsigned short Digit, Inverse, Bits;
    {
      for (unsigned short i = 0; i<Period; i++){
        Digit = Period;
        Inverse = 0;

        for (Bits = i; Bits; Bits >>= 1){
          Digit >>= 1;
          if (Bits & 1)
            Inverse += Digit;
        }

        Sigma[i] = Inverse;
      }
    }

    // These next two are for DOF
    // Initialize scrambled digit permutation for radical inverse base 5
    // from "Strictly Deterministic Sampling" page 5
    unsigned short Sigma5[5] = {0,3,2,1,4};
    unsigned short Sigma7[7] = {0,2,5,3,1,4,6};
      
    // for motion blur
    CanvasParameters myCanvas = canvas;
    SoPerspectiveCamera *myCamera = new SoPerspectiveCamera();
    myCamera->ref();

    // stage completed
    if (gv->preview)
      gv->preview->resetCounters(ctx->id);

    // figure out order of completion
    // FIXME - more comments .. this is determining thread order for
    // progressive iteration to draw blocks in right order...
    // Also, check into creating/adding EmissiveTriangles.
    pthread_mutex_lock(&completionMutex);
    int newId=0;
    {
      for (int i=0; i<maxThreads; i++)
        if (completionOrder[i]>=0)
          newId++;
    }
    completionOrder[ctx->id] = newId;
    pthread_mutex_unlock(&completionMutex);
    
    // new iterator
    PixelIterator *myIter;
    if (gv->preview)
      myIter = new ProgressiveIterator(gv->imgDim, lowerLeft, upperRight,
                                       ctx->id, gv->numThreads,
                                       completion, completion_stage);
    else
      myIter = new PixelIterator(gv->imgDim, lowerLeft, upperRight,
                                 ctx->id, gv->numThreads,
                                 completion, completion_stage);

    auto_ptr<PixelIterator> pIter(myIter);
    pIter->setOrderId(completionOrder[ctx->id]);
    pIter->setStage('f');

    while (pIter->next(i,j)){

      // store pixel info for diagnostic purposes
      ctx->setPixel(i,j);

      // for removing randomness
      if (gv->nonRandomPixels)
        ctx->mydrand48.setSeed((i+1)*(j+1));

      // initialize color
      color = SbColor(0.0, 0.0, 0.0);
      SbColor avgColor(0,0,0);

      // reset depth
      if (gv->depthImage == GlobalVars::DEPTH_AVERAGE){
        pixDepthAvg = 0;
        pixDepthCnt = 0;
      } else {
        pixDepthAvg = -1;
        pixDepthCnt = 1;
      }

      for (int sampx = 0; sampx<sqrtSamples; sampx++){
        for (int sampy = 0; sampy<sqrtSamples; sampy++){

          // random
          //sample[X] = drand48();
          //sample[Y] = drand48();
      
          // quasi random
          //sobvect(2, samples);
	    
          // jittered
          //samples[X] = (sampx + ctx->mydrand48()) * invSqrtSamples;
          //samples[Y] = (sampy + ctx->mydrand48()) * invSqrtSamples;

          // Quasi random
          //int n = sampx*sqrtSamples+sampy; // FIXME
          //samples[X] = n/(float)(canvas.sqrt_samples*canvas.sqrt_samples);
          //samples[Y] = RI_vdC(n+sampleSet*canvas.sqrt_samples*canvas.sqrt_samples);
          //samples[X] = RI_S(quasiSamp);
          //samples[Y] = RI_vdC(quasiSamp);
          //quasiSamp++;

          //problem with quasiSamp = 260, -g747,169 on scene_jensen.iv

          // box filter
          //samples[X] -= 0.5;
          //samples[Y] -= 0.5;

          // tent filter from Realistic Ray Tracing 2nd Edition, p.60
          // transforms [0,1) -> (-1,1)
          //if (samples[X]<0.5)
          //  samples[X] = sqrt(2.0*samples[X]) - 1.0;
          //else
          //  samples[X] = 1.0 - sqrt(2.0 - 2.0*samples[X]);
          //if (samples[Y]<0.5)
          //  samples[Y] = sqrt(2.0*samples[Y]) - 1.0;
          //else
          //  samples[Y] = 1.0 - sqrt(2.0 - 2.0*samples[Y]);

          // offset calc  [0, 1)       [-0.5, 0.5)
          // transforms (-1,1)  ->   (-0.5,1.5)
          // (traces pixel centers)
	    
          // from Strictly Deterministic Sampling, Sec. 4.1
          Halton halt3;
          {
            int globalSampX = i*sqrtSamples+sampx; // translate local subpix id
            int globalSampY = j*sqrtSamples+sampy; // to global (w.r.t. total image)
            int m = globalSampX & (Period - 1);   // Period is 2^n
            int n = globalSampY & (Period - 1);
            r.ldInstance = m*Period + Sigma[n];   // index into 2^n * 2^n hammersly point set
	    
            halt3.Number(r.ldInstance, 3);

            if (gv->numSamples==1){
              offset[X] = 0.5;
              offset[Y] = 0.5;
            }
            // -s4 --subpix=2 may be a little noisy.
            // You can either turn on pixel center sampling below,
            // or use env PANE_SUBPIX_FILTER=2 to get the old behavior.
            //else if (sqrtSamples==gv->subPixDim && gv->subPixDim==2){
            //  offset[X] = (sampx+0.5f)/sqrtSamples;
            //  offset[Y] = (sampy+0.5f)/sqrtSamples;
            //}
            else {
              // This is sort of like a stratified Hammersley point set
              // It's a Hammersly point set, but each sample is entirely in its own
              // subpixel.
              offset[X] = (sampx + Sigma[n]/(double)Period)/(double)sqrtSamples;
              offset[Y] = (sampy + Sigma[m]/(double)Period)/(double)sqrtSamples;
            }

            // shift sample pattern randomly for PPM
            if (gv->ppmIter>=0) {
              offset[X] += ppmImgOffset[X];
              if (offset[X]>=1.f) offset[X] -= 1.f;
              offset[Y] += ppmImgOffset[Y];
              if (offset[Y]>=1.f) offset[Y] -= 1.f;
            }

          }
          r.ldDimension = 1; // Pixel sampling uses i/N and base dim=0 (b=2) of
                             // halton bases (so it's hammersley point set)

          //temp
          //offset[X] = samples[X];
          //offset[Y] = samples[Y];

          // create Ray

          // pixel space coordinates
          //      0   1   2      N-1    pixel index         i
          //    | . | . | . |
          //   0.0  _   _  1.0   1.0    image coordinates
          //   0.0  _   _  3.0    N     pixel coordinate    y
          //               c.w
          //   
          //  y = (i+0.5) * pixelW    pixelW = 3.0/N = 1.0
          //  i = y/pixelW - 0.5
          //
          //  y = i+0.5
          //  i = y-0.5
          //  idx = (int)(i+0.5) = (int)y
          SbVec2f pixCrd;
          pixCrd[0] = i+offset[X];
          pixCrd[1] = j+offset[Y];
	    
          // 3D image space coordinates
          // position in space that ray will go through,
          // in screen-aligned coordinates
          SbVec2f imgCrd;
          imgCrd[0] = pixCrd[0]*pixelDelta[0];
          imgCrd[1] = pixCrd[1]*pixelDelta[1];

          // motion blur
          SbColor tColor(0,0,0);
          color.setValue(0,0,0);
	  
          for (int ti=0; ti<tSamps; ti++){

            // calculate time sample
            if (gv->timeSamples>0){

              // from Strictly Deterministic Sampling, Sec. 4.2,
              // by Alexander Keller
              float time = halt3+ti/(double)tSamps;
              if (time>=1.0f) time -= 1.0f;
              float exposure = 0.5f;
              time = gv->timeMin + time*exposure*(gv->timeMax - gv->timeMin);
              r.ldDimension = 2; // time sampling uses base dim=1
              //print("Time: %f  ", time);

              // only adjust camera if main camera is compatible,
              // and result will set a camera, for consistency.
              // otherwise, if viewer camera is orthographic,
              // render will be different than viewer, or,
              // if no cameras are defined in camlist, myCamera will not be set.
              if (gv->cam->isOfType(SoPerspectiveCamera::getClassTypeId()) &&
                  gv->camList->numCams()>0){
		  
                // position camera
                gv->camList->positionCameraCubic( myCamera, (1.0-cos(time*M_PI))/2.0);
		
                // setup canvas
                canvasSetup(myCanvas, myCamera);
              }
            }
	    
            // world coordinates
            SbVec3f target;
            target[0] = myCanvas.width  * (imgCrd[0] - 0.5f); // sideways component
            target[1] = myCanvas.height * (imgCrd[1] - 0.5f); // vertical component
            target[2] = -myCanvas.s;                          // distance ahead component
            // convert from camera space to world space offset
            r.dir =
              target[0]*myCanvas.u +
              target[1]*myCanvas.v +
              target[2]*myCanvas.w;

            // adjust for orthographic camera
            if (myCanvas.orthographic){
              r.pos = myCanvas.e + r.dir + myCanvas.s*myCanvas.w;
              r.dir = -myCanvas.w;
            }

            static bool latLongCam = getenv("PANE_LATLONG_CAM");
            if (latLongCam){
              float phi   = (imgCrd[0])*2*M_PI; // .5=0 deg   =straight ahead
              float theta = (1-imgCrd[1])*M_PI;      // .5=pi/2 deg=straight ahead
              r.dir.setValue(
                sin(theta)*sin(phi),
                cos(theta),
                sin(theta)*cos(phi)
                );
            }
            static bool tinyPlanetCam = getenv("PANE_TINYPLANET_CAM");
            if (tinyPlanetCam){

              // Solve t^2*d.d + 2*t*(o-c).d + (o-c).(o-c)-R^2 = 0
              float rad = 0.125;
              SbVec3f c(0,rad,0);         //center
              SbVec3f o(0,rad*2,0);
              SbVec3f d = SbVec3f(imgCrd[0]-0.5, 0, -(imgCrd[1]-0.5)) - o;
              d.normalize();
              SbVec3f oc = c-o;
              float b = oc.dot(d);
              float det = sqrtf(b*b-oc.dot(oc)+rad*rad);
              float t = b + det;
              SbVec3f spherePt = o+t*d;
              
              r.dir = spherePt-c;

              // opposite

//               // Solve t^2*d.d + 2*t*(o-c).d + (o-c).(o-c)-R^2 = 0
//               float rad = 0.125;
//               SbVec3f c(0,rad,0);         //center
//               SbVec3f o(0,rad*2,0);
//               SbVec3f d = SbVec3f(imgCrd[0]-0.5, 0, -(imgCrd[1]-0.5)) - o;
//               d.normalize();
//               SbVec3f oc = c-o;
//               float b = oc.dot(d);
//               float det = sqrtf(b*b-oc.dot(oc)+rad*rad);
//               float t = b + det;
//               SbVec3f spherePt = o+t*d;
              
//               r.dir = spherePt-2*c;
//               r.dir = r.dir-o;

              
            }

            static bool invLatLongCam = getenv("PANE_INVLATLONG_CAM");
            if (invLatLongCam){
              // phi   .5=0 deg   =straight ahead
              // theta .5=pi/2 deg=straight ahead
              float phi   = (imgCrd[0])*2*M_PI;      // full image
              float theta = (1-imgCrd[1])*M_PI;   
              //float phi   = (imgCrd[0])/**2*/*M_PI;  // just the left half of the image
              //float theta = (1-imgCrd[1])*M_PI;      //
              //float phi   = (imgCrd[0]*(.25-.1)+.1)        *M_PI; // fractal pound 
              //float theta = (1-(imgCrd[1]*(.405-.2)+.2))   *M_PI; // corner detail
              r.dir.setValue(
                sin(theta)*sin(phi),
                cos(theta),
                sin(theta)*cos(phi)
                );

              r.pos = r.dir*-2;
            }

            static bool convexCam = getenv("PANE_CONVEX_CAM");
            if (convexCam){

              float angle = ((SoPerspectiveCamera *)gv->cam)->heightAngle.getValue();
              float rad = ((SoPerspectiveCamera *)gv->cam)->focalDistance.getValue();
//               float angle = M_PI/180. * 30;
//               float rad = 1;
              
              float phi   = M_PI + (imgCrd[0]-0.5)*angle;
              float theta = M_PI_2 + (1-imgCrd[1]-0.5)*angle;
              r.dir.setValue(
                sin(theta)*sin(phi),
                cos(theta),
                sin(theta)*cos(phi)
                );
              r.dir.normalize();

              ONB onb;
              onb.constructFromUW(myCanvas.u, myCanvas.w);
              r.dir = onb.toWorldSpace(r.dir);

              r.pos = myCanvas.e;
              r.pos += r.dir*rad;
              r.pos += myCanvas.w*rad;

            }

            static bool concaveCam = getenv("PANE_CONCAVE_CAM");
            if (concaveCam){

              float angle = ((SoPerspectiveCamera *)gv->cam)->heightAngle.getValue();
              float rad = ((SoPerspectiveCamera *)gv->cam)->focalDistance.getValue();
//               float angle = M_PI/180. * 30;
//               float rad = 1;
              
              float phi   = (imgCrd[0]-0.5)*angle;
              float theta = M_PI_2 + (imgCrd[1]-0.5)*angle;
              r.dir.setValue(
                sin(theta)*sin(phi),
                cos(theta),
                sin(theta)*cos(phi)
                );
              r.dir.normalize();
              r.dir = -r.dir;

              ONB onb;
              onb.constructFromUW(myCanvas.u, myCanvas.w);
              r.dir = onb.toWorldSpace(r.dir);

              r.pos = myCanvas.e;
              r.pos -= r.dir*rad;
              r.pos -= myCanvas.w*rad;

            }

            // this contracts or expands the camera rays wrt to center with a
            // (nonlinear) scale wrt to angular distance
            static bool warpCam = getenv("PANE_WARP_CAM");
            if (warpCam){
              r.dir.normalize();
              SbVec3f fwd = -myCanvas.w;
              float dot = r.dir.dot(fwd);
              if (dot<.9999){
                SbVec3f orth = r.dir-dot*fwd;
                orth.normalize();
                //dot = 1./(1.-dot*.95);
                //dot = 1.f-sqrtf(1.f-dot*dot);
                //dot = 1-powf(1-powf(dot, (1/.3)), .3);
                float d=.3;
                dot = 1-powf(1-powf(dot, (1/d)), d);
                r.dir = dot*fwd + sqrtf(1-dot*dot)*orth;
              }
            }

            // warps plane to look spherical
            static float sphericalCam = getenv("PANE_SPHERICAL_CAM") ?
              atof(getenv("PANE_SPHERICAL_CAM")) : 0;
            if (sphericalCam){

              // initialize r
              r.pos = myCanvas.e;
              // r.dir already initialized
              
              float rad = sphericalCam;
              float height = 0.9;

              // Solve t^2*d.d + 2*t*(o-c).d + (o-c).(o-c)-R^2 = 0
              SbVec3f c(0,height-rad,0);         //center
              SbVec3f o = r.pos;
              SbVec3f d = r.dir;
              d.normalize();
              SbVec3f oc = c-o;
              float b = oc.dot(d);
              float det = b*b-oc.dot(oc)+rad*rad;
              SbVec3f newp;
              bool adjustDir = false;
              if (det>0){
                det = sqrtf(det);
                float t = b - det;
                SbVec3f spherePt = o+t*d;

                // find new end pt
                SbVec3f cen2pt = spherePt-c;
                float costhetap = cen2pt[1]/rad;
                costhetap = max(min(costhetap, 1.0f), -1.0f);
                float thetap = acosf(costhetap);
                float len = rad*thetap;
                SbVec3f radial = cen2pt;
                radial[1] = 0;
                radial.normalize();
                SbVec3f top = SbVec3f(0, height, 0);
                newp = top + radial*len;
                adjustDir = true;
              }

              // find new start pt
              SbVec3f cen2pt = r.pos-c;
              rad = cen2pt.length();
              float costhetap = cen2pt[1]/rad;
              costhetap = max(min(costhetap, 1.0f), -1.0f);
              float thetap = acosf(costhetap);
              float len = rad*thetap;
              SbVec3f radial = cen2pt;
              radial[1] = 0;
              radial.normalize();
              SbVec3f top = c+SbVec3f(0,rad,0);
              SbVec3f newo = top + radial*len;

              if (adjustDir){
                r.dir = newp-newo;
                r.dir.normalize();
              } else {
                // ray missed
                // aim ray above horizon
                float cos_a_max = sqrtf(1 - sphericalCam*sphericalCam/(rad*rad));
                float a = acosf(min(cos_a_max, 1.f));
                r.dir.normalize();
//                cerr << "r.dir[1]=" << r.dir[1] << endl;
                float b = acosf(-r.dir[1]);
                float c = b-a;
                SbVec3f dir = r.dir;
                dir[1] = 0;
                dir.normalize();
                dir[1] += tan(c);
                dir.normalize();
                r.dir = dir;
//                cerr << "newdir=" << dir << "\n";
              }
              r.pos = newo;
            }

            static bool visCam = getenv("PANE_VISUALIZE_CAM");
            if (visCam){
              // visualization
              Ray(r.pos,r.dir,r.ctx).show(1, SbColor(1.0, 1.0, 1.0));
              static int centerShown = false;
              if (r.ctx->id == 0 && !centerShown){
                centerShown = true;
                showPoint(myCanvas.e, .1);
              }
            }
            
            
            // depth of field (DOF)
            SbColor aperColor(1,1,1);
            if (gv->fNumber != 0.0){
	    
              // get point on unit disc
              //float e1 = ctx->mydrand48();
              //float e2 = ctx->mydrand48();
              
              // from "Strictly Deterministic Sampling" page 23
              r.ldDimension = 4; // DOF sampling uses base dim=2 and 3
              int k = sampx*sqrtSamples+sampy;
              Halton halt2, halt5, halt7;
              halt2.Number(k,2);
              halt5.Number(r.ldInstance+ti,5,Sigma5);
              halt7.Number(r.ldInstance+ti,7,Sigma7);
              float e1 = fmod(halt5 + k/float(numSamples), 1);
              float e2 = fmod(halt7 + halt2, 1);
              float p = sqrt(e1);
              float t = (2.0*M_PI)*e2;

              // convert polar (p,t) to (x,y)
              // t is angle from right 
              float x = p*cos(t);
              float y = p*sin(t);

              // turn into lens offset
              SbVec3f offset = x*myCanvas.u + y*myCanvas.v;

              // aperture is degees off axis of lens edge, from focal plane
              offset *= myCanvas.aperture;
              
              // move eye
              r.pos = myCanvas.e + offset;

              // adjust direction
              r.dir -= offset;
              
              // chromatic aberration
              static const SbVec3f dist(.6, .8, 1);
              static const SbColor colr(1.f/(dist[0]*dist[0]),
                                        1.f/(dist[1]*dist[1]),
                                        1.f/(dist[2]*dist[2]));

              // disabled for now (adds noise at low samples)
//               aperColor[0] = p<dist[0] ? colr[0] : 0;
//               aperColor[1] = p<dist[1] ? colr[1] : 0;
//               aperColor[2] = p<dist[2] ? colr[2] : 0;

              // bokeh effect
              //
              // gives a N-sided aperture effect
              // could precompute this for speed
              //
              // NOTE: This screws up the normalizations in colr above!
              
//               const static int S = 6; // sides
//               const static float bt = M_PI*2.0f/S;
//               const static float m = -sin(bt)/(1-cos(bt));
//               float nt = fmodf(t,bt);
//               float nx = p*cosf(nt);
//               float ny = p*sinf(nt);
//               if (ny>(m*(nx-dist[0]))) aperColor[0] = 0;
//               if (ny>(m*(nx-dist[1]))) aperColor[1] = 0;
//               if (ny>(m*(nx-dist[2]))) aperColor[2] = 0;
            }

            r.dir.normalize();

            SbVec3f spectralMult(1,1,1);
            if (gv->spectral){
//              double eps = ctx->mydrand48();
              
              int k = sampx*sqrtSamples+sampy;
              int ldDim1 = r.ldDimension;
              r.ldNextDim();
              int bd1 = Halton::prime(ldDim1);
              vector<unsigned short> &sigma1 = Halton::sigmas[ldDim1];
              Halton halt1;
              halt1.Number(r.ldInstance, bd1, &sigma1[0]);
              float eps = fmod(halt1 + k/float(numSamples), 1);
              
              //double eps = fmod(halt11 + k/float(numSamples), 1);//halt11;

              Spectrum::getSpectralMult(eps, spectralMult, r.wavelen);
            }

            ctx->subPixSampId = sampx*sqrtSamples+sampy;
	  
            //////////////////////////
            // Get color along ray
            //////////////////////////

            // debugging a single ray
            if (canvas.trace){
              
              DEBUG_STATE=1;
//               if (sampx==11 && sampy==30){
//                 DEBUG_STATE=2;
//                 cerr << "trace this\n";
//               }
              tColor = (*this.*getRadiance)(r, canvas.flags);
              DEBUG_STATE=0;

//               pthread_mutex_lock(&mutexes[LOCK_STDERR]);
//               cerr << "r: " << r << "\n";
//               print("Sample color: ");
//               printCol(tColor);
//               print("\n");
//               pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
              
//               if (tColor[0]>120){
//                 cerr << " sampx=" << sampx << " sampy=" << sampy << endl;
//               }
              
              // add in color
              color += vecScale(tColor, aperColor);

              if (r.t>=0){
                pixDepthAvg += r.t;
                pixDepthCnt++;
              }
            }

            // normal calculation
            else {

              if (1){
                // DEBUG_STATE=1;
                tColor = (this->*getRadiance)(r, canvas.flags);
                // DEBUG_STATE=0;
              }
              else {
                // shirley aliasing example, RRT 2nd Ed., p. 51
                double x = (double)i + offset[X] + 1.0;
                double y = (double)j + offset[Y] + 1.0;
                double func1 = 0.5*(1.0 + sin((x*x+y*y)*(1/100.)));
                tColor = SbColor(func1, func1, func1);
              }

              if (0&&!finite(tColor[0])){
                cerr << "Retrying pixel " << i << "," << j << "\n";
                DEBUG_STATE=1;
                tColor = (this->*getRadiance)(r, canvas.flags);
                DEBUG_STATE=0;
              }

              vecScaleItself(tColor, aperColor);

              if (gv->spectral)
                vecScaleItself(tColor, spectralMult);
              
              // add in color
              color += tColor;

              // store depth
              if (r.t>=0){
                if (gv->depthImage == GlobalVars::DEPTH_AVERAGE){
                  pixDepthAvg += r.t;
                  pixDepthCnt++;
                } else { // DEPTH_NEAREST
                  if (r.t>=0 && (pixDepthAvg<0 || r.t<pixDepthAvg))
                    pixDepthAvg = r.t;
                }
              }
	      
              if (STOP_RENDERING)
                goto RENDER_TILE_STOP;
            }

          } // end time sampling

	  // add color to framebuffer, using filtering
          color *= invtSamps;  // do not divide by numSamples, the pixel
                               // weighting will handle it
          
          setPixWeighted(pixCrd, color);
          avgColor += color;

        } // end for (int sampy...
      } // end for (int sampx...

      avgColor *= invNumSamples;

      float depth = pixDepthCnt ? pixDepthAvg/pixDepthCnt : -1;
      
      // if debugging a single pixel, say color
      if (canvas.trace){
        pthread_mutex_lock(&mutexes[LOCK_STDERR]);
        print("Avg color: ");
        printCol(avgColor);
        print(" depth=%g\n", depth);
        pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
      }

      // set depth
      setDepth(i, j, depth);

      // draw block at (i,j) + (tileHalfSize,tileHalfSize) within render area
      if (gv->preview && gv->ppmIter<0)
        gv->preview->setPreviewPix(i, j, ctx->id, avgColor,
                                   pIter->getTileHalfSize(),
                                   pIter->getTileIteration()+32,
                                   lowerLeft, upperRight);
    } // end pixel loop

    RENDER_TILE_STOP:
    
    // destroy motionblur camera
    if (myCamera){
      myCamera->unref();
    }

    if (STOP_RENDERING){
      if (gv->numThreads != 1) pthread_exit(0);
      else return;
    }
    
  } // end if (sampmap)...

  // completion meter
  completion[ctx->id] = 100.0;
  pthread_mutex_lock(&mutexes[LOCK_STDERR]);
  print("\r");
  for (int i=0; i<gv->numThreads; i++)
    print("%6.2f%%%c ", completion[i], completion_stage[i]);
  pthread_mutex_unlock(&mutexes[LOCK_STDERR]);

  // quit threads
  if (gv->numThreads != 1){

    // time statistics
    struct tms renderTimes;
    times(&renderTimes);
    ctx->ticks += renderTimes.tms_utime + renderTimes.tms_cutime;
    
    pthread_exit(0);
  }
  
  return;
}

//''''''''''''''''''''''''WW''''WW'''''''''''WW''W''''''WWW'''''''''''
//                        WWW  WWW           WW  W       WW           
// WWWW  WWW WWWW W  WWW  W WWW WW  WWW WWW  WW  W  WWW  WW  W WW  WW 
//WWW   W WW WW WW W WW W W  W  WW W WW WW W WW  W W WW  WW  W WW WWWW
//  WWW W WW WW WW W WW W W     WW W WW WW W WW W  W WW  WW  W WW WW  
//WWWW   WWW WW WW W WWW  W     WW  WWW WWW   WW    WWW WWWW  WWW  WWW
//                   WW                 WW                            
//'''''''''''''''''''WW'''''''''''''''''WW''''''''''''''''''''''''''''
//
// returns value in sample map (gray scale, 0<=x<1.0)
// uses hermite smoothing to return smoothly varying result.
//
float Pane::sampMapValue(float x, float y){

  float g;
  float s, t;
  float i_f, j_f;
  float sp, tp;
  int i, j, i1, j1;

  int size[2] = { sampMapWidth, sampMapHeight };

  // calculate image coordinates (interpolate)
  s = x; 
  t = y;

  //  clamp
  if (s<0.0 || s>1.0 || t<0.0 || t>1.0)
    return(0.0);
  
  // calculate texture coordinates
  i_f = size[0]*s-0.5;
  j_f = size[1]*t-0.5;

  // calculate indices
  i  = (int)floor(i_f);
  j  = (int)floor(j_f);
  i1 = i+1;
  j1 = j+1;
	 
  i  = clamp( i, 0, size[0]-1);
  i1 = clamp(i1, 0, size[0]-1);
  j  = clamp( j, 0, size[1]-1);
  j1 = clamp(j1, 0, size[1]-1);
	 
  // calculate texel ratios
  sp = i_f-floor(i_f);
  tp = j_f-floor(j_f);
	 
  // perform hermite smoothing
  // (see Peter Shirley's Realistic Ray Tracing 1st ed.)
  sp = 3.0*sp*sp-2.0*sp*sp*sp;
  tp = 3.0*tp*tp-2.0*tp*tp*tp;

  g = (1.0-sp)*(1.0-tp)*sampMap[size[0]*j +i ] +
      sp*(1.0-tp)      *sampMap[size[0]*j +i1] +
      (1.0-sp)*tp      *sampMap[size[0]*j1+i ] +
      sp*tp            *sampMap[size[0]*j1+i1];
      
  return(g);
}


float Pane::mitchellFilter(float r){
  if (r<=1.f){
    float t1 = 1-r;
    return ( -15*t1*t1*t1 + 18*t1*t1 +9*t1 +2 ) * (1.f/18.f);
  }
  else if (r<=2.){
    float t2 = 2-r;
    return ( 5*t2*t2*t2 - 3*t2*t2 ) * (1.f/18.f);
  }
  else
    return 0;
}

float Pane::boxFilter(float r){
  const float d = .5f;  // in pixels
  return r<d ? 1.f : 0.f;
}

float Pane::tentFilter(float r){
  const float d = 1.0f;  // in pixels
  if (r<d){
    //return (3./M_PI) * (1. - r/d);  // for non-separated version I believe
    //return 1. - r/d;
    return 1. - r;  // d==1
  } else
    return 0;
}

void Pane::setPixWeighted(SbVec2f &pixCrd_, const SbColor &color){

  // use non-filtering function
//   setPix((int)pixCrd[X], (int)pixCrd[Y], color);
//   return;

  SbVec2f pixCrd = pixCrd_*gv->subPixDim;
  
//  const float radf = 2.0f * gv->subPixFilterSize; // mitchell
  const float radf = 1.0f * gv->subPixFilterSize;   // tent
//  const float radf = 0.0f; // tent  TEMPORARY, for debugging only
  int i0 = (int)(pixCrd[0] - radf);
  int j0 = (int)(pixCrd[1] - radf);
  int i1 = (int)(pixCrd[0] + radf);
  int j1 = (int)(pixCrd[1] + radf);

  if (i0<0) i0 = 0;
  if (j0<0) j0 = 0;
  if (i1>=gv->imgTotalDim[0]) i1 = gv->imgTotalDim[0]-1;
  if (j1>=gv->imgTotalDim[1]) j1 = gv->imgTotalDim[1]-1;

  const int iDelta = i1-i0;
  
  // buffer weights
  float *weights = (float *) alloca( (iDelta+1)*sizeof(float) );
  float *aweights = (float *) alloca( (iDelta+1)*sizeof(float) );
		 
  for (int j=j0; j<=j1; j++){
    
    for (int i=0; i<=iDelta; i++){
      SbVec2f p(i0+i+0.5, j+0.5);
//      float dist = (p-pixCrd).length();
//      float weight = mitchellFilter(dist);         // IF YOU CHANGE THIS,
//      float weight = tentFilter(dist);             // BE SURE TO ADJUST radf ABOVE
//      float weight = boxFilter(dist);              // !!!!

      SbVec2f dist = p-pixCrd;
      dist[0] = fabs(dist[0]*gv->invSubPixFilterSize);
      dist[1] = fabs(dist[1]*gv->invSubPixFilterSize);
      float weight = tentFilter(dist[0]) * tentFilter(dist[1]);  
      
      weights[i] = weight;
      aweights[i] = fabs(weight);
    }

    const int pixIdx = gv->imgTotalDim[X]*j+i0;
    
    // lock pixel buffer
    pthread_mutex_lock(pixMutexes[j]);
    
    for (int i=0; i<=iDelta; i++){
      int idx = pixIdx+i;
      pix[idx]       += color*weights[i];
      pixWeight[idx] += aweights[i];
    }

    pthread_mutex_unlock(pixMutexes[j]);
  }

  
}

void Pane::renderPPM(int trace) {
  currentlyRenderingPPMRange = true;
  for (int i=gv->ppmIter; i<=gv->ppmIterStop; i++) {
    cerr << "\nPPM frame " << i << "\n";
    TIME(gv->pane->shootPhotons());
    render(trace);
  }
  currentlyRenderingPPMRange = false;
  if (writeWhenDone) {
    writeWhenDone = false;
    writeImg();
  }
}
