#include <map>
#include <Inventor/nodes/SoNormal.h>
#include "Pane.h"
// #include "debug.h"
#include "Fractal.h"
#include "Isosurface.h"
#include "Julia.h"
#include "Plane.h"
#include "Photon_map.h"
#include "BIH.h"

//#include <google/heap-profiler.h>
#include "inventorHack.h"


using namespace std;


// Rebuild octree (for animation and demo purposes)
void Pane::rescanScene(){
  rayAccel->reset();
  litPnts.resize(1);    // remove all but headlight
  litTris.clear();

  // clear matList
  while (matList.size()>1){  // leave default material
    delete matList.back();
    matList.pop_back();
  }
  
  readScene();
}   


//                     _ ____
//  _ __ ___  __ _  __| / ___|  ___ ___ _ __   ___
// | '__/ _ \/ _` |/ _` \___ \ / __/ _ \ '_ \ / _ \
// | | |  __/ (_| | (_| |___) | (_|  __/ | | |  __/
// |_|  \___|\__,_|\__,_|____/ \___\___|_| |_|\___|
//
// (banner made with figlet)
//
// Read the scene into internal data structure
//

// global variables (within this file)
namespace {
  map <SoIndexedShape*, RayAccel*> instShapeAccel;   // rayAccel for instanced shape
  map <SoIndexedShape*, int> instShapeCount;         // reference count
  int numInstancedShapes = 0;
  int numInstancedTris = 0;
  int numTotalShapes = 0;
}

// Count shapes with this callback to see what should be instanced
SoCallbackAction::Response
CountShapesPreCB(void *data, SoCallbackAction *action,
                 const SoNode *node){
  
  Pane *pane = (Pane *)data;
  SoIndexedShape *shape = (SoIndexedShape *)node;

  instShapeCount[shape]++;
  numTotalShapes++;

  if (instShapeCount[shape] == 2){ // consider it instanced second time it's seen
    instShapeAccel[shape] = NULL;
    numInstancedShapes++;
  
    // count triangles in shape
    bool isStripSet =
      shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId());
    int numCoordIdxs = shape->coordIndex.getNum();
    if (numCoordIdxs>=3) {
      int v[3];
      int base = 0;
      bool isStripSet =
        shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId());
      for (int j=base+1; j<numCoordIdxs-1; j++) {
        v[0] = shape->coordIndex[base];
        if (v[0]<0) { j = ++base; continue; }
        v[1] = shape->coordIndex[j];
        if (v[1]<0) { j = base = j+1; continue; }
        v[2] = shape->coordIndex[j+1];
        if (v[2]<0) { j = base = j+2; continue; }
        if (isStripSet) base++;
        // detect same vertices
        if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2]) continue;
        numInstancedTris++;
      }
    }
  }

  return SoCallbackAction::PRUNE;
}

int Pane::readScene(){
  
  static bool doInstancing = !getenv("PANE_DISABLE_INSTANCING");
  if (doInstancing){
    SoCallbackAction countCB;
    countCB.addPreCallback(SoIndexedShape::getClassTypeId(),
                           CountShapesPreCB, this);
    countCB.apply(gv->scene);
  
    cerr << "Instanced shapes: " << numInstancedShapes << " (" << numTotalShapes << " total)\n";
    cerr << "Instanced tris (original): " << numInstancedTris << "\n";
    instanceTris.reserve(numInstancedTris);
  }

  TriMesh::initMeshs();
  
  SoCallbackAction cb;

  // handle Separators
  cb.addPreCallback(SoSeparator::getClassTypeId(),
		    Pane::sepPreCB, this);
  cb.addPostCallback(SoSeparator::getClassTypeId(),
  		     Pane::sepPostCB, this);

  // track Materials/MatPluses
  cb.addPreCallback(SoMaterial::getClassTypeId(),
		    Pane::matPreCB, this);
  cb.addPreCallback(SoMatPlus::getClassTypeId(),
		    Pane::matPlusPreCB, this);

  // track Coordinate3, Normal, VertexProperty
  cb.addPreCallback(SoCoordinate3::getClassTypeId(),
		    Pane::coord3PreCB, this);
  cb.addPreCallback(SoNormal::getClassTypeId(),
		    Pane::normalPreCB, this);
  cb.addPreCallback(SoVertexProperty::getClassTypeId(),
		    Pane::vPropPreCB, this);
  
  // add all Triangles, EmissiveTriangles
  cb.addTriangleCallback(SoShape::getClassTypeId(),
			 Pane::triCBHandler, this);

  // add PointLights
  cb.addPreCallback(SoPointLight::getClassTypeId(),
		    Pane::pntLtCBHandler, this);

  // handle IndexedFaceSets/IndexedTriangleStripSets
  cb.addPreCallback(SoIndexedShape::getClassTypeId(),
		    Pane::indexedShapePreCB, this);
  cb.addPostCallback(SoIndexedShape::getClassTypeId(),
  		     Pane::indexedShapePostCB, this);
  
  // handle Spheres
  cb.addPreCallback(SoSphere::getClassTypeId(),
		    Pane::spherePreCB, this);
  cb.addPostCallback(SoSphere::getClassTypeId(),
  		     Pane::spherePostCB, this);

  // handle Fractals
  cb.addPreCallback(SoLinearFractal::getClassTypeId(),
		    linearFractalPreCB, this);
  
  // Go!
  TIME(cb.apply(gv->scene););

  cerr << "Materials:" << matList.size() << endl;
  cerr << "Separators in stack:" << sepStack.size() << endl;

  // turn off shininess
  if (gv->allowReflection == 0){
    gv->sceneHasShininess = 1;

    for (list <Material *>::iterator p = matList.begin(); p != matList.end(); p++){
      Material *mat = *p;
      mat->setShininess(0);
    }

    Plane::getDefaultMat()->setShininess(0);
    Sphere::getDefaultMat()->setShininess(0);
    Julia::getDefaultMat()->setShininess(0);
  }
  
  // optionally free scene memory
  static bool freeScene = getenv("PANE_FREE_SCENE");
  if (freeScene || gv->renderQuit && ! gv->renderIv){

    // try turning off notification, maybe this make deleting the scene faster?
    cerr << "Disabling notification...\n";
    SoCallbackAction myCB;
    myCB.addPreCallback(SoNode::getClassTypeId(), Pane::turnOffNotifyCB, this);
    myCB.apply(gv->scene);

    // HACK
    //
    // disable auditor checking using modified SGI Open Inventor
    // library calls. Otherwise SoSeparator::removeAllChildren()
    // can be painfully slow for nodes with lots of children.
    // See inventorHack.cxx for details.
    cerr << "Freeing scene graph...";
    //HeapProfilerDump("heapProf1");
    setInventorAuditorChecking(false);
    TIME(gv->scene->removeAllChildren(););
    setInventorAuditorChecking(true);
    //HeapProfilerDump("heapProf2");
    cerr << "done.\n";
  }

  // adjust transparency, so we can see 100% transparent objects in viewer
  SoCallbackAction trancb;
  trancb.addPostCallback(SoMaterial::getClassTypeId(),
			 Pane::transMatCBHandler, this);
  trancb.apply(gv->scene);

  // trim arrays
  vector<Triangle>(regTris).swap(regTris);
  vector<DMTri>(dmTris).swap(dmTris);
  vector<EmissiveTriangle>(litTris).swap(litTris);
  vector<Julia>(julias).swap(julias);
  vector<Sphere>(spheres).swap(spheres);
  vector<Instance>(instances).swap(instances);

  //HeapProfilerDump("heapProf3");
  
  // Now that scene is loaded, add to acceleration structure.
  cerr << "Adding primitives to " << rayAccelName.c_str() << "...";
  rayAccel->reserve(
    regTris.size() +
    litTris.size() +
    dmTris.size() +
    julias.size() +
    spheres.size() +
    instances.size()
    );

  // compute normals if requested
  if (gv->computeNormals)
    TriMesh::computeAllNormals();
  
  for (int i=0; i<regTris.size(); i++)
    rayAccel->add(&regTris[i]);
  for (int i=0; i<litTris.size(); i++)
    rayAccel->add(&litTris[i]);
  for (int i=0; i<dmTris.size(); i++)
    rayAccel->add(&dmTris[i]);
  for (int i=0; i<julias.size(); i++)
    rayAccel->add(&julias[i]);
  for (int i=0; i<spheres.size(); i++)
    rayAccel->add(&spheres[i]);
  for (int i=0; i<instances.size(); i++)
    rayAccel->add(&instances[i]);


  // setup Isosurface material
  Isosurface::getDefaultMat()->setShininess(0);
  
  // setup Plane material
  Plane::getDefaultMat()->setShininess(0);
//  Plane::getDefaultMat()->setDiffuse(SbColor(.5,.5,.5));  
Plane::getDefaultMat()->setDiffuse(SbColor(.18,.18,.18));
//  Plane::getDefaultMat()->setDiffuse(SbColor(0,0,0));  
//   Plane::getDefaultMat()->setSpecular(SbColor(.05, .05, .05));
//   Plane::getDefaultMat()->setDiffuse(SbColor(.5, .5, .5));
//   Plane::getDefaultMat()->rough1 = 40000;
//   Plane::getDefaultMat()->rough2 = 40000;
//   Plane::getDefaultMat()->model = Material::ASHIKHMIN;
  Plane::getDefaultMat()->resolveParams();
    
  // setup DMPlane material
  DMPlane::getDefaultMat()->setShininess(0);
//  DMPlane::getDefaultMat()->setDiffuse(SbColor(.18,.18,.18));
    DMPlane::getDefaultMat()->setDiffuse(SbColor(.02,.18,.02));
//   DMPlane::getDefaultMat()->setDiffuse(SbColor(0.041452, 0.121986, 0.420508));

  // setup Sphere material
  Sphere::getDefaultMat()->setSpecular(SbColor(.05, .05, .05));
  Sphere::getDefaultMat()->setDiffuse(SbColor(.18, .18, .18));
//    Sphere::getDefaultMat()->setEmission(SbColor(.18, 0, 0));
  Sphere::getDefaultMat()->setShininess(0);
//   Sphere::getDefaultMat()->rough1 = 40000;
//   Sphere::getDefaultMat()->rough2 = 40000;
//  Sphere::getDefaultMat()->model = Material::ASHIKHMIN;
  Sphere::getDefaultMat()->resolveParams();
  
  if (gv->addPlane){

    Plane *plane = new Plane();
    
    ((PrimList*)scenePrim)->add(plane);

    float planeHeight = getenv("PANE_PLANE_HEIGHT") ?
      atof(getenv("PANE_PLANE_HEIGHT")) : 0;
    plane->setDist(planeHeight);

    // add texture
    static bool addPlaneTexture = getenv("PANE_PLANE_TEXTURE");
    if (addPlaneTexture){


      Plane::getDefaultMat()->setShininess(0);
      Plane::getDefaultMat()->setSpecular(SbColor(.05, .05, .05));
      Plane::getDefaultMat()->setDiffuse(SbColor(.5, .5, .5));
      Plane::getDefaultMat()->rough1 = 40000;
      Plane::getDefaultMat()->rough2 = 40000;
      Plane::getDefaultMat()->model = Material::ASHIKHMIN;
      Plane::getDefaultMat()->resolveParams();
      
      int numComps;
      SbVec2s textureSize;
      SoTexture2 *text2 = new SoTexture2();
      text2->ref();
      text2->filename.setValue("ashShirPattern2.rgb");
      const unsigned char *img = text2->image.getValue(textureSize, numComps);

      Texture *texture = new Texture(SoTexture2::REPEAT,
                                     SoTexture2::REPEAT,
                                     SoTexture2::MODULATE,
                                     SbColor(0,0,0),
                                     numComps, textureSize, img,
                                     0,
                                     0.0f);

      // HACK ALERT: stuff the matrix into stex/ttex
      const float scale = 0.15;
      float *stex = texture->stex;
      float *ttex = texture->ttex;
      stex[0] = scale;    // 0 0
      stex[1] = 0;        // 1 0
      stex[2] = 0;        // 3 0
      ttex[0] = 0;        // 0 1
      ttex[1] = scale;    // 1 1
      ttex[2] = 0;        // 3 1
      
      plane->setTexture(texture);
      
      cerr << "\nPlane: Added texture.\n";
    }

  }
  
  if (gv->addDMPlane){
    ((PrimList*)scenePrim)->add(new DMPlane(gv));
  }

  Plane *addedWater = NULL;
  static bool addWater = getenv("PANE_ADD_WATER");
  if (addWater){
    // add water
    Plane *water = new Plane();
    ((PrimList*)scenePrim)->add(water);
    addedWater = water;

    float planeHeight = getenv("PANE_WATER_HEIGHT") ?
      atof(getenv("PANE_WATER_HEIGHT")) : 0;
    water->setDist(planeHeight);
    
  }

  static bool addDMWater = getenv("PANE_ADD_DMWATER");
  if (addDMWater){
    // add water
    DMPlane *water = new DMPlane(gv);
    ((PrimList*)scenePrim)->add(water);
    addedWater = water;

    Noise noise;
    noise.octaves = 10;
    noise.amplitude = .025;
    float scale = 4;
    noise.frequency = SbVec3f(scale,scale,scale);
//     noise.bias = .4257;
//     noise.gain = .8769;
    noise.offset = .9;
    //noise.freqJump = 2.029;
    //noise.ampJump = 1;
    noise.sign=true;
    noise.funcPtr = &Noise::monofractal;

    noise.initExponents();
    
    water->getNoise() = noise;
    water->setDoColorMap(false);
  }

  if (addedWater){
    Plane *water = addedWater;
    water->makeNewMat();
    water->getMat()->setDiffuse(SbColor(0,0,.18));
    water->getMat()->setSpecular(SbColor(1,1,1));
    water->getMat()->setShininess(1);
    water->getMat()->transparency = 1;
    // To scale atten and maintain color, say same color at 10x distance:
    //   newAtten = [r_old, g_old, b_old]^(1/10)
    //water->getMat()->atten.setValue(0.94698, 0.96217, 0.97546);
    water->getMat()->atten.setValue(.58, .68, .78);
    water->getMat()->ior2 = 1.33;
    water->getMat()->resolveParams();

    // if camera is underwater, do the next two lines
    canvas.atten = 1;
    canvas.attenColor = water->getMat()->atten;
  }
  
  if (getenv("PANE_ISO"))
    ((PrimList*)scenePrim)->add(new Isosurface());
  
  if (getenv("PANE_FRACT")){
    Fractal *fract = new Fractal( ((PrimList*)scenePrim)->prim(0));
    
    SbMatrix matrix;
    matrix.makeIdentity();
    matrix.setRotate(SbRotation(SbVec3f(1,0,0), -90*M_PI/180)); // ompf

    // swap rayAccel with instance
    (*(PrimList*)scenePrim)[0] =
      new Instance(fract, matrix);
  
  }
  
  cerr << "done." << endl;

  cerr << "\nPrimitives: " << rayAccel->size() << endl;
  
  // refine tree
  cerr << "Refining " << rayAccelName.c_str() << "..." << endl;
  rayAccel->setParams(gv->octreeMaxDepth, gv->octreeMaxItems,
                      gv->octreeDebug);
  rayAccel->refine(gv->simpleOctree);
  rayAccel->printStats();

  // Set default meters per unit for scene size of 10m
  if (gv->metersPerUnit<0.0)
    gv->setUnitsPerMeter(getMaxDist()/10.0);
  
  // set up irradiance cache
  irradCache.setDims(SbVec3f(rayAccel->getMin()),
		     rayAccel->getMaxSideLength(), 0);
  irradCache.setA(gv->irradCacheError);
  irradCache.setResolution(gv->ambientResolution);
  irradCache.setGV(gv);

  // set photon map parameter
  if (!gv->sceneHasShininess && gv->useGlassModel)
    gv->numCPhotons = 0;

  Triangle::printGlobalStats();
  
  return rayAccel->size();
}

// PointLight CB Handler
// Stores PointLight information when traversing scene graph
SoCallbackAction::Response
Pane::pntLtCBHandler(void *data, SoCallbackAction *action,
		     const SoNode *node){
  Pane *p = (Pane *)data;
  if (p->gv->ambientIllum) return SoCallbackAction::CONTINUE;

  // copy light so it doesnt go away when scene is removed
  SoPointLight *oldLt = (SoPointLight *)node;
  SoPointLight *newLt = new SoPointLight();
  newLt->location  = oldLt->location;
  newLt->on        = oldLt->on;
  newLt->intensity = oldLt->intensity;
  newLt->color     = oldLt->color;
  p->garbageDump->addChild(newLt); // add to delete-on-exit list
  
  p->litPnts.push_back(newLt);
  p->gv->sceneHasLights = 1;
  p->numLights++;
  return SoCallbackAction::CONTINUE;
}

// SoMaterial CB Handler
// reduces transparency for viewing purposes
// Note: If the node is referenced more than once
// the reduction will accumulate.
SoCallbackAction::Response
Pane::transMatCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  
  SoMaterial *mat = (SoMaterial *)node;
  Pane *pane = (Pane *)data;

  // reduce material transparency by 25%
  int num = mat->transparency.getNum();
  float *trans = mat->transparency.startEditing();
  for (int i=0; i<num; i++)
    trans[i] *= pane->gv->transpScale;
  mat->transparency.finishEditing();
  
  return SoCallbackAction::CONTINUE;
}

// Material CB Handler
// turns surfaces black, for photon viewing
SoCallbackAction::Response
Pane::matBlackCBHandler(void *data, SoCallbackAction *action,
                        const SoNode *node){
  SoMaterial *mat = (SoMaterial *)node;

  for (int i=0; i<mat->ambientColor.getNum(); i++)
    mat->ambientColor.set1Value(i, 0.0, 0.0, 0.0);
  for (int i=0; i<mat->diffuseColor.getNum(); i++)
    mat->diffuseColor.set1Value(i, 0.0, 0.0, 0.0);
  for (int i=0; i<mat->specularColor.getNum(); i++)
    mat->specularColor.set1Value(i, 0.0, 0.0, 0.0);
  for (int i=0; i<mat->emissiveColor.getNum(); i++)
    mat->emissiveColor.set1Value(i, 0.0, 0.0, 0.0);
  
  return SoCallbackAction::CONTINUE;
}

SoCallbackAction::Response
Pane::turnOffNotifyCB(void *data, SoCallbackAction *action,
                      const SoNode *node){
//  cerr << "disabling notify\n";
  SoNode *myNode = (SoNode *) node; // cast away const
  
//  cerr << "\nNotify  is: " << node->isNotifyEnabled() << "\n";
  bool on = myNode->enableNotify(false);
//   cerr << "Notify was: " << on << "\n";
//   cerr << "Notify now: " << node->isNotifyEnabled() << "\n";
  return SoCallbackAction::CONTINUE;
}

// Triangle CB Handler
// Puts triangle into octtree
void Pane::triCBHandler(void *data, SoCallbackAction *action,
			const SoPrimitiveVertex *v0,
			const SoPrimitiveVertex *v1,
			const SoPrimitiveVertex *v2){

  Pane *pane = (Pane *)data;

  if (!pane->processTriangles) return;

  bool okay = false;
  bool emis = pane->getCurrentMat()->emissive;

  if (emis) {
    EmissiveTriangle t(pane, action, v0, v1, v2, okay);
    if (okay) {
      pane->litTris.push_back(t);

      // record keeping
      pane->numLights++;
      pane->totalEmisAvg += t.getEmisAvg();
      if (t.getEmisAvg() > pane->maxEmisAvg)
        pane->maxEmisAvg = t.getEmisAvg();
    }
  }
  else {
    Triangle t(pane, action, v0, v1, v2, okay);
    if (okay)
      pane->regTris.push_back(t);
//     DMTri t(pane, action, v0, v1, v2, okay);
//     if (okay)
//       pane->dmTris.push_back(t);
  }
}




// Separator CB Handler
// Stores pointer to most recent SoSeparator node in
// scene graph traversal.
SoCallbackAction::Response
Pane::sepPreCB(void *data, SoCallbackAction *action,
	       const SoNode *node){
  Pane *pane = (Pane *)data;
  SoSeparator *sep = (SoSeparator *)node;
  
  // place group copy on stack
  pane->sepStack.push_front(new sepListItem_t(*pane->sepStack.front()));

  return SoCallbackAction::CONTINUE;
}

SoCallbackAction::Response
Pane::sepPostCB(void *data, SoCallbackAction *action,
		const SoNode *node){
  Pane *pane = (Pane *)data;
  SoSeparator *sep = (SoSeparator *)node;
  
  // remove group from stack
  delete pane->sepStack.front();
  pane->sepStack.pop_front();

  return SoCallbackAction::CONTINUE;
}

// Sphere CB Handler
// Stores sphere to sphere list, disables triangle processing temporarily.
SoCallbackAction::Response
Pane::spherePreCB(void *data, SoCallbackAction *action,
                  const SoNode *node){
  Pane *pane = (Pane *)data;
  SoSphere *sphere = (SoSphere *)node;
  
  SbVec3f s,t;
  SbRotation r,so;
  SbMatrix matrix = action->getModelMatrix();
  matrix.getTransform(t,r,s,so);
  Material *mat = pane->getCurrentMat();

  // find texture
  int numComps;
  SbVec2s textureSize;
  const unsigned char *img = action->getTextureImage(textureSize, numComps);
  SbMatrix tmatrix = action->getTextureMatrix();
  Texture *texture = NULL;
  
  // create texture (this is still very inefficient space-wise)
  if (img){

    // only works for rgb textures
    if (numComps != 3 && numComps != 1){
      cerr << "Illegal number of components: " << numComps << endl;
    } else {
      
      texture = new Texture(action->getTextureWrapS(),
                            action->getTextureWrapT(),
                            action->getTextureModel(),
                            action->getTextureBlendColor(),
                            numComps, textureSize, img,
                            pane->gv->bumpMapping,
                            pane->gv->bumpScale);

      // HACK ALERT: stuff the matrix into stex/ttex
      float *stex = texture->stex;
      float *ttex = texture->ttex;
      stex[0] = tmatrix[0][0];
      stex[1] = tmatrix[1][0];
      stex[2] = tmatrix[3][0];
      ttex[0] = tmatrix[0][1];
      ttex[1] = tmatrix[1][1];
      ttex[2] = tmatrix[3][1];
    }
  }


  // decide what kind of sphere to make (instanced or not)
  SbVec3f axis;
  float angle;
  r.getValue(axis, angle);
  
  if (s[0] != s[1] || s[0] != s[2] || s[1] != s[2] || angle != 0) {

    // TODO: fold radius into matrix so Sphere class doesnt have to store radius
    // ...
    
    // Storing position(0,0,0) is kind of wasteful when doing many spheres,
    // but oh well.
    Primitive *p = new Sphere(SbVec3f(0,0,0), sphere->radius.getValue(),
                              mat->diffuse, mat, texture);
    
    pane->instances.push_back(Instance(p, matrix));
  }
  else {
    float radius = s[0]*sphere->radius.getValue();
    pane->spheres.push_back(Sphere(t, radius, mat->diffuse, mat, texture));
  }

  // Get material
  if (mat->isShiny())
    pane->gv->sceneHasShininess = 1;

  if (mat->isEmissive()){
    pane->gv->sceneHasLights = 1;
    pane->litSpheres.push_back(pane->spheres.size()-1);
  }

  // disable triangle processing (or we'll add the triangulated sphere)
  pane->processTriangles = false;

  return SoCallbackAction::CONTINUE;
}

SoCallbackAction::Response
Pane::spherePostCB(void *data, SoCallbackAction *action,
		const SoNode *node){
  Pane *pane = (Pane *)data;
  SoSphere *sphere = (SoSphere *)node;

  // re-enable triangle processing
  pane->processTriangles = true;
  
  return SoCallbackAction::CONTINUE;
}

// SoIndexedFaceSet and SoIndexedTriangleStripSet CB Handler.
//
// Returns SoCallbackAction::PRUNE to skip the node (if appropriate),
// skipping generatePrimitive() and generateNormals().
//
// Returns SoCallbackAction::CONTINUE if inventor should traverse the node.
//
//
// This code is tricky in several ways.
// Since the node hasn't been traversed yet (and may not be ever
// if PRUNE is returned), that means pane->getCurrentCoord()/getCurrentNorm()
// etc. is inaccurate if the node contains its own VertexProperty.
// Therefore, this condition is checked. In general, if a SoCoordinate3
// or VertexProperty has values, then it overrides coord*, norms*,
// diffuse* etc. However, the code also checks getNum()==numVerts
// to be valid.
//
// For example, the coordinates could come from:
//
//   A previous SoCoordinate3 node       // if getNum() is non-zero
//   A previous SoVertexProperty node    // if getNum() is non-zero
//   An internal SoVertxProperty node    // if getNum() is non-zero
//   action->getCoordinate3(i)           // Invalid since this doesn't
//                                       // include the internal vertexPropety
//
//   Normals are similar, as are diffuse which may be defined
//   by SoVertexProperty::orderedRGBA.
//
// Another tricky thing is the vertex ordering for faceSets vs. triangle
// strips. Face sets should deal with vertex lists as a fan, while strip sets
// as a strip (obviously), and flip every other triangle.
//
// All of this is converted into a simple indexed triangle set (TriMesh).
//
SoCallbackAction::Response
Pane::indexedShapePreCB(void *data, SoCallbackAction *action,
                        const SoNode *node){
  
  Pane *pane = (Pane *)data;
  SoIndexedShape *shape = (SoIndexedShape *)node;

  // Is this an already processed instance?
  bool createOriginalInstance=false;
  bool copyExistingInstance=false;
  RayAccel *rayAccel=NULL;
  map<SoIndexedShape*, RayAccel*>::iterator rai = instShapeAccel.find(shape);
  if (rai != instShapeAccel.end()){
    if (rai->second) { // already created
      rayAccel = rai->second;
      copyExistingInstance = true; 
    } else {           // this is the first instance
      createOriginalInstance = true;
      rayAccel = new BIH();
      instShapeAccel[shape] = rayAccel;
    }
  }
  
  static struct timeval theTime;
  int oldSec = theTime.tv_sec;
  gettimeofday(&theTime, NULL);
  if (theTime.tv_sec != oldSec)
    cerr << "indexed shape (face set)\n";

  // only handle IndexedFaceSet for now
  bool isStripSet =
    shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId());
  bool isFaceSet =
    shape->isOfType(SoIndexedFaceSet::getClassTypeId());

  SoIndexedFaceSet *faceSet = isFaceSet ?
    (SoIndexedFaceSet*)node : NULL;
  SoIndexedTriangleStripSet *stripSet = isStripSet ?
    (SoIndexedTriangleStripSet*)node : NULL;
  
  if (!isStripSet && !isFaceSet)
    return SoCallbackAction::CONTINUE;

  // dont render/include invisible objects
  if (action->getDrawStyle() != SoDrawStyle::FILLED)
    return SoCallbackAction::PRUNE;
  
  // find vertex and normal lists
  SoMFVec3f *verts = pane->getCurrentVerts();
  SoMFVec3f *norms = pane->getCurrentNorms();

  // get material properties
  SbMatrix matrix = action->getModelMatrix();
  Material *mat = pane->getCurrentMat();
  SoMaterial *soMat = pane->getCurrentSoMat();
  SoVertexProperty *vProp = pane->getCurrentVProp();

  if (createOriginalInstance){
    // create instance
    cerr << "Creating new instance\n";
    pane->instances.push_back(Instance(rayAccel, matrix, mat));
  } else if (copyExistingInstance){
    // copy instance
    static int cnt=0;
    static int reportCnt=1;
    if (++cnt==reportCnt){
      reportCnt*=10;
      cerr << "Copying instance (" << cnt << ")\n";
    }
    pane->instances.push_back(Instance(rayAccel, matrix, mat));
    return SoCallbackAction::PRUNE;
  }
  
  // check for local vertexProperty
  SoVertexProperty *localVProp = faceSet ?
    (SoVertexProperty*)(faceSet->vertexProperty.getValue()) :
    (SoVertexProperty*)(stripSet->vertexProperty.getValue());

  if (localVProp){
    if (localVProp->vertex.getNum())
      verts = &localVProp->vertex;
    if (localVProp->normal.getNum())
      norms = &localVProp->normal;

    if (localVProp->orderedRGBA.getNum()){
      soMat = NULL;
      vProp = localVProp;
    }
  }
  SoMFVec2f *localTexCoord = (localVProp && localVProp->texCoord.getNum()) ?
    &localVProp->texCoord : NULL;

  if (!verts->getNum())
     return SoCallbackAction::PRUNE;
 
  //------------------------------------------------------------------
  // Loop over triangles
  //------------------------------------------------------------------

  // variables
  int triCount = 0;
  int numTris = 0;
  int smallAreas = 0;
  int colinearVerts = 0;
  int numVerts = verts->getNum();
  
  // create tri mesh
  TriMesh *mesh = new TriMesh();
  TriMesh::meshs.push_back(mesh);
  mesh->mat = mat;

  // add points
  mesh->verts.reserve(numVerts);
  for (int i=0; i<numVerts; i++) {

    if (createOriginalInstance){
      // store vertex
      mesh->verts.push_back((*verts)[i]);
    } else {
      // transform to world space and store vertex
      SbVec3f vert;
      matrix.multVecMatrix((*verts)[i], vert);
      mesh->verts.push_back(vert);
    }
  }

  // Compute mesh bounding box.
  // Needed for default texture coordinates
  int numComps;
  SbVec2s textureSize;
  const unsigned char *img = action->getTextureImage(textureSize, numComps);

  bool hasTextCoords =
    action->getTextureCoordinateBinding() ==
    SoTextureCoordinateBinding::PER_VERTEX_INDEXED &&
    ((localTexCoord && localTexCoord->getNum() == numVerts) ||
     (!localTexCoord && action->getNumTextureCoordinates() == numVerts));
  
  static const bool genDefTextCoordsDef = (getenv("PANE_GEN_DEF_TEXT_COORDS") != NULL);
  bool genDefTextCoords = (img || genDefTextCoordsDef) && !hasTextCoords;
  //    || pane->gv->computeNormals
  
  BBox minMax;
  int tDim1=0, tDim2=1;     // longest and 2nd longest dimensions of bbox
  SbVec3f invBboxSize;
  
  if (genDefTextCoords){

    // Default texture coordinates are computed the same way inventor does it.
    // Compute the bounding box, map largest dimension to s, 2nd largest to t.
    // See Inventor Mentor Chapter 7 for more information.

    // compute inverse size
    minMax.grow(&mesh->verts[0], numVerts);
    SbVec3f size = minMax.size();
    invBboxSize = SbVec3f(
      (size[0] ? 1/size[0] : 0),
      (size[1] ? 1/size[1] : 0),
      (size[2] ? 1/size[2] : 0)
      );

    // compute longest dims
    if (size[0]>=size[1] && size[0]>=size[2]){
      tDim1 = 0;
      tDim2 = size[1]>=size[2] ? 1 : 2;
    }
    else if (size[1]>=size[0] && size[1]>=size[2]){
      tDim1 = 1;
      tDim2 = size[0]>=size[2] ? 0 : 2;
    }
    else if (size[2]>=size[0] && size[2]>=size[1]){
      tDim1 = 2;
      tDim2 = size[0]>=size[1] ? 0 : 1;
    }
  }
  
  // experimental
  // can delloc (not safe) if desperate...
  // if (pane->gv->renderQuit)
  //   verts->deleteValues(0);

  // add normals
  bool nonIndexedNormals = false;
  if (norms && norms->getNum() == numVerts &&
      action->getNormalBinding() == SoNormalBinding::PER_VERTEX_INDEXED) {

    SbMatrix matInvTrans = matrix.inverse().transpose();
    mesh->normals.reserve(numVerts);
    for (int i=0; i<numVerts; i++) {
      SbVec3f norm;
      matInvTrans.multDirMatrix((*norms)[i], norm);
      norm.normalize();

      // dont let there be zero normal
      if (norm.dot(norm) < 0.9){
        cerr << __FILE__ << ":" <<  __LINE__ << " (" << __FUNCTION__ << "): "
             << "Detected zero normal.\n";
        norm.setValue(0,1,0);
      }
      
      mesh->normals.push_back(norm);
    }
  } else if (norms && action->getNormalBinding() == SoNormalBinding::PER_VERTEX) {      
    nonIndexedNormals = true;
    SbMatrix matInvTrans = matrix.inverse().transpose();
    int numNorms = norms->getNum();
    mesh->normals.reserve(numNorms);
    for (int i=0; i<numNorms; i++) {
      SbVec3f norm;
      if (createOriginalInstance){
        norm = (*norms)[i];
      } else {
        matInvTrans.multDirMatrix((*norms)[i], norm);
      }
      norm.normalize();

      // dont let there be zero normal
      if (norm.dot(norm) < 0.9){
        cerr << __FILE__ << ":" <<  __LINE__ << " (" << __FUNCTION__ << "): "
             << "Detected zero normal.\n";
        norm.setValue(0,1,0);

        // show the position of the crazy normal
//         static int size=1;
//         static bool shown = false;
//         if (!shown) {
//           shown = true;
//           norm.setValue(
//             drand48()*2-1,
//             drand48()*2-1,
//             drand48()*2-1
//             );
//           Ray(mesh->verts[i],norm,NULL).show2(size++, SbColor(drand48(), drand48(), drand48()));
//           cerr << "Showing zero normal.\n";
//         }
        
      }
      
      mesh->normals.push_back(norm);
    }
  }

  // Get material
  if (mat->isShiny())
    pane->gv->sceneHasShininess = 1;
  
  // add diffuse colors
  if (action->getMaterialBinding() == SoMaterialBinding::PER_VERTEX_INDEXED &&
      (soMat && soMat->diffuseColor.getNum() == numVerts ||
       vProp && vProp->orderedRGBA.getNum() == numVerts) ) {

    mesh->diffuse.reserve(numVerts);
    for (int i=0; i<numVerts; i++) {
    
      SbColor ambient, specular, emission, diffuse;
      float shininess, transparency;
      action->getMaterial(ambient, diffuse, specular, emission,
                          shininess, transparency, i);

      // Since the IndexedShape has not been traversed yet in the
      // above getMaterial call, diffuse will not be correct.
      // Get accurrate diffuse color now.
      if (soMat)
        diffuse = soMat->diffuseColor[i];
      else {
        uint32_t pc = vProp->orderedRGBA[i];
        char r,g,b,t;
        r = (pc & 0xFF000000) >> (8*3);
        g = (pc & 0x00FF0000) >> (8*2);
        b = (pc & 0x0000FF00) >> (8*1);
        t = (pc & 0x000000FF) >> (8*0);
        diffuse = SbColor(r/255., g/255., b/255.);
        transparency = t/255.;
      }
      
      // bad color?
      for (int j=0; j<3; j++){
        if (!finite(diffuse[j])){
          diffuse = SbColor(1.0, 0.0, 0.0);
          break;
        }
      }

      // temporary reflection capping
      const static bool reflCap = (bool)getenv("PANE_REFLECTION_CAP");
      if (reflCap)
          diffuse = BBox::vecMin(diffuse, SbVec3f(.9,.9,.9));
      
      const static bool reflScale = (bool)getenv("PANE_REFLECTION_SCALE");
      if (reflScale)
        diffuse *= .8;
      
      mesh->diffuse.push_back(diffuse);
    }
  }


  // for now, the rest is handled per tri, see below
  
  // loop over indices
  int numCoordIdxs = shape->coordIndex.getNum();
  if (numCoordIdxs<3)
    return SoCallbackAction::PRUNE;

  int v[3];
  int base = 0;

  // count triangles to do single alloc
  for (int j=base+1; j<numCoordIdxs-1; j++) {

    v[0] = shape->coordIndex[base];
    if (v[0]<0) { j = ++base; continue; }
    v[1] = shape->coordIndex[j];
    if (v[1]<0) { j = base = j+1; continue; }
    v[2] = shape->coordIndex[j+1];
    if (v[2]<0) { j = base = j+2; continue; }

    if (isStripSet)
      base++;
    
    // detect same vertices
    if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2])
      continue;
    
    numTris++;
  }

  //cerr << "allocating space for " << numTris << " triangles...\n";
  mesh->indexes.reserve(numTris*3);
  if (nonIndexedNormals)
    mesh->normalIndexes.reserve(numTris*3);
  if (createOriginalInstance)
    rayAccel->reserve(numTris);

  // create texture
  if (img && (hasTextCoords || genDefTextCoords) ){

    // only works for rgb textures
    if (numComps != 3 && numComps != 1){
      cerr << "Illegal number of components: " << numComps << endl;
    } else {
      
      // add texture coordinates
      Texture *texture = new Texture(action->getTextureWrapS(),
                                     action->getTextureWrapT(),
                                     action->getTextureModel(),
                                     action->getTextureBlendColor(),
                                     numComps, textureSize, img,
                                     pane->gv->bumpMapping,
                                     pane->gv->bumpScale);

      // add texture coordinates
      mesh->uvs.reserve(numVerts);
      for (int k=0; k<numVerts; k++) {
        SbVec2f t2;
        if (localTexCoord){
          t2 = (*localTexCoord)[k];
        }
        else if (action->getNumTextureCoordinates()==numVerts){
          t2  = action->getTextureCoordinate2(k);
        }
        else if (genDefTextCoords){
          const SbVec3f &v = mesh->verts[k];
          t2[0] = (v[tDim1] - minMax.min[tDim1])*invBboxSize[tDim1];
          t2[1] = (v[tDim2] - minMax.min[tDim2])*invBboxSize[tDim2];
        }
        else {
          GIVE_ERROR("No texture coordinates available!\n");
        }
          
        SbVec3f t3;
        t3.setValue(t2[0], t2[1], 0.0);
        SbMatrix tmatrix = action->getTextureMatrix();
        tmatrix.multVecMatrix(t3, t3);
        mesh->uvs.push_back(SbVec2f(t3[0], t3[1]));
      }

      mesh->texture = texture;
    }
  }
  
//   if (mat->emissive)
//     pane->litTris.reserve(pane->litTris.size()+numTris);
//   else
//     pane->regTris.reserve(pane->regTris.size()+numTris);

  //cerr << "continuing processing tris...\n";
  numTris = 0;
  colinearVerts = 0;
  base = 0;
  int numDelim=0;
  
  for (int j=base+1; j<numCoordIdxs-1; j++) {

    v[0] = shape->coordIndex[base];
    if (v[0]<0) {
      j = ++base;
      triCount = 0;
      numDelim++;
      continue;
    }
    v[1] = shape->coordIndex[j];
    if (v[1]<0) {
      j = base = j+1;
      triCount = 0;
      numDelim++;
      continue;
    }
    v[2] = shape->coordIndex[j+1];
    if (v[2]<0) {
      j = base = j+2;
      triCount = 0;
      numDelim++;
      continue;
    }

    int ni[3] = { base, j, j+1 };
    
    if (isStripSet)
      base++;
    
    // detect same vertices
    if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2]){
      if (colinearVerts++ == 0)
        cerr << "Warning! Colinear vertices in triangle!\n";
      continue;
    }

    // triangle found
    numTris++;             // for counting triangles
    triCount++;            // for flipping odd triangles

    if (numTris%500000==0){
      oldSec = theTime.tv_sec;
      gettimeofday(&theTime, NULL);
      if (theTime.tv_sec != oldSec)
        fprintf(stderr, "\r%6.2f%%    ", j/(float)(numCoordIdxs)*100);
    }

    // decimate
    static int decimate = getenv("PANE_DECIMATE") ? // only use every Nth tri
      atoi(getenv("PANE_DECIMATE")) : 0;
    if (decimate && numTris%decimate) continue;  
    
    int indexesN = mesh->indexes.size();
    int myv[3] = { v[0], v[1], v[2] };
    
    // flip triangles if odd and IndexedTriangleStripSet or reverse ordering
    if (isStripSet && triCount%2==0 ||
        action->getVertexOrdering() == SoShapeHints::CLOCKWISE) {

      // reverse order
      myv[0] = v[2];
      myv[2] = v[0];
    }

    mesh->indexes.push_back(myv[0]);
    mesh->indexes.push_back(myv[1]);
    mesh->indexes.push_back(myv[2]);

    if (nonIndexedNormals){
      mesh->normalIndexes.push_back(ni[0]-numDelim);
      mesh->normalIndexes.push_back(ni[1]-numDelim);
      mesh->normalIndexes.push_back(ni[2]-numDelim);
    }

    if (mat->emissive &&
        !createOriginalInstance) { // instanced explicit lit triangles not supported
      EmissiveTriangle t(pane, mesh, indexesN);
      pane->litTris.push_back(t);

      // record keeping
      pane->numLights++;
      pane->totalEmisAvg += t.getEmisAvg();
      if (t.getEmisAvg() > pane->maxEmisAvg)
        pane->maxEmisAvg = t.getEmisAvg();
    }
    else {
      if (createOriginalInstance){
        if (pane->instanceTris.size() == pane->instanceTris.capacity()){
          GIVE_ERROR("Exceeded allocated instance memory! Programmer error.\n");
          exit(1);
        }
        Triangle t(mesh, indexesN);
        pane->instanceTris.push_back(t);
        rayAccel->add(&pane->instanceTris.back());
      } else {
        Triangle t(mesh, indexesN);
        pane->regTris.push_back(t);
        // DMTri t(mesh, n, mat, texture);
        // pane->dmTris.push_back(t);
      }
    }
  }

  if (createOriginalInstance){
    cerr << "Created BIH of instance: ";
    rayAccel->setParams(30, 6, false);
    rayAccel->refine(false);
    rayAccel->printStats();
    cerr << "\n";
  }

  oldSec = theTime.tv_sec;
  gettimeofday(&theTime, NULL);
  if (theTime.tv_sec != oldSec)
    fprintf(stderr, "\r%6.2f%%    ", 100.0);
  
  // can delloc (not safe) if desperate...
  // if (pane->gv->renderQuit)
  //   shape->coordIndex.deleteValues(0);
  
  // disable triangle processing (or we'll add the shape as individual triangles)
  pane->processTriangles = false;

//  return SoCallbackAction::CONTINUE;
    return SoCallbackAction::PRUNE;
}

SoCallbackAction::Response
Pane::indexedShapePostCB(void *data, SoCallbackAction *action,
                           const SoNode *node){
  Pane *pane = (Pane *)data;
  SoIndexedShape *shape = (SoIndexedShape *)node;

  // re-enable triangle processing
  pane->processTriangles = true;
  
  return SoCallbackAction::CONTINUE;
}

// material CB Handler
SoCallbackAction::Response
Pane::matPreCB(void *data, SoCallbackAction *action,
	       const SoNode *node){
  Pane *pane = (Pane *)data;
  SoMaterial *mat = (SoMaterial *)node;

  if (mat->diffuseColor.getNum()) {
    pane->sepStack.front()->soMat = mat;
    pane->sepStack.front()->vProp = NULL;
  }
  
  // create new material node
  SoMatPlus *matPlus = pane->sepStack.front()->matPlus;
  Material *newMat = new Material(mat, matPlus);
  pane->sepStack.front()->mat = newMat; // update stack 
  
  pane->matList.push_back(newMat);
  return SoCallbackAction::CONTINUE;
}

// LinearFractal CB Handler
// Currently Material must come AFTER SoMatPlus.
SoCallbackAction::Response
Pane::linearFractalPreCB(void *data, SoCallbackAction *action,
                         const SoNode *node){
  Pane *pane = (Pane *)data;
  SoLinearFractal *linFract = (SoLinearFractal *)node;
  
  SbMatrix matrix = action->getModelMatrix();
  pane->fractals.push_back(
    Fractal(((PrimList*)pane->scenePrim)->prim(0),
            linFract)
    );

  // swap rayAccel with instance
  (*(PrimList*)pane->scenePrim)[0] =
    new Instance(&pane->fractals.back(), matrix);
  
  return SoCallbackAction::CONTINUE;
}

// matPlus CB Handler
// Currently Material must come AFTER SoMatPlus.
SoCallbackAction::Response
Pane::matPlusPreCB(void *data, SoCallbackAction *action,
		   const SoNode *node){
  Pane *pane = (Pane *)data;
  SoMatPlus *matPlus = (SoMatPlus *)node;
  
  // place matPlus on stack
  pane->sepStack.front()->matPlus = matPlus;

  // FIXME mat plus needs to register with material
  return SoCallbackAction::CONTINUE;
}

// SoCoordinate3 CB Handler
SoCallbackAction::Response
Pane::coord3PreCB(void *data, SoCallbackAction *action, const SoNode *node){
  Pane *pane = (Pane *)data;
  SoCoordinate3 *coord = (SoCoordinate3 *)node;
  if (coord->point.getNum())
    pane->sepStack.front()->verts = &(coord->point);
  return SoCallbackAction::CONTINUE;
}

// SoNormal CB Handler
SoCallbackAction::Response
Pane::normalPreCB(void *data, SoCallbackAction *action, const SoNode *node){
  Pane *pane = (Pane *)data;
  SoNormal *normal = (SoNormal *)node;
  if (normal->vector.getNum())
    pane->sepStack.front()->norms = &(normal->vector);
  return SoCallbackAction::CONTINUE;
}

// SoVertexProperty CB Handler
SoCallbackAction::Response
Pane::vPropPreCB(void *data, SoCallbackAction *action, const SoNode *node){
  Pane *pane = (Pane *)data;
  SoVertexProperty *vProp = (SoVertexProperty *)node;

  if (vProp->orderedRGBA.getNum()){
    pane->sepStack.front()->mat = NULL;
    pane->sepStack.front()->vProp = vProp;
  }
  
  if (vProp->vertex.getNum())
    pane->sepStack.front()->verts = &(vProp->vertex);
  if (vProp->normal.getNum())
    pane->sepStack.front()->norms = &(vProp->normal);
  return SoCallbackAction::CONTINUE;
}

