#include "Pane.h"
#include "PreviewQuad.h"
#include "debug.h"
#include "FileUtils.h"

#include <srfftw.h>

using namespace std;

//'WW''''''''''''''''''WW''''WW''''''''''
// WW                  WWW  WWW          
//WWWW  WW   WWW   WW  W WWW WW  WWW WWW 
// WW  WW W  WW W WWWW W  W  WW W WW WW W
// WW  WW W  WW W WW   W     WW W WW WW W
// WW   WW   WW W  WWW W     WW  WWW WWW 
//                                   WW  
//'''''''''''''''''''''''''''''''''''WW''
//
// Tone maps the image, turning radiance
// values into illuminance ones, then finally
// into image pixel colors.
//
// The algorithm used is described in
// Peter Shirley's Realistic Ray Tracing 1st Ed. Ch. 18 p. 152.
// Radiance values are turned into illuminance values,
// dark values are de-saturated to simulate the human
// visual system's b&w night vision, and Greg Ward's
// contrast based scalefactor is used to compress the illuminance
// values into the displayable range.
//
// I tried my best to choose physically accurate primaries.
void Pane::toneMap(SbColor *&pixOut){

//   Values in pix[] are W/m^2-sr in three channels.
//
//   What spectra are these channels? Good question. First I take
//   the matrices for RGB->XYZ (and the inverse) from www.srgb.com.
  

  // Tone mapping algorithm is taken from
  // Realistic Ray Tracing by Peter Shirley page 152
       
  SbColor *c;
  double r, g, b;
    
  // tone mapping variables
  const double k = 683.0;  // k constant (lm/W) (Peter Shirley)
  double V;
  //const double Yw = 100.0; // maximum display luminance (cd/m^2) (PS)
  //const double Yw = 110.5; // maximum display luminance (cd/m^2)
  const double Yw = 250.0; // maximum display luminance (cd/m^2) (cinema)
  //const double Yw = 80.0; // srgb standard
  double Yi = 0.0;         // world adaptation luminance
  double avgYi = 0.0;      // world average luminance
  double cY;       // tristimulous values
  double x, y, W, s;       // misc
  //const double l[3] = { 620.0, 545.0, 465.0 }; // lambda (nm)
  const double dl[3] = { 1.0, 1.0, 1.0 }; // dlambda

  // From Chris R. Schoenenman
  // http://research.microsoft.com/~hollasch/cgindex/color/freq-rgb.html
  // Values for xbar, byar, zbar

//   const double xb[3] = { 0.8563, 0.4516, 0.2541 }; // xbar
//   const double yb[3] = { 0.3981, 0.9822, 0.1528 }; // ybar
//   const double zb[3] = { 0.0000, 0.0079, 1.5549 }; // zbar
//   const double vb[3] = { 0.0332, 0.4810, 0.3281 }; // vbarn



//   // from apple cinema display numbers
//   // www.cis.rit.edu/research/mcsl/research/PDFs/LCD.pdf
//   // computed by dividing matrix given by highest element
//   // so that components are relative.
//   const double l[3] = { 618.0, 555.0, 485.0 }; // lambda (nm)
//   const double xb[3] = { 0.7863, 0.4772, 0.2265 }; // xbar
//   const double yb[3] = { 0.4103, 1.0000, 0.1638 }; // ybar
//   const double zb[3] = { 0.0142, 0.1197, 1.2792 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar

  
  // sRGB standard, www.srgb.com
  // these are not real single-wavelength primaries
  // however, they have similar photopic responses
  // to real wavelengths.
  // i.e., red causes a red tristimulous response
  // with the same luminance (Y) as a wavelength of 635nm,
  // but with different chromacity (X, Z).

  // Each of "r", "g", "b" do not have a single wavelength
  // equivalent. Instead, each causes a response
  // according to the table below, yanked from www.srgb.com.
  // While not exactly realistic, the red channel causes
  // a red response, and similarly for green and blue.
  // Additionally, RGB(1,1,1) produces XYZ(.95,1,1.089)

  // In our RGB space, the r coordinate, expressed
  // in W/m^2/sr, is not the power of a single (red)
  // wavelength of light, but rather the power of a
  // distribution evoking a red response. For example,
  // a value of 1.0 evokes a response of XYZ(.4124,.2126,.0193),
  // or, said another way, .4124*(the X stimuli,


//  const double l[3] = { 635.0, 520.0, 466.0 }; // lambda (nm)


//   const double xb[3] = { 0.4124, 0.3576, 0.1805 }; // xbar
//   const double yb[3] = { 0.2126, 0.7152, 0.0722 }; // ybar
//   const double zb[3] = { 0.0193, 0.1192, 0.9505 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar


  // dominant wave lengths
  // const double l[3] = { 635.0, 520.0, 466.0 }; // lambda (nm)
  
  // primary spectra
  //const double l[3] = { 611.0, 549.0, 460.0 }; // lambda (nm)

  // xyz functions scaled by (0.595, 1.0, 0.85) to make white
//   const double xb[3] = { 0.5899, 0.4183, 0.2472 }; // xbar
//   const double yb[3] = { 0.2918, 0.9928, 0.0510 }; // ybar
//   const double zb[3] = { 0.0002, 0.0095, 1.4188 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar


  const double l[3] = { 611.0, 549.0, 467.0 }; // lambda (nm)
  
  // note: these multipliers highly dependent on the white point (xw, yw)!!
//   // this is centered about the sRGB white point but it
//   // desaturates the blue wall of jensen's cornell box
//   const double xw=0.3127, yw=0.3290;
//   const double d = 0.825;
//   const double e = 1.0;
//   const double f = 0.985;

//   // no white-point centering
//   const double xw=0.3333, yw=0.3333;
//   const double d = 0.65;
//   const double e = 0.94;
//   const double f = 1.0;

//   // responses taken from interpolated CIE 1931 xb yb zb data (wyszecki stiles)
//   const double xb[3] = { 0.9914*d, 0.4183*e,    0.2408*f }; // xbar
//   const double yb[3] = { 0.4905*d, 0.9928*e,    0.0770*f }; // ybar
//   const double zb[3] = { 0.0003*d, 0.0095*e,    1.4861*f }; // zbar 466
//   const double vb[3] = { 0.01477, 0.497, 0.631 }; // vbar


  //const double xw=0.3127, yw=0.3290;
  const double xw=0.3333, yw=0.3333;
  // scale fake primaries to match brightness (luminance) of
  // balanced real primaries: xb,yb,zb *= sum(real_yb)/sum(fake_yb)
//   const double d = (0.4905*0.65 + 0.9928*0.94 + 0.0770*1.0) /
//                    (0.2126+0.7152 +0.0722);  // 1.3291
//   const double e = d;
//   const double f = d;
  
//   const double xb[3] = { 0.4124*d, 0.3576*e, 0.1805*f }; // xbar
//   const double yb[3] = { 0.2126*d, 0.7152*e, 0.0722*f }; // ybar
//   const double zb[3] = { 0.0193*d, 0.1192*e, 0.9505*f }; // zbar
  const double vb[3] = { 0.01477, 0.497, 0.631 }; // vbar

  // inverse of sRGB, from http://www.w3.org/Graphics/Color/sRGB
//   const double xb[3] = {  3.2410, -1.5374, -0.4986 }; // xbar
//   const double yb[3] = { -0.9692,  1.8760,  0.0416 }; // ybar
//   const double zb[3] = {  0.0556, -0.2040,  1.0570 }; // zbar
  // NOTE  these values are *slightly* different than
  // http://www.w3.org/Graphics/Color/sRGB
  const double xb[3] = { 0.4124, 0.3576, 0.1805 }; // xbar
  const double yb[3] = { 0.2126, 0.7151, 0.0721 }; // ybar
  const double zb[3] = { 0.0193, 0.1192, 0.9505 }; // zbar
  
  // use to convert XYZ -> to my RGB (i.e., inv([xb,yb,zy]*d))
  const double inv[3][3] =
    {
      {  2.4383,   -1.1566,   -0.3752 },
      { -0.7290,    1.4113,    0.0312 },
      {  0.0419,   -0.1535,    0.7953 }
    };

//   const double xb[3] = { 0.6742, 0.4183, 0.2297 }; // xbar
//   const double yb[3] = { 0.3335, 0.9928, 0.0474 }; // ybar
//   const double zb[3] = { 0.0002, 0.0095, 1.3187 }; // zbar
//   const double vb[3] = { 0.00862, 0.402, 0.851 }; // vbar

  // Chromaticity values from Peter Shirley
  //const double xw = 0.283, yw = 0.298; // (xw, yw) for trinitron
  //const double xw=0.3328, yw=0.3516; // cinema display
  // my own chromacity values
  //const double xw = 0.34, yw = 0.37; // (xw, yw) for trinitron
  //const double xw = 0.284, yw = 0.3; // (xw, yw) for trinitron


  // make sure pixel colors are not inf, nan, or negative
  SbVec3f maxPix(0,0,0);
  int pixErrs=0;
  for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){

    SbColor &c = pix[i];
    
    if ( !finite(c) ){
      int j = i/gv->imgTotalDim[X];
      int ii = i%gv->imgTotalDim[X];
      if (pixErrs++<10){
        cerr << "Pixel at (" << ii << ", " << j << ") is invalid: "
             << c << endl;
      }
      
      if (!finite(c[R])) c[R] = 0.0;
      if (!finite(c[G])) c[G] = 0.0;
      if (!finite(c[B])) c[B] = 0.0;
    }

    if ( c[0]<0 || c[1]<0 || c[2]<0 ){
      int j = i/gv->imgTotalDim[X];
      int ii = i%gv->imgTotalDim[X];
      if (pixErrs++<10){
        cerr << "Pixel at (" << ii << ", " << j << ") is negative: "
             << c << endl;
      }
      
      if (c[R]<0) c[R] = 0.0;
      if (c[G]<0) c[G] = 0.0;
      if (c[B]<0) c[B] = 0.0;
    }

    for (int d=0; d<3; d++){
      bool said = false;
      if (c[d]>maxPix[d]){
        maxPix[d] = c[d];
//         if (!said){
//           int y = i/gv->imgTotalDim[X];
//           int x = i-(gv->imgTotalDim[X]*y);
//           y = gv->imgTotalDim[Y]-1-y;
//           cerr << "Max pixel at " << x << "," << y << endl;
//           said = true;
//         }
      }
    }
  }
  if (pixErrs>10)
    cerr << "Warning: " << pixErrs << " total pixel errors!\n";

  // adjust exposure
  cerr << "Maximum radiance: " << maxPix;
  if (gv->exposure != 1.0)
    for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++)
      pix[i] *= gv->exposure;
  cerr << " (" << maxPix*gv->exposure << " after exposure compensation)\n";
  
  // calculate Yi, log average, and maxY
  float maxY = 0.0;
  float rgbAvg = 0;
  int numpix = 0;
  int numpixRgb = 0;
  for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){

    SbColor c = pix[i];
    cY = k * ( yb[R]*c[R]*dl[R] + yb[G]*c[G]*dl[G] + yb[B]*c[B]*dl[B] );

    // track regular rgb average for TM_AVERAGE
    rgbAvg += vecAvg(c);
    numpixRgb++;

    // minimum value of cY taken from Tumblin Rushmeier '93
    if (cY>3.18e-7 && (gv->stepSize || pixDepth[i]>0.0) ){
      if (cY > maxY){
	maxY = cY;
      }
      Yi += log10(cY);
      avgYi += cY;
      numpix++;
    }
  }

  // Log average of Y idea of Jack Tumblin and Holly Rushmeier
  // finish calculating Yi
  if (numpix){
    Yi /= numpix;
    avgYi /= numpix;
  }
  if (numpixRgb)
    rgbAvg /= numpixRgb;
  
  Yi = pow(10.0, Yi);

  cerr << "Max luminance =         " << maxY << " cd/m^2\n"; //cd = lm/sr
  cerr << "Log average luminance = " << Yi   << " cd/m^2\n"; //lm=638*W*eff(wl)
  
  //Yi += 0.84;        // from Tumblin Rushmeier '93
                       // log10(Lw) = E{log10(Lin)}+0.84

  // use user specified Yi
  // if you set Yi = maxY/2 then CBSF matches interactive preview
  if (gv->luminanceLogMean>0.0)
    Yi = gv->luminanceLogMean;
  
  lastImageYi = Yi;
  lastImageAvgYi = avgYi;

  cerr << "Adaptation luminance = " << Yi << " cd/m^2\n";
  cerr << "Average luminance = " << avgYi << " cd/m^2\n";

  // do tone mapping
  if (gv->toneMapping){

    static bool disableDesat = getenv("PANE_NO_DESATURATE");
    if (disableDesat){
      cerr << "Desaturation disabled." << endl;
    }
    

    // convert spectra to CIE XYZ, then to Yxy
    for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){

      SbColor c = pix[i];
      SbColor L = c;

      // convert rgb to XYZ
      L[X] = k * ( xb[R]*c[R]*dl[R] + xb[G]*c[G]*dl[G] + xb[B]*c[B]*dl[B] );
      L[Y] = k * ( yb[R]*c[R]*dl[R] + yb[G]*c[G]*dl[G] + yb[B]*c[B]*dl[B] );
      L[Z] = k * ( zb[R]*c[R]*dl[R] + zb[G]*c[G]*dl[G] + zb[B]*c[B]*dl[B] );
      
      V = 1700.0 * ( vb[R]*c[R]*dl[R] + vb[G]*c[G]*dl[G] + vb[B]*c[B]*dl[B] );
      
      if (L[Y]<3.18e-7)
	L[Y] = 3.18e-7;

      // choose desaturation depending on brightness (human visual system modeling)
      if (log10(L[Y]) < -2.0)
	s = 0.0;
      else if (log10(L[Y]) < 0.6){
	double f = (log10(L[Y])+2.0)/2.6;
	s = 3.0*f*f -  2.0*f*f*f;
      }
      else
	s = 1.0;

      // convert to Yxy
      W = L[Y] + L[X] + L[Z];
      x = L[X]/W;
      y = L[Y]/W;

      // desaturate
      // xw = yw = 0.25 for nighttime rendering. ("Night Rendring" by Jensen et al.)
      if (!disableDesat &&
          gv->toneMapping != GlobalVars::TM_AVERAGE &&
          gv->toneMapping != GlobalVars::TM_MAX &&
          gv->toneMapping != GlobalVars::TM_NONE
        ) {
        x = (1.0-s)*xw + s*(x+xw-0.3333);
        y = (1.0-s)*yw + s*(y+yw-0.3333);
        L[Y] = 0.4468*(1.0-s)*V + s*L[Y];
      }
	
      // store transformed colors (Yxy)
      pix[i][X] = x;
      pix[i][Y] = L[Y];
      pix[i][Z] = y;
    }


    if (gv->toneMapping == GlobalVars::TM_AVERAGE && gv->luminanceLogMean>0){

      cerr << "Applying given scale factor: " << gv->luminanceLogMean << endl;
      // this is a hack to match TM_AVERAGE
      double sf = gv->luminanceLogMean;
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	pix[i][Y] *= sf;
      }
    }
    else if (gv->toneMapping == GlobalVars::TM_WARDCBSF){
      // Ward's contrast-based scale-factor
      
      // This scalefactor maps Yi to Yw/2.0, then divides by Yw
      // to map Yw to 1.0. This works in conjunction with
      // the XYZ->RGB matrix later, which maps
      // a luminance of 1.0 to an rgb of about (1,1,1)
      // which gives a physical luminance of Yw.
      //
      // Example input: Yi -> Yw/2 -> 0.5 -> Yw/2 physical output
      double sf = 1.0/Yw * pow( ( 1.219 + pow(Yw/2.0, 0.4) ) /
			      ( 1.219 + pow(Yi, 0.4) ), 2.5);

      cerr << "Contrast-based scalefactor = " << sf << endl;

      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){

	// scale luminance
	// after this scaling, 1.0 is really
	// Yw (the real monitor brightness) in real life
	pix[i][Y] *= sf;
      }
    }
    
    else if (gv->toneMapping == GlobalVars::TM_WARDHIST)
      histogramAdjust(0);
    
    else if (gv->toneMapping == GlobalVars::TM_WARDHUMAN)
      histogramAdjust(1);

    else if (gv->toneMapping == GlobalVars::TM_NONLINEAR){

      // RRT 2nd ed
      // do linear scale followed by nonlinear scale

      // map Yi to .18 (pg. 1180)
      static const char *keyStr = getenv("PANE_EXPOSURE_KEY");
      static const float key = keyStr ? atof(keyStr) : .18; // can play around with this..
      double sf = key/Yi;
      maxY *= sf;

      // do nonlinear scale (pg. 181)
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	pix[i][Y] *= sf;
	if (maxY)
	  pix[i][Y] *= (1.0+pix[i][Y]/(maxY*maxY)) / (1.0+pix[i][Y]);
      }
    } else if (gv->toneMapping == GlobalVars::TM_EXPOSURE){

      // Exposure mapping
      // http://freespace.virgin.net/hugo.elias/graphics/x_posure.htm
      // Also: "Photorealistic rendering of scenes with physically-based sky
      // light" by Andreas Agvard and Fredrik Lanker

      // remap Yi to key
      static const char *keyStr = getenv("PANE_EXPOSURE_KEY");
      static const float key = keyStr ? atof(keyStr) : 0.5; // can play around with this..
      if (keyStr){
        cerr << "Exposure mapping using key = " << key << "." << endl;
      }
      
      double sf = key/Yi;
      maxY *= sf;

      // do scaling
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	pix[i][Y] *= sf;
	if (maxY)
	  pix[i][Y] = 1-exp(-pix[i][Y]);
      }
    } else if (gv->toneMapping == GlobalVars::TM_AVERAGE){

      // this approximately matches the preview scaling
      
      // find average color
      float avgLum=0;
      int cnt=0;
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	if (pix[i][Y]>3.18e-7){
	  avgLum += pix[i][Y];
          cnt++;
        }
      }

      // remap avg to 0.5
      if (cnt)
        avgLum /= cnt;
      double sf = 1;

      // we apply exposure again because dividing by average removes its effect
      float mult=0;
      if (gv->preview) {
        sf *= gv->exposure*gv->preview->getMult()/k;
        mult = gv->preview->getMult();
      }
      else if (rgbAvg) {
        sf *= gv->exposure*0.5f/(rgbAvg*k);
        mult = 0.5f/rgbAvg;
      }

      cerr << "ToneMap_Average: scaleFactor=" << sf << " mult=" << mult << " avgLum=" << avgLum << endl;
      cerr << "Equivalent to: -Tnone --exposure=" << sf*k << "\n";
      // do scaling
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	pix[i][Y] *= sf;
      }
    } else if (gv->toneMapping == GlobalVars::TM_MAX){
      double sf = gv->exposure/maxY;

      cerr << "ToneMap_Max: scaleFactor=" << sf << endl;
      // do scaling
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	pix[i][Y] *= sf;
      }

    } else if (gv->toneMapping == GlobalVars::TM_NONE){
      // do nothing... but remove k luminance conversion factor
      float sf = 1.f/k;
      for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
	pix[i][Y] *= sf;
      }
    } else {
      cerr << "Programmer error." << endl;
      GIVE_ERROR("Programmer error");
    }

    // calculate max luminance
    float maxY2 = 0.0;
    float minY2 = 1e99;
    for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
      if (pix[i][Y] > maxY2) maxY2 = pix[i][Y];
      if (pix[i][Y] < minY2) minY2 = pix[i][Y];
    }

    static bool stretchContrast = getenv("PANE_STRETCH_CONTRAST");
    if (stretchContrast){

      //     |---------------|          // screen [0,1]
      //         |--------|             // stretch
      // |---|----|
      //              |------|--|
      // |---|---------------|--|
      //|--|
      //                       |--|
      //

      cerr << "Auto-stretching contrast: " << flush;
      
      if ((minY2>0 || maxY2<0) && minY2<maxY2 ){
        float offset = min2(-minY2,0.f);
        float scale = 1 / (min2(1.f,maxY2)-max2(0.f,minY2));
        cerr << "offset=" << offset << " scale=" << scale << ".\n";
        float newMinY2=1e99, newMaxY2=0;
        for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++){
          pix[i][Y] = (pix[i][Y] + offset) * scale;
          if (pix[i][Y] > newMaxY2) newMaxY2 = pix[i][Y];
          if (pix[i][Y] < newMinY2) newMinY2 = pix[i][Y];
        }
        cerr << "  [" << minY2    << ", " << maxY2    << "] -> "
             << "  [" << newMinY2 << ", " << newMaxY2 << "]\n";
      } else {
        cerr << "not needed.\n";
      }
    }
    
    cerr << "Maximum scaled luminance = " << maxY2 << endl;
  
    // convert Yxy ->rgb
    // two methods:
    // glare, and no glare
    
    // with glare  
    if (gv->useGlare)
      // glareFilter convolves a model of the eye's point-spread function
      // against all above a given thresh hold.
      // in this case, the threshhold is the monitor's maximum luminance,
      // after scaling, which is Yw*sf.
      //
      // This is really convoluted:
      // rgb->XYZ->Yxy->scale (tonemap)->
      //        glare filter->rgb->combine with glare->gamma correct->rgb
      glareFilter(Yi, 1.0, pixOut); //Yw*sf); 
  
    // without glare
    else {

      // Yxy -> RGB
      convertYxyToRgb(pix, gv->imgTotalDim[0], gv->imgTotalDim[1]);
    }
  }

  gammaCorrect(pix, gv->imgTotalDim[X], gv->imgTotalDim[1]);
    
  toneMapped = 1;
  return;
}

float Pane::jnd(float La){
  float lLa = log10(La);
  float lLt;
  if (lLa<-3.94)
    lLt = -2.86;
  else if (lLa<-1.44)
    lLt = pow(0.405*lLa+1.6, 2.18)-2.86;
  else if (lLa<-0.0184)
    lLt = lLa-0.395;
  else if (lLa<1.9)
    lLt = pow(0.249*lLa+0.65,2.7)-0.72;
  else
    lLt = lLa-1.255;
  return (pow(10.0f, lLt));
}

// input is pix array, containing Lxy (stored as xLy) values
void Pane::histogramAdjust(int human){

  // downsample image so each pixel represents 1 degree field of view

  // compute new image size
  float aspectRatio = ABSF(gv->imgTotalDim[X]/(double)gv->imgTotalDim[Y]);
  float heightAngle = gv->heightAngle;
  float widthAngle = aspectRatio*heightAngle;

  int dim[2];
  dim[0] = (int)(2.0 * tan(widthAngle/2.0)/(2.0*M_PI/360.0) );
  dim[1] = (int)(2.0 * tan(heightAngle/2.0)/(2.0*M_PI/360.0) );
  dim[0] = gv->imgTotalDim[0]/2;
  dim[1] = gv->imgTotalDim[0]/2;

  // scale image
  vector<float> scaleImg(dim[0]*dim[1]);
  for (int j=0; j<dim[1]; j++){
    for (int i=0; i<dim[0]; i++){

      // compute boundaries in previous image
      float imgCrd[2];
      int pixCrd[2];
      int i0, j0, i1, j1;

      // lower left
      imgCrd[0] = 1.0/dim[0]*i;
      imgCrd[1] = 1.0/dim[1]*j;
      pixCrd[0] = (int)(imgCrd[0]*gv->imgTotalDim[0]);
      pixCrd[1] = (int)(imgCrd[1]*gv->imgTotalDim[1]);
      i0 = MAX2(0,pixCrd[0]);
      j0 = MAX2(0,pixCrd[1]);

      // top right
      imgCrd[0] += 1.0/dim[0];
      imgCrd[1] += 1.0/dim[1];
      pixCrd[0] = (int)(imgCrd[0]*gv->imgTotalDim[0]);
      pixCrd[1] = (int)(imgCrd[1]*gv->imgTotalDim[1]);
      i1 = MIN2(gv->imgTotalDim[0]-1,pixCrd[0]);
      j1 = MIN2(gv->imgTotalDim[1]-1,pixCrd[1]);
      
      // and apply box filter
      scaleImg[j*dim[0]+i] = 0.0;
	
      for (int y=j0; y<=j1; y++)
	for (int x=i0; x<=i1; x++)
	  scaleImg[j*dim[0]+i] += pix[y*gv->imgTotalDim[0]+x][Y];

      scaleImg[j*dim[0]+i] /= (j1-j0+1)*(i1-i0+1);
    }
  }

#if 0
  // dump scaled image
  const char *fn = "scale.dump";
  FILE *file = fopen(fn, "wb");
  if (!file){
    cerr << "Unable to open file " << fn << ": " << strerror(errno) << endl;
    return;
  }
  for (int i=1; i<dim[0]*dim[1]; i++)
    fwrite(&scaleImg[i], sizeof(float), 1, file);
  fclose(file);
  cerr << "Wrote " << dim[0]*dim[1] << " floats to " << fn << endl;
#endif
  
  // create histogram of luminance values
  const int bins = 100;
  const int binmul = 20;
  int goodpix = 0;

  // find min/max luminances
  float minLum = 1.0e-4;  // minimum luminance value perceivable
  float Lwmin, Lwmax;
  Lwmin = Lwmax = 0.0;
  
  for (int i=0; i<dim[0]*dim[1]; i++){
    
    if (scaleImg[i] <= 0.0)
      continue;
    
    if (Lwmin <= 0.0)
      Lwmin = scaleImg[i];
    else if (scaleImg[i] < Lwmin)
      Lwmin = scaleImg[i];

    if (scaleImg[i] > Lwmax)
      Lwmax = scaleImg[i];

    goodpix += binmul;
  }
  if (Lwmin <= 0.0){
    cerr << "Image is black (great!).\n";
    return;
  }
  Lwmin = MAX2(Lwmin, minLum);

  float Bwmin = log(Lwmin);
  float Bwmax = log(Lwmax);
  if (Bwmax == Bwmin)
    Bwmax += 0.0001*fabs(Bwmax);

  float Ldmax = 100.0;
  float Ldmin = 1.0;
  float Bdmax = log(Ldmax);
  float Bdmin = log(Ldmin);

  if ( (Bdmax-Bdmin)>(Bwmax-Bwmin) ){
    cerr << "Compression unnecessary. Using linear scalefactor.";

    float sf = (Bdmax-Bdmin)/(Bwmax-Bwmin);
    for (int i=0; i<gv->imgTotalDim[0]*gv->imgTotalDim[1]; i++)
      pix[i][Y] = pix[i][Y] * sf + Bdmin;
    
    return;
  }
  
  // initialize histogram
  // f[0]=0, bins are f[1]...f[bins]
  vector<int> f(bins+1, 0);
  vector<float> P(bins+1, 0.0);
  float Lwavg = 0.0;
  int goodScaledPix = 0;

  // count luminances (log scale)
  for (int i=0; i<dim[0]*dim[1]; i++){
    if (scaleImg[i] <= 0.0)
      continue;
    
    int b = (int)( ( log(scaleImg[i]) - Bwmin ) / (Bwmax-Bwmin) * bins );
    b++;
    if (b<0)
      b = 0;
    if (b>bins)
      b = bins;

    f[b] += binmul;

    Lwavg += log(scaleImg[i]);
    goodScaledPix++;
  }

  // compute average luminance
  if (goodScaledPix)
    Lwavg = exp(Lwavg/goodScaledPix);
  if (Lwavg < minLum)
    Lwavg = minLum;

  // compute c.d.f.
  float sum = 0;
  for (int j=1; j<=bins; j++){
    sum += f[j];
    P[j] = sum/(float)goodpix;
  }

#if 0
  // dump c.d.f.
  fn = "cdf.dump.0";
  file = fopen(fn, "wb");
  if (!file){
    cerr << "Unable to open file " << fn << ": " << strerror(errno) << endl;
    return;
  }
  for (int i=0; i<=bins; i++)
    fwrite(&P[i], sizeof(float), 1, file);
  fclose(file);
  cerr << "Wrote " << dim[0]*dim[1] << " floats to " << fn << endl;
#endif
  
  // do ceiling on histogram
  int total = dim[0]*dim[1];
  float tolerance = 0.025 * goodpix;
  const float deltab = (Bwmax-Bwmin)/bins;
  int trimmings;
  do {
    trimmings = 0;

    // count good pix
    goodpix = 0;
    for (int j=1; j<=bins; j++)
      goodpix += f[j];
    
    if (goodpix < tolerance){

      double sf = 1.0/Ldmax * pow( ( 1.219 + pow(Ldmax/2.0f, 0.4f) ) /
				   ( 1.219 + pow(Lwavg*180.f/683.f, 0.4f) ), 2.5);
      
      cerr << "ERROR: No compression needed, "
	   << "should have used linear scale earlier.\n"
	   << "Using contrast-based scalefactor: " << sf << "\n";
      
      for (int i=0; i<gv->imgTotalDim[0]*gv->imgTotalDim[1]; i++)
	pix[i][Y] *= sf;
      
      return;
    }

    // recompute c.d.f.
    float sum = 0;
    for (int j=1; j<=bins; j++){
      sum += f[j];
      P[j] = sum/(float)goodpix;
    }
    
    for (int i=1; i<=bins; i++){
    
//       // recount good pix
//       goodpix = 0;
//       for (int j=1; j<=bins; j++)
// 	goodpix += f[j];
      

      // linear ceiling
      float ceiling = goodpix*deltab/(Bdmax-Bdmin);

      // human ceiling
      if (human) {
	
	// compute Bde
	float Bde = Bdmin + (Bdmax-Bdmin)*P[i];
	float Ld = exp(Bde);

	float Bw = i/(float)bins*(Bwmax-Bwmin)+Bwmin;
	float Lw = exp(Bw);
	ceiling *= jnd(Ld)/jnd(Lw)*Lw/Ld;
	//ceiling += 1.;
      }
      //cerr << "ceiling: " << ceiling << " " << f[i] << endl;
      
      if (f[i] > ceiling){
	trimmings += f[i]-(int)ceiling;
	f[i] = (int)ceiling;
      }
    }

    //cerr << "ITERATING." << endl;
    
  } while (trimmings > tolerance);

  //cerr << "ITERATION DONE." << endl;


  // count pix
  goodpix = 0;
  for (int i=1; i<=bins; i++)
    goodpix += f[i];

  // create c.d.f.
  sum = 0;
  for (int i=1; i<=bins; i++){
    sum += f[i];
    P[i] = sum/(float)goodpix;
  }

#if 0
  // dump c.d.f.
  fn = "cdf.dump.1";
  file = fopen(fn, "wb");
  if (!file){
    cerr << "Unable to open file " << fn << ": " << strerror(errno) << endl;
    return;
  }
  for (int i=0; i<=bins; i++)
    fwrite(&P[i], sizeof(float), 1, file);
  fclose(file);
#endif
  
  // perform tone mapping
  for (int i=0; i<gv->imgTotalDim[0]*gv->imgTotalDim[1]; i++){

    if (pix[i][Y] <= 0.0)
      continue;
    
    float Bw = log(pix[i][Y]);

    // linearly interpolate P(Bw)
    float bf = ( Bw - Bwmin ) / (Bwmax-Bwmin) * bins;
    bf = clamp(bf, 0.0f, (float)bins);
    int b0 = (int)bf;
    int b1 = MIN2(b0+1, bins);
    float s = bf-b0;
    float PBw = (1-s)*P[b0] + s*P[b1];

    // compute Bde
    float Bde = Bdmin + (Bdmax-Bdmin)*PBw;
    pix[i][Y] = exp(Bde);
    pix[i][Y] /= Ldmax;  // scale so display luminance is 1.0
  }

  float lmin = 1e30;
  float lmax = -1e30;
  for (int i=0; i<gv->imgTotalDim[0]*gv->imgTotalDim[1]; i++){
    if (pix[i][Y] > 0.0){
      if (pix[i][Y] > lmax)
	lmax = pix[i][Y];
      if (pix[i][Y] < lmin)
	lmin = pix[i][Y];
    }
  }
  cerr << "Hist: lmin=" << lmin << " lmax=" << lmax << endl;
  

}


inline void setSuperPix(SbColor &c, int x, int y, GlobalVars *gv, SbColor *buf){
  for (int j=0; j<gv->subPixDim; j++){
    for (int i=0; i<gv->subPixDim; i++){
      int idx =
        (y*gv->subPixDim+j)*gv->imgTotalDim[0] +
        (x*gv->subPixDim+i);
      buf[idx] = c;
    }
  }
}

inline void addSuperPix(SbColor &c, int x, int y, GlobalVars *gv, SbColor *buf){
  for (int j=0; j<gv->subPixDim; j++){
    for (int i=0; i<gv->subPixDim; i++){
      int idx =
        (y*gv->subPixDim+j)*gv->imgTotalDim[0] +
        (x*gv->subPixDim+i);
      buf[idx] += c;
    }
  }
}

inline SbColor getSuperPix(int x, int y, GlobalVars *gv, SbColor *buf){
  SbColor c(0,0,0);
  for (int j=0; j<gv->subPixDim; j++){
    for (int i=0; i<gv->subPixDim; i++){
      int idx =
        (y*gv->subPixDim+j)*gv->imgTotalDim[0] +
        (x*gv->subPixDim+i);
      c += buf[idx];
    }
  }
  c /= gv->subPixDim*gv->subPixDim;
  return c;
}

// glare filter that accounts for pupil diffraction
// uses two techniques:
//
// Physically-based glare effects for digital images
// by Spencer, Shirley, Zimmerman, Greenberg
// and
// A lighting model aiming at drive simulators
// by Nakamae, Kaneda, Okamoto, Nishita
//
// Input: pix[] array filled with Yxy luminance values,
// where pix[Y] is luminance and pix[x] pix[z] are x and y
// Yi is the adaptation luminance, and
// Yw is the maximum luminance of the display.
// toneMap() calls with Yw=1.0 because since it scales the monitor max luminance to 1
void Pane::glareFilter(float Yi, float Yw, SbColor *&pixOut){

  cerr << "Performing simple glare filter...\n";
  
  // shirley
  int fdim[2];
  if (gv->useGlare == GlobalVars::GLARE_SHIRLEY){

    // Compute desired filter type and read filter.
    int result = 0;
    string type;
    
    if (log10(Yi) < -2.0)
      type = "scotopic";
    else if (log10(Yi) < 0.6)
      type = "mesopic";
    else
      type = "photopic";

    string fn = type + ".hdr";

    float degrees = 45.0;
    if (gv->cam->isOfType(SoPerspectiveCamera::getClassTypeId())){
      degrees = ((SoPerspectiveCamera *)gv->cam)->heightAngle.getValue();
      degrees *= 180.0/M_PI;
    }
    
    cerr << "Glare should have been created using:\n"
	 << "makeFilter  -d " << degrees << " -i " << gv->imgTotalDim[Y]
	 << " " << type << "\n";
    
    result = readGlareFilter(fn, fdim);
    if (result != 1){

      cerr << "Error: Could not open filter image: " << fn << "\n"
	   << "Not applying any glare!\n"
	   << "Use above command to generate appropriate glare filter.\n";
      
      // Yxy -> RGB
      convertYxyToRgb(pix, gv->imgTotalDim[0], gv->imgTotalDim[1]);

      return;
     }
  }

  // Yxy -> RGB
  convertYxyToRgb(pix, gv->imgTotalDim[0], gv->imgTotalDim[1]);

  // Efficient convolution.
  static bool doFFT = getenv("PANE_GLARE_FFT");
  if (doFFT){

    cerr << "Convolving glare filter using FFT.\n";

    // create final output buffer for convolveFFT downsampled result
    pixOut = new SbColor[gv->imgDim[X] * gv->imgDim[Y]];
    
    convolveFFT(pix, gv->imgTotalDim[0], gv->imgTotalDim[1],
                filtpix, fdim[0], fdim[1],
                pixOut, gv->subPixDim);

//     for (int j=0; j<gv->imgDim[1]; j++)
//       for (int i=0; i<gv->imgDim[0]; i++)
//         addSuperPix(pixOut[gv->imgDim[0]*j+i], i, j, gv, pix);
//     delete [] pixOut;
//     pixOut = NULL;

    gammaCorrect(pixOut, gv->imgDim[X], gv->imgDim[1]);
    
    return;
  }

  // create glare buffer
  int numPix = gv->useGlare == GlobalVars::GLARE_SHIRLEY ? 
    gv->imgDim[X] * gv->imgDim[Y] :
    gv->imgTotalDim[X] * gv->imgTotalDim[Y];
  SbColor *bPix = new SbColor[numPix];
  for (int i=0; i<numPix; i++)
    bPix[i] = SbColor(0.0, 0.0, 0.0);
  
  
  // loop over high intensity pixels in lowres framebuffer
  for (int y=0; y<gv->imgDim[Y]; y++){

    // status meter
    fprintf(stderr, "\r%6.2f%%  ", y/(float)(gv->imgDim[Y]-1)*100.0);
    
    for (int x=0; x<gv->imgDim[X]; x++){

      SbColor c = getSuperPix(x, y, gv, pix);

      // only process glare for bright pixels
      if (!(c[R]>1 || c[G]>1 || c[B]>1))
	continue;
       
//       bPix[m] = SbColor(0.0, -1.0, -1.0);
//       continue;

      //cerr << "Convolving filter with remaining pixel value: " << pixm << "\n";

      // mark high pixels
      if (gv->tempInt == 33){
        SbColor g = SbColor(1,0,0)-c;
        addSuperPix(g, x, y, gv, bPix);
        continue;
      }

      // find max channel
      // This could be used to reduce the added color to only the portion of
      // the color > 1, but not sure at the moment how to do that.
//       int chan=R;
//       if (c[G]>c[chan]) chan = G;
//       if (c[B]>c[chan]) chan = B;
//       float amount = c[chan]-1;
      
      // shirley
      if (gv->useGlare == GlobalVars::GLARE_SHIRLEY){

	// convolve
	const int di = fdim[0]/2;
	for (int j=MAX2(0, y-di); j<MIN2(gv->imgDim[Y], y+di+1); j++){
	  for (int i=MAX2(0, x-di); i<MIN2(gv->imgDim[X], x+di+1); i++){

	    // skip center pixel
	    if (j==y && i==x)
	      continue;
	    
	    int l = i-x+fdim[0]/2;
	    int m = j-y+fdim[1]/2;
	    int f = fdim[0]*m+l;

            SbColor g = vecScale(filtpix[f], c);
            bPix[j*gv->imgDim[0]+i] += g;
	  }
	}
      }

      // nakamae
      else {

	const float a = 2.5  * 1.0/1000.0; // slit diameter (pupil, meters)
	const float d = 30   * 1.0/100.0;  // focusing distance (of eye, meters)
	const float h = 30   * 1.0/100.0;  // image plane size (height, meters)

	SbColor k = SbColor(
	  2.0*M_PI/(611e-9), // red
	  2.0*M_PI/(549e-9), // green
	  2.0*M_PI/(467e-9)  // blue
	  );
	
        const int de = MAX2(gv->imgTotalDim[X], gv->imgTotalDim[Y]);
	for (int j=MAX2(0, y-de); j<MIN2(gv->imgTotalDim[Y], y+de); j++){

	  int dyi = abs(j-y);
	
	  for (int i=MAX2(0, x-de); i<MIN2(gv->imgTotalDim[X], x+de); i++){
	    int n = gv->imgTotalDim[X]*j+i;
            
            
	    int dxi = abs(i-x);
            float di = sqrt(dyi*dyi + dxi*dxi);
            float theta = atan2(di/gv->imgTotalDim[Y]*h, d);

	    if (theta <= 0.0)
	      continue;

	    SbColor alpha = 0.5*k*a*sin(theta);
            
            const float avg = 2/M_PI;
	    SbColor f0 = SbColor(
// 	      sin(alpha[R])/alpha[R],    // Can you say aliasing?
// 	      sin(alpha[G])/alpha[G],    // Replaced sin with avg value
// 	      sin(alpha[B])/alpha[B]     // in next lines
	      avg/alpha[R],
	      avg/alpha[G],
	      avg/alpha[B]
	      );
	      
	    f0[R] *= f0[R];
	    f0[G] *= f0[G];
	    f0[B] *= f0[B];

	    bPix[n][R] += c[R]*f0[R];
	    bPix[n][G] += c[G]*f0[G];
	    bPix[n][B] += c[B]*f0[B];
	
	  }
	}
      }
    }
  }

  // combine buffers
  if (gv->useGlare == GlobalVars::GLARE_SHIRLEY){

    for (int j=0; j<gv->imgDim[Y]; j++)
      for (int i=0; i<gv->imgDim[X]; i++)
        addSuperPix(bPix[j*gv->imgDim[X]+i], i, j, gv, pix);
    
    
  } else {

    for (int i=0; i<gv->imgTotalDim[X]*gv->imgTotalDim[Y]; i++)
      pix[i] += bPix[i];

  }
  delete [] bPix;

  // backup hi-res glare image hdr
  // create comment (will not include accurate tone mapping info but better than nothing)
  string comment = createComment();
  FileUtils::writeHDR(gv->imgTotalDim, "glareTmp.hdr",
                      pix, gv->createBackup, comment.c_str());
}

static rfftwnd_plan plan_rc, plan_cr;

void init_FFT(int x, int y){
  plan_rc = rfftw2d_create_plan(x, y, FFTW_REAL_TO_COMPLEX, FFTW_IN_PLACE);
  plan_cr = rfftw2d_create_plan(x, y, FFTW_COMPLEX_TO_REAL, FFTW_IN_PLACE);
}

#define FFT(s,u)\
if (s==1) rfftwnd_one_real_to_complex(plan_rc, (fftw_real *)u, (fftw_complex *)u);\
else rfftwnd_one_complex_to_real(plan_cr, (fftw_complex *)u, (fftw_real *)u)

void Pane::convolveFFT(SbColor *in, int inX, int inY,
                       SbColor *filter, int filtX, int filtY,
                       SbColor *out, int subPixDim) {

  int resResize[2] = { inX/subPixDim, inY/subPixDim };
  SbColor *inResize = NULL;
  downSampleImg(in, inX, inY, inResize, resResize[0], resResize[1], subPixDim);

  // make twice as big as (max) to prevent wrap-around
  int maxDim[2];
  maxDim[0] = max(resResize[0]*2, filtX*2); 
  maxDim[1] = max(resResize[1]*2, filtY*2);
  int extra = maxDim[0]%2==0 ? 2 : 1;

  int inOffset[2], filtOffset[2];
  inOffset[0] = (maxDim[0]-resResize[0])/2;
  inOffset[1] = (maxDim[1]-resResize[1])/2;
  filtOffset[0] = (maxDim[0]-filtX)/2+1;
  filtOffset[1] = (maxDim[1]-filtY)/2+1;

  // form fft-ready array for input image
  init_FFT(maxDim[0], maxDim[1]);

//   srand48(3);
//   for (int i=0; i<resResize[0]*resResize[1]; i++){
//     inResize[i]=SbColor(drand48(), drand48(), drand48());
//   }
  
  // loop over channels
  for (int c=0; c<3; c++){

    cerr << "Channel " << c << ". ";

    // prepare in image for FFT
    int size =(maxDim[0]+extra)*maxDim[1];
    float *img = new float[size];
    for (int i=0; i<size; i++)
      img[i] = 0;
    for (int j=0; j<resResize[1]; j++)
      for (int i=0; i<resResize[0]; i++){

        // copy image data where at least one channel > 1
        int idx1 = (maxDim[0]+extra)*(j+inOffset[1]) + (i+inOffset[0]);
        int idx2 = resResize[0]*j + i;
//         float v = MAX3(inResize[idx2][0],
//                        inResize[idx2][1],
//                        inResize[idx2][2]);
        float v = inResize[idx2][c];
        /*
        if (v>1){
          img[idx1] = v-1;
          //inResize[idx2][c] * (v-1)/v;
          //out[idx2][c] = inResize[idx2][c]* (v-1);
        }
        else
          img[idx1] = 0;
        */
      }

    cerr << "fft(image). ";
//    FFT(1, img);
     // prepare filter image for FFT
    float *imgFilt = new float[size];
    for (int i=0; i<size; i++)
      imgFilt[i] = 0;

    int filtCenterX = filtX/2;
    int filtCenterY = filtY/2;
    
    for (int j=0; j<filtY; j++)
      for (int i=0; i<filtX; i++){

        // skip center pixel
//         if (abs(j-filtCenterY)<=0 && abs(i-filtCenterX)<=0)
//          continue;
/*        
        imgFilt[(maxDim[0]+extra)*(j+filtOffset[1]) + (i+filtOffset[0])] =
          filter[filtX*j + i][c];
*/      
      }

//     cerr << "Filter (center):\n"
//          << "0=" << filter[filtCenterY*filtY+filtCenterX+0] << "\n"
//          << "1=" << filter[filtCenterY*filtY+filtCenterX+1] << "\n"
//          << "2=" << filter[filtCenterY*filtY+filtCenterX+2] << "\n"
//          << "3=" << filter[filtCenterY*filtY+filtCenterX+3] << "\n";
//     cerr << "Filter (center+1):\n"
//          << "0=" << filter[(filtCenterY+1)*(filtY)+filtCenterX+0] << "\n"
//          << "1=" << filter[(filtCenterY+1)*(filtY)+filtCenterX+1] << "\n"
//          << "2=" << filter[(filtCenterY+1)*(filtY)+filtCenterX+2] << "\n"
//          << "3=" << filter[(filtCenterY+1)*(filtY)+filtCenterX+3] << "\n";

/*    
    cerr << "fft(filter). ";
    FFT(1, imgFilt);

    // multiply
    for (int j=0; j<maxDim[1]; j++)
      for (int i=0; i<maxDim[0]+extra-1; i+=2){

        int idx = (maxDim[0]+extra)*j + i;
        float a=img[idx], b=img[idx+1];
        float c=imgFilt[idx], d=imgFilt[idx+1];
        img[idx]   = a*c - b*d;
        img[idx+1] = b*c + a*d;
      }
*/
    // convert back to image
    cerr << "fft^-1(image). ";
//    FFT(-1, img);

    float f = 1.0/(maxDim[0]*maxDim[1]);

    for (int j=0; j<resResize[1]; j++)
      for (int i=0; i<resResize[0]; i++) {

        // shift image
        int x, y;

        int oddx = maxDim[0]%2;
        int oddy = maxDim[0]%2;
        
        if (i<resResize[0]/2+oddx/*+1*/) {
          x = maxDim[0]-resResize[0]/2+i-oddx/*-1*/;
        } else {
          x = i-resResize[0]/2-oddx/*-1*/;
        }
        
        if (j<resResize[1]/2+oddy/*+1*/) {
          y = maxDim[1]-resResize[1]/2+j-oddy/*-1*/;
        } else {
          y = j-resResize[1]/2-oddy/*-1*/;
        }

//         out[resResize[0]*j + i][c] =
//           inResize[resResize[0]*j + i][c] +
//           f * max(img[(maxDim[0]+extra)*y + x], 0.f);
        
      }

    delete [] img;
    delete [] imgFilt;

    cerr << "\n";
  }

  delete [] inResize;
}

void Pane::gammaCorrect(SbColor *buf, int x, int y) {

  for (int i=0; i<x*y; i++){

    double r, g, b;
    r = buf[i][R];
    g = buf[i][G];
    b = buf[i][B];

    // gamma correct
    r = clamp(r, 0.0, 1.0);
    g = clamp(g, 0.0, 1.0);
    b = clamp(b, 0.0, 1.0);

    if (fabs(gv->gamma-2.2)>0.001){

      // regular gamma correction
      r = pow(r, 1.0/gv->gamma);
      g = pow(g, 1.0/gv->gamma);
      b = pow(b, 1.0/gv->gamma);
    }
    else if (fabs(gv->gamma-1.0)>0.001){

      // sRGB standard gamma correction (similar to regular w/2.2)
      r = r<=0.0031308 ? 12.92*r : 1.055*pow(r, 1.0/2.4) - 0.055;
      g = g<=0.0031308 ? 12.92*g : 1.055*pow(g, 1.0/2.4) - 0.055;
      b = b<=0.0031308 ? 12.92*b : 1.055*pow(b, 1.0/2.4) - 0.055;

    }

    // store transformed colors
    buf[i].setValue(r,g,b);
  }

}

void Pane::convertYxyToRgb(SbColor *buf, int x, int y){
  
  for (int i=0; i<x*y; i++){
  
    SbColor L = buf[i];

    // transform Yxy -> XYZ
    float x=L[X];
    float y=L[Z];
    L[X] = x*L[Y]/y;
    L[Z] = (1.0-x-y)*L[Y]/y;

    // sRGB standard, http://www.w3.org/Graphics/Color/sRGB
    // physical sRGB standard, based on Yw=110.5
    buf[i][R] =  3.2406*L[X] +  -1.5372*L[Y] +  -0.4986*L[Z];
    buf[i][G] = -0.9689*L[X] +   1.8758*L[Y] +   0.0415*L[Z];
    buf[i][B] =  0.0557*L[X] +  -0.2040*L[Y] +   1.0570*L[Z];

    // Technically this matrix is off by a factor of 1/Yw.
    // The given matrix expects a maximum luminance (Y)
    // of 1.0. This is fine since our scale factor
    // scales everything so the maximum luminance is 1.
    //
    // If we weren't using a scalefactor to scale our
    // luminances, and we wanted to convert (X,Y,Z)
    // to the actual (r,g,b) signal that produces
    // that exact stimuli, then you need to use
    // the above matrix and then divide each of
    // the r,g,b by Yw (or divide matrix by Yw).
    //
    // For example:
    //
    // r /= Yw;
    // g /= Yw;
    // b /= Yw;      
  }

}

