#include "Plane.h"
#include "Main.h"
#include "debug.h"
#include "GlobalVars.h"

Material Plane::defaultMat;
vector<Material *> Plane::matsList;


SbColor Plane::getIntDiffColor(const Ray &r) const {

  SbColor diffColor;
  const SbVec3f &bc = r.bc;

  diffColor = r.prim->getMat(r)->diffuse;

  if (texture && texture->bumpMap != 2)
    return getColor(bc, diffColor);

  return diffColor;
}



Material DMPlane::dmDefaultMat;

const int DMPlane::cacheDim[2] = {512,512};
const int DMPlane::cacheLevels = 11;

DMPlane::DMPlane(GlobalVars *gv_, float vertFOV, int vertRes) : Plane() {
  gv = gv_;
  mat = &dmDefaultMat;
  maxHeight = 1;
  doColorMap = false;//true;
  errorAt1 = tanf(vertFOV*M_PI/180. / vertRes);

//   noise.octaves = 10;
//   noise.amplitude = 1;
//   float scale = 3;
//   noise.frequency = SbVec3f(scale,scale,scale);
//   //noise.translation = -offset;
//   //noise.bias = clamp(bias,0,1);
//   //noise.gain = clamp(gain,0,1);
//   noise.ampJump = 1;
//   noise.freqJump = 2;
//   noise.sign=false;

  noise.octaves = 8;
  noise.amplitude = .99/1.92651;
  float scale = 2.638*.173;
  noise.frequency = SbVec3f(scale,scale,scale);
  noise.bias = .4257;
  noise.gain = .8769;
  noise.translation = -SbVec3f(.3742, .135, -.5825);
  noise.freqJump = 2.029;
  noise.ampJump = 1;
  noise.sign=true;
  
  noise.initExponents();


  // initalize height map

  // mount everest = 8848m
  // tree line kilaminjaro 3000m
  heightMap = Spline<SbColor>(0,8848);  
  heightMap.Add(0,      SbColor(0.180, 0.141, 0.0353));          // brown
  heightMap.Add(.6,     SbColor(0.0549, 0.180, 0.0353)*.5);      // green
  heightMap.Add(.7-1e-5, SbColor(0.0549, 0.180, 0.0353)*.7);      // green end
  heightMap.Add(.7,      SbColor(0.180, 0.141, 0.0353));          // brown
  heightMap.Add(.75-1e-5, SbColor(0.180, 0.141, 0.0353)*.8);       // brown end
  heightMap.Add(.75,     SbColor(1,1,1)*.30);                     // snow



  // 256x256 each level
  // 0 [.5,.5]
  // 1 [1,1]
  // 2 [2,2]
  // 3 [4,4],
  // 4 [8,8]
  // 5 16,
  // 6 32,
  // 7 64,
  // 8 128,
  // 9 256,
  // 10 512,
  // 11 1024
  
  // initialize cache
  heightCache.resize(cacheLevels);
  for (int i=0; i<cacheLevels; i++){
    heightCache[i].resize(cacheDim[0]*cacheDim[1], -1.f);
  }

  // find center position
  SbVec3f camPos = gv->cam->position.getValue();
  //cerr << "Plane:: camPos=" << camPos << endl;
  camPos = onb.toONBSpace(camPos);
  camCenter2d.setValue(camPos[0], camPos[1]);
  
}

// This code is slow as hell now
bool DMPlane::intersectRay(Ray &r, double t0, double t1) {
  r.t = -1;
  r.prim = NULL;
  r.bc = SbVec3f(1,0,0);

  //if (r.ignorePrim == this) return false;

  float h = maxHeight;
  SbVec3f p = onb.toONBSpace(r.pos);
  SbVec3f d = onb.toONBSpace(r.dir);

  const double myEps = 0.001; // tanscale(fov_h/res_h)   // error at 1u
  const double stept = myEps;//1./100;//2048.;//1/256.;//0.01;

  double h0, h1, hm;
  double d0, d1, dm;
  Vec3d p0, p1, pm; // in ONB space


  // intersect ray with slab
  float posDotN = p[2]; //p.dot(onb.getW());
  float dirDotN = d[2]; //d.dot(onb.getW());

  // ray parallel and outside slab?
  static const float min = 0.00001;//FLT_MIN*32;
  bool parallel = fabs(dirDotN) < min;
  if (parallel &&
      (posDotN<dist || posDotN>(dist+h)) )
    return false;

  // clamp t0 t1 to volume  
  if (!parallel){
    double m0 = -(posDotN-(dist-min))/dirDotN;
    double m1 = -(posDotN-(dist+h+min))/dirDotN;

    // order m0, m1
    if (m1<m0){
      double t = m1;
      m1 = m0;
      m0 = t;
    }

    t0 = MAX2(t0, m0);
    t1 = MIN2(t1, m1);
  }
  
  // more near and far clipping
  float nearClip = t0;//0.0001;
  // 0.01 causes missed intersections for rays near the surface
  // 0.00001-0.00002 causes too many self intersections
  
  float farClip = 10000;
  t0 = MAX2(t0, nearClip);
  t1 = MIN2(t1, farClip);
  
  if (t1<=t0)
    return false;
  
  d0 = t0;
  
  p0 = p+d*d0;
  h0 = height(p0, d0, true);
  
  bool above = p0[2]>h0;

  const double maxGrad = 2;
  const double dx = sqrt(max(1-d[2]*d[2],0.f));
  const double stepSizePre = 1/((above?maxGrad:-maxGrad)*dx-d[2]);
  

  static bool debuggingEnabled = false;
  if (debuggingEnabled && DEBUG_STATE)
    cerr << "Begin marching\n";

  // iterate
  d1 = d0;
  p1 = p0;
  h1 = h0;
  unsigned int steps = 0;
  while (d1<t1) {
    
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";

    d0 = d1;    // distance
    p0 = p1;    // position (u,v, height)
    h0 = h1;    // heightfield height
    
    double stepsize = (p1[2]-h1)*stepSizePre;
    
    if (stepsize<-myEps)
      return false;  // surface is *behind* us, so miss
    
    d1 += MAX2(MAX2(d1*myEps, myEps), stepsize);      //stept;        
    steps++;
    
    if (d1>t1){
      if (debuggingEnabled && DEBUG_STATE)
        cerr << __LINE__ << ": Reached end\n";
      d1 = t1;
    }

    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";
    
    // get height
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": p1= " << p1 << "\n";
    
    p1 = p+d*d1;
    h1 = height(p1, d1, true);
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": h1=" << h1 << "\n";
    
    // detect crossing
    if (above && p1[2]<=h1 ||
        !above && p1[2]>=h1) {


      if (debuggingEnabled && DEBUG_STATE)
        cerr << "hit... now refine...\n";

      // refine intersection
      double diff;
      do {
        dm = (d0+d1)*0.5;
        diff = d1-dm;
        pm = p+d*dm;
        hm = height(pm, dm, true);

        if (above && pm[2]<hm ||
            !above && pm[2]>hm) {
          d1=dm;
          p1=pm;
        } else {
          d0=dm;
          p0=pm;
        }
          
      } while (diff>0.5e-5); //  0.5e-9
        
      // hit heightfield
      r.t = d0; // choose nearer hit, give next ray a better
                // chance of not self-intersecting
      r.prim = this;
      r.bc = p0;

      if (debuggingEnabled && DEBUG_STATE)
        cerr << "returning t=" << d0 << "\n";
      return true;
    }

  }

  if (debuggingEnabled && DEBUG_STATE)
    cerr << __LINE__ << ": no hit\n";
  return false;
}

SbVec3f DMPlane::getNormal(const SbVec3f &bc) const {

  //return SbVec3f(0,1,0);

  const float du = 0.001;
  const float duInv = 1/du;
  const float dv = du;
  const float dvInv = duInv;

  float h = maxHeight;

  SbVec3f p0 = bc; 
  SbVec3f pu = bc; pu[0]+=du; // + onb.getU()*du;
  SbVec3f pv = bc; pv[1]+=dv; //+ onb.getV()*dv;

  bool useCache=true;
  float h0 = height(Vec3d(p0[0], p0[1], 1.), 1, useCache);
  float hu = height(Vec3d(pu[0], pu[1], 1.), 1, useCache);
  float hv = height(Vec3d(pv[0], pv[1], 1.), 1, useCache);

  SbVec3f norm(-(hu-h0)*duInv, -(hv-h0)*dvInv, 1);
  norm = onb.toWorldSpace(norm); // if we were using a full 4x4 transformation
                                 // matrix (SbMatrix), we would do:
                            // m.inverse().transpose().multDirMatrix(in, out)
  norm.normalize();
  return norm;
}

SbVec3f DMPlane::getIntNormal(const SbVec3f &bc) const {

  //return SbVec3f(0,1,0);

  const float du = 0.001;  // 0.00001
  const float duInv = 1/du;
  const float dv = du;
  const float dvInv = duInv;

  float h = maxHeight;

  SbVec3f p0 = bc; 
  SbVec3f pu = bc; pu[0]+=du; // + onb.getU()*du;
  SbVec3f pv = bc; pv[1]+=dv; //+ onb.getV()*dv;

  bool useCache=false;
  float h0 = height(Vec3d(p0[0], p0[1], 1.), 1, useCache);
  float hu = height(Vec3d(pu[0], pu[1], 1.), 1, useCache);
  float hv = height(Vec3d(pv[0], pv[1], 1.), 1, useCache);

  SbVec3f norm(-(hu-h0)*duInv, -(hv-h0)*dvInv, 1);
  norm = onb.toWorldSpace(norm); // if we were using a full 4x4 transformation
                                 // matrix (SbMatrix), we would do:
                            // m.inverse().transpose().multDirMatrix(in, out)
  norm.normalize();
  return norm;
}

SbColor DMPlane::getIntDiffColor(const Ray &r) const {
//  return SbColor(.5,.5,.5);
  const SbVec3f &bc = r.bc;

  if (doColorMap){
    SbVec3f norm = this->getIntNormal(bc);
    return heightMap.eval(bc[2]-.2*norm[1]-.1);
  } else {
    return getMat()->diffuse;
  }
  
//   float v = fmodf(bc[2],10000.)/10000.;
//   SbColor c(v, 0, 0);
//   //cerr << "bc=" << bc << " c=" << c << "\n";
//   return c;
}


// number of leading zeroes (from Hacker's Delight)
inline int nlz(unsigned x){
  unsigned y;
  int n;

  n=32;
  y = x>>16; if (y!=0) { n = n-16; x=y; }
  y = x>> 8; if (y!=0) { n = n- 8; x=y; }
  y = x>> 4; if (y!=0) { n = n- 4; x=y; }
  y = x>> 2; if (y!=0) { n = n- 2; x=y; }
  y = x>> 1; if (y!=0) return n-2;
  return n-x;
}


namespace {
  
inline
Vec3d idx2pos(int i, int j, const int dim[2], SbVec2f &min, SbVec2f &max){
  return Vec3d(i/float(dim[0]-1)*(max[0]-min[0])+min[0],
               j/float(dim[1]-1)*(max[1]-min[1])+min[1],
               0);
}
  
}

float DMPlane::bilinearLookup(vector<float> &v, const int dim[2],
                              SbVec2f min, SbVec2f max, SbVec2f pos) const {

  int i, j, i1, j1;
  float s, t,sp, tp;

  s = fabs((pos[0]-min[0])/(max[0]-min[0]));
  t = fabs((pos[1]-min[1])/(max[1]-min[1]));

  // wrap
  if (s>1){
    s = fmodf(s, 2);
    if (s>1)
      s = 2-s;
  }
  if (t>1){
    t = fmodf(t, 2);
    if (t>1)
      t = 2-t;
  }
  
  i  = (int)floor(s*(dim[0]-1));
  j  = (int)floor(t*(dim[1]-1));
  i1 = i+1;
  j1 = j+1;

  i  = clamp(i,  0, dim[0]-1);
  j  = clamp(j,  0, dim[1]-1);
  i1 = clamp(i1, 0, dim[0]-1);
  j1 = clamp(j1, 0, dim[1]-1);

  // make sure these are defined
  float v00, v01, v10, v11;
  int idx;
  
  v00 = v[idx=dim[0]*j +i ];
  if (v00<0) v00 = v[idx] = height(idx2pos(i , j , dim, min, max));
  
  v01 = v[idx=dim[0]*j1+i ];
  if (v01<0) v01 = v[idx] = height(idx2pos(i , j1, dim, min, max));
  
  v10 = v[idx=dim[0]*j +i1];
  if (v10<0) v10 = v[idx] = height(idx2pos(i1, j , dim, min, max));
  
  v11 = v[idx=dim[0]*j1+i1];
  if (v11<0) v11 = v[idx] = height(idx2pos(i1, j1, dim, min, max));
  
  sp = s*(dim[0]-1)-i;
  tp = t*(dim[1]-1)-j;
  
  // perform hermite smoothing
//   sp = 3.0*sp*sp-2.0*sp*sp*sp;
//   tp = 3.0*tp*tp-2.0*tp*tp*tp;

  float r;
  r = (1.0-sp)*(1.0-tp)*v00 +
            sp*(1.0-tp)*v10 +
      (1.0-sp)*tp      *v01 +
            sp*tp      *v11;

  return r;
}

double DMPlane::heightCached(const Vec3d &x) const {
  
  // determine cache level
  float maxVal = max(fabs(x[0]-camCenter2d[0]), fabs(x[1]-camCenter2d[1]));

  // 256x256 each level
  // 
  // 0 [.5,.5]
  // 1 [1,1]
  // 2 [2,2]
  // 3 [4,4],
  // 4 [8,8]
  // 5 16,
  // 6 32,
  // 7 64,
  // 8 128,
  // 9 256,
  // 10 512,
  // 11 1024

//   unsigned int i = (unsigned int)(fabs(maxVal)*2.f);
//   int cacheLevel = 32-nlz(i);
  
  float log2 = logf(maxVal)*(1.f/logf(2.f));
  int n = (int)ceil(log2);
//   int n2 = fastlog2(maxVal)+1;
//   if (n!=n2)
//     cerr << "n=" << n << " n2=" << n2 << endl;
  int cl0 = clamp(n+1, 0, cacheLevels-1);
  int cl1 = clamp(n+2, 0, cacheLevels-1);
  float cs0 = powf(2,cl0-1);
  float cs1 = powf(2,cl1-1);
  float cs0half = cs0*0.5f;
  float alpha = 1.f-(maxVal-cs0half)/cs0half;

//   cerr << "Querying ( cached) (" << x[0] << ", " << x[1] << ") n=" << n << "\n";
//   cerr << "cl0=" << cl0 << " cs0=" << cs0 << endl;
//   cerr << "cl1=" << cl1 << " cs1=" << cs1 << endl;
  float h0 = bilinearLookup(heightCache[cl0], cacheDim,
                            SbVec2f(-cs0, -cs0)+camCenter2d, SbVec2f(cs0, cs0)+camCenter2d,
                            SbVec2f(x[0], x[1]));

  //return h0;
  if (cl0==cl1) return h0;

  float h1 = bilinearLookup(heightCache[cl1], cacheDim,
                            SbVec2f(-cs1, -cs1)+camCenter2d, SbVec2f(cs1, cs1)+camCenter2d,
                            SbVec2f(x[0], x[1]));

  float h = h0*alpha + (1-alpha)*h1;
                            
  return h;
}
