#ifndef PLANE_H
#define PLANE_H

#include <iostream>
#include <iomanip>
#include "Primitive.h"
#include "Material.h"
#include "Misc.h"
#include "Noise.h"
#include "debug.h"
#include "Main.h"
#include "Texture.h"

using namespace std;

class Plane : public Primitive {

private:
  static vector<Material *> matsList;  // global material list
  
protected:

  static Material defaultMat;
  
  SbVec3f norm;
  float dist;
  Material *mat;
  Texture *texture;
  ONB onb;
  
public:

  Plane() {
    norm = SbVec3f(0,1,0);
    dist = 0;
    mat = &defaultMat;
    texture = NULL;
    //onb.constructFromW(norm);
    onb.constructFromUW(SbVec3f(0,0,1), norm);
  }

  void setDist(float d){ dist = d; }
  
  virtual PrimType getType() const { return PRIMTYPE_PLANE; }
  virtual string getTypeName() const { return "Plane"; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8) {
    static const float min = 0.00001;//FLT_MIN*32;
    
    r.t = -1;
    r.prim = NULL;
    r.bc = SbVec3f(1,0,0);

    if (r.ignorePrim == this) return false;
    
    float dirDotN = r.dir.dot(norm);
    if (fabs(dirDotN) < min)
      return false;

    float posDotN = r.pos.dot(norm);
    double t = (dist-posDotN)/dirDotN;

    if (t<t0 || t>t1)
      return false;

    r.t = t;
    r.prim = this;
    r.bc = r.point();
    return true;
  }

  SbVec3f getMin() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(-1000, 0, -1000);
    
    float v = FLT_MAX;
    return SbVec3f(v,v,v);
  }
  SbVec3f getMax() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(1000, 0, 1000);
    
    float v = -FLT_MAX;
    return SbVec3f(v,v,v);
  }

  void makeNewMat() {
    mat = new Material();
    matsList.push_back(mat);
  }
  
  Material* getMat() { return mat; }
  const Material* getMat() const { return mat; }
  SbVec3f getNormal(const SbVec3f &bc)  const { return norm; }


  // temporary
  void setTexture(Texture *t){ texture = t; }
  Texture* getTexture() { return texture; }
  void getSTfromBC(const SbVec3f &bc, float &s, float &t) const {
    s = bc[0];
    t = bc[2];

    // transform by texture matrix..
    const Texture *text = texture;
    SbMatrix m;
    m.makeIdentity();
    m[0][0] = text->stex[0];
    m[1][0] = text->stex[1];
    m[3][0] = text->stex[2];
    m[0][1] = text->ttex[0];
    m[1][1] = text->ttex[1];
    m[3][1] = text->ttex[2];
    SbVec3f coord;
    m.multVecMatrix(SbVec3f(s,t,0), coord);
    s = coord[0];
    t = coord[1];
  }
  SbColor getColor(const SbVec3f &bc) const {
    float s, t;
    getSTfromBC(bc, s, t);
    return texture->getColor(s, t);
  }
  SbColor getColor(const SbVec3f &bc, const SbColor &diffuse) const {
    float s, t;
    getSTfromBC(bc, s, t);
    return texture->getColor(s, t, diffuse);
  }
  SbColor getIntDiffColor(const Ray &r) const;

  static Material* getDefaultMat() { return &defaultMat; }

};

class GlobalVars;
class DMPlane : public Plane {

private:
  float maxHeight;
  float errorAt1;  // allowable error at d=1 (needed for octaves)
                   // = tan(fov/res)
  GlobalVars *gv;
  bool doColorMap;
  SbVec2f camCenter2d;    // this where the highest detail is
  
  Spline<SbColor> heightMap;

  mutable Noise noise;
  mutable vector< vector< float > > heightCache;
  static const int cacheDim[2];
  static const int cacheLevels;
  
  static Material dmDefaultMat;

public:
  DMPlane(GlobalVars *gv, float vertFOV=45, int vertRes=768);

  static Material* getDefaultMat() { return &dmDefaultMat; }
  virtual PrimType getType() const { return PRIMTYPE_DMPLANE; }
  virtual string getTypeName() const { return "DMPlane"; }
  
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);

  SbVec3f getMin() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(-1000, 0, -1000);
    
    float v = FLT_MAX;
    return SbVec3f(v,v,v);
  }
  SbVec3f getMax() const {
    // temporary until I fix pane->rayAccel to be Primitive*
    return SbVec3f(1000, maxHeight, 1000);
    
    float v = -FLT_MAX;
    return SbVec3f(v,v,v);
  }

  void setDoColorMap(bool b){ doColorMap = b; }
  
  SbVec3f getIntNormal(const SbVec3f &bc) const;
  SbVec3f getNormal(const SbVec3f &bc) const;

  // testing purposes
  SbColor getIntDiffColor(const Ray &r) const;

  Noise& getNoise() { return noise; }
  
  double height(const Vec3d &x) const {
    //return 0;
    //return fabs(sin((x[0]*x[0]+x[1]*x[1])*10))*.1;
    //cerr << "Querying (uncached) (" << x[0] << ", " << x[1] << ")\n";
    
    return fabs(noise.noise(x[0], x[1], 1.)*maxHeight);
  }
  float height(const Vec3d &x, float d, bool useCache=false) const {

    static bool enableCache = getenv("PANE_PLANE_CACHE");
    if (enableCache && useCache)
       return heightCached(x);
    
    return height(x); // don't use adaptive octaves for now

    // clamp d
    if (d<1) d = 1;

//    static int minOct = 30;
    // from QAEB paper by K. Musgrave, bottom of page 2
//    cerr << __LINE__ << " wtf errorAt1*d=" << errorAt1*d << "\n";
//    int oct = (int)( 0.5 + -3.5-fastlog2(float(errorAt1*d)) );
//     int oct = (int)( 0.5 + -3.5 - log(errorAt1*d)*(1/log(2)) );
//    cerr << __LINE__ << " wtf oct= " << oct << "\n";

//     if (oct<1)
//       oct = 1;
    
//     if (oct>10){
//       cerr << "oct=" << noise.octaves << " d=" << d << " errorAt1=" << errorAt1 <<"\n";
//       exit(1);
//     }

//     if (oct<minOct){
//       minOct = oct;
//       cerr << "minOct = " << minOct << "\n";
//     }

//    noise.octaves = oct;
    return height(x);
  }

  double heightCached(const Vec3d &x) const;

  float bilinearLookup(vector<float> &v, const int dim[2],
                       SbVec2f min, SbVec2f max, SbVec2f pos) const;
  
};


#endif
