#ifndef PRIMITIVE_H
#define PRIMITIVE_H

#include<float.h>
#include<vector>
#include "Ray.h"
#include "Misc.h"
#include "Material.h"

using namespace std;

class SbVec3f;

typedef struct PrimDerivs {
  SbVec3f dPdU;
  SbVec3f dPdV;
  SbVec3f dNdU;
  SbVec3f dNdV;
  PrimDerivs() {}
  PrimDerivs(const SbVec3f &a, const SbVec3f &b,
             const SbVec3f &c, const SbVec3f &d) :
    dPdU(a), dPdV(b), dNdU(c), dNdV(d) {}
} PrimDerivs;

class Primitive {

public:

  static const SbColor Black;
  static const SbColor White;
  static const SbVec3f XDir, YDir, ZDir;
  static const PrimDerivs DefaultDerivs;
  static Material defaultMat;

  typedef unsigned short PrimType;
  enum {
    PRIMTYPE_NONE          = 0x00000000,  // 0
    PRIMTYPE_TRIANGLE      = 0x00000001,  // 1
    PRIMTYPE_EMISTRIANGLE  = 0x00000002,  // 2 (1|2 = 3)
    PRIMTYPE_TRIANGLELIST  = 0x00000004,  // 4
    PRIMTYPE_PRIMITIVELIST = 0x00000008,  // 8
    PRIMTYPE_INSTANCE      = 0x00000010,  // 16
    PRIMTYPE_OCTREE        = 0x00000020,  // 32
    PRIMTYPE_SPHERE        = 0x00000040,  // 64
    PRIMTYPE_JULIA         = 0x00000080,  // 128
    PRIMTYPE_PLANE         = 0x00000100,  // 256
    PRIMTYPE_ISOSURFACE    = 0x00000200,  // 512
    PRIMTYPE_WORLEY        = 0x00000400,  // 1024
    PRIMTYPE_FRACTAL       = 0x00000800,  // 2048
    PRIMTYPE_BIH           = 0x00001000,  // 4096
    PRIMTYPE_TEST          = 0x00002000,  // 8192     DualAccel debugging object
    PRIMTYPE_ACCELLIST     = 0x00004000,  // 
    PRIMTYPE_DMPLANE       = 0x00008000,  // 
  }; // possible values for PrimType

  virtual ~Primitive(){ /*cerr << "~Primitive: "  << this << endl;*/ }
  virtual PrimType getType() const = 0;
  virtual bool isOfType(PrimType p) const { return getType() & p; }
  virtual string getTypeName() const { return "unknown"; }

  virtual bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8) = 0;
  virtual bool intersectRaySimple(Ray &r, double t0=1e-5, double t1=1e8) { // for shadow rays
    return intersectRay(r, t0, t1);
  }
  // single sided intersection
  bool intersectRaySS(Ray &r,  bool ss=true,
                      double t0=1e-5, double t1=1e8){

    // this sucks and should be modified. doesn't even look like it works..
    bool retVal = intersectRay(r, t0, t1);

    if (retVal && r.depth == 1 && !ss){
      while (r.prim->getNormal(r).dot(r.dir)>0.0){
        // (gv->tempInt != 1024 || r.prim->getEmission()[0]>0.0)){
        Ray new_r(r.pos+r.t*r.dir, r.dir, r, r.prim);
        new_r.depth = r.depth;
        if (!intersectRay(new_r, 0, 1-r.t))
          return false;
        r.t += new_r.t;
        r.prim = new_r.prim;
        r.prim2 = new_r.prim2;
        r.bc = new_r.bc;
      }
    }

    return retVal;
  }

  virtual bool intersectCube(const double min[3], const double max[3]) const;
  virtual bool intersectCubeCenterHalf(const double center[3],
                                       const double half[3]) const {
    double min[3], max[3];
    min[0] = center[0] - half[0];
    min[1] = center[1] - half[1];
    min[2] = center[2] - half[2];
    max[0] = center[0] + half[0];
    max[1] = center[1] + half[1];
    max[2] = center[2] + half[2];
    return intersectCube(min, max);
  }
  
  virtual SbVec3f  getMin()      const = 0;
  virtual SbVec3f  getMinShiny() const {
    return getMat()->isShiny() ?
      getMin() : SbVec3f(FLT_MAX, FLT_MAX, FLT_MAX);
  }
  virtual SbVec3f  getMax()      const = 0;
  virtual SbVec3f  getMaxShiny() const {
    return getMat()->isShiny() ?
      getMax() : SbVec3f(-FLT_MAX, -FLT_MAX, -FLT_MAX);
  }
  virtual float    getMin(int dim) const { return getMin()[dim]; }
  virtual float    getMax(int dim) const { return getMax()[dim]; }

  virtual SbVec3f getCenter() const {      return (getMin()     +getMax()     )*0.5f; }
  virtual SbVec3f getCenterShiny() const { return (getMinShiny()+getMaxShiny())*0.5f; }
  virtual float getRadius() const {        return (getMax()     -getMin()     ).length(); }
  virtual float getRadiusShiny() const {   return (getMaxShiny()-getMinShiny()).length(); }

  virtual Material* getMat() { return &defaultMat; }
  virtual const Material* getMat() const { return &defaultMat; }
  virtual Material* getMat(const Ray &r) { return getMat(); }
  virtual const Material* getMat(const Ray &r) const { return getMat(); }
  
  virtual SbVec3f getNormal(      const SbVec3f &bc) const{ return YDir; }
  virtual SbVec3f getIntNormal(   const SbVec3f &bc) const{ return getNormal(bc); }
  virtual SbVec3f getNormal(      const Ray      &r) const{ return getNormal(r.bc); }
  virtual SbVec3f getIntNormal(   const Ray      &r) const{ return getIntNormal(r.bc); }
  
  virtual SbColor getIntDiffColor(const Ray &r) const{ return getMat(r)->diffuse;}
  virtual SbColor getIntEmission( const Ray &r) const{ return getMat(r)->emission;}
  virtual bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer)const{return 0;}

};

class PrimRef {
private:
  /*
  uint32_t i; // 4 bits for type   (16 possible)
              // 28 bits for index (268,435,456 possible)

  static typeMask = 0xF0000000;
  static idxMask =  0x0FFFFFFF;

  static vector
  
  inline Primitive& operator*() {
    int pType = (i & typeMask) >> 28;
    int idx = (i & idxMask);
  }
  */

  Primitive *p;

public:
  PrimRef(Primitive *base);

};

class PrimList : public Primitive {

private:
  vector<Primitive *> prims;

public:
  PrimList(){ }
  virtual ~PrimList(){ }

  int size() const { return prims.size(); }
  const Primitive* prim(int i) const { return &*prims[i]; }
  Primitive* prim(int i) { return &*prims[i]; }
  virtual void add(Primitive *p) { prims.push_back(p); }
  virtual void reserve(int n) { prims.reserve(n); }
  virtual void clear() { prims.clear(); }
  void swap(int i, int j){
    Primitive *tmp = prims[i];
    prims[i] = prims[j];
    prims[j] = tmp;
  }
  void resizeToFit(){ vector<Primitive *>(prims).swap(prims); }
  Primitive*& operator[](int i){ return prims[i]; }
  const Primitive* operator[](int i) const { return prims[i]; }

  virtual PrimType getType() const { return PRIMTYPE_PRIMITIVELIST; }
  virtual string getTypeName() const { return "PrimitiveList"; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8){
    double minT = -1;
    Primitive *minPrim = NULL;
    SbVec3f minBc(1,0,0);
    bool hit = false;
    FastPrimStack minPrim2;

    // loop over triangles
    for (int i=0; i<prims.size(); i++){

      // skip triangle
      // do this test in child's
      //if (prims[i] == r.ignorePrim) continue;

      if (prims[i]->intersectRay(r, t0, t1)){
        hit = true;
        minT = t1 = r.t;
        minPrim = r.prim;
        minBc = r.bc;
        minPrim2 = r.prim2;
      }
    }

    r.t = minT;
    r.prim = minPrim;
    r.bc = minBc;
    r.prim2 = minPrim2;
    return hit;
  }
  
  virtual SbVec3f  getMin() const {
    SbVec3f myMin(FLT_MAX, FLT_MAX, FLT_MAX);
    for (int i=0; i<prims.size(); i++)
      myMin = BBox::vecMin(myMin, prims[i]->getMin());
    return myMin;
  }
  virtual SbVec3f  getMinShiny() const {
    SbVec3f myMin(FLT_MAX, FLT_MAX, FLT_MAX);
    for (int i=0; i<prims.size(); i++)
      myMin = BBox::vecMin(myMin, prims[i]->getMinShiny());
    return myMin;
  }
  virtual SbVec3f  getMax() const {
    SbVec3f myMax(FLT_MIN, FLT_MIN, FLT_MIN);
    for (int i=0; i<prims.size(); i++)
      myMax = BBox::vecMax(myMax, prims[i]->getMax());
    return myMax;
  }
  virtual SbVec3f  getMaxShiny() const {
    SbVec3f myMax(FLT_MIN, FLT_MIN, FLT_MIN);
    for (int i=0; i<prims.size(); i++)
      myMax = BBox::vecMax(myMax, prims[i]->getMaxShiny());
    return myMax;
  }

};

class Instance : public Primitive {

private:
  Primitive *prim;
  SbMatrix m, mi;   // Matrix m and matrix m_inverse
  Material *mat;

public:
  Instance(Primitive *p, const SbMatrix &m_, Material *mat_=NULL) :
    prim(p), m(m_), mi(m_.inverse()), mat(mat_) { }
  
  Instance(){} // needs to be fast for Fractal, so no initializing is done.
  ~Instance(){ /* cerr << "~Instance: " << this << endl; */ }

  void deletePrim(){ delete prim; }
  Primitive* getPrim() { return prim; }

  virtual PrimType getType() const { return PRIMTYPE_INSTANCE; }
  virtual string getTypeName() const { return "Instance"; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  
  virtual SbVec3f  getMin() const;
  virtual SbVec3f  getMinShiny() const;
  virtual SbVec3f  getMax() const;
  virtual SbVec3f  getMaxShiny() const;

  virtual Material* getMat(const Ray &r) {
    if (mat) return mat;
//  return r.prim->getMat();    
    return r.prim2.front()->getMat(r);
  }
  virtual const Material* getMat(const Ray &r) const {
    if (mat) return mat;
//  return r.prim->getMat();    
    return r.prim2.front()->getMat(r);
  }

  // uses stack of primitive pointers to allow
  // instances of instances (I think...)
  virtual SbVec3f getNormal(const Ray &r) const{
    SbVec3f n;
    SbVec3f pn = r.prim2[r.prim2.idx--]->getNormal(r);
    mi.transpose().multDirMatrix(pn, n);
    r.prim2.idx++;
    n.normalize();
    return n;
  }
  virtual SbVec3f getIntNormal(const Ray &r) const{
    SbVec3f n;
    SbVec3f pn = r.prim2[r.prim2.idx--]->getIntNormal(r);
    mi.transpose().multDirMatrix(pn, n);
    r.prim2.idx++;
    n.normalize();
    return n;
  }
  
  virtual SbVec3f getNormal(const SbVec3f &bc) const{
    SbVec3f n;
    mi.transpose().multDirMatrix(prim->getNormal(bc), n);
    n.normalize();
    return n;
  }
  virtual SbVec3f getIntNormal(const SbVec3f &bc) const { // is this even used?
    SbVec3f n;
    mi.transpose().multDirMatrix(prim->getIntNormal(bc), n);
    n.normalize();
    return n;
  }
  
  virtual SbColor getIntDiffColor(const Ray &r) const {
    return r.prim2.front()->getIntDiffColor(r);
  }
  virtual SbColor getIntEmission( const Ray &r) const {
    return r.prim2.front()->getIntEmission(r);
  }
  virtual bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer) const{
    if (!prim->getDerivs(bc, pDer)) return false;
    SbMatrix tr = mi.transpose();
    tr.multDirMatrix(pDer.dPdU, pDer.dPdU);
    tr.multDirMatrix(pDer.dPdV, pDer.dPdV);
    tr.multDirMatrix(pDer.dNdU, pDer.dNdU);
    tr.multDirMatrix(pDer.dNdV, pDer.dNdV);
    return true;
  }
  
};

#endif  /* PRIMITIVE_H */
