//WWWW''''''''''''''''''''WWW''''''''''''''''''
//WW  W                    WW                  
//WW  W  WWW W WW     WWW  WW   WWW  WWWW  WWWW
//WWWW  W WW W WW    WW W  WW  W WW WWW   WWW  
//WW  W W WW W WW    WW    WW  W WW   WWW   WWW
//WW  W  WWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//             WW                              
//'''''''''''WWW'''''''''''''''''''''''''''''''

#ifndef RAYCLASS
#define RAYCLASS

#include <iostream>
#include <vector>
#include <Inventor/SbLinear.h>
#include <Inventor/SbColor.h>
#include "Halton.h"

// for piggyback variables
class Triangle;  
class LineSet;
class Primitive;

typedef struct thread_context_st thread_context;

using namespace std;

class FastPrimStack {
public:
  Primitive *prims[16];   // FIXME: Copying this a billion times seems really slow
  int idx;
  FastPrimStack() : idx(-1) { }
  void push_back(Primitive *p){
    prims[++idx] = p;
  }
  Primitive* front() { return prims[0]; }
  Primitive*& operator[](int i) { return prims[i]; }
};

class Ray {
private:

  static LineSet *ls;
  
public:

  static Primitive *NO_PRIM;
  
  SbVec3f pos;
  SbVec3f dir;             // must be normal (unit length)

  int depth;               // depth from eye of hit point (first is 1)

  int atten;               // Used by *dielectric* media to track state.
                           // True if in inside dielectric and Beer's law
                           // should be applied. Used by getRadiancePM, PT,
                           // emitPhotons.
  SbColor attenColor;      // The attenuation to use.
  
  int inside;              // Used by *volumetric* media to track state.
                           // Used by volumeIntegrate to not have density
                           // inside glass.
  
  float time;              // time of ray
  
  float importance;        // importance of result of ray
                           // (1.0 == very important, primary)
                           // Used by refractionContrib and getRadiancePM.

  float wavelen;           // wavelength in nm

  // "Dimension" refers to the dimension of a global generalized scrambled
  // Hammersly point set.
  // "Hammersley" means the first dimension's sequence of points is just i/N,
  // and the remaining dimensions are a Halton sequence, which are sequences
  // determined by taking the radical inverses of increasing natural numbers,
  // Radical inversion means taking the digits of a natural number in some
  // base b, mirroring them about the decimal point, and evaluating that number.
  // For the Halton sequence, subsequent dimensions use ascending
  // prime numbers starting with 2 for the base b (one per dimension).
  // "Scrambled" means there is a (fixed) scrambling of the digits 0..b-1
  // during the inversion, which gets rid of the patterns found in higher dimensions.
  // "Generalized" means instead of starting by taking the radical inverse of
  // 0, 1, 2, 3... you start at a random integer (the "instance").
  // For the global generalized scrambled Hammersly point set, the
  // generalization is actually done by adding (modulo one) a point
  // determined by the corresponding dimension of the i'th Halton sequence,
  // where i is the "instance".
  // "Trajectory splitting" is just replacing one monte carlo sample with a sum
  // of them, e.g., evaluating shadow samples with multiple shadow rays. This
  // is like the opposite of Russian Roulette, which stochastically *reduces* the
  // number of rays.

  // Functions that accept rays may modify ldDimension and ldInstance.
  // Callers should be aware of this.
  int ldDimension;     // "d" First available dimension of the (low discrepancy)
                       // global generalized scrambled Hammersley point set.
                       // Starts at 0. 
  int ldInstance;      // "i" The current instance. This determines the
                       // generalization using a Halton sequence and module addition.

  // If a functions uses low discrepancy (ld) values (delta_d != 0):
  //    (use dimensions ldDimensionNext...ldDimensionNext+delta_d-1)
  //    ldDimension += delta_d
  //    d' = ldDimension
  //    i' = i + j + 1
  // Otherwise, if it doesn't use a (ld) values, but branches:
  //    ldDimension ++
  //    d' = ldDimension
  //    i' = i + j + 1
  //    We increment dimension to prevent instance collisions
  //    for multiple sets of j. They could just add onto each other,
  //    which is valid but messy if #j's is not always the same.
  //    We increment the instance +1 to distinguish it from the current ray.
  
  // piggy back variables for intersection
  thread_context *ctx;     // context for access to thread global vars/RNG's

  // input
  Primitive *ignorePrim;            // id of triangle ray should not hit, if >=0
  
  // early output
  //
  // These should be set by intersectRay
  //
  // No hit:
  //   intersectRay returns false
  //   t = -1
  //   tri = NULL
  //   XXX triId = Ray::NO_TRI          (triId not currently used)
  //   bc = UNDEFINED
  //
  // Hit:
  //   intersectRay returns true
  //   t = dist to closest intersection
  //   tri = Primitive* of primitive intersected
  //   XXX triId = id of primitive      (triId not currently used)
  //   bc = bary of prim
  //
  double t;                // length along ray to hit point 
  double tMedia;           // length until first participating media scatter.
                           // Used to detect light contribution,
                           // by multipleDiffuseContrib for irradianceCaching
  Primitive *prim;         // pointer to primitive that this ray hits

  // auxillary primitive information, for tracking instances
  mutable FastPrimStack prim2; // Used by Instance. List of hit primitives ending with "final" hit prim.
                               // Used by Instance::getNormal() for walking list.
  
  SbVec3f bc;              // barycentric coordinates of hit point

  // eventual output
  SbVec3f x;               // position in space of the hit point
  SbColor diffCol;         // diffuse color of hit point surface (includes texture)

  // debugging
  int accelTravs;
  int accelInts;

  // Used in ray initialization areas like renderTile()
  Ray() : pos(0,0,0), dir(0,0,-1), importance(1), atten(0), inside(0),
          ignorePrim(NO_PRIM), ldDimension(0), ldInstance(0), wavelen(520) { }

  // primarily used for shadow rays and show'ed rays
  Ray(const SbVec3f & pos_, const SbVec3f & dir_, thread_context *ctx_,
      Primitive *ignorePrim_=NO_PRIM, int depth_=1):
    pos(pos_), dir(dir_), ctx(ctx_), depth(depth_),
    importance(1),
    atten(0),
    inside(0),
    ignorePrim(ignorePrim_),
    ldDimension(0), ldInstance(0), wavelen(520) { }

  // Used for secondary (reflected, etc.) rays
  Ray(const SbVec3f &pos_, const SbVec3f &dir_, const Ray &r,
      Primitive *ignorePrim_=NO_PRIM) :
    pos(pos_), dir(dir_),
    ctx(r.ctx),
    depth(r.depth+1),                        // depth auto. incremented
    importance(r.importance),
    atten(r.atten),
    attenColor(r.attenColor),
    inside(r.inside),
    ignorePrim(ignorePrim_),
    ldDimension(r.ldDimension), 
    ldInstance(r.ldInstance),
    prim(NULL),
    time(r.time),
    wavelen(r.wavelen)
    { }

  ~Ray();
  
  // returns point that is parameter t along the ray
  SbVec3f point(float t){ return pos + t*dir; }
  SbVec3f point(){ return pos + t*dir; }

  void ignore(Primitive *p){ ignorePrim = p; }

  // draws cylinder representation
  void show(double height = 100.0, SbColor color = SbColor(-1.0, -1.0, -1.0),
	    int clear = 0);

  
  // draws line representation
  void show2(double height = 100.0, SbColor color = SbColor(-1.0, -1.0, -1.0),
	     int clear = 0);


  int ldNextDim() {
    ldDimension++;
    if (ldDimension == Halton::numPrimes)
      ldDimension = 0;
    return ldDimension;
  }
  
  friend ostream& operator<< (ostream& os, const Ray& r);
};

#endif /* RAYCLASS */
