
#ifndef RAYACCELCLASS
#define RAYACCELCLASS

#include "Primitive.h"

class Ray; // forward declaration of Ray class

// maybe this should inherit from PrimList
class RayAccel : public Primitive {
  
private:

  void error(const char *f);
  
protected:
  int numPrims;
  double maxSideLength;
  SbVec3f min, max;

  // keep bounding sphere of primitives
  SbVec3f center;
  float radius;
  float shiny_radius;
  SbVec3f shiny_center;
  
public:

  RayAccel(){
    //error("RayAccel");
    numPrims = 0;
    maxSideLength = 0;
    min[0] = min[1] = min[2] = 0;
    max[0] = max[1] = max[2] = 0;
    center.setValue(0,0,0);
    radius = shiny_radius = 0;
    shiny_center.setValue(0,0,0);
  }
  ~RayAccel(){ }
  virtual void reset(){ error("reset"); }
  virtual int add(Primitive *t){ error("add"); return 0; }
  virtual void reserve(int n) { error("reserve"); }
  virtual void resizeToFit() { error("resizeToFit"); }
  virtual void printStats(){ error("printStats"); }
  virtual void setParams(int maxD, int maxI, int debug){ error("setParams"); }
  virtual void refine(int simpleDiv = 0){ error("refine"); }
  virtual void reorderNodes(){ cerr << "error: RayAccel::reorderNodes()!\n"; }
  virtual double& getMaxSideLength(){ return maxSideLength; }  
  virtual int size(){ return numPrims; }

  // Primitive methods
  SbVec3f getMin() const { return min; }
  SbVec3f getMax() const { return max; }
  SbVec3f getCenter() const { return center; }
  SbVec3f getCenterShiny() const { return shiny_center; }
  float getRadius() const { return radius; }
  float getRadiusShiny() const { return shiny_radius; }
};


// RayAccel comparison class
// Use -R3 to activate.
class DualAccel : public RayAccel {

private:
  RayAccel *ra0, *ra1;
  
public:

  DualAccel();
  virtual ~DualAccel(){
    delete ra0;
    delete ra1;
  }
  virtual void reset(){
    ra0->reset();
    ra1->reset();
  }
  virtual int add(Primitive *t){
    ra0->add(t);
    ra1->add(t);
  }
  virtual void reserve(int n) {
    ra0->reserve(n);
    ra1->reserve(n);
  }
  virtual void printStats(){
    ra0->printStats();
    ra1->printStats();
  }
  virtual void setParams(int maxD, int maxI, int debug){
    ra0->setParams(maxD, maxI, debug);
    ra1->setParams(maxD, maxI, debug);
  }
  virtual void refine(int simpleDiv = 0){
    ra0->refine(simpleDiv);
    ra1->refine(simpleDiv);
  }
  virtual void reorderNodes(){
    ra0->reorderNodes();
    ra1->reorderNodes();
  }

  // Primitive methods
  SbVec3f getMin() const {
    SbVec3f r0 = ra0->getMin();
    SbVec3f r1 = ra1->getMin();
    if (r0!=r1)
      SAY_HELLO;
    return r0;
  }
  SbVec3f getMax() const {
    SbVec3f r0 = ra0->getMax();
    SbVec3f r1 = ra1->getMax();
    if (r0!=r1)
      SAY_HELLO;
    return r0;
  }
  
  int size(){
    if (ra0->size() != ra1->size())
      SAY_HELLO;
    return ra0->size();
  }
  double& getMaxSideLength(){
    if (ra0->getMaxSideLength() != ra1->getMaxSideLength())
      SAY_HELLO;
    return ra0->getMaxSideLength();
  }
  
  SbVec3f getCenter() const {
    if (ra0->getCenter() != ra1->getCenter())
      SAY_HELLO;
    return ra0->getCenter();
  }
  float getRadius() const {
    if (ra0->getRadius() != ra1->getRadius())
      SAY_HELLO;
    return ra0->getRadius();
  }
  SbVec3f getCenterShiny() const {
    if (ra0->getCenterShiny() != ra1->getCenterShiny())
      SAY_HELLO;
    return ra0->getCenterShiny();
  }
  float getRadiusShiny() const {
    if (ra0->getRadiusShiny() != ra1->getRadiusShiny())
      SAY_HELLO;
    return ra0->getRadiusShiny();
  }


  virtual PrimType getType() const { return PRIMTYPE_TEST; }
  virtual string getTypeName() const { return "DualAccel"; }

  virtual bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8){
    bool r1 = ra1->intersectRay(r, t0, t1);
    double rt1 = r.t;
    bool r0 = ra0->intersectRay(r, t0, t1);
    double rt0 = r.t;
    if (r0 != r1)
      SAY_HELLO;
    if (rt0 != rt1)
      SAY_HELLO;
    return r0;
  }
  
};


// Plain list
class AccelList : public RayAccel {

private:
  PrimList primList;
  SbVec3f shiny_min, shiny_max;
  
public:

  AccelList() {
    reset();
  }
  virtual ~AccelList(){
  }
  virtual void reset(){
    primList.clear();
    min[0] = min[1] = min[2] = 1e30;
    max[0] = max[1] = max[2] = -1e30;
    shiny_min.setValue(1e30f, 1e30f, 1e30f);
    shiny_max.setValue(-1e30f, -1e30f, -1e30f);
  }
  virtual int add(Primitive *t){
    primList.add(t);
  }
  virtual void reserve(int n) {
    primList.reserve(n);
  }
  virtual void printStats(){
  }
  virtual void setParams(int maxD, int maxI, int debug){
  }
  virtual void refine(int simpleDiv = 0){

    // adjust first node boundaries
    // (make global bounding box)
    for (int ti=0; ti<primList.size(); ti++){
      const Primitive *prim = primList.prim(ti);
      min = BBox::vecMin(min, prim->getMin());
      max = BBox::vecMax(max, prim->getMax());
      shiny_min = BBox::vecMin(shiny_min, prim->getMinShiny());
      shiny_max = BBox::vecMax(shiny_max, prim->getMaxShiny());
    }
    
    // setup boundaries of first node
    for (int i=0; i<3; i++){
      if (max[i]<min[i]){
        min[i] = -1;
        max[i] = 1;
      }
      if (shiny_max[i]<shiny_min[i]){
        shiny_min[i] = -1;
        shiny_max[i] = 1;
      }
      if (max[i]-min[i] > maxSideLength)
        maxSideLength = max[i]-min[i];
    }

    center = (min+max)/2.0;
    radius = (max-center).length();

    // shiny bounding sphere
    shiny_center = (shiny_min+shiny_max)/2.0;
    shiny_radius = (shiny_max-shiny_center).length();
  }
  virtual void reorderNodes(){
  }

  // Primitive methods
  SbVec3f getMin() const {
    return primList.getMin();
  }
  SbVec3f getMax() const {
    return primList.getMax();
  }
  
  int size(){
    return primList.size();
  }
  double& getMaxSideLength(){
    return maxSideLength;
  }

  // Primitive methods  
  virtual PrimType getType() const { return PRIMTYPE_ACCELLIST; }
  virtual string getTypeName() const { return "AccelList"; }

  virtual bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8){
    return primList.intersectRay(r, t0, t1);
  }
  SbVec3f getMinShiny() const { return primList.getMinShiny(); }
  SbVec3f getMaxShiny() const { return primList.getMaxShiny(); }
  
};

#endif /* RAYACCELCLASS */
