#include <stdlib.h>
#include <string>
#include "SkyConstants.h"
#include "Sky.h"
#include "Spectrum.h"
#include "DataSet.h"
#include "debug.h"
#include "Misc.h"
#include <cmath>

using namespace std;

//     Y N
//     |/
//  W--x--X(E)
//    /|
//   Z(S)

// Implementation of "A Practical Analytic Model for Daylight"
// by  A. J. Preetham, Peter Shirley, Brian Smits
// http://www.cs.utah.edu/vissim/papers/sunsky/

// by Kevin Beason

Sky::Sky(){
  sunBasisU = SbVec3f(1,0,0);
  sunBasisV = SbVec3f(0,0,-1);
  sunBasisW = SbVec3f(0,1,0);
  sunRadius = sunRadiusAngle = sunDist = sunOmega = 0;
  turbidity = thetaSolar = thetaSolarSafe = 0;
  sunRadiance.setValue(0,0,0);
  AY=BY=CY=DY=EY=Ax=Bx=Cx=Dx=Ex=Ay=By=Cy=Dy=Ey=0;
  Yz=xz=yz=0;
  S_0_m = S_0_p = NULL;
}

int Sky::ti_S0pre=10;
int Sky::pi_S0pre=Sky::ti_S0pre*2;
int Sky::ti_Lpre=8;
int Sky::pi_Lpre=Sky::ti_Lpre*4;

Spectrum Sky::neta[1801];

Sky::~Sky(){
  delete S_0_m;
  delete S_0_p;
}

void Sky::init(GlobalVars *gv_, float stdTime) {

  // set GlobalVars pointer
  gv = gv_;
  getType() = SKY;

//  float stdTime = stdTimeStr ? atof(stdTimeStr) : 12+5;
//  float stdTime = 12+3;     // hours, decimal. 3pm
//  float stdTime = 12+5;     // hours, decimal. 6pm
//  float julianDate = 322;   // 11/18/07  see: `date +%j`
  float julianDate = 66;   // 3/6/08  see: `date +%j`
  float stdMeridianPST = -8.*15 * M_PI/180;  // radians
  float stdMeridian = stdMeridianPST;
  float latitude = 34.017 * M_PI/180;  // los angeles
  float longitude = -118.41 * M_PI/180; // http://www.gorissen.info/Pierre/maps/googleMapLocationv3.php

  // in decimal hours
  float solarTime = stdTime +
    0.170 * sin(4.*M_PI*(julianDate-80)/373.) -
    0.129 * sin(2.*M_PI*(julianDate-8)/355.) +
    12. * (stdMeridian-longitude)/M_PI;

  float solarDeclination = 0.4093 * sin(2.*M_PI*(julianDate-81)/368.); // radians "delta"
  float tmp = M_PI*solarTime/12.;

  // solar angle from zenith in radians
  // [0, pi/2]
  thetaSolar = M_PI/2. - asin(sin(latitude)*sin(solarDeclination) -
                              cos(latitude)*cos(solarDeclination) *
                              cos(tmp));

  // solar azimuth in radians, west from South
  // [0, 2pi]
  float opp = -cos(solarDeclination)*sin(tmp);
  float adj = cos(latitude)*sin(solarDeclination) -
              sin(latitude)*cos(solarDeclination)*cos(tmp);
  float phiSolar = atan2(opp, adj);

  //phiSolar=thetaSolar=0;   // straight overhead
  //thetaSolar = 60 * M_PI/180.;
  //phiSolar = 0;

  // sun distance, in meters. perihelion/aphelion
  // distances from http://www.physicalgeography.net/fundamentals/6h.html
  float sunDistJul4 = 152.1e9; // far
  float sunDistJan3 = 147.3e9; // near

  // lerp... not accurate but whateva
  float yearFrac = (julianDate-3+365%365)/365.;
  if (yearFrac>.5) yearFrac = 1-yearFrac;
  sunDist = (1-yearFrac)*sunDistJan3 + yearFrac*sunDistJul4;


  float sunSizeScale = getenv("PANE_SUN_SIZE") ?
    atof(getenv("PANE_SUN_SIZE")) : 1.0f;
  sunRadius = 695500e3 * sunSizeScale; // in meters (thanks google)
  //sunDist = 1.5e11;
  sunRadiusAngle = atan2(sunRadius, sunDist);  //sunSize/sunDist;

  // compute solid angle
  sunOmega = M_PI*sunRadius*sunRadius / (4*M_PI*sunDist*sunDist) * 4*M_PI;
  
  // orthonormal basis (W) pointed towards sun
  // phi = 0 = noon = south = z
  // phi = pi/2 = evening = west = -x
  // up = y
  cerr << "phiSolar = " << phiSolar << endl;
  float phi = phiSolar; //1.5*M_PI-phiSolar;
  cerr << "phi = " << phi << endl;
  sunBasisW.setValue(sin(thetaSolar)*-sin(phi), // phi is with respect to S (z)
		     cos(thetaSolar),           // increasing towards W (-x)
		     sin(thetaSolar)*cos(phi));
  sunBasisV.setValue(sunBasisW[1], sunBasisW[2], sunBasisW[0]);
  sunBasisV -= sunBasisV.dot(sunBasisW)*sunBasisW;
  sunBasisV.normalize();
  sunBasisU = sunBasisV.cross(sunBasisW);

  cerr << "Sun W = " << sunBasisW << endl;

  Spectrum sunSpec(solAmplitudes, 38, 380, 750);
  sunSpec.scaleSpec(10); // values were off by 10
  sunSpec.scaleSpec(10); // convert to W/(m^2 sr nm)
  //sunSpec.dump("utah.csv");
  
  // modulate color based on atmosphere absorbtion

  // clamp to pi/2
  float thetaSolarSafe = thetaSolar;
  if (thetaSolarSafe>M_PI/2)
    thetaSolarSafe = M_PI/2;
  
  // reverse at pi/2
//   if (thetaSolarSafe>M_PI/2)
//     thetaSolarSafe = M_PI-thetaSolarSafe;
  
  cerr << "thetaSolar = " << thetaSolar << endl;
  cerr << "thetaSolarSafe = " << thetaSolarSafe << endl;


  // min turbidity is such that BY>=0 => 1.20284186831739
//  turbidity = 1.203;
//  turbidity = 1.5;
  turbidity = 2;
//    turbidity = 6;
//  turbidity = 10;
//  turbidity = 30;

  static bool overrideTurbid = getenv("PANE_SKY_TURBID");
  static float overrideTurbidVal = overrideTurbid ? atof(getenv("PANE_SKY_TURBID")) : 0;
  if (overrideTurbid)
    turbidity = overrideTurbidVal;
  
  float beta = 0.04608*turbidity-0.04586;
  float alpha = 1.3;
  float l = 0.35;
  float w = 2;
  float m = 1./(cos(thetaSolarSafe)+0.15*
                pow(93.885-thetaSolarSafe*180/M_PI, -1.253));

  // need wl, k_o, k_g, k_wa
  SpectralCurveIrreg k_oFunc(k_oWavelengths, k_oAmplitudes, 64);
  SpectralCurveIrreg k_gFunc(k_gWavelengths, k_gAmplitudes, 4);
  SpectralCurveIrreg k_waFunc(k_waWavelengths, k_waAmplitudes, 13);

  int num = Spectrum::getNumWavelens();
  float max = Spectrum::getMaxWavelen();
  float min = Spectrum::getMinWavelen();
  cerr << "m=" << m << endl;
  cerr << "num=" << num << " max=" << max << " min=" << min << endl;
  for (int i=0; i<num; i++){
    float wl = (max-min) * i/(float)(num-1) + min;

    // apparently lambda should be in um
    float wlum = wl/1000;
    //cerr << "wl=" << wl << endl;

    float k_o = k_oFunc(wl);
    float k_g = k_gFunc(wl);
    float k_wa = k_waFunc(wl);
    
    float t_air = exp(-m * 0.008735 * powf(wlum, -4.08));
    float t_ang = exp(-m * beta * pow(wlum, -alpha));
    float t_ozo = exp(-m * k_o * l);
    float t_gas = exp(-1.41 * k_g * m / pow(1 + 118.93 * k_g * m, 0.45));
    float t_h2o = exp(-0.2385 * k_wa * w * m /
                      pow(1 + 20.07 * k_wa * w * m, 0.45));

    float t_tot;
    t_tot = t_air*t_ang*t_ozo*t_gas*t_h2o;
//  t_tot = t_air;
//    cerr << "t_tot[" << i << "=" << wl << "] = " << t_tot << endl;
//     cerr << "[" << i << "=" << wl << "] " <<
//       "k_o="  << k_o << " " <<
//       "k_wa=" << k_wa << " " <<
//       "k_g="  << k_g << endl;
      
      
    sunSpec.scaleAmp(i, t_tot);
  }
  
  sunRadiance = sunSpec.getRGB();

  if (sunSizeScale != 1.0)
    sunRadiance *= 1./(sunSizeScale*sunSizeScale);
  
  // constants
  // these vales taken from source rather than paper
  // http://www.cs.utah.edu/vissim/papers/sunsky/code
  float T=turbidity;
  AY =  0.17872*T + -1.46303;
  BY = -0.35540*T +  0.42749;
  CY = -0.02266*T +  5.32505;
  DY =  0.12064*T + -2.57705;
  EY = -0.06696*T +  0.37027;

  Ax = -0.01925*T + -0.25922;
  Bx = -0.06651*T +  0.00081;
  Cx = -0.00041*T +  0.21247;
  Dx = -0.06409*T + -0.89887;
  Ex = -0.00325*T +  0.04517;

  Ay = -0.01669*T + -0.26078;
  By = -0.09495*T +  0.00921;
  Cy = -0.00792*T +  0.21023;
  Dy = -0.04405*T + -1.65369;
  Ey = -0.01092*T +  0.05291;

  float t = thetaSolar;
  float t2 = t*t, t3=t*t*t;
  float chi = (4./9.-T/120.)*(M_PI-2*t);
  Yz = (4.0453*T-4.9710)*tan(chi) - 0.2155*T + 2.4192;
  xz =
    ( 0.00165*t3 + -0.00374*t2 +  0.00208*t + 0.0    )*T*T +
    (-0.02902*t3 +  0.06377*t2 + -0.03202*t + 0.00394)*T   +
    ( 0.11693*t3 + -0.21196*t2 +  0.06052*t + 0.25885);
  yz =
    ( 0.00275*t3 + -0.00610*t2 +  0.00316*t + 0.0    )*T*T +
    (-0.04214*t3 +  0.08970*t2 + -0.04153*t + 0.00515)*T   +
    ( 0.15346*t3 + -0.26756*t2 +  0.06669*t + 0.26688);

  // convert Kcd/m^2 to cd/m^2 to W/m^2sr
  Yz *= 1000/683.;

  // render sky view to environment map
  cerr << "Precomputing sky HDR map... ";
  SbVec3f ndc;
  int dim[2] = {512, 512};
  float *texture = new float[3 * dim[0]*dim[1]];

  for (int j=0; j<dim[1]; j++){
    for (int i=0; i<dim[0]; i++){

      ndc[0] =  (((double)i + 0.5) / dim[0] - 0.5)*2.0;
      ndc[1] =  (((double)j + 0.5) / dim[1] - 0.5)*2.0;
      ndc[2] = 0;

      int k = (j*dim[0]+i)*3;
      
      float theta = ndc.length() * M_PI; //  /2;
      
      if (theta>M_PI_2*258/256.) {     // go a little past horizon to completely
                                       // fill in pixels seen on horizon
        texture[k+0]=texture[k+1]=texture[k+2] = 0;
        continue;
      }

      //         sky coord    ==> texture coord
      //
      //            Y               ___
      //          __|___           |   |   center of texure is +z
      //         /__|__/__ X  ==>  |___|   in EnvironMap, so sky
      //       a/__/__/  (east)   a        adds a (1,0,0,-pi/2) rotation to
      //          /                        rotate +z to +y
      //         Z (south)
      //
      // phi is angle between vector from center to pixel and x axis (looking down along y)
      float phi = atan2(ndc[1], ndc[0]);
      SbVec3f dir = SbVec3f(
        sin(theta)*cos(phi),
        cos(theta),
        sin(theta)*-sin(phi)
        );

      SbColor col = getRadiance(dir, false);
      texture[k+0] = col[0];
      texture[k+1] = col[1];
      texture[k+2] = col[2];
    }
  }

  cerr << "done.\n";

  // calculate rotation
  float angle = gv->envMapRotation*M_PI/180.0; // convert degrees to radians
  SbRotation userSpin(SbVec3f(0.0, 1.0, 0.0), angle); // user angle about Y axis
  SbRotation skyCorrection(SbVec3f(1,0,0),-M_PI/2); // 90 degree fix about X
  SbRotation skyRot = skyCorrection*userSpin;
  
  // initialize texture subclass
  sky.init(gv, texture, dim[0], dim[1], &skyRot);
  string file = "skyEnvMap.hdr";
  static bool write = getenv("PANE_WRITE_SKY");
  if (write) sky.writeDataHdr(file);

  // now use rotation
  setRotateSky(userSpin);
  
  delete [] texture;

  static bool addAerial = getenv("PANE_SKY_AERIAL");
  if (!addAerial) return;
  
  ComputeOutScatteringCoeffs();
  ComputeInScatteringCoeffs();
}

SbColor Sky::impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple) {

  int samples = 4;
  if (samples<gv->lightSamples*2)
    samples = gv->lightSamples*2;
  
  int sqrt_samples = (int)sqrt((float)samples); // this should be precomputed

  // use one sample in final gather
  if (simple)
   sqrt_samples = 1;
  
  if (sqrt_samples<1) sqrt_samples = 1;
  int numVis = 0;
  SbColor color(0,0,0);

  float sun_w = M_PI*sunRadius*sunRadius / (4*M_PI*sunDist*sunDist) *4*M_PI;
//   float density = 1/sun_w;
//   float densityInv = 1/density;
  float pdf1 = 1/sun_w;
  
  for (int sampy=0; sampy<sqrt_samples; sampy++){
    for (int sampx=0; sampx<sqrt_samples; sampx++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;
      Material *mtl = r.prim->getMat(r);
      BRDF *brdf = mtl->getBRDF();
      SbVec3f dir;
      SbVec3f brdfCosT;
      float value = 1;
      float newPdf;
      
      //const float t_eps = 0.002;
      //Ray lr = Ray(r.x+dir*t_eps, dir, r.ctx);
      Ray lr = Ray(r.x, dir, r.ctx);
      lr.ignore(r.prim);

      // Multiple Importance Sampling (MIS)
      float lightWeight = mtl->getGlossiness() ? .9 : 1;
      float brdfWeight = 1-lightWeight;
      if (1<=lightWeight || r.ctx->mydrand48()<lightWeight){
      
        // sample sun disc uniformly
        // Shirley's RRT 1e pg. 123
        double cosAlpha = 1+eps1*(cos(sunRadiusAngle)-1);
        double sinAlpha = sqrt(1-cosAlpha*cosAlpha);
        double phi = 2*M_PI*eps2;
        dir = 
          sinAlpha*cos(phi)*sunBasisU +
          sinAlpha*sin(phi)*sunBasisV +
          cosAlpha*sunBasisW;

        // rotate
        textureToWorldRot.multDirMatrix(dir, dir);

        // preetham scales sun radiance down at horizon anyway
        if (dir[1]<=0)
          continue;

        // dont illuminate lights below horizon
        if (ng.dot(dir) <= 0.0)
          continue;
      
        // don't illuminate area that think they are facing away
        float dot1 = n.dot(dir);
        if (dot1<=0.0)
          continue;      
		  
        lr.dir = dir;
        if (gv->pane->scenePrim->intersectRay(lr)){
          continue; // hits something (and not background)
        }
        
        float pdf2;
        brdf->GetBRDF(r.dir, dir, n, ng, r, brdfCosT, pdf2);
        //newPdf = pdf1;    // no MIS
        //newPdf = .5*pdf1 + .5*pdf2;
        newPdf = lightWeight*pdf1 + brdfWeight*pdf2;        
      }
      else {
        double brdfVal, pdfVal, brdfOverPdfVal, brdfOverPdfCosTval;
        SbVec3f brdfColor;
        double eps3 = r.ctx->mydrand48();
        brdf->GetSampleDir(eps1, eps2, eps3, r.dir, n, ng, r, dir,
                           brdfVal, pdfVal, brdfOverPdfVal, brdfOverPdfCosTval,
                           brdfColor);
        
        if (pdfVal<=0)
          continue;
        
        // check direction is pointing towards sun ;)
        SbVec3f sunDirWorld;
        textureToWorldRot.multDirMatrix(sunBasisW, sunDirWorld);  // rotate
        float dirDotSunW = dir.dot(sunDirWorld);
        float cosSunRadiusAngle = cos(sunRadiusAngle);
        if (dirDotSunW<cosSunRadiusAngle)
          continue;
        
        lr.dir = dir;

        float dot1 = n.dot(dir);
        if (dot1<=0.0)
          continue;      
        
        if (gv->pane->scenePrim->intersectRay(lr)){
          continue; // hits something (and not background)
        }

        brdfCosT = brdfColor*brdfVal*dot1;
        newPdf = lightWeight*pdf1 + brdfWeight*pdfVal;
      }

      // ray marching absorbtion and attenuation
      float len = 1e6; //gv->pane->dataSet1->getVolMaxDist();
      SbVec3f lx = r.x+len*dir;
      SbColor Le = sunRadiance;
      gv->pane->applyAttenuation(r, r.x, lx, len, Le);

      value /= newPdf;
	  
      // compute color
      // shirley p.120 eq. 15.5
      color += vecScale(brdfCosT*value, Le);
      
    }
  }

  // cerr << "sunRadiance = " << sunRadiance << endl;
  // cerr << "sunDist = " << sunDist << endl;
  // cerr << "sunW = " << sun_w << endl;
//  color *= densityInv / (sqrt_samples*sqrt_samples);
  if (sqrt_samples)
    color /= (sqrt_samples*sqrt_samples);  

  static bool doSky = !getenv("PANE_SKYLIGHT_OFF");
  if (doSky)
    color += sky.impSample(r, n, ng, simple);

  return color;
}

float Sky::getSample(double eps1, double eps2,
                     SbVec3f &dir, SbColor &c){

  if (eps2>=0.5){  // sun
    // sample sun disc uniformly
    // Shirley's RRT 1e pg. 123
    double cosAlpha = 1+eps1*(cos(sunRadiusAngle)-1);
    double sinAlpha = sqrt(1-cosAlpha*cosAlpha);
    double phi = 2*M_PI*((eps2-0.5f)*2.f);
    dir = 
      sinAlpha*cos(phi)*sunBasisU +
      sinAlpha*sin(phi)*sunBasisV +
      cosAlpha*sunBasisW;

    // rotate
    textureToWorldRot.multDirMatrix(dir, dir);

    c = sunRadiance;
    return 1/sunOmega*0.5f;
  } else {        // sky
    return sky.getSample(eps1, eps2*2.0, dir, c)*0.5;
  }
}

SbColor Sky::volImpSample(Ray &r, int samples){

  int sqrt_samples = (int)sqrt((float)samples);
  
  if (sqrt_samples<1) sqrt_samples = 1;
  int numVis = 0;
  SbColor color(0,0,0);

  float sun_w = M_PI*sunRadius*sunRadius / (4*M_PI*sunDist*sunDist) *4*M_PI;
  float density = 1/sun_w;
  float densityInv = 1/density;
  
  for (int sampy=0; sampy<sqrt_samples; sampy++){
    for (int sampx=0; sampx<sqrt_samples; sampx++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;

      // sample sun disc uniformly
      // Shirley's RRT 1e pg. 123
      double cosAlpha = 1+eps1*(cos(sunRadiusAngle)-1);
      double sinAlpha = sqrt(1-cosAlpha*cosAlpha);
      double phi = 2*M_PI*eps2;
      SbVec3f dir = 
 	sinAlpha*cos(phi)*sunBasisU +
 	sinAlpha*sin(phi)*sunBasisV +
 	cosAlpha*sunBasisW;

      // rotate
      textureToWorldRot.multDirMatrix(dir, dir);

      const float t_eps = 0.002;
      Ray lr = Ray(r.x+dir*t_eps, dir, r.ctx);
      lr.ignore(r.prim);
      if (gv->pane->scenePrim->intersectRay(lr)){
	continue; // hits something (and not background)
      }

      // ray marching absorbtion and attenuation
      float len = 1e6; //gv->pane->dataSet1->getVolMaxDist();
      SbVec3f lx = r.x+len*dir;
      SbColor Le = sunRadiance;
      gv->pane->applyAttenuation(r, r.x, lx, len, Le);
	  
      // compute color
      // shirley p.120 eq. 15.5
      color += Le;

    }
  }

  // cerr << "sunRadiance = " << sunRadiance << endl;
  // cerr << "sunDist = " << sunDist << endl;
  // cerr << "sunW = " << sun_w << endl;
  color *= densityInv / (sqrt_samples*sqrt_samples);

  color += sky.volImpSample(r, samples);

  return color;
}

namespace {

  float Perez(float theta, float gamma, float zenVal, float thetaSun,
              float A, float B, float C, float D, float E){
    double cosgamma = cos(gamma);
    double costheta = cos(theta);
    double cosThetaSun = cos(thetaSun);

    double BoverCT = B/costheta;
    if (BoverCT > 20)
      BoverCT = 20;

    double num =
      (1. + A*exp(BoverCT)) *
      (1. + C*exp(D*gamma) + E*cosgamma*cosgamma);

    double den =
      (1. + A*exp(B)) *
      (1. + C*exp(D*thetaSun) + E*cosThetaSun*cosThetaSun);

    double rat = num/den;
    if (rat>1e2)
      rat = 1e2;
    else if (rat<-1e2)
      rat = -1e2;

    return rat * zenVal;
    
  }
}

SbColor Sky::getRadiance(SbVec3f &dir, bool includeSun){

  // when would I ever want to include the sun??
  includeSun = false;
  
  SbColor color(0,0,0);

  //return color;
  
  // rotate
  worldToTextureRot.multDirMatrix(dir, dir);
  
  float theta = acosf(dir.dot(SbVec3f(0,1,0)));

  float dirDotSunW = dir.dot(sunBasisW);
  // clip to [-1,1]
  if (dirDotSunW > 1) dirDotSunW = 1;
  else if (dirDotSunW < -1) dirDotSunW = -1;
  float gamma = acosf(dirDotSunW);

  if (includeSun) {        // direct ray

    // black below horizon
    if (dir[1]<-.01)
      return SbColor(0,0,0); 

    if (gamma<sunRadiusAngle)           // include sun
      color += sunRadiance;

    if (theta>M_PI_2)                   // work below horizon
      theta = M_PI-theta;
  }
  else if (dir[1]<=0) {      // for precomputing sky
    
    if (dir[1]<-.01)         // go a little past horizion to fill in texture completely
      return SbColor(0,0,0); // nothing below horizon

    if (theta>M_PI_2)        // work below horizon
      theta = M_PI-theta;
  }

  //return color; // disable everything but sun

  static float Ym[] = { 1e9,-1e9 };
  static float xm[] = { 1e9,-1e9 };
  static float ym[] = { 1e9,-1e9 };

  float Y =  Perez(theta,gamma,Yz,thetaSolar,AY,BY,CY,DY,EY);
  if (Y<0)
    return SbColor(0,0,0);

  bool say = false;
  if (Y<Ym[0]){
    Ym[0] = Y;
    say = true;
  } else if (Y>Ym[1]) {
    Ym[1] = Y;
    say = true;
  }
  
  float x = Perez(theta,gamma,xz,thetaSolar,Ax,Bx,Cx,Dx,Ex);

  if (x<xm[0]){
    xm[0] = x;
    say = true;
  } else if (x>xm[1]) {
    xm[1] = x;
    say = true;
  }
  
  float y = Perez(theta,gamma,yz,thetaSolar,Ay,By,Cy,Dy,Ey);

  if (y<ym[0]){
    ym[0] = y;
    say = true;
  } else if (y>ym[1]) {
    ym[1] = y;
    say = true;
  }

  color += Spectrum::YxyToRgb(Y,x,y);
  
  return color;
}

SbColor Sky::getEmissiveFlux(float receiverArea){
  
  // L = Phi/(Area*cosTheta*Omega) -> Phi = L*Area*cosTheta*Omega
  //
  // (where Area = pi*worldRad^2, cosTheta = 1, Omega={sunW,4*pi} )
  //
  // EmisSun = sunRadiance * AreaDisc * 1 * sunOmega    (theta is unknown)
  // EmisSky = sky.Lavg    * AreaDisc * 1 * 4*pi
  //
  // EmisTot = EmisSun+EmisSky

  // L = Phi/(Area*cosTheta*Omega) -> Phi = L*Area*cosTheta*Omega
  SbVec3f sunFlux = sunRadiance * receiverArea * sunOmega;
  SbVec3f skyFlux = sky.getEmissiveFlux(receiverArea);
  return sunFlux+skyFlux;
}

void Sky::ComputeOutScatteringCoeffs(){
  double refrIdx_air = 1.0003;
  double a = (refrIdx_air*refrIdx_air-1);
  double N = 2.545e25;
  double p_n = 0.035;
  
  double c = (0.6544*turbidity-0.6510)*1e-16;
  double v = 4;

  Spectrum col1(6500), col2(6500);
  col1.scaleSpec(1/4.73223e+06); // normalize by red channel
  col2.scaleSpec(1/4.73223e+06); // normalize by red channel
  
  Spectrum sunSpec(solAmplitudes, 38, 380, 750);
  Spectrum K(KAmplitudes, 501, 300, 800);
  
  int num = Spectrum::getNumWavelens();
  float max = Spectrum::getMaxWavelen();
  float min = Spectrum::getMinWavelen();
  for (int i=0; i<num; i++){
    float wl = (max-min) * i/(float)(num-1) + min;
    wl *= 1e-9; // convert nm to m

    double Bm = 8*M_PI*M_PI*M_PI * a*a/(3*N* wl*wl*wl*wl)*(6+3*p_n)/(6-7*p_n);
    col1.scaleAmp(i, Bm);

    double Bp = 0.434*c*M_PI*pow(2*M_PI/wl, v-2)*K.getAmp(i);
    Bp *= pow(.01, v-3); // WTF! Corrective factor in code but not paper!
    col2.scaleAmp(i, Bp);
  }

  B_m_0 = col1.getRGB(); // Rayleigh (gas)
  B_p_0 = col2.getRGB(); // Mie (haze)

//   cerr << "B_m_0 = " << B_m_0 << "\n";
//   cerr << "B_p_0 = " << B_p_0 << "\n";
}

SbColor Sky::ComputeAerialPersp(const Ray &r, SbColor &c){

  static bool doAP = getenv("PANE_SKY_AP");
  if (!doAP)
    return c;
  
  SbVec3f dir;
  worldToTextureRot.multDirMatrix(r.dir, dir);
  double cosT = dir[1]; //dir.dot(SbVec3f(0,1,0));
  double h0 = r.pos[1];
  h0 = gv->meters(h0)+1000;
  double t = r.t<0 ? 1e30 : r.t;
  t = gv->meters(t);
  
    
  SbColor atten, inscatter;
  if (h0<0) // under earth
    return c;

//   if (cosT<1e-7)
//     cosT = 1e-7;

  // h0+t*cosT = eps
  // ==> t = (eps-h0)/cosT
  if (h0+t*cosT<0)
    t = ((h0*1e-7)-h0)/cosT;

  if (h0+t*cosT<0)
    GIVE_ERROR("Check Sky.cxx");
  
  double phi = atan2(dir[2], dir[0]); // must be consistent with
  // cache and constants creation
  ComputeAttenAndInScatter(h0, cosT, phi, t, atten, inscatter);

  static bool attenOnly = getenv("PANE_SKY_ATTEN_ONLY");
  if (attenOnly)
    return vecScale(c, atten);
  
  return vecScale(c, atten) + inscatter;
}

namespace {
  inline void reduceRow4x4(double *r1, double *r2, int j){
    double x = -r2[j]/r1[j];   // x*a+b=0  => x=-b/a
    for (int i=j; i<5; i++){
      r2[i] += x*r1[i];
    }
  }

  inline void scaleRow4x4(double *r1, double x, int j){
    for (int i=j; i<5; i++)
      r1[i] *= x;
  }

  void solve4x4(double m[4][5]){
    // reduce to upper triangular form
    scaleRow4x4(m[0], 1/m[0][0], 0);
    reduceRow4x4(m[0], m[1], 0);
    reduceRow4x4(m[0], m[2], 0);
    reduceRow4x4(m[0], m[3], 0);
  
    scaleRow4x4(m[1], 1/m[1][1], 1);
    reduceRow4x4(m[1], m[2], 1);
    reduceRow4x4(m[1], m[3], 1);

    scaleRow4x4(m[2], 1/m[2][2], 2);
    reduceRow4x4(m[2], m[3], 2);

    scaleRow4x4(m[3], 1/m[3][3], 3);

    // back substitute
    reduceRow4x4(m[3], m[0], 3);
    reduceRow4x4(m[3], m[1], 3);
    reduceRow4x4(m[3], m[2], 3);
  
    reduceRow4x4(m[2], m[0], 2);
    reduceRow4x4(m[2], m[1], 2);

    reduceRow4x4(m[1], m[0], 1);
  }

  // g(v)    =  Av^3+Bv^2+Cv+D
  // g'(v)   = 3Av^2+2Bv+c
  // g''(v)  = 6Av+2B
  // g'''(v) = 6A
  
  inline double g_v(double v, double a[4]){
    return a[0]*v*v*v + a[1]*v*v + a[2]*v + a[3];
  }
  inline double gp_v(double v, double a[4]){
    return 3*a[0]*v*v + 2*a[1]*v + a[2];
  }
  inline double gpp_v(double v, double a[4]){
    return 6*a[0]*v + 2*a[1];
  }
  inline double gppp_v(double v, double a[4]){
    return 6*a[0];
  }
  
}

void Sky::ComputeInScatteringCoeffs(){

  // load neta
  float tmp[50];
  for (int i=0; i<1801; i++){
    for (int j=0; j<50; j++)
      tmp[j] = netaLambdaTable[j][i];
    neta[i] = Spectrum(tmp, 50, 300, 790);
  }

//   int th[] = {1, 4, 7, 10, 30, 60 , 80, 90 , 110, 120, 130, 150, 180};
//   int wls[] = {400, 450, 550, 650, 850};

//   cerr << "wtf table:\n";
//   for (int t=0;t<13; t++){
//     btheta = th[t]*M_PI/180;
//     Spectrum &neta_theta = neta[int(btheta*180/M_PI*10)];
//     cerr  << "theta = " << th[t] << " | ";
//     for (int w=0;w<5; w++){
//       cerr << neta_theta.getVal(wls[w]) << " ";
//     }
//     cerr << "  neta\n";
//   }
  
  delete S_0_m;
  delete S_0_p;

  S_0_m = new SbColor[(ti_S0pre+1)*pi_S0pre];
  S_0_p = new SbColor[(ti_S0pre+1)*pi_S0pre];

  string updateMsg = "Precompute sky in-scattering coeffs S_0_m, S_0_p...";
  
  // add sky light in-scattering
  for (int ti=0; ti<=ti_S0pre; ti++){    // theta
    cerr << "\r" << updateMsg << ti/float(ti_S0pre)*100. << "%";
    for (int pi=0; pi<pi_S0pre; pi++){  // phi

      // find theta at corners of regions in [0,ti] in [0, pi]
      float costheta_a = 1 - 2*ti/float(ti_S0pre); // cosT=1-2ti/TI => ti/TI = (1-cosT)/2
      float theta_a = acosf(costheta_a);

      // find phi corners of regions [0,pi) in [0, 2pi)
      float phi_a = pi/float(pi_S0pre)*M_PI*2;

      SbVec3f dir(
 	sin(theta_a)*cos(phi_a),  // phi is angle in XZ plane from x axis,
 	cos(theta_a),             // increasing towards +z
 	sin(theta_a)*sin(phi_a));

      SbColor S_0_m_theta, S_0_p_theta;
      ComputeInScatteringForOneDir(dir, S_0_m_theta, S_0_p_theta);
      int idx = ti*pi_S0pre+pi;
      S_0_m[idx] = S_0_m_theta;
      S_0_p[idx] = S_0_p_theta;
    }
  }
  cerr << "\r" << updateMsg << "done.\n";

}

void Solve2(double m[4][5]){
  double A, B, C, D;
  double m00=m[0][0], m01=m[0][1], m02=m[0][2], m03=m[0][3];
  double m10=m[1][0], m11=m[1][1], m12=m[1][2], m13=m[1][3];
  double m20=m[2][0], m21=m[2][1], m22=m[2][2], m23=m[2][3];
  double m30=m[3][0], m31=m[3][1], m32=m[3][2], m33=m[3][3];
  double E1=m[0][4], E2=m[1][4], E3=m[2][4], E4=m[3][4];
  
A = - (m12*(m21*E4 - m31*E3) + m02*(m31*E3 - m21*E4) + m01*(E4 - E3)
 + m11*(E3 - E4) + (m31 - m21)*E2 - m31 + m21)
/(m12*(m20*m31 - m21*m30) + m02*(m21*m30 - m20*m31) + m00*(m31 - m21)
 + m10*(m21 - m31) + m11*(m30 - m20) + m01*(m20 - m30));
 
B = (m12*(m20*E4 - m30*E3) + m02*(m30*E3 - m20*E4) + m00*(E4 - E3)
 + m10*(E3 - E4) + (m30 - m20)*E2 - m30 + m20)
/(m12*(m20*m31 - m21*m30) + m02*(m21*m30 - m20*m31) + m00*(m31 - m21)
 + m10*(m21 - m31) + m11*(m30 - m20) + m01*(m20 - m30));
 
C = - (m00*(m21*E4 - m31*E3) + m10*(m31*E3 - m21*E4) + m11*(m20*E4 - m30*E3)
 + m01*(m30*E3 - m20*E4) + (m21*m30 - m20*m31)*E2 + m20*m31 - m21*m30)
/(m12*(m20*m31 - m21*m30) + m02*(m21*m30 - m20*m31) + m00*(m31 - m21)
 + m10*(m21 - m31) + m11*(m30 - m20) + m01*(m20 - m30));
 
D = (m00*(m12*(m21*E4 - m31*E3) + m11*(E3 - E4) + (m31 - m21)*E2)
 + m02*(m10*(m31*E3 - m21*E4) + m11*(m20*E4 - m30*E3) + (m21*m30 - m20*m31)*E2)
 + m01*(m12*(m30*E3 - m20*E4) + m10*(E4 - E3) + (m20 - m30)*E2)
 + m12*(m20*m31 - m21*m30) + m10*(m21 - m31) + m11*(m30 - m20))
/(m12*(m20*m31 - m21*m30) + m02*(m21*m30 - m20*m31) + m00*(m31 - m21)
 + m10*(m21 - m31) + m11*(m30 - m20) + m01*(m20 - m30));

m[0][4] = A;
m[1][4] = B;
m[2][4] = C;
m[3][4] = D;
}

void myRiCalculateABCD(double a, double b, double c, double d, double e,
			    double den, double &A, double &B, double &C, double &D)
{
    A = (-b*d -2 + 2*c + a*e - b*e + a*d)/den;
    B = -(2*a*a*e + a*a*d - 3*a - a*b*e +
	  3*a*c + a*b*d - 2*b*b*d - 3*b - b*b*e + 3*b*c)/den;
    C =( -b*b*b*d - 2*b*b*a*e -   b*b*a*d + a*a*b*e +
	2*a*a*b*d - 6*a*b     + 6*b*a*c   + a*a*a*e)/den;
    D = -(   b*b*b - b*b*b*a*d - b*b*a*a*e + b*b*a*a*d
	  -3*a*b*b + b*e*a*a*a - c*a*a*a + 3*c*b*a*a)/den;
}

void Solve3(double m[4][5], double H1, double u1s, double fs, double fp0, double fps){
  double A,B,C,D;
  double a1 = H1;
  double b1 = u1s;
  double fdash0 = fp0;
  double fdashs = fps;
  double den1 = (a1-b1)*(a1-b1)*(a1-b1);
  myRiCalculateABCD(a1,b1,fs,fdash0,fdashs,den1,A,B,C,D);
  m[0][4] = A;
  m[1][4] = B;
  m[2][4] = C;
  m[3][4] = D;
}

void Check(double m[4][5], double a[4]){
  double b[4]={0};
  cerr << "Checking abcd:\n";
  for (int i=0; i<4; i++){
    for (int j=0; j<4; j++){
      b[i] += m[i][j]*a[j];
    }
    cerr << "Res: " << b[i] << "  Act: " << m[i][4]
         << "  Dif: " << m[i][4]-b[i] << "  Rel: " << (m[i][4]-b[i])/m[i][4] << "\n";
  }
}

// Ganked from Preetham code as a temporary fix
inline void RiCalculateABCD(double a, double b, double c, double d, double e,
			    double den, double &A, double &B, double &C, double &D)
{
    A = (-b*d -2 + 2*c + a*e - b*e + a*d)/den;
    B = -(2*a*a*e + a*a*d - 3*a - a*b*e +
	  3*a*c + a*b*d - 2*b*b*d - 3*b - b*b*e + 3*b*c)/den;
    C =( -b*b*b*d - 2*b*b*a*e -   b*b*a*d + a*a*b*e +
	2*a*a*b*d - 6*a*b     + 6*b*a*c   + a*a*a*e)/den;
    D = -(   b*b*b - b*b*b*a*d - b*b*a*a*e + b*b*a*a*d
	  -3*a*b*b + b*e*a*a*a - c*a*a*a + 3*c*b*a*a)/den;
}

// Ganked from Preetham code as a temporary fix
inline double RiHelper1(double A, double B, double C, double D,
			double H, double K, double u)
{
    double t = exp(-K*(H-u));
    return (t/K)*((A*u*u*u + B*u*u + C*u +D) -
		  (3*A*u*u + 2*B*u + C)/K +
		  (6*A*u + 2*B)/(K*K) -
		  (6*A)/(K*K*K));
}

SbColor Sky::SolveInScatter(double cosT,
                            double a_haze, const SbColor &K1, double H1, double u1s, const SbColor &t_haze, const SbVec3f &B_p,
                            double a_mol,  const SbColor &K2, double H2, double u2s, const SbColor &t_mol, const SbVec3f &B_m){

  SbColor I1;
  
  for (int i=0; i<3; i++){

    // h(x) = h0+x*cos(theta)
    // B(h(x))= B0*exp(-a*h) = B0*u(x)
    // K = B0/(alpha*cosT)
    // H = exp(-alpha*h0)
    // tau = exp(-K*(H-u(s)))
    // u(x)  = exp(-a*(h0+x*cos(theta)))
    // u'(x) = exp(-a*(h0+x*cos(theta))) * -a*cos(theta)
    // v=u1(x)
    // f(x)  = exp(-K2(H2-u2(x)))
    // f'(x) = -K2*-u2'(x) * exp(-K2(H2-u2(x)))
    // f'(x) = -K2*-  (exp(-a_2*(h0+x*cos(theta))) * -a_2*cos(theta))  * exp(-K2(H2-u2(x)))
    //
    
    double u20 = H2;
    
    double fs  = t_mol[i];
//     double fp0 = -K2[i]* -(u20 * -a_mol*cosT); // * exp(-K2[i]*(H2-u20));
//     double fps = -K2[i]* -(u2s * -a_mol*cosT) * fs;
//     double fp0 = K2[i]* (u20 * -a_mol*cosT); // * exp(-K2[i]*(H2-u20));
//     double fps = K2[i]* (u2s * -a_mol*cosT) * fs;


    // Here I give up and just use the Preetham code instead of mine.
    // At some point I should figure out why mine fails.
    double fp0 = -B_m[i]*H2;
    double fps = -fs*B_m[i]*u2s;
    double a1 = H1;
    double b1 = u1s;
    double u_f1 = b1;
    double u_i1 = H1;
    double a2 = H2;
    double b2 = u2s;
    double fdash0 = fp0;
    double fdashs = fps;
    double den1 = (a1-b1)*(a1-b1)*(a1-b1);
    double den2 = (a2-b2)*(a2-b2)*(a2-b2);
    double K = B_p[i]/(a_haze*cosT);
    
    double A,B,C,D;
    RiCalculateABCD(a1,b1,fs,fdash0,fdashs,den1,A,B,C,D);
    double int_f = RiHelper1(A,B,C,D,H1,K, u_f1);
    double int_i = RiHelper1(A,B,C,D,H1,K, u_i1);
    I1[i] = (int_f - int_i)/(-a_haze*cosT);
    continue;

    // The rest is my (broken) solution
    
    double m[4][5] = { // row, col
      { H1*H1*H1,      H1*H1,  H1, 1,   1   },
      { u1s*u1s*u1s, u1s*u1s, u1s, 1,   fs  },
      { 3*H1*H1,        2*H1,   1, 0,   fp0 },
      { 3*u1s*u1s,     2*u1s,   1, 0,   fps }
    };

    // copy m
    double origm[4][5];
    for (int a=0; a<4; a++)
      for (int b=0; b<5; b++)
        origm[a][b] = m[a][b];

    //solve4x4(m);
    //Solve2(m);
    Solve3(m, H1, u1s, fs, fp0, fps);
    double abcd[4] = { m[0][4], m[1][4], m[2][4], m[3][4] };
    //Check(origm, abcd);

     I1[i] = -1/(a_haze*cosT)*(
      t_haze[i]*(g_v(u1s,abcd)/K1[i] - gp_v(u1s,abcd)/(K1[i]*K1[i]) +
              gpp_v(u1s,abcd)/(K1[i]*K1[i]*K1[i]) - gppp_v(u1s,abcd)/(K1[i]*K1[i]*K1[i]*K1[i])) -
      (g_v(H1,abcd)/K1[i] - gp_v(H1,abcd)/(K1[i]*K1[i]) +
       gpp_v(H1,abcd)/(K1[i]*K1[i]*K1[i]) - gppp_v(H1,abcd)/(K1[i]*K1[i]*K1[i]*K1[i]) ) );

     if (I1[i]<0 && DEBUG_STATE)
       cerr << "wtf";
  }

  return I1;
}

void Sky::ComputeAttenAndInScatter(double h0, double cosT, double phi, double t,
                                   SbColor &atten, SbColor &inscatter){

//   float minCos = 1e-6;
//   if (cosT>=0){
//     if (cosT<minCos)
//       cosT = minCos;
//   }
//   else {
//     if (cosT>-minCos)
//       cosT = -minCos;
//   }

  // haze = _p _1
  double a_haze = .8333 * 1e-3; // convert km^-1 to m^-1
  SbVec3f K1 = B_p_0/(a_haze*cosT);
  double H1 = exp(-a_haze*h0);
  double u1s = exp(-a_haze*(h0+t*cosT));
  SbColor t_haze = vecExp(-K1*(H1-u1s));

  // gas molecules = _m _2
  float a_mol = .1136 * 1e-3; // convert km^-1 to m^-1
  SbVec3f K2 = B_m_0/(a_mol*cosT);
  double H2 = exp(-a_mol*h0);
  double u2s = exp(-a_mol*(h0+t*cosT));
  SbColor t_mol = vecExp(-K2*(H2-u2s));

//  atten=SbColor(1,1,1);
//  atten=t_haze;
//  atten=t_mol;  
  atten = vecScale(t_haze, t_mol);

  SbColor S_0_m_i, S_0_p_i;

  // compute S_0_m,p ti, pi
  double ti0d = (1-cosT)/2*ti_S0pre;
  int ti0 = (int)ti0d;
  int ti1 = (ti0+1)%(ti_S0pre+1);
  double ta = ti0d-ti0;
    
  if (phi<0) phi += 2*M_PI;
  double pi0d = phi/(2*M_PI)*pi_S0pre;
  int pi0 = (int)pi0d;
  int pi1 = (pi0+1)%pi_S0pre;
  double pa = pi0d-pi0;

  // bilinear interpolate from precomputed cache
  S_0_p_i =
    S_0_p[ti0*pi_S0pre+pi0]*(1-ta)*(1-pa) + S_0_p[ti1*pi_S0pre+pi0]*ta*(1-pa) +
    S_0_p[ti0*pi_S0pre+pi1]*(1-ta)*pa     + S_0_p[ti1*pi_S0pre+pi1]*ta*pa;

  S_0_m_i =
    S_0_m[ti0*pi_S0pre+pi0]*(1-ta)*(1-pa) + S_0_m[ti1*pi_S0pre+pi0]*ta*(1-pa) +
    S_0_m[ti0*pi_S0pre+pi1]*(1-ta)*pa     + S_0_m[ti1*pi_S0pre+pi1]*ta*pa;

  static bool skyApprox = !getenv("PANE_SKY_NO_APPROX");
//  if (0){
//  if (1){
//  if (skyApprox && fabs(a_haze*cosT*t)<0.3){
  if (skyApprox && fabs(cosT*t)<0.0001){

    const SbColor one1(1,1,1);
    SbColor I_haze = -H1* //exp(-H1) *   // paper shows wrong result
      vecDiv(one1 - vecExp(-(one1*a_haze*cosT + B_p_0*H1 + B_m_0*H2)*t),
             one1*a_haze*cosT + B_p_0*H1 + B_m_0*H2);

    SbColor I_mol = -H2 * //exp(-H2) *   // paper shows wrong result
      vecDiv(one1 - vecExp(-(one1*a_mol*cosT + B_p_0*H1 + B_m_0*H2)*t),
             one1*a_mol*cosT + B_p_0*H1 + B_m_0*H2);

//    inscatter = vecScale(S_0_p_i, I_haze);
//    inscatter = vecScale(S_0_p_i, I_mol);
    inscatter = vecScale(S_0_p_i, I_haze) + vecScale(S_0_m_i, I_mol);
  }
  else {


    SbColor I1 = SolveInScatter(cosT,
                                a_haze, K1, H1, u1s, t_haze, B_p_0,
                                a_mol,  K2, H2, u2s, t_mol, B_m_0); // haze (_p)
    SbColor I2 = SolveInScatter(cosT,
                                a_mol,  K2, H2, u2s, t_mol, B_m_0,
                                a_haze, K1, H1, u1s, t_haze, B_p_0); // gas
                                                                          // (_m)

//    inscatter = SbColor(0,0,0);
//    inscatter = vecScale(S_0_p_i, I1);
//    inscatter = vecScale(S_0_m_i, I2);    
    inscatter = vecScale(S_0_p_i, I1) + vecScale(S_0_m_i, I2);
  }

}

void Sky::ComputeInScatteringForOneDir(const SbVec3f &out,
                                       SbVec3f &S_0_m_theta, SbVec3f &S_0_p_theta) {

  double refrIdx_air = 1.0003;
  double a = (refrIdx_air*refrIdx_air-1);
  double N = 2.545e25;
  double p_n = 0.035;
  
  double c = (0.6544*turbidity-0.6510)*1e-16;
  double v = 4;

  S_0_m_theta = S_0_p_theta = SbVec3f(0,0,0);

  Spectrum whiteSpec(6500);
  whiteSpec.scaleSpec(1/4.73223e+06); // normalize by red channel
  
  // add sky light in-scattering
  // integrate by sampling midpoints of theta/phi subdivision
  for (int ti=0; ti<ti_Lpre; ti++){    // theta
    for (int pi=0; pi<pi_Lpre; pi++){  // phi

      // find theta as midpoint of delta theta region ti [ti,ti+1]
      float costheta_a = 1-ti/float(ti_Lpre);
      float costheta_b = 1-(ti+1)/float(ti_Lpre);
//       float costheta_a = 1-2*ti/float(ti_Lpre);
//       float costheta_b = 1-2*(ti+1)/float(ti_Lpre);
      float theta_a = acosf(costheta_a);
      float theta_b = acosf(costheta_b);
      float theta_m = (theta_a+theta_b)*0.5; // midpoint
      float dtheta = theta_b - theta_a;

      // find phi as midpoint of delta phi region ti [ti,ti+1]
      float phi_a = pi/float(pi_Lpre)*M_PI*2;
      float phi_b = (pi+1)/float(pi_Lpre)*M_PI*2;
      float phi_m = (phi_a+phi_b)*0.5; // midpoint
      float dphi = phi_b - phi_a;

      float dw = sin(theta_m)*dtheta*dphi;

      SbVec3f dir(
 	sin(theta_m)*cos(phi_m),
 	cos(theta_m),
 	sin(theta_m)*sin(phi_m));

      float bcostheta = out.dot(dir);
      float btheta = acosf(bcostheta);

      // rotate: this is to counteract matrix mult in getRadiance()...wasteful
      textureToWorldRot.multDirMatrix(dir, dir);
      
      SbColor Ls = getRadiance(dir, false);

      Spectrum col1, col2;
      col1 = col2 = whiteSpec;

      int num = Spectrum::getNumWavelens();
      float max = Spectrum::getMaxWavelen();
      float min = Spectrum::getMinWavelen();
      for (int i=0; i<num; i++){
        float wl = (max-min) * i/(float)(num-1) + min;
        wl *= 1e-9; // convert nm to m

//         float B_m_theta =            // gas molecules (Rayleigh)
//           M_PI*M_PI*a*a/(2*N*wl*wl*wl*wl) *
// //          (6+3*p_n)/(6-7*p_n)*(1+bcostheta*bcostheta);
//             (6+3*p_n)/(6-7*p_n)*(1+.9324*bcostheta*bcostheta);

        // from Preetham CODE (not paper)
        float B_m_theta =            // gas molecules (Rayleigh)
          2*M_PI*M_PI*a*a/(3*N*wl*wl*wl*wl) *.7629*
//          (6+3*p_n)/(6-7*p_n)*(1+bcostheta*bcostheta);
            (6+3*p_n)/(6-7*p_n)*(1+.9324*bcostheta*bcostheta);
        col1.scaleAmp(i, B_m_theta);

        Spectrum &neta_theta = neta[int(btheta*180/M_PI*10)];
          
        float B_p_theta =            // haze (Mie)
          0.434*c*pow(2*M_PI/wl,v-2)*0.5*neta_theta.getAmp(i);
        B_p_theta *= pow(.01, v-3); // WTF! Corrective factor in code but not paper!
        col2.scaleAmp(i, B_p_theta);
      }

      SbColor prod_m = vecScale(Ls, col1.getRGB()); // Rayleigh (gas)
//      SbColor prod_m =  col1.getRGB();
      S_0_m_theta += prod_m*dw;
      
      SbColor prod_p = vecScale(Ls, col2.getRGB()); // Mie (haze)
//      SbColor prod_p = col2.getRGB();
      S_0_p_theta += prod_p*dw;
    }
  }

  // add sun in-scattering
  Spectrum col1(6500), col2(6500);
  col1.scaleSpec(1/4.73223e+06); // normalize by red channel
  col2.scaleSpec(1/4.73223e+06); // normalize by red channel

  float bcostheta = out.dot(sunBasisW);
  float btheta = acosf(bcostheta);
      
  int num = Spectrum::getNumWavelens();
  float max = Spectrum::getMaxWavelen();
  float min = Spectrum::getMinWavelen();
  for (int i=0; i<num; i++){
    float wl = (max-min) * i/(float)(num-1) + min;
    wl *= 1e-9; // convert nm to m

//     float B_m_theta =            // gas molecules (Rayleigh)
//       M_PI*M_PI*a*a/(2*N*wl*wl*wl*wl) *
// //      (6+3*p_n)/(6-7*p_n)*(1+bcostheta*bcostheta);
//       (6+3*p_n)/(6-7*p_n)*(1+.9324*bcostheta*bcostheta);      
//     col1.scaleAmp(i, B_m_theta);
    
    // from Preetham CODE (not paper)
    float B_m_theta =            // gas molecules (Rayleigh)
      2*M_PI*M_PI*a*a/(3*N*wl*wl*wl*wl) * .7629 *
//      (6+3*p_n)/(6-7*p_n)*(1+bcostheta*bcostheta);
      (6+3*p_n)/(6-7*p_n)*(1+.9324*bcostheta*bcostheta);      
    col1.scaleAmp(i, B_m_theta);

    Spectrum &neta_theta = neta[int(btheta*180/M_PI*10)];
          
    float B_p_theta =            // haze (Mie)
      0.434*c*pow(2*M_PI/wl,v-2)*0.5*neta_theta.getAmp(i);
    B_p_theta *= pow(.01, v-3); // WTF! Corrective factor in code but not paper!
    col2.scaleAmp(i, B_p_theta);
  }

  SbColor prod_m = vecScale(sunRadiance, col1.getRGB()); // Rayleigh (gas)
  S_0_m_theta += prod_m*sunOmega;
      
  SbColor prod_p = vecScale(sunRadiance, col2.getRGB()); // Mie (haze)
  S_0_p_theta += prod_p*sunOmega;
}
    
