#include "Main.h"

#ifndef SPECTRUMCLASS
#define SPECTRUMCLASS

class Spectrum {

  protected:
  
  static const int numWavelens = 90;
  static const float minWavelen;
  static const float maxWavelen;
  static const float dy;
  static const float dyInv;
  static const float ciecols[numWavelens][3];
  // from http://white.stanford.edu/~brian/scielab/scielab1-1-1/cmatrix.m
  // by Xuemei Zhang
  static const SbMatrix xyz2lms;
  static SbColor maxLMS;

  float powerDist[numWavelens]; // W/(m^2 sr nm)  (SI units)
  
  public:

  Spectrum(){}
  Spectrum(float temp){ setBlackbody(temp); }
  Spectrum(float y[], int n, float start, float stop) {
    setArray(y, n, start, stop);
  }
  Spectrum(float x[], float y[], int n) {
    setArray(x, y, n);
  }
  void setBlackbody(float temp);
  void setArray(float y[], int n, float start, float stop);
  void setArray(float x[], float y[], int n);
  bool dump(const string &file); // returns true on success
  void scaleSpec(float s);
  void scaleSpec(Spectrum s){
    for (int i=0; i<numWavelens; i++)
      powerDist[i] *= s.powerDist[i];
  }
  void scaleAmp(int i, float s){ powerDist[i] *= s; }
  float getAmp(int i) const { return powerDist[i]; }

  // helper function, converts wavelength into indexes i0, i1, alpha
  static bool getiy(float lambda, int &i, int &i1, float &a){
    //     |  |       |
    //  |  |  |       |  |  |
    //  +--+--+-- ... +--+--+
    //  380                 825
    //
    //  Each powerDist[i] is measured value (density) at that wavelength
    //  For the power in a band you have to interpolate (average) neighbors
    //  and multiply by width of band. It's a density function, not a
    //  histogram. A histogram is a sum per band which is basically the same
    //  as the density but multiplied by the band-width.
    
    if (lambda<minWavelen || lambda>maxWavelen)
      return false;

    float x = lambda-minWavelen;
    i = (int)(x*dyInv);
    i1 = std::min(i+1, numWavelens-1);
    float y = x-dy*i;
    a = y*dyInv;
    return true;
  }

  static float getWavelen(int i, float a){
    return minWavelen+(i+a)*dy;
  }
  
  float getVal(float lambda) const { // wl is in nm
    float a;
    int i, i1;
    if (!getiy(lambda, i, i1, a))
      return 0;

    return powerDist[i]*(1.0f-a) + powerDist[i1]*a;
  }

  static int getNumWavelens() { return numWavelens; }
  static float getMinWavelen() { return minWavelen; }
  static float getMaxWavelen() { return maxWavelen; }
  static float getRandomWavelen(float eps) {
    return eps*(maxWavelen-minWavelen)+minWavelen;
  }
  
  friend ostream& operator<<(ostream &os, const Spectrum &s);
  
  static SbColor cieResponse(float lambda) {
    float a;
    int i, i1;
    if (!getiy(lambda, i, i1, a))
      return SbColor(0.0, 0.0, 0.0);

    return (SbColor(ciecols[i][0],  ciecols[i][1],  ciecols[i][2]) *(1.0f-a) +
	    SbColor(ciecols[i1][0], ciecols[i1][1], ciecols[i1][2])*a);
  }
  
  SbColor getXYZ(){
    SbColor c(0,0,0);
    for (int i=0; i<numWavelens; i++){
      c[0] += ciecols[i][0]*powerDist[i];
      c[1] += ciecols[i][1]*powerDist[i];
      c[2] += ciecols[i][2]*powerDist[i];
    }

    return c*dy;
  }

  static void setMaxLMS(float temp){
    Spectrum s(temp);
    maxLMS = s.getLMS();
  }
  
  SbColor getLMS(){
    SbColor xyz = getXYZ();
    xyz2lms.multMatrixVec(xyz, xyz);
    return xyz;
  }

  SbColor getAdaptedXYZ(){

    // from Nguyen at al.
    SbVec3f invMaxLMS(
      maxLMS[0] ? 1/maxLMS[0] : 0,
      maxLMS[1] ? 1/maxLMS[1] : 0,
      maxLMS[2] ? 1/maxLMS[2] : 0
      );
    SbMatrix diag(
      invMaxLMS[0], 0, 0, 0,
      0, invMaxLMS[1], 0, 0,
      0, 0, invMaxLMS[2], 0,
      0, 0, 0, 1
      );

    SbColor xyz = getXYZ();

    // save luminance
    float lum = xyz[1];

    // get adapted XYZ
    (xyz2lms.inverse()*diag*xyz2lms).multMatrixVec(xyz, xyz);

    // scale new XYZ to old luminance
/*     float W = xyz[0]+xyz[1]+xyz[2];   // convert to Lxy */
/*     float x = xyz[0]/W;               // from RRT v1 by Peter Shirley */
/*     float y = xyz[2]/W; */
/*     xyz[0] = x*lum/y;                 // convert back to XYZ */
/*     xyz[1] = lum; */
/*     xyz[2] = (1-x-y)*lum/y; */
    
    return xyz;
  }
    
  static SbColor xyzToRgb(SbColor xyz) {

    SbColor rgb;
    
    // sRGB standard, http://www.w3.org/Graphics/Color/sRGB
    // physical sRGB standard, based on Yw=110.5
    rgb[0] =  3.2410*xyz[0] +  -1.5374*xyz[1] +  -0.4986*xyz[2];
    rgb[1] = -0.9692*xyz[0] +   1.8760*xyz[1] +   0.0416*xyz[2];
    rgb[2] =  0.0556*xyz[0] +  -0.2040*xyz[1] +   1.0570*xyz[2];
    if (rgb[0]<0) rgb[0] = 0;
    if (rgb[1]<0) rgb[1] = 0;
    if (rgb[2]<0) rgb[2] = 0;
    return (rgb);///110.5f);

    // inverse of RGB2XYZ in Pane.cxx
    // use to convert XYZ -> to my RGB (i.e., inv([xb,yb,zy]*d))
    const double inv[3][3] =
      {
	{  2.4383,   -1.1566,   -0.3752 },
	{ -0.7290,    1.4113,    0.0312 },
	{  0.0419,   -0.1535,    0.7953 }
      };
    rgb[0] = inv[0][0]*xyz[0] + inv[0][1]*xyz[1] + inv[0][2]*xyz[2];
    rgb[1] = inv[1][0]*xyz[0] + inv[1][1]*xyz[1] + inv[1][2]*xyz[2];
    rgb[2] = inv[2][0]*xyz[0] + inv[2][1]*xyz[1] + inv[2][2]*xyz[2];
    return rgb;
  }

  static SbColor YxyToRgb(float Y, float x, float y) {
    SbColor xyz;
    xyz[0] = x*Y/y;
    xyz[1] = Y;
    xyz[2] = (1-x-y)*Y/y;
    return xyzToRgb(xyz);
  }
  
  SbColor getRGB(){
    return xyzToRgb(getXYZ());
  }
    
  SbColor getAdaptedRGB(){
    return xyzToRgb(getAdaptedXYZ());
  }

  static void getSpectralMult(double eps, SbVec3f &mult, float &wavelen);
  
};

inline ostream& operator<<(ostream &os, const Spectrum &s){
  for (int i=0; i<Spectrum::numWavelens; i++){
    os << s.powerDist[i] << " ";
  }
  return os;
}

class SpecSampler : public Spectrum {
public:
  SpecSampler();
  void sampleWavelen(float eps, float &wavelen, float &prob);
};

#endif /* SPECTRUMCLASS */
