#include <stdlib.h>
#include "Sphere.h"
#include "debug.h"

SphereSet Sphere::defaultSphereSet;
SphereSet Sphere::defaultTextureSphereSet;
SphereSet Sphere::unitSphereSet;
SphereSet Sphere::unitTextureSphereSet;
Material  Sphere::defaultMat;
SbVec3f   Sphere::origin(0,0,0);

bool Sphere::intersectRay(Ray &r, double t0, double t1){
  
  if (r.ignorePrim == this)
    t0 = 1e-2; // expand tolerance
  
  r.t = -1;
  r.prim = NULL;

  Vec3d oc = Vec3d(r.pos) - Vec3d(pos());
  Vec3d dd = r.dir;
  double ddoc = dd.dot(oc);
  double dd2 = dd.dot(dd);
  double det = ddoc*ddoc - dd2*(oc.dot(oc)-rad()*rad());
  
  if (det<0) return false;
  det = sqrt(det);
  double t = (-ddoc - det); // /dd2;
  if (t>t1) return false;
  if (t>=t0){
    r.t = t;
    r.prim = this;
    r.bc = r.pos+r.dir*r.t;
    return true;
  }
  t = (-ddoc + det);  // /dd2;
  if (t>=t0 && t<=t1) {
    r.t = t;
    r.prim = this;
    r.bc = r.pos+r.dir*r.t;
    return true;
  }
    
  return false;
}

SbColor Sphere::getIntDiffColor(const Ray &r) const {
  const SbColor &color = col();
  const SbVec3f &bc = r.bc;
  
  const Texture *text = texture();
  if (text && text->bumpMap !=2){
    float s, t;
    getSTfromBC(bc, s, t);
    return text->getColor(s, t, color);
  }

  return color;
}

SbVec3f Sphere::getIntNormal(const SbVec3f &bc) const {
  SbVec3f n = getNormal(bc);

  // do bump mapping?
  const Texture *text = texture();
  if (! (text && text->bumpMap))
    return n;

  // get geometric derivs if we don't have them already
  PrimDerivs myPrimDerivs;
  PrimDerivs *pDer;
//  if (!pDer){
  if (!getDerivs(bc, myPrimDerivs))
    return n;
  pDer = &myPrimDerivs;
//  }

  float s, t;
  getSTfromBC(bc, s, t);
  SbVec3f dh = text->getDeriv(s,t); // returns dh/du, dh/dv, h
  float dhdu = dh[0];        // either of these might be zero
  float dhdv = dh[1];
  float h0 = dh[2];
    
  SbVec3f dppdu = pDer->dPdU + text->bumpScale*dhdu*n + h0*pDer->dNdU;
  SbVec3f dppdv = pDer->dPdV + text->bumpScale*dhdv*n + h0*pDer->dNdV;

  SbVec3f n2 = dppdu.cross(dppdv);
  n2.normalize();
    
  return n2;
}

bool Sphere::getDerivs(const SbVec3f &bc, PrimDerivs &pDer) const {
  SbVec3f r = bc-pos();
  pDer.dPdU = SbVec3f(0,1,0).cross(r);
  pDer.dPdU.normalize();
  pDer.dPdV = r.cross(pDer.dPdU);
  pDer.dPdV.normalize();
  pDer.dNdV = pDer.dNdU = SbVec3f(0,0,0);
  return true;
}

void Sphere::getPhoton(double eps1, double eps2,
                       double eps3, double eps4,
                       SbVec3f &position, SbVec3f &dir, SbColor &emission){

  // find emission point
  float phi_e = eps2*(float)(2*M_PI);
  float z = clamp((float)(eps1*2.f-1.f), -1.f, 1.f);
  position[0] = sqrtf(1-z*z)*cosf(phi_e);
  position[1] = sqrtf(1-z*z)*sinf(phi_e);
  position[2] = z;
  position.normalize();
  ONB onb(position);
  position *= rad();
  position += pos();
  
  // phong dist from RRT pg. 98
  static float lightExp = getenv("PANE_LIGHT_EXP") ? atof(getenv("PANE_LIGHT_EXP")) : 1;
  float phi = eps3 * (float)(2*M_PI);                   // rotation about normal
  float theta = acosf(powf(eps4, 1.f/(lightExp+1.f)));  // angle from normal
  SbVec3f hemiDir;
  hemiDir.setValue(sinf(theta)*cosf(phi),
                   sinf(theta)*sinf(phi),
                   cosf(theta));
  dir = onb.toWorldSpace(hemiDir);

  emission = getMat()->getEmission();
}


// Todo:
//
// sphereLightContrib:
//   MIS
//   reducing number of samples based on solid angle
//
// getPhoton:
//   projection maps / rejection sampling using angle/costheta
//
// global photon map doesn't need to allocate same as caustic if num is less.
// (grep for "max_photons")
//
// Octree:
//   do not increase global size


void Sphere::getPhotonImpSample(double eps1, double eps2,
                                double eps3, double eps4,
                                Primitive *prim, bool shiny,
                                SbVec3f &position, SbVec3f &dir,
                                SbColor &emission, float &scale){

  // get scene center, radius
  SbVec3f center;
  float radius;
  if (shiny){
    center = prim->getCenterShiny();
    radius = prim->getRadiusShiny();
  } else {
    center = prim->getCenter();
    radius = prim->getRadius();
  }
  SbVec3f w = center-pos();
  float w2 = w.dot(w);
  float radius2 = radius*radius;
  float dist = radius+rad();
  if (w2<dist*dist) {
    static bool warned = false;
    if (!warned){
      warned = true;
      cerr << "\nWarning: Sphere: Importance sampling cannot be done because emissive sphere\n"
           << "is too close to scene center. Resorting to regular sampling.\n";
    }

    getPhoton(eps1, eps2, eps3, eps4, position, dir, emission);
    scale = 1;
    return;
  }
  
  // find emission point
  float phi_e = eps2*(float)(2*M_PI);
  float z = clamp((float)(eps1*2.f-1.f), -1.f, 1.f);
  position[0] = sqrtf(1-z*z)*cosf(phi_e);
  position[1] = sqrtf(1-z*z)*sinf(phi_e);
  position[2] = z;
  position.normalize();
  SbVec3f norm = position;
  position *= rad();
  position += pos();

  // compute light direction, l
  ONB onb(w);
  float cos_a_max = sqrtf(1 - radius2/w2);
  float frac = 1-cos_a_max;
  float omega = 2*M_PI*frac;  // exact solid angle, computed analytically

  float cos_a = 1-eps3+eps3*cos_a_max;
  float sin_a = sqrtf(1-cos_a*cos_a);
  float phi = (float)(2*M_PI)*eps4;
  SbVec3f hemiDir(cosf(phi)*sin_a,
                  sinf(phi)*sin_a,
                  cos_a);
  dir = onb.toWorldSpace(hemiDir);
  dir.normalize();
  
  emission = getMat()->getEmission();
  float pdf = 1.f/omega;

  scale = dir.dot(norm);
  
  if (scale<=0){
    scale = 0;
    return;
  }

  scale /= M_PI*pdf; // divide by pdf and scale by (now missing) cos(theta)/pi
}

