//WWWWWW'''''''WW''''''''''''''''WWW'''''''''''''''WWW''''''''''''''''''
//  WW                            WW                WW                  
//  WW    W WW WW  WWW WWW   WWW  WW   WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WW   WW W WW WW W W WW  WW  WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW   WW W WW WW W W WW  WW  WW      WW    WW  W WW   WWW   WWW
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//                            WW                                        
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''

#include "GlobalVars.h"
#include "Pane.h"
#include "Triangle.h"
#include "Ray.h"
#include "Texture.h"
#include "debug.h"
#include "Noise.h"

using namespace std;

vector<TriMesh*> TriMesh::meshs;

int Triangle::errorTris = 0;
float Triangle::emissiveArea = 0;

Triangle::Triangle() {
  init();
}

void Triangle::init() {
  mesh = NULL;
  n = -1;
}

void Triangle::createTriangle(Pane *pane, SoCallbackAction *action,
                              const SoPrimitiveVertex *v0_,
                              const SoPrimitiveVertex *v1_,
                              const SoPrimitiveVertex *v2_,
                              bool &okay){

  init();
  mesh = new TriMesh();
  TriMesh::meshs.push_back(mesh);
  n = mesh->add();

  // dont render/include invisible objects
  okay = true;
  if (action->getDrawStyle() != SoDrawStyle::FILLED){
    okay = false;
    return;
  }

  // reverse vertices
  const SoPrimitiveVertex *v0 = v0_, *v1 = v1_, *v2 = v2_;
  if (action->getVertexOrdering() == SoShapeHints::CLOCKWISE){
    v0 = v2_;
    v1 = v1_;
    v2 = v0_;
  }

  SbVec3f &vert0 = vert(0);
  SbVec3f &vert1 = vert(1);
  SbVec3f &vert2 = vert(2);
  
  // translate and store vertices
  SbMatrix matrix = action->getModelMatrix();
  matrix.multVecMatrix(v0->getPoint(), vert0);
  matrix.multVecMatrix(v1->getPoint(), vert1);
  matrix.multVecMatrix(v2->getPoint(), vert2);

  // calculate edges
  SbVec3f edge1 = vert1-vert0;
  SbVec3f edge2 = vert2-vert0;

  // calculate area
  float area = 0.5*(edge1.cross(edge2)).length();

  if (area <= 0.0){
    if (errorTris++ == 0)
      cerr << "\n";
    static int num = 1;
    if (errorTris == num){
      cerr << "\rWarning! " << errorTris
           << " triangle(s) with 0 area!       ";
      num *= 10;
    }
    okay = false;
    return;
  }

  SbVec3f &normals0 = normals(0);
  SbVec3f &normals1 = normals(1);
  SbVec3f &normals2 = normals(2);
  
  // calculate normals
  SbMatrix matInvTrans = matrix.inverse().transpose();
  matInvTrans.multDirMatrix(v0->getNormal(), normals0);
  matInvTrans.multDirMatrix(v1->getNormal(), normals1);
  matInvTrans.multDirMatrix(v2->getNormal(), normals2);
  normals0.normalize();
  normals1.normalize();
  normals2.normalize();

  // set geometric normal
  SbVec3f normal = getNormal();

  // don't let there be zero normals
  if (normals0.dot(normals0) < 0.9)
    normals0 = normal;
  if (normals1.dot(normals1) < 0.9)
    normals1 = normal;
  if (normals2.dot(normals2) < 0.9)
    normals2 = normal;

  // dont let n.ng < 0
  float dot[3];
  dot[0] = normals0.dot(normal);
  dot[1] = normals1.dot(normal);
  dot[2] = normals2.dot(normal);
  if (dot[0]<0.0 || dot[1]<0.0 || dot[2]<0.0){
    
    // flip normals
    if (dot[0]<0.0 && dot[1]<0.0 && dot[2]<0.0){
      normals0 = -normals0;
      normals1 = -normals1;
      normals2 = -normals2;
    }
    else {
      // push negative normals up until they're flat
      for (int i=0; i<3; i++)
	if (dot[i]<0.0){
          SbVec3f &normalsi = normals(i);
	  normalsi += (0.0015-dot[i]) * normal;
	  normalsi.normalize();
	}
    }
  }

  // Get material
  Material *mat = pane->getCurrentMat();
  mesh->mat = mat;
  //cerr << "Triangle material:" << *mat << "\n";
  SbColor ambient, specular, emission;
  float shininess, transparency;
  action->getMaterial(ambient, diffuse(0), specular, emission,
		      shininess, transparency,
		      v0->getMaterialIndex());
  action->getMaterial(ambient, diffuse(1), specular, emission,
		      shininess, transparency,
		      v1->getMaterialIndex());
  action->getMaterial(ambient, diffuse(2), specular, emission,
		      shininess, transparency,
		      v2->getMaterialIndex());

  if (mat->isShiny())
    pane->gv->sceneHasShininess = 1;
  
  // bad color?
  for (int i=0; i<3; i++) {
    SbVec3f &diffusei = diffuse(i);
    for (int j=0; j<3; j++)
      if (!finite(diffusei[j])){
	//cerr << "BAD COLOR VALUE!\n";
	diffusei = SbColor(1.0, 0.0, 0.0);
      }
  }

  // temporary reflection capping
  const static bool reflCap = (bool)getenv("PANE_REFLECTION_CAP");
  if (reflCap){
    for (int i=0; i<3; i++)
      diffuse(i) = BBox::vecMin(diffuse(i), SbVec3f(.9,.9,.9));
  }

  const static bool reflScale = (bool)getenv("PANE_REFLECTION_SCALE");
  if (reflScale){
    for (int i=0; i<3; i++)
      diffuse(i) *= .8;
  }
  
  // get texturing info
  int numComps;
  SbVec2s size;
  const unsigned char *img = action->getTextureImage(size, numComps);
  
  if (img){

    // only works for rgb textures
    if (numComps == 3 || numComps == 1){
      
      Texture *texture = new Texture(action->getTextureWrapS(),
			 action->getTextureWrapT(),
			 action->getTextureModel(),
			 action->getTextureBlendColor(),
			 numComps, size, img,
			 pane->gv->bumpMapping,
			 pane->gv->bumpScale);

      SbMatrix tmatrix = action->getTextureMatrix();
      SbVec3f t3;
      SbVec4f t4;
      SbVec2f uv;
      mesh->texture = texture;
      mesh->uvs.reserve(3);

      t4 = v0->getTextureCoords();
      t3.setValue(t4[0], t4[1], 0.0);
      tmatrix.multVecMatrix(t3, t3);
      uv = SbVec2f(t3[0], t3[1]);
      mesh->uvs.push_back(uv);

      t4 = v1->getTextureCoords();
      t3.setValue(t4[0], t4[1], 0.0);
      tmatrix.multVecMatrix(t3, t3);
      texture->stex[1] = t3[0];
      texture->ttex[1] = t3[1];
      uv = SbVec2f(t3[0], t3[1]);
      mesh->uvs.push_back(uv);

      t4 = v2->getTextureCoords();
      t3.setValue(t4[0], t4[1], 0.0);
      tmatrix.multVecMatrix(t3, t3);
      texture->stex[2] = t3[0];
      texture->ttex[2] = t3[1];
      uv = SbVec2f(t3[0], t3[1]);
      mesh->uvs.push_back(uv);
    }
    else {
      cerr << "Illegal number of components: " << numComps << endl;
      exit(1);
    }
  }

}

Triangle::~Triangle(){
//   static int x=0;
//   x++;
}

SbVec3f Triangle::getNormal() const {

    // calculate edges
  const SbVec3f &vert0 = vert(0);
  SbVec3f edge1 = vert(1)-vert0;
  SbVec3f edge2 = vert(2)-vert0;

  // calculate normal
  SbVec3f normal = edge1.cross(edge2);
  normal.normalize();
  return(normal);
}
  

// calculate barycentric coordinates t
void Triangle::calcBaryCoords(const SbVec3f &pnt, SbVec3f &bc) const{
    
  SbVec3f u, v, w; // basis for plane of triangle
  SbVec3f x, y, z; // coordinates of tri verts wrt above basis
  SbVec3f d;       // coordinates of center
  SbVec3f t;       // barycentric coordinates

  const SbVec3f &vert0 = vert(0);
  SbVec3f edge1 = vert(1)-vert0;
  SbVec3f edge2 = vert(2)-vert0;
  
  u = edge1;
  u.normalize();

  v = edge2;
  v = v - v.dot(u)*u;
  v.normalize();

  x = SbVec3f(0.0, 0.0, 0.0);
  y = SbVec3f(u.dot(edge1), v.dot(edge1), 0.0);
  z = SbVec3f(u.dot(edge2), v.dot(edge2), 0.0);
  d = SbVec3f(u.dot(pnt-vert0), v.dot(pnt-vert0), 0.0);

  // meat of the algorithm. taken from a post by David Watson:
  // http://research.microsoft.com/~hollasch/cgindex/math/barycentric.html
  float b0;
  b0    =   (y[X]-x[X]) * (z[Y]-x[Y]) - (z[X]-x[X]) * (y[Y]-x[Y]);
  bc[0] = ( (y[X]-d[X]) * (z[Y]-d[Y]) - (z[X]-d[X]) * (y[Y]-d[Y]) ) / b0;
  bc[1] = ( (z[X]-d[X]) * (x[Y]-d[Y]) - (x[X]-d[X]) * (z[Y]-d[Y]) ) / b0;
  bc[2] = ( (x[X]-d[X]) * (y[Y]-d[Y]) - (y[X]-d[X]) * (x[Y]-d[Y]) ) / b0;
}

// or let user specify them
void Triangle::useBaryCoords(const double a, const double b,
			     const double c, SbVec3f &bc){
  bc[0] = a;
  bc[1] = b;
  bc[2] = c;
}  

// calculate interpolated normal
SbVec3f Triangle::getIntNormal(const SbVec3f &bc,
                               const PrimDerivs *pDer) const{

  if (!hasNormals())
    return getNormal();
  
  SbVec3f n(bc[0]*normals(0) + bc[1]*normals(1) + bc[2]*normals(2));
  n.normalize();

  // do bump mapping?
  if (! (texture() && texture()->bumpMap))
    return n;

  // get geometric derivs if we don't have them already
  PrimDerivs myPrimDerivs;
  if (!pDer){
    if (!getDerivs(bc, myPrimDerivs))
      return n;
    pDer = &myPrimDerivs;
  }

  
  SbVec3f dh = getDeriv(bc); // returns dh/du, dh/dv, h
  float dhdu = dh[0];        // either of these might be zero
  float dhdv = dh[1];
  float h0 = dh[2];
    
  SbVec3f dppdu = pDer->dPdU + texture()->bumpScale*dhdu*n + h0*pDer->dNdU;
  SbVec3f dppdv = pDer->dPdV + texture()->bumpScale*dhdv*n + h0*pDer->dNdV;

  SbVec3f n2 = dppdu.cross(dppdv);
  n2.normalize();
    
  return n2;
}

// calculate interpolated diffuse color
SbColor Triangle::getIntDiffColor(const Ray &r) const {

  SbColor diffColor;
  const SbVec3f &bc = r.bc;

  static bool applyProcText = getenv("PANE_TRI_PROC_TEXTURE");

  if (applyProcText) {
    
    SbColor col1(.15, .15, .15);
    SbColor col2(0.001, 0.001, 0.001);
    static Noise noise;
    static bool noiseInit = false;
    static Spline<SbColor> colorMap(0,1);  
    if (r.ctx->id == 0){
      if (!noiseInit) {
        noise.octaves = 4;
        noise.amplitude = .99/1.92651;
        float scale = 20;
        noise.frequency = SbVec3f(scale,scale,scale);
        noise.bias = .4257;
        noise.gain = .8769;
        noise.translation = -SbVec3f(.3742, .135, -.5825);
        noise.freqJump = 2.029;
        noise.ampJump = 1;
        noise.sign=false;
        noise.initExponents();

//        colorMap.SetInterp(false);
        colorMap.Add(0,     SbColor(0.3, 0.3, 0.3));
//        colorMap.Add(.4,    SbColor(0.15, 0.15, 0.15));
//         colorMap.Add(.4,    SbColor(0.15, 0.15, 0.15));
//         colorMap.Add(.5,     SbColor(0.01, 0.01, 0.01));
//         colorMap.Add(.6,     SbColor(0.05, 0.05, 0.05));
        colorMap.Add(1,     SbColor(0.01, 0.01, 0.01));
        
        noiseInit = true;
      }
    } else {
      while (!noiseInit){};
    }

    float a = noise.noise(r.x);
//     a*=a;
// //    a=1-a;
//     a = clamp(a, 0.f, 1.f);
//     diffColor = a*col1 + (1-a)*col2;
    diffColor = colorMap.eval(a);
    //diffColor = a<.5 ? col1 : col2;
    return diffColor;
  }

  if (hasDiffuse())
    diffColor = bc[0]*diffuse(0) + bc[1]*diffuse(1) + bc[2]*diffuse(2);
  else
    diffColor = r.prim->getMat(r)->diffuse;
      
  if (texture() && texture()->bumpMap != 2)
    return getColor(bc, diffColor);

  return diffColor;
}

// FIXME: For efficiency, I *should* be passing
// in a precompute PrimDerivs, but this allows for not passing that.
bool Triangle::getDerivs(const SbVec3f &bc, PrimDerivs &pDer) const {

  if (!texture()) return false;

  // calculate dpdu and dpdv
  // This derivation follows PBRT pg. 129
  //
  //     v2
  //
  //   v
  //
  //  v0        v1
  //       u
  //
  // p(ui,vi) = v0 + ui*dpdu + vi*dpdv
  //
  // p1-p0 = (u1*dpdu +v1*dpdv) - (u0*dpdu+v0*dpdv) = (u1-u0)*dpdu + (v1-v0)*dpdv
  // p2-p0 = (u2*dpdu +v2*dpdv) - (u0*dpdu+v0*dpdv) = (u2-u0)*dpdu + (v2-v0)*dpdv
  //
  // [p1-p0] = [ u1-u0  v1-v0 ] [dpdu]
  // [p2-p0]   [ u2-u0  v2-v0 ] [dpdv]                      
  //
  // Invert:          (A)
  //
  // [dpdu] = [ u1-u0  v1-v0 ]^-1 [p1-p0]
  // [dpdv]   [ u2-u0  v2-v0 ]    [p2-p0]
  //
  //       (a)
  //
  //  A^-1 = Adj(A) =   (C_ij)^T   C_ij = -1^(i+j)M_ij
  //         ------     --------   M_ij = det|matrix w/o row i col j|
  //         det(A)      det(A)
  //
  // [ A00 A01 ]^-1 = [ a00 a01 ] = 1/(A00*A11 - A01*A10)  [  A11 -A01 ]
  // [ A10 A11 ]      [ a10 a11 ]                          [ -A10  A00 ]

  // partial derivatives from PBR by Pharr & Humpreys
  SbVec2f uv0=uv(0), uv1=uv(1), uv2=uv(2);
  float A00 = uv1[0] - uv0[0];
  float A01 = uv1[1] - uv0[1];
  float A10 = uv2[0] - uv0[0];
  float A11 = uv2[1] - uv0[1];

  float det = A00*A11 - A01*A10;
  if (det == 0.0)
    return false;

  float invDet = 1/det;
    
  const SbVec3f &vert0 = vert(0);
  SbVec3f edge1 = vert(1)-vert0;
  SbVec3f edge2 = vert(2)-vert0;
    
  SbVec3f dpdu = ( A11*edge1 - A01*edge2) * invDet;
  SbVec3f dpdv = (-A10*edge1 + A00*edge2) * invDet;

  SbVec3f dndu, dndv;

  if (hasNormals()){
    const SbVec3f &normals0 = normals(0);
    const SbVec3f &normals1 = normals(1);
    const SbVec3f &normals2 = normals(2);
  
    // Compute normal derivative (very similar)
    //
    // n(ui,vi) = n0 + ui*dndu + vi*dndv
    //
    // n1-n0 = (u1-u0)*dndu + (v1-v0)*dndv
    // n2-n0 = (u2-u0)*dndu + (v2-v0)*dndv
    //
    // [n1-n0] = [ u1-u0  v1-v0 ] [dndu]
    // [n2-n0]   [ u2-u0  v2-v0 ] [dndv]                      
    //
    // [dndu]  = [ u1-u0  v1-v0 ]^-1 [n1-n0]
    // [dndv]    [ u2-u0  v2-v0 ]    [n2-n0]                      
    SbVec3f ne1 = normals1-normals0;
    SbVec3f ne2 = normals2-normals0;
    
    dndu = ( A11*ne1 - A01*ne2) * invDet;
    dndv = (-A10*ne1 + A00*ne2) * invDet;
  } else {
    dndu = dndv = SbVec3f(0,0,0);
  }

  pDer = PrimDerivs(dpdu, dpdv, dndu, dndv);
  return true;
}

// calculate interpolated diffuse emission
SbColor Triangle::getIntEmission(const Ray &r) const{
  SbColor color = getEmission();
  const SbVec3f &bc = r.bc;
  
  // modulate emission by texture
  if (texture()){

    SbColor diffColor = bc[0]*diffuse(0) + bc[1]*diffuse(1) + bc[2]*diffuse(2);
    SbColor tcolor = getColor(bc, diffColor);

    vecScaleItself(color, tcolor);
  }

  return color;
}

// point debugging

namespace {
  pthread_mutex_t lock_show = PTHREAD_MUTEX_INITIALIZER;
}

void Triangle::show(SbColor c){
  static double incr = 1.0;
  double scale = 10.0;
  const int showVerts = 0;

  pthread_mutex_lock(&lock_show);
  double myincr = incr;
  incr *= 0.999;
  pthread_mutex_unlock(&lock_show);
  
  // dont show triangles more than once
//   if (shown==1)
//     return;
//   shown = 1;
  
  // Tri debugging
  SoMaterial *mat = new SoMaterial();
  mat->emissiveColor.setValue(c);
  SoSeparator *sep;
  SoTransform *trn;
  SoCube *sph = new SoCube();
  sph->width.setValue(myincr*scale);
  sph->depth.setValue(myincr*scale);
  sph->height.setValue(myincr*scale);

  // show vertex normals
  if (1){

    SbVec3f edge1 = vert(1)-vert(0);
    double l = edge1.length()/4.0;
    if (hasNormals()) {
      (Ray(vert(0), normals(0), NULL)).show(l, SbColor(1.0, 0.0, 0.0));
      (Ray(vert(1), normals(1), NULL)).show(l, SbColor(0.0, 1.0, 0.0));
      (Ray(vert(2), normals(2), NULL)).show(l, SbColor(0.0, 0.0, 1.0));
    }
  }

  // first vert
  sep = new SoSeparator();
  trn = new SoTransform();
  trn->translation.setValue(vert(0));
  sep->addChild(trn);
  sep->addChild(mat);
  sep->addChild(sph);
  if (showVerts)
    scnRoot->addChild(sep);

  // second vert
  sep = new SoSeparator();
  trn = new SoTransform();
  trn->translation.setValue(vert(1));
  trn->scaleFactor.setValue(SbVec3f(0.9, 0.9, 0.9));
  sep->addChild(trn);
  sep->addChild(mat);
  sep->addChild(sph);
  if (showVerts)
    scnRoot->addChild(sep);

  // third vert
  sep = new SoSeparator();
  trn = new SoTransform();
  trn->translation.setValue(vert(2));
  trn->scaleFactor.setValue(SbVec3f(0.8, 0.8, 0.8));
  sep->addChild(trn);
  sep->addChild(mat);
  sep->addChild(sph);
  if (showVerts)
    scnRoot->addChild(sep);

  // triangle
  sep = new SoSeparator();
  SoCoordinate3 *cor = new SoCoordinate3();
  cor->point.set1Value(0, vert(0));
  cor->point.set1Value(1, vert(1));
  cor->point.set1Value(2, vert(2));
  SoTriangleStripSet *tset = new SoTriangleStripSet();
  tset->numVertices.set1Value(0, 3);
  sep->addChild(mat);
  sep->addChild(cor);
  //sep->addChild(tset);
  scnRoot->addChild(sep);
  
}


// edges as inset lines representation
void Triangle::show2(){
  Ray r;
  SbVec3f dir;
  float l;

  // dont show triangles more than once
//   if (shown==1)
//     return;
//   shown = 1;

  showPhoton(vert(0), SbColor(1,0,0));
  showPhoton(vert(1), SbColor(0,1,0));
  showPhoton(vert(2), SbColor(0,0,1));
  
  SbVec3f v0 = vert(0)*0.9 + vert(1)*0.05 + vert(2)*0.05;
  SbVec3f v1 = vert(1)*0.9 + vert(2)*0.05 + vert(0)*0.05;
  SbVec3f v2 = vert(2)*0.9 + vert(0)*0.05 + vert(1)*0.05;
      
  r.pos = v0;
  dir = v1-v0;
  l = dir.normalize();
  r.dir = dir;
  r.show2(l, SbColor(1,0,0));

  r.pos = v1;
  dir = v2-v1;
  l = dir.normalize();
  r.dir = dir;
  r.show2(l, SbColor(0,1,0));

  r.pos = v2;
  dir = v0-v2;
  l = dir.normalize();
  r.dir = dir;
  r.show2(l, SbColor(0,0,1));
}

void Triangle::printGlobalStats(){
  cerr << "Triangles with zero area: " << errorTris << "\n"
       << "Total area of emissive triangles: " << emissiveArea << "\n";
}

ostream& operator<< (ostream& os, const Triangle& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x.vert(0) << ", " << x.vert(1) << ", " << x.vert(2) << ")";
  resetios(os);
  return(os);
}


EmissiveTriangle::EmissiveTriangle(){
  area = 0.0;
  emisAvg = 0.0;
  SbColor emission = SbColor(0.0, 0.0, 0.0);
  SbVec3f normal = SbVec3f(0.0, 1.0, 0.0);
  map = NULL;
  celluv = NULL;
  mapDeltaCDF = 0.0;
  mapPDF = 1.0;
  cells = 0;
}

void EmissiveTriangle::init(Pane *pane) {
  map = NULL;
  celluv = NULL;
  
  emission = getMat()->emission;
  emisAvg = vecAvg(emission);
  
  // calculate edges
  const SbVec3f &vert0 = vert(0);
  SbVec3f edge1 = vert(1)-vert0;
  SbVec3f edge2 = vert(2)-vert0;

  // calculate area and precompute normal
  normal = edge1.cross(edge2);
  area = 0.5*normal.normalize();

  // emission should be flux, but flux density is given.
  // Multiply by area to convert flux density to flux.
  // This is useful if you want lights to have same brightness
  // regardless of size.
  if (pane->gv->useEmisFluxDensity){

    // make sure unitsPerMeter is okay
    if (pane->gv->unitsPerMeter<0.0){
      cerr << "\nSettings unitsPerMeter to 1.0\n";
      pane->gv->setUnitsPerMeter(1.0);
    }
    
    emission *= area;
  }
  
  if (getMat()->emissive){
    pane->gv->sceneHasLights++;
    emissiveArea += area;
  }

  onb.constructFromUV(edge1, edge2);
}
 
const int EmissiveTriangle::mapSize[2] = { 64, 64 };

void EmissiveTriangle::genImpMap(const SbVec3f &center, float rad){

  const SbVec3f &vert0 = vert(0);
  const SbVec3f &vert1 = vert(1);
  const SbVec3f &vert2 = vert(2);
  SbVec3f triCenter = (vert0+vert1+vert2) * (1./3.f);

  // find triangle radius
  float triR, nextR;
  triR = (vert0-triCenter).length();
  nextR = (vert1-triCenter).length();
  if (nextR>triR) triR = nextR;
  nextR = (vert2-triCenter).length();
  if (nextR>triR) triR = nextR;

  // find sphere dir
  SbVec3f sphereDir = center-triCenter;
  sphereDir = onb.toONBSpace(sphereDir);
  
  float dist = sphereDir.normalize();
  //rad += triR; // account for area of triangle

  mapDeltaCDF = 0.0;
  mapPDF = 1.0;
  cells = 0;

  if (rad<=0) {
    cerr << "Projection map creation: ERROR: negative radius!\n";
    return;
  }
  
  // don't use projection maps if sphere is too close
  if (dist<rad){
    cerr << "Projection map creation: Light too close, map not created.\n";
    return;
  }

  // find max angle theta
  // theta is determined from conic projection of sphere
  // theta = asec(h/R)    h=dist R=rad
  // asec(x) = acos(1/x)
  // theta = acos(R/h)
  // theta = pi/2 - theta     Because we need the OTHER angle in the right triangle  
  float theta = M_PI_2 - acos(rad/dist); 
  float cosProjTheta = cos(theta);
  
  // x = i  theta = 0->pi/2 = acos(sqrt(x))
  // y = j  phi   = 0->2pi  = 2*pi*y
  //
  // initialize map
  map = new char[mapSize[0]*mapSize[1]];
  
  for (int j=0; j<mapSize[1]; j++)
    for (int i=0; i<mapSize[0]; i++)
	map[j*mapSize[0]+i] = 0;

  float dx = 1.0f/(mapSize[0]);
  float dy = 1.0f/(mapSize[1]);
  
  for (int j=0; j<=mapSize[1]; j++){
    for (int i=0; i<=mapSize[0]; i++){
      
      float x = clamp(i*dx, 0.0f, 1.0f);
      float y = clamp(j*dy, 0.0f, 1.0f);
      float theta = acos(sqrt(x));
      float phi = 2.0*M_PI*y;

      SbVec3f r(sin(theta)*cos(phi),
		sin(theta)*sin(phi),
		cos(theta));

      float cosTheta = r.dot(sphereDir);

      // see if points at bounding volume
      if (cosTheta>cosProjTheta) {   // true if vector is within theta
                                     // degrees of sphereDir

	// j wraps
	int j0 = (j-1+mapSize[1])%mapSize[1];
	int j1 = (j             )%mapSize[1];

	// i doesn't 
        int i0 = i-1; 
	int i1 = i;   
	
	if (i0>=0){
	  map[j0*mapSize[0]+i0] = 1;
	  map[j1*mapSize[0]+i0] = 1;
	}
	if (i1<mapSize[0]){
	  map[j0*mapSize[0]+i1] = 1;
	  map[j1*mapSize[0]+i1] = 1;
	}
      }
      
    }
  }

  // collapse array, to only store filled cells, idea from Pooya Karimian
  // http://www.sfu.ca/~pkarimia/courses/cmpt770graphics/lab4/
  for (int j=0; j<mapSize[1]; j++){
    for (int i=0; i<mapSize[0]; i++){
      if (map[j*mapSize[0]+i])
	cells++;
    }
  }

  if (cells==0)
    return;

  // create new array to hold precomputed u,v values
  // stored as a sequence of pairs
  celluv = new float[cells*2];
  int idx = 0;
  for (int j=0; j<mapSize[1]; j++){
    for (int i=0; i<mapSize[0]; i++){
      if (!map[j*mapSize[0]+i])
	continue;
      celluv[idx++] = i*(1.0f/mapSize[0]);
      celluv[idx++] = j*(1.0f/mapSize[1]);
    }
  }

  mapDeltaCDF = 1.0f/cells;
  // pdf = 1/a  a=cells/total => pdf = total/cells
  mapPDF = (mapSize[0]*mapSize[1])/(float)cells;
  //printMap();
}

void EmissiveTriangle::printMap(){

  if (!map) return;
  
  // count filled boxes
  cerr << "Projection map (x=elevation, from flat to up  y=around axis):\n";
  
  for (int j=0; j<mapSize[1]; j++){
    for (int i=0; i<mapSize[0]; i++){
      if (map[j*mapSize[0]+i])
	cerr << "*";
      else
	cerr << " ";
    }
    cerr << "|\n";
  }
}

void EmissiveTriangle::sampleDir(SbVec3f &dir, float &pdf,
				 float e1, float e2){

  if (cells == 0){
    pdf = 0.0;
    return;
  }
  // use e2 to choose cell
  float cellf = (e2*cells);
  int cell = (int)cellf;
  e2 = cellf-cell;
  
  e1 = celluv[cell*2+0]+e1*(1.0f/mapSize[0]);
  e2 = celluv[cell*2+1]+e2*(1.0f/mapSize[1]);

  float theta = acos(sqrt(e1));
  float phi = 2.0f*M_PI*e2;

  SbVec3f triDir(sin(theta)*cos(phi),
		 sin(theta)*sin(phi),
                 cos(theta));

  // convert to world space
  if (triDir[2]<0.0f)
    triDir[2] = 0.0;
  dir = onb.toWorldSpace(triDir);
  
  pdf = mapPDF;
}

namespace DMTriSect {
  typedef struct isect_t {
//class isect_t {
//public:
    int i;
    double t;
    double u, v;      // boundary tri uv's (useful only for top and bottom)
    Vec3d p;          // prism texture space coordinates (u,v,w)

    isect_t() { }
    isect_t(const int &i_, const double &t_, const double &u_, const double &v_) :
      i(i_), t(t_), u(u_), v(v_) { }
      
//     bool operator<( const isect_t &a ) const {
//       return t < a.t;
//     }
  } isect_t;

ostream& operator<< (ostream& os, const isect_t& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(i=" << x.i << " t=" << x.t << " u=" << x.u << " v=" << x.v
     << " pu=" << x.p[0] << " pv=" << x.p[1] << " pw=" << x.p[2] << ")";
  resetios(os);
  return(os);
}
  
//};
}
using namespace DMTriSect;

namespace {
  
//   inline SbVec3f normFromW(const SbVec3f *v,
//                            const SbVec3f *n,
//                            float w){
//     SbVec3f myv[3] = {
//       v[0]+w*n[0],
//       v[1]+w*n[1],
//       v[2]+w*n[2]
//     };
//     SbVec3f myn = (myv[1]-myv[0]).cross(myv[2]-myv[0]);
//     myn.normalize();
//     return myn;
//   }


  // returns offset of Tw relative to p
  inline float distToTw(const SbVec3f *v,
                        const SbVec3f *n,
                        const SbVec3f &p,
                        float w){
    SbVec3f Tv[3] = {
      v[0] + w*n[0],
      v[1] + w*n[1],
      v[2] + w*n[2]
    };
    SbVec3f Tn = (Tv[1]-Tv[0]).cross(Tv[2]-Tv[0]);
    Tn.normalize();
    
    return Tn.dot(Tv[0]) - Tn.dot(p);
  }

  // moller triangle intersection but just finds u and v
  inline bool intersectUV(const SbVec3f *v,
                          const SbVec3f *n,
                          const SbVec3f &rpos,
                          const SbVec3f &rdir,
                          Vec3d &p){
//     SbVec3f rdir = n[0];
//     SbVec3f rpos = p-n[0];

    SbVec3f myv[3] = {
      v[0] + p[2]*n[0],
      v[1] + p[2]*n[1],
      v[2] + p[2]*n[2]
    };
    SbVec3f edge1 = myv[1]-myv[0];
    SbVec3f edge2 = myv[2]-myv[0];
    
    // begin calculating determinant - also used to calculate U parameter
    SbVec3f pvec = rdir.cross(edge2);
    
    // if determinant is near zero, ray lies in plane of triangle
    float det = edge1.dot(pvec);

    const float EPSILON = 1e-5;
    if (det < EPSILON && det > -EPSILON) // ray is parallell to the plane of the triangle
      return false;
     
    // calculate distance from vert0 to ray origin 
    SbVec3f tvec = rpos - myv[0];
      
    // calculate U parameter and test bounds
    p[0] = tvec.dot(pvec);
     
    // prepare to test V parameter
    SbVec3f qvec = tvec.cross(edge1);
      
    // calculate V parameter and test bounds
    p[1] = rdir.dot(qvec);

    float inv_det = 1.0f / det;
   
    // calculate t, ray intersects triangle
    p[0] *= inv_det;
    p[1] *= inv_det;

    return true;
  }
  
}

bool DMTri::intersectRay(Ray &r, double t0, double t1) {

//  return Triangle::intersectRay(r,t0,t1);

  if (r.ignorePrim == this) return false;

  r.t = -1;
  r.prim = NULL;
  
  // assume three normals
  // assume texture
  
  // test Prism
  float h = maxHeight;
  SbVec3f n[3] = { normals(0), normals(1), normals(2) };
  SbVec3f v[6] = { vert(0), vert(1), vert(2),
                   vert(0)+h*n[0], vert(1)+h*n[1], vert(2)+h*n[2] };
  // Find entrance and exit point of prism
  // From Interactive Smooth and Curved Shell Mapping  
  // Okay if slightly outside of prism,
  // So test both configurations of triangulating sides and take
  // min, max.
  static const int tris[14][3] = {
    {0,1,2}, {3,4,5},  //  0  1  top, bottom
    {0,1,4}, {0,4,3},  //  2  3  01 side face (Rising)
    {0,1,3}, {1,4,3},  //  4  5  01 side face (Falling)
    {1,2,5}, {1,5,4},  //  6  7  12 side face (Rising)
    {1,2,4}, {2,5,4},  //  8  9  12 side face (Falling)
    {2,0,3}, {2,3,5},  // 10 11  20 side face (Rising)
    {2,0,5}, {0,3,5},  // 12 13  20 side face (Falling)
  };

  isect_t minMax[2];
  int numIs=0;
  minMax[0].t = FLT_MAX;
  minMax[1].t = -FLT_MAX;

  //cerr << __LINE__ << ": BEGINNING CONVEX HULL INTERSECTION:\n";
  for (int i=0; i<14; i++){

    double t, uu, vv;
    if (intersect_ray_tri(r.pos, r.dir,
                          v[tris[i][0]],
                          v[tris[i][1]],
                          v[tris[i][2]],
                          &t, &uu, &vv)){

      if (t<minMax[0].t)
        minMax[0] = isect_t(i, t, uu, vv);
      if (t>minMax[1].t)
        minMax[1] = isect_t(i, t, uu, vv);
      numIs++;
    }
  }

  if (minMax[0].t >= minMax[1].t) return false;

  // prism is behind us
  if (minMax[1].t<t0) return false;

//   r.t = minMax[1].t;
//   r.prim = this;
//   r.bc = SbVec3f(1,0,0);
//   return true;
  
  // prism texture coordinates are
  // (u,v) of base triangle and w (from 0 to h)

  const double eps = 0.00001; // Is this epsilon a safe value?
  const double hEps = h*eps; // Is this epsilon a safe value?
  
  // compute entrance and exit prism texture coordinates
  for (int i=0; i<2; i++){

    // top or bottom tri are easy
    if (minMax[i].i<2) {
      minMax[i].p = Vec3d(minMax[i].u, minMax[i].v, minMax[i].i*h);
      continue;
    }

    // otherwise we hit side
    SbVec3f p = r.pos + minMax[i].t*r.dir;
    
    // perform bisectional search for pw
    float w0 = 0;
    float w1 = h;
    float d0, d1;
    d0 = distToTw(v,n,p,w0);
    d1 = distToTw(v,n,p,w1);

    //cerr << "d0=" << d0 << " d1=" << d1 << "\n";
    
    // order d0<d1
    if (d0>d1){

      // swap w0, w1
      float tmp_w = w1;
      float tmp_d = d1;
      w1 = w0;
      d1 = d0;
      w0 = tmp_w;
      d0 = tmp_d;
    }
    
    // The point should definitely be *between* the top and bottom triangles
    if (d0>0 || d1<0) {
      cerr << "WTF at line " << __LINE__ << "\n";
      return false;
    }

    do {
      float wm = (w0+w1)*0.5f;
      float dm = distToTw(v,n,p,wm);
      if (dm<0) {
        w0 = wm;
        d0 = dm;
      } else {
        w1 = wm;
        d1 = dm;
      }
    } while (d1>hEps); 

    // w1 is our w.
    //cerr << "w=" << w1 << "\n";
    minMax[i].p[2] = w1;
    
    // what is our u and v?
    //cerr << "DEBUGGING: " << r.pos+r.dir*minMax[i].t << "\n";
    

    if (!intersectUV(v, n, r.pos,r.dir, minMax[i].p)){

      // oddly this gives more acne. Maybe ought to do calculation in double precision?
      if (!intersectUV(v, n, (r.pos+r.dir*minMax[i].t-n[0]), n[0], minMax[i].p)){ 
      
        cerr << "WTF at line " << __LINE__ << "\n";
        return false;
      }
    }
  }

  // now we have entrance and exit u,v,w
  // march...

  // clip t0,t1
  t0 = MAX2(t0, minMax[0].t);
  t1 = MIN2(t1, minMax[1].t);

  // FIXME move this earlier
  if (t1<t0) {
//     cerr << "WTF at line " << __LINE__ << "\n";
//     cerr << "min=" << minMax[0] << "\nmax=" << minMax[1] << "\n"
//          << "t0=" << t0 << " t1=" << t1 << "\n";
    return false;
  }


  // uv's equal if p[2] equal?
  if (minMax[0].p[2] == minMax[1].p[2]) {
//     if ((minMax[1].t-minMax[0].t)<hEps)
//       return false;
    if ((minMax[1].t-minMax[0].t)<eps*(t0+t1)*0.5)
      return false;
    cerr << __LINE__ << ": height equal!! Unhandled case. Fix me!\n";
    r.t = minMax[0].t;
    r.prim = this;
    r.bc = SbVec3f(0,1,0);
    return true;
    //return false;
  }

  

//   cerr << "i0=" << minMax[0].i << " uvh0= " << uvh0 << "  "
//        << "i1=" << minMax[1].i << " uvh1= " << uvh1 << "\n";
//  SbVec3f duvh = (uvh0-uvh1)/(minMax[1].t-minMax[0].t);
  double tDiff = minMax[1].t-minMax[0].t;
  double invTdiff = 1/tDiff;

  const double stept = 1/2048.;//1/256.;//0.01;
  Vec3d cuvh;
  double tt[2] = {t0, t1};

  // clip t0 t1 to actual triangle (u,v) in [0,1]
  if (DEBUG_STATE){
    cerr << __LINE__ << ": Begin uv hull clamping\n";
  }

  
  // loop over t0, t1
  bool care=0;

  // do it twice because once isn't enough
  for (int iter=0; iter<2; iter++){
    for (int ti=0; ti<2; ti++){

      Vec3d uvw;
      uvw[0] = minMax[ti].p[0];
      uvw[1] = minMax[ti].p[1];
      uvw[2] = 1-uvw[0]-uvw[1];
    
      if (uvw[0]>=0 && uvw[0]<=1 &&
          uvw[1]>=0 && uvw[1]<=1 &&
          uvw[2]>=0 && uvw[2]<=1) {
        continue;
      }

      // adjust stept
      double myStept = 0;
      for (int d=0; d<3; d++){
        if (uvw[d]<0){
          double dp = d<2 ?
            minMax[1].p[d] - minMax[0].p[d] :  // du or dv
            (minMax[0].p[0] + minMax[0].p[1]) - (minMax[1].p[0] + minMax[1].p[1]); // dw
          if (dp){
            double dtdp = tDiff/dp;
            double dt = -uvw[d]*dtdp;
            if (dt>0 && dt>myStept || // further
                dt<0 && dt<myStept) { // nearer
//             if (ti==1 && d==2)
//               care = 1;
              myStept = dt;
            }
          }
        }      
      }
    
      tt[ti] += myStept;
    }
  }

  if (care){
    cerr << "Refined:\n" << setprecision(12)
         << "t0=" << t0    << " t1=" << t1 << "\n"
         << "  =" << tt[0] << "   =" << tt[1] << "\n";
  }
      
  t0 = tt[0];
  t1 = tt[1];

  if (t0==t1 || t1<t0)
    return false; //cerr << "WTF!!!!!!!!!!! not supposed to happen at " << __LINE__ << "\n";
  
  if (DEBUG_STATE)
    cerr << __LINE__ << ": Done uv cull clamping\n";

  double a = (t0-minMax[0].t) * invTdiff;
  cuvh = minMax[0].p*(1-a) + minMax[1].p*a;


  double h0, h1, hm;
  double d0, d1, dm;
  SbVec3f bc(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
  d0 = t0;
  h0 = getColor(bc)[0]*h;

  if (DEBUG_STATE){
    cerr << "min=" << minMax[0] << "\nmax=" << minMax[1] << "\n"
         << "t0=" << t0 << " t1=" << t1
         << " cuvh=" << cuvh << " h0=" << h0 << "\n";
  }
  
  // check for already near
  const double myEps = 1e-5;
  if (fabs(cuvh[2]-h0)<myEps) {

      if (DEBUG_STATE)
        cerr << "hit!\n";
      
      // hit heightfield
      r.t = d0;
      r.prim = this;
      r.bc = bc;
      return true;
  }
  
  bool above = cuvh[2]>h0;
  
  //if (DEBUG_STATE)
  //cerr << __LINE__ << ": h0= " << h0 << "\n";

  static bool debuggingEnabled = false;
  if (debuggingEnabled && DEBUG_STATE)
    cerr << "Begin marching\n";

  // iterate
  d1=d0;
  while (d1<t1) {
    
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";

    d0 = d1;
    h0 = h1;
    d1 += stept;
    if (d1>t1){
      if (debuggingEnabled && DEBUG_STATE)
        cerr << __LINE__ << ": Reached end\n";
      d1 = t1;
    }

    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";
    
    // get height
    double a = (d1-minMax[0].t) * invTdiff;
    cuvh = minMax[0].p*(1-a) + minMax[1].p*a;

    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": cuvh= " << cuvh << "\n";
    
    // check uv to be in tri
    const double uvEps = 0; //1e-6;
    const double uvMin = -uvEps;
    const double uvMax = 1+uvEps;
    if (cuvh[0]<uvMin || cuvh[0]>uvMax ||
        cuvh[1]<uvMin || cuvh[1]>uvMax ||
        (cuvh[0]+cuvh[1])>uvMax ) {

      if (debuggingEnabled && DEBUG_STATE)
        cerr << __LINE__ << ": Left tri's uv space. SHOULD NOT HAPPEN cuvh=" << cuvh << "\n";
      
      // shouldn't happen, but can
        return false;
    }

    // maybe clamp u,v to be in triangle?
    
    bc = SbVec3f(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
    double h1 = getColor(bc)[0]*h;
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": h1=" << h1 << "\n";
    
    // detect crossing
    if (above && cuvh[2]<h1 ||
        !above && cuvh[2]>h1 ||
        fabs(cuvh[2]-h1)<myEps) {

      if (debuggingEnabled && DEBUG_STATE)
        cerr << "hit... now refine...\n";

      double diff;
      SbVec3f bc;
      do {
        dm = (d0+d1)*0.5;
        diff = d1-dm;
        double a = (dm-minMax[0].t) * invTdiff;
        cuvh = minMax[0].p*(1-a) + minMax[1].p*a; // FIXME only need h, not u&v
        bc = SbVec3f(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
        hm = getColor(bc)[0]*h;

        if (above && cuvh[2]<hm ||
            !above && cuvh[2]>hm ||
            fabs(cuvh[2]-hm)<myEps) {
          d1=dm;
          h1=hm;  // is this needed?
        } else {
          d0=dm;
          h0=hm;  // is this needed?
        }
          
      } while (diff>myEps);
        
        
      // hit heightfield
      r.t = d0; // choose nearer hit, give next ray a better chance of not self-intersecting
      r.prim = this;
      r.bc = bc;
      if (debuggingEnabled && DEBUG_STATE)
        cerr << "returning t=" << d0 << "\n";
      return true;
    }

  }

  if (debuggingEnabled && DEBUG_STATE)
    cerr << __LINE__ << ": no hit\n";
  return false;
}

/*
// Smits et. al now...
bool DMTri::intersectRay(Ray &r, double t0, double t1) {

//  return Triangle::intersectRay(r,t0,t1);

  if (r.ignorePrim == this) return false;

  r.t = -1;
  r.prim = NULL;
  
  // assume three normals
  // assume texture
  
  // test Prism
  float h = maxHeight;
  SbVec3f n[3] = { normals(0), normals(1), normals(2) };
  SbVec3f v[6] = { vert(0), vert(1), vert(2),
                   vert(0)+h*n[0], vert(1)+h*n[1], vert(2)+h*n[2] };
  // Find entrance and exit point of prism
  // From Interactive Smooth and Curved Shell Mapping  
  // Okay if slightly outside of prism,
  // So test both configurations of triangulating sides and take
  // min, max.
  static const int tris[14][3] = {
    {0,1,2}, {3,4,5},  //  0  1  top, bottom
    {0,1,4}, {0,4,3},  //  2  3  01 side face (Rising)
    {0,1,3}, {1,4,3},  //  4  5  01 side face (Falling)
    {1,2,5}, {1,5,4},  //  6  7  12 side face (Rising)
    {1,2,4}, {2,5,4},  //  8  9  12 side face (Falling)
    {2,0,3}, {2,3,5},  // 10 11  20 side face (Rising)
    {2,0,5}, {0,3,5},  // 12 13  20 side face (Falling)
  };

  isect_t minMax[2];
  int numIs=0;
  minMax[0].t = FLT_MAX;
  minMax[1].t = -FLT_MAX;

  //cerr << __LINE__ << ": BEGINNING CONVEX HULL INTERSECTION:\n";
  for (int i=0; i<14; i++){

    double t, uu, vv;
    if (intersect_ray_tri(r.pos, r.dir,
                          v[tris[i][0]],
                          v[tris[i][1]],
                          v[tris[i][2]],
                          &t, &uu, &vv)){

      if (t<minMax[0].t)
        minMax[0] = isect_t(i, t, uu, vv);
      if (t>minMax[1].t)
        minMax[1] = isect_t(i, t, uu, vv);
      numIs++;
    }
  }

  if (minMax[0].t >= minMax[1].t) return false;

  // prism is behind us
  if (minMax[1].t<t0) return false;
  
  // prism texture coordinates are
  // (u,v) of base triangle and w (from 0 to h)

  const double eps = 0.00001; // Is this epsilon a safe value?
  const double hEps = h*eps; // Is this epsilon a safe value?

  // This part could be replaced by Bilinear patch intersection
  
  // compute entrance and exit prism texture coordinates
  for (int i=0; i<2; i++){

    // top or bottom tri are easy
    if (minMax[i].i<2) {
      minMax[i].p = Vec3d(minMax[i].u, minMax[i].v, minMax[i].i*h);
      continue;
    }

    // otherwise we hit side
    SbVec3f p = r.pos + minMax[i].t*r.dir;
    
    // perform bisectional search for pw
    float w0 = 0;
    float w1 = h;
    float d0, d1;
    d0 = distToTw(v,n,p,w0);
    d1 = distToTw(v,n,p,w1);

    //cerr << "d0=" << d0 << " d1=" << d1 << "\n";
    
    // order d0<d1
    if (d0>d1){

      // swap w0, w1
      float tmp_w = w1;
      float tmp_d = d1;
      w1 = w0;
      d1 = d0;
      w0 = tmp_w;
      d0 = tmp_d;
    }
    
    // The point should definitely be *between* the top and bottom triangles
    if (d0>0 || d1<0) {
      cerr << "WTF at line " << __LINE__ << "\n";
      return false;
    }

    do {
      float wm = (w0+w1)*0.5f;
      float dm = distToTw(v,n,p,wm);
      if (dm<0) {
        w0 = wm;
        d0 = dm;
      } else {
        w1 = wm;
        d1 = dm;
      }
    } while (d1>hEps); 

    // w1 is our w.
    //cerr << "w=" << w1 << "\n";
    minMax[i].p[2] = w1;
    
    // what is our u and v?
    //cerr << "DEBUGGING: " << r.pos+r.dir*minMax[i].t << "\n";
    

    if (!intersectUV(v, n, r.pos,r.dir, minMax[i].p)){

      // oddly this gives more acne. Maybe ought to do calculation in double precision?
      if (!intersectUV(v, n, (r.pos+r.dir*minMax[i].t-n[0]), n[0], minMax[i].p)){ 
      
        cerr << "WTF at line " << __LINE__ << "\n";
        return false;
      }
    }
  }

  // now we have entrance and exit u,v,w
  // march...

  // clip t0,t1
  t0 = MAX2(t0, minMax[0].t);
  t1 = MIN2(t1, minMax[1].t);

  // FIXME move this earlier
  if (t1<t0) {
//     cerr << "WTF at line " << __LINE__ << "\n";
//     cerr << "min=" << minMax[0] << "\nmax=" << minMax[1] << "\n"
//          << "t0=" << t0 << " t1=" << t1 << "\n";
    return false;
  }


  // uv's equal if p[2] equal?
  if (minMax[0].p[2] == minMax[1].p[2]) {
//     if ((minMax[1].t-minMax[0].t)<hEps)
//       return false;
    if ((minMax[1].t-minMax[0].t)<eps*(t0+t1)*0.5)
      return false;
    cerr << __LINE__ << ": height equal!! Unhandled case. Fix me!\n";
    r.t = minMax[0].t;
    r.prim = this;
    r.bc = SbVec3f(0,1,0);
    return true;
    //return false;
  }

  

//   cerr << "i0=" << minMax[0].i << " uvh0= " << uvh0 << "  "
//        << "i1=" << minMax[1].i << " uvh1= " << uvh1 << "\n";
//  SbVec3f duvh = (uvh0-uvh1)/(minMax[1].t-minMax[0].t);
  double tDiff = minMax[1].t-minMax[0].t;
  double invTdiff = 1/tDiff;

  const double stept = 1/2048.;//1/256.;//0.01;
  Vec3d cuvh;
  double tt[2] = {t0, t1};

  // clip t0 t1 to actual triangle (u,v) in [0,1]
  if (DEBUG_STATE){
    cerr << __LINE__ << ": Begin uv hull clamping\n";
  }

  
  // loop over t0, t1
  bool care=0;

  // do it twice because once isn't enough
  for (int iter=0; iter<2; iter++){
    for (int ti=0; ti<2; ti++){

      Vec3d uvw;
      uvw[0] = minMax[ti].p[0];
      uvw[1] = minMax[ti].p[1];
      uvw[2] = 1-uvw[0]-uvw[1];
    
      if (uvw[0]>=0 && uvw[0]<=1 &&
          uvw[1]>=0 && uvw[1]<=1 &&
          uvw[2]>=0 && uvw[2]<=1) {
        continue;
      }

      // adjust stept
      double myStept = 0;
      for (int d=0; d<3; d++){
        if (uvw[d]<0){
          double dp = d<2 ?
            minMax[1].p[d] - minMax[0].p[d] :  // du or dv
            (minMax[0].p[0] + minMax[0].p[1]) - (minMax[1].p[0] + minMax[1].p[1]); // dw
          if (dp){
            double dtdp = tDiff/dp;
            double dt = -uvw[d]*dtdp;
            if (dt>0 && dt>myStept || // further
                dt<0 && dt<myStept) { // nearer
//             if (ti==1 && d==2)
//               care = 1;
              myStept = dt;
            }
          }
        }      
      }
    
      tt[ti] += myStept;
    }
  }

  if (care){
    cerr << "Refined:\n" << setprecision(12)
         << "t0=" << t0    << " t1=" << t1 << "\n"
         << "  =" << tt[0] << "   =" << tt[1] << "\n";
  }
      
  t0 = tt[0];
  t1 = tt[1];

  if (t0==t1 || t1<t0)
    return false; //cerr << "WTF!!!!!!!!!!! not supposed to happen at " << __LINE__ << "\n";
  
  if (DEBUG_STATE)
    cerr << __LINE__ << ": Done uv cull clamping\n";

  double a = (t0-minMax[0].t) * invTdiff;
  cuvh = minMax[0].p*(1-a) + minMax[1].p*a;


  double h0, h1, hm;
  double d0, d1, dm;
  SbVec3f bc(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
  d0 = t0;
  h0 = texture->getColor(bc)[0]*h;

  if (DEBUG_STATE){
    cerr << "min=" << minMax[0] << "\nmax=" << minMax[1] << "\n"
         << "t0=" << t0 << " t1=" << t1
         << " cuvh=" << cuvh << " h0=" << h0 << "\n";
  }


  // subdivision N?
  const int N = 5;

  // enter edge?
  //
  //  v1
  //
  //  e2   e1
  // 
  //  v0  e0   v1
  //
  int exitEdge, enterEdge;
  switch (minMax[0].i){
  case 0:
  case 1:
    // exit edge?
    switch (minMax[0].i){
    case 0:
    case 1:
      // ray enters and leaves top and bottom,
      // enterEdge does not matter
      enterEdge = 0;
      break;
    case 2:
    case 3:
    case 4:
    case 5:
      exitEdge = 0;
      enterEdge = (exitEdge+1)%3;
      break;
    case 6:
    case 7:
    case 8:
    case 9:
      exitEdge = 1;
      enterEdge = (exitEdge+1)%3;
      break;
    case 10:
    case 11:
    case 12:
    case 13:
      exitEdge = 2;
      enterEdge = (exitEdge+1)%3;
      break;
    default:
      cerr << __LINE__ << " unknown triangle error!\n";
      break;
    }
    break;
  case 2:
  case 3:
  case 4:
  case 5:
    enterEdge = 0;
    break;
  case 6:
  case 7:
  case 8:
  case 9:
    enterEdge = 1;
    break;
  case 10:
  case 11:
  case 12:
  case 13:
    enterEdge = 2;
    break;
  default:
    cerr << __LINE__ << " unknown triangle error!\n";
    break;
  }

  // Which vertex is "a", "b", "c"?
  int ci[3];
  switch (enterEdge){
  case 0:
    ci[0] = 0;
    ci[1] = 1;
    ci[2] = 2;
    break;
  case 1:
    ci[0] = 1;
    ci[1] = 2;
    ci[2] = 0;
    break;
  case 1:
    ci[0] = 2;
    ci[1] = 0;
    ci[2] = 1;
    break;
  default:
    cerr << __LINE__ << " unknown edge error!\n";
    break;
  }
  
  // compute enter and exit cells
  int enterCell[3], exitCell[3];
  enterCell[0] = clamp((int)bc[0], 0, N-1);
  enterCell[1] = clamp((int)bc[1], 0, N-1);
  enterCell[2] = clamp((int)bc[2], 0, N-1);

  {
    double a = (t1-minMax[0].t) * invTdiff;
    cuvh = minMax[0].p*(1-a) + minMax[1].p*a;
    SbVec3f bc(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
    exitCell[0] = clamp((int)bc[0], 0, N-1);
    exitCell[1] = clamp((int)bc[1], 0, N-1);
    exitCell[2] = clamp((int)bc[2], 0, N-1);
  }

  int enterSum = enterCell[0]+enterCell[1]+enterCell[2];
  bool enterLower = enterSum==(N-1);

  SbVec3f uva, uvb, uvc;
  if (enterLower){
    
  } else {

  }
  

  
  // check for already near
  const double myEps = 1e-5;
  if (fabs(cuvh[2]-h0)<myEps) {

      if (DEBUG_STATE)
        cerr << "hit!\n";
      
      // hit heightfield
      r.t = d0;
      r.prim = this;
      r.bc = bc;
      return true;
  }
  
  bool above = cuvh[2]>h0;
  
  //if (DEBUG_STATE)
  //cerr << __LINE__ << ": h0= " << h0 << "\n";

  if (DEBUG_STATE)
    cerr << "Begin marching\n";

  // iterate
  d1=d0;
  while (d1<t1) {
    
    if (DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";

    d0 = d1;
    h0 = h1;
    d1 += stept;
    if (d1>t1){
      if (DEBUG_STATE)
        cerr << __LINE__ << ": Reached end\n";
      d1 = t1;
    }

    if (DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";
    
    // get height
    double a = (d1-minMax[0].t) * invTdiff;
    cuvh = minMax[0].p*(1-a) + minMax[1].p*a;

    if (DEBUG_STATE)
      cerr << __LINE__ << ": cuvh= " << cuvh << "\n";
    
    // check uv to be in tri
    const double uvEps = 0; //1e-6;
    const double uvMin = -uvEps;
    const double uvMax = 1+uvEps;
    if (cuvh[0]<uvMin || cuvh[0]>uvMax ||
        cuvh[1]<uvMin || cuvh[1]>uvMax ||
        (cuvh[0]+cuvh[1])>uvMax ) {

      if (DEBUG_STATE)
        cerr << __LINE__ << ": Left tri's uv space. SHOULD NOT HAPPEN cuvh=" << cuvh << "\n";
      
      // shouldn't happen, but can
        return false;
    }

    // maybe clamp u,v to be in triangle?
    
    bc = SbVec3f(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
    double h1 = texture->getColor(bc)[0]*h;
    if (DEBUG_STATE)
      cerr << __LINE__ << ": h1=" << h1 << "\n";
    
    // detect crossing
    if (above && cuvh[2]<h1 ||
        !above && cuvh[2]>h1 ||
        fabs(cuvh[2]-h1)<myEps) {

      if (DEBUG_STATE)
        cerr << "hit... now refine...\n";

      double diff;
      SbVec3f bc;
      do {
        dm = (d0+d1)*0.5;
        diff = d1-dm;
        double a = (dm-minMax[0].t) * invTdiff;
        cuvh = minMax[0].p*(1-a) + minMax[1].p*a; // FIXME only need h, not u&v
        bc = SbVec3f(1-cuvh[0]-cuvh[1], cuvh[0], cuvh[1]);
        hm = texture->getColor(bc)[0]*h;

        if (above && cuvh[2]<hm ||
            !above && cuvh[2]>hm ||
            fabs(cuvh[2]-hm)<myEps) {
          d1=dm;
          h1=hm;  // is this needed?
        } else {
          d0=dm;
          h0=hm;  // is this needed?
        }
          
      } while (diff>myEps);
        
        
      // hit heightfield
      r.t = d0; // choose nearer hit, give next ray a better chance of not self-intersecting
      r.prim = this;
      r.bc = bc;
      if (DEBUG_STATE)
        cerr << "returning t=" << d0 << "\n";
      return true;
    }

  }

  if (DEBUG_STATE)
    cerr << __LINE__ << ": no hit\n";
  return false;
}
*/

SbVec3f DMTri::getNormal(const SbVec3f &bc) const {
//  return Triangle::getIntNormal(bc);
 
  // calculate dpdu and dpdv
  // This derivation follows PBRT pg. 129
  //
  //     v2
  //
  //   v
  //
  //  v0        v1
  //       u
  //
  // p(ui,vi) = v0 + ui*dpdu + vi*dpdv
  //
  // v1-v0 = (u1*dpdu +v1*dpdv) - (u0*dpdu+v0*dpdv) = (u1-u0)*dpdu + (v1-v0)*dpdv
  // v2-v0 = (u2*dpdu +v2*dpdv) - (u0*dpdu+v0*dpdv) = (u2-u0)*dpdu + (v2-v0)*dpdv
  //
  // [v1-v0] = [ u1-u0  v1-v0 ] [dpdu]
  // [v2-v0]   [ u2-u0  v2-v0 ] [dpdv]                      
  //
  // Invert       ^^^^^^^
  //
  // [ u1-u0  v1-v0 ]^-1 [v1-v0] = [dpdu]
  // [ u2-u0  v2-v0 ]    [v2-v0]   [dpdv]
  //
  // Cramers rule: Inverse of A:   x_i = det(A_i)
  //                                     --------
  //                                      det(A)

  // calc det
  SbVec2f uv0=uv(0), uv1=uv(1), uv2=uv(2);
  float A00 = uv1[0] - uv0[0];
  float A01 = uv1[1] - uv0[1];
  float A10 = uv2[0] - uv0[0];
  float A11 = uv2[1] - uv0[1];
  
  float det = A00*A11 - A01*A10;

  SbVec3f dpdu, dpdv, dndu, dndv;

  SbVec3f n = Triangle::getNormal();
  
  if (!det){
    ONB onb(n);
    dpdu = onb.getU();
    dpdv = onb.getV();
    dndu = SbVec3f(0,0,0);
    dndv = SbVec3f(0,0,0);
    cerr << __LINE__ << " oh noes\n";
  } else {
    float invDet = 1/det;

    const SbVec3f &vert0 = vert(0);
    const SbVec3f e1 = vert(1)-vert0;
    const SbVec3f e2 = vert(2)-vert0;

    // simplified
    dpdu = ( A11*e1 - A01*e2) * invDet;
    dpdv = (-A10*e1 + A00*e2) * invDet;

    const SbVec3f &normal0 = normals(0);
    SbVec3f ne1 = normals(1)-normal0;
    SbVec3f ne2 = normals(2)-normal0;
    
    dndu = ( A11*ne1 - A01*ne2) * invDet;
    dndv = (-A10*ne1 + A00*ne2) * invDet;
  }

//   float du = (fabs(A00)+fabs(A10))*0.01; // should use du,dv from ray differentials
//   float dv = (fabs(A01)+fabs(A11))*0.01; // 

  // returns dh/du, dh/dv, h
  SbVec3f dh = getDeriv(bc);
  float dhdu = dh[0]; // either of these might be zero
  float dhdv = dh[1];
  float h0 = dh[2];
  
  // using formula at bottom of PBRT pg.477
  SbVec3f dppdu = dpdu + maxHeight*dhdu*n + maxHeight*h0*dndu;
  SbVec3f dppdv = dpdv + maxHeight*dhdv*n + maxHeight*h0*dndv;

  SbVec3f n2 = dppdu.cross(dppdv);
  n2.normalize();
  return n2;
}


// calculate interpolated diffuse color
SbColor DMTri::getIntDiffColor(const Ray &r) const {
  SbColor color;
  const SbVec3f &bc = r.bc;

  const SbColor &diffuse0 = diffuse(0);
  const SbColor &diffuse1 = diffuse(1);
  const SbColor &diffuse2 = diffuse(2);
  
  color.setValue(
    bc[0]*diffuse0[R] + bc[1]*diffuse1[R] + bc[2]*diffuse2[R],
    bc[0]*diffuse0[G] + bc[1]*diffuse1[G] + bc[2]*diffuse2[G],
    bc[0]*diffuse0[B] + bc[1]*diffuse1[B] + bc[2]*diffuse2[B]
    );

  return color;
}

DMTri::DMTri() : Triangle(), maxHeight(.5) {
}
  
DMTri::DMTri(Pane *pane, SoCallbackAction *action,
             const SoPrimitiveVertex *v0,
             const SoPrimitiveVertex *v1,
             const SoPrimitiveVertex *v2,
             bool &okay) :
  Triangle(pane,action,v0,v1,v2,okay),
  maxHeight(.3) {
}

// This is the default method of computing normals.
// Also see computeNormalsIv() in Main.cxx to compute
// normals for an Inventor scene graph.
void TriMesh::computeNormals(){

  // data structures
  int numT = numTris();
  int numV = numVerts();
  vector<SbVec3f>(numV, SbVec3f(0,0,0)).swap(normals);
  int smallAreas = 0;
  int numI = indexes.size();

  // creaseAngle and other ideas from inventor's SoNormalGenerator::generate
  const float creaseAngle = M_PI/6;
  const float cosCreaseAngle = cos(creaseAngle);
  
  for (int j=0; j<numI; j+=3){

    // get center
    Vec3d v0 = verts[indexes[j+0]];
    Vec3d v1 = verts[indexes[j+1]];
    Vec3d v2 = verts[indexes[j+2]];
	
    // get normal
    Vec3d normal = (v1-v0).cross(v2-v1);
    float area2 = normal.normalize();

    // add normal to normal lists
    for (int k=0; k<3; k++){
//      Vec3d vertNorm = normals[indexes[j+k]];
//       if (vertNorm.dot(vertNorm) == 0 ||
//           normal.dot(normal) < 1e-4 ||
//           vertNorm.norm().dot(normal) > cosCreaseAngle)
        normals[indexes[j+k]] += normal*area2;
    }
	
  }

  for(int j=0; j<numV; j++){
    SbVec3f &normal = normals[j];
    float len2 = normal.dot(normal);
    if (len2) normal *= 1/sqrtf(len2);
    else {
      cerr << "Zero normal found.\n";
      normal = SbVec3f(0,1,0);
    }
  }
}
